#
# Convert HYG and Hipparcos + Tycho-2 catalogs to AstroInfo v2 catalogs
# The Tycho-2 catalog lacks the very brightest stars, so any Hipparcos
# stars not in Tycho-2 are included in Tycho-2 conversions.
#

# To save memory, skip names of objects with magnitude this magnitude or
# dimmer
$skipNameMagnitude=8;
$skippedName = "--";

use POSIX qw(floor);

$outputname = $ARGV[0];

if ($outputname =~ /hyg(([0-9.]*)-)?([0-9.]*)\.txt/) {
    $minMag = $2 if ($1 && $2);
    $maxMag = $3 if $3;
    $hyg    = 1;
} elsif ( $outputname =~ /tycho(([0-9.]*)-)?([0-9.]*)\.txt/) {
    $minMag = $2 if ($1 && $2);
    $maxMag = $3 if $3;

    $hyg    = 0;
}
else {
    print STDERR "Usage: HygToTXT.pl [hyg|tycho][frommag]-[tomag].txt\n";
}

open HYG, "<hygxyz.csv" or die;
$_=<HYG>;
s/[\r\n]//g;
s/ +/ /g;
@line = split /,/;
for $i (0..@line-1) {
    $hygcol{$line[$i]} = $i;
}
$hygcount = 0;
while(<HYG>) {
    s/[\r\n]//g;
    $hygline[ $hygcount ] = $_;
    @line = split /,/;
    $hip = $line[$hygcol{HIP}];
    if ($hip) {
        $hipToHyg[$hip] = $hygcount;
    }
    $hygcount++;
}
close HYG;

open OUTPUT, ">$outputname" or die;
if ($minMag || $maxMag) {
    $catalogname = ( $hyg ? "HYG|" : "Tycho2|" ). ($minMag ? "$minMag<=" : "") . "m"
	. ($maxMag ? "<$maxMag" : "");
} else {
    if ( $hyg ) {
        $catalogname = "HYG catalog";
    }
    else {
        $catalogname = "Tycho-2 catalog";
    }
}
print OUTPUT "$catalogname\n";

print "Processing $catalogname\n";

if ( $hyg ) {
    for $i (0..$hygcount-1) {
        @line = split ( /,/, $hygline[ $i ] );
        $mag = $line[$hygcol{Mag}];
        if ( $line[$hygcol{ProperName}] eq "Sol" ||
             ! CheckMag( $mag ) ) {
            next;
        }
        print OUTPUT HYGName($i)."\t".
                     SplitDMS($line[$hygcol{RA}])."\t".
                     SplitDMS($line[$hygcol{Dec}])."\t".
                     $mag."\t--\n";
    }
}
else {
    open TYCH, "<../build/tycho-11.csv" or die;
    $_=<TYCH>;
    s/[\r\n]//g;
    s/\s+/ /g;
    @line = split /,/;
    for $i (0..@line-1) {
        $col{$line[$i]} = $i;
    }

    while(<TYCH>) {
        s/[\r\n]//g;
        @line = split /,/;
        $mag  = $line[$col{"Mag"}];

        if ( $line[$col{HIP}] ) {
            $didHip[$line[$col{HIP}]] = 1;
        }

        if ( ! CheckMag( $mag ) ) {
            next;
        }

        if ( $line[$col{HIP}] && defined( $hipToHyg[$line[$col{HIP}]] ) ) {
            $name = HYGName($hipToHyg[$line[$col{HIP}]]);
        }
        else {
            $name = "Tyc2 ". $line[$col{TYC1}]. " " . $line[$col{TYC2}] . " " . $line[$col{TYC3}];
        }
        
        if ( $skipNameMagnitude <= $mag ) {
            $name = $skippedName;
        }
        
        print OUTPUT $name."\t".
                     SplitDMS($line[$col{RA}]/360.*24.)."\t".
                     SplitDMS($line[$col{Dec}])."\t".
                     $mag."\t--\n";
    }

    for $i (0..$hygcount-1) {
        @line = split( /,/, $hygline[ $i ] );
        $mag = $line[$hygcol{Mag}];
        if ( ! CheckMag( $mag ) || ! $line[$hygcol{HIP}] || $didHip[$line[$hygcol{HIP}]] ) {
            next;
        }
        
        $name = HYGName($i);

        if ( $skipNameMagnitude <= $mag ) {
            $name = $skippedName;
        }

        print OUTPUT $name."\t".
                     SplitDMS($line[$hygcol{RA}])."\t".
                     SplitDMS($line[$hygcol{Dec}])."\t".
                     $mag."\t--\n";
    }
}

sub SplitDMS {
    my $in = shift;
    if ( $in < 0 ) {
        $sign = "-";
        $in   = -$in;
    }
    else {
        $sign = "";
    }
    my $d = floor($in);
    $in = 60 * ( $in - $d );
    my $m = floor($in);
    $in = 60 * ( $in - $m );
    my $s = sprintf("%.4f", $in);

    return "$sign$d\t$m\t$s";
}

sub CheckMag {
   my $mag = shift;
   return ! ( ( defined($maxMag) && $maxMag <= $mag ) ||
              ( defined($minMag) && $mag < $minMag  ) )
}

sub HYGName {
    my $i = shift;
    my @line = split(/,/, $hygline[$i] );
    my $bf = $line[$hygcol{BayerFlamsteed}];
    my $name = $line[$hygcol{ProperName}];
    my $hip = $line[$hygcol{HIP}];
    my $hd = $line[$hygcol{HD}];
    my $gliese = $line[$hygcol{Gliese}];

    if ( $name ne "" ) {
        if ( $bf ne "" ) {
            return "$bf: $name";
        }
        else {
            return "$name";
        }
    }
    elsif ( $bf ne "" ) {
        return $bf;
    }
    elsif ( $hd ne "" ) {
        return "HD $hd";
    }
    elsif ( $hip ) {
        return "HIP $hip";
    }
    elsif ( $gliese ) {
        return "$gliese";
    }
    else {
        return "--";
    }
}
