/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: BurrowsWheeler.java 984 2010-04-16 10:00:24Z hoenicke $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.sf.astroinfo;
import net.sf.astroinfo.pdb.*;
import java.io.*;
import java.util.*;
import java.text.MessageFormat;

/**
 * Compress the Names of the entries in the catalogue.  The compression
 * algorithm involves the 
 * <a href="http://en.wikipedia.org/wiki/Burrows-Wheeler_transform">Burrows-Wheeler transform</a>.
 * 
 * The code is a not very well designed and I'm not proud about it 
 * but it works and is fast enough.
 *
 * The array is first reversed, to enable forward decompressing and search.
 * There is another non-standard trick here: the zeros that seperate records
 * are sorted by their position, which corresponds to their record number.
 * Thus the first characters in the sorted array are the first characters of
 * the records in their record order. This makes it possible to find a record
 * by its index.
 */
public class BurrowsWheeler implements AstroInfoConstants {
    public byte[] uncompressed;
    public byte[] lowercase;
    public BWIndex[] indices;
    public int[]  charfreq;
    public byte[] freqlen;
    public static int BLOCKSIZE=2048;
    public static int SUBBLOCKSIZE=256;
    public List<Record> entries;
    public Catalog catalog;
    public byte[] order;

    static final boolean GOOD_COMPRESSION = false;

    private static char toLower(byte c) {
	return Character.toLowerCase((char) (c & 0xff));
    }

    byte getUncompressed(int offset) {
	return uncompressed[indices[offset < indices.length 
				    ? offset : indices.length-1].index];    
    }

    class Record extends PDBRecord {
	int    index;
	int[] cf;
	byte[][] suborder   = new byte[BLOCKSIZE/SUBBLOCKSIZE][];
	byte[][] subcount   = new byte[BLOCKSIZE/SUBBLOCKSIZE][];
	byte[][] compressed = new byte[BLOCKSIZE/SUBBLOCKSIZE][];

	Record(int index) {
	    this.index = index;
	}

	Record(int index, int[] cf) {
	    this.index = index;
	    this.cf = cf;
	}

	int bits;
	byte numbits;
	public void writeBits(DataOutputStream out, int value, int nb) 
	    throws IOException 
	{
	    while (numbits + nb >= 16) {
		bits = ((bits << (16 - numbits)) 
			| (value >>> (numbits + nb - 16)));
		out.writeShort(bits);
		nb -= 16 - numbits;
		numbits = 0;
	    }
	    numbits += nb;
	    bits = ((bits << nb) | value);
	}
	public void flushBits(DataOutputStream out) throws IOException 
	{
	    if (numbits > 0) {
		if (numbits <= 8) {
		    out.writeByte(bits << (8 - numbits));
		} else {
		    out.writeShort((bits << (16-numbits)));
		}
		numbits = 0;
	    }
	}

	private void encodeRepetition(DataOutputStream out, int repeat) 
	    throws IOException 
	{
	    if (repeat >= 4) {
		writeBits(out, 15, 4);
		writeBits(out, repeat-4, 8);
	    } else {
		while (repeat-- > 0) {
		    writeBits(out, 0, 4);
		}
	    }
	}

	public void read(PalmDataInputStream out) throws IOException 
	{
	    throw new IOException("Not Implemented!");
	}

	public byte[] compressSubblock(int subblock) throws IOException{
	    byte[] suborder = new byte[order.length];
	    byte[] currorder = new byte[order.length];
	    byte[] subcount = new byte[256];
	    int    sublen = 0;
	    byte[] compressed = new byte[SUBBLOCKSIZE];
	    ByteArrayOutputStream bas = new ByteArrayOutputStream();
	    DataOutputStream out = new DataOutputStream(bas);
	    
	    int offset = index*BLOCKSIZE + subblock * SUBBLOCKSIZE;
	    int repeat = 0;
	    for (int i = 0; i < SUBBLOCKSIZE; i++) {
		byte c = getUncompressed(offset+i);
		byte lc = (byte) toLower(c);
		int j;
		for (j = 0; j < sublen; j++)
		    if ((byte) toLower(currorder[j]) == lc)
			break;
		if (j == sublen) {
		    suborder[sublen++] = c;
		    currorder[j] = c;
		}
		subcount[lc & 0xff]++;
		int compressedValue = (c == currorder[j] ? j : (j | 0x80));
		if (compressedValue == 0) {
		    repeat++;
		} else {
		    encodeRepetition(out, repeat);
		    repeat = 0;
		    if (compressedValue < 14) {
			writeBits(out, compressedValue, 4);
		    } else {
			writeBits(out, 14, 4);
			writeBits(out, compressedValue, 8);
		    }
		}
		if (j > 0) 
		    System.arraycopy(currorder, 0, currorder, 1, j);
		currorder[0] = c;
	    }
	    encodeRepetition(out, repeat);
	    this.suborder[subblock] = new byte[sublen];
	    this.subcount[subblock] = new byte[sublen];
	    for (int j = 0; j < sublen; j++) {
		byte lc = (byte) toLower(suborder[j]);
		int k;
		for (k = 0; order[k] != lc; k++)
		    ;
		this.suborder[subblock][j] = (byte)
		    (order[k] == suborder[j] ? k : (0x80 | k));
		this.subcount[subblock][j] = 
		    subcount[order[k] & 0xff];
	    }
	    flushBits(out);
	    out.flush();
	    return bas.toByteArray();
	}

	public boolean subblockEqual(int subblock1, int subblock2) {
	    int offset1 = index * BLOCKSIZE + subblock1 * SUBBLOCKSIZE;
	    int offset2 = index * BLOCKSIZE + subblock2 * SUBBLOCKSIZE;
	    for (int i = 0; i < SUBBLOCKSIZE; i++) {
		if (getUncompressed(offset1+i) != getUncompressed(offset2+i))
		    return false;
	    }
	    return true;
	}
	
	public void write(PalmDataOutputStream out) throws IOException 
	{
	    if (index == -2) {
		out.writeByte(order.length);
		for (int i = 0 ; i < order.length; i++)
		    out.writeByte(order[i]);
		if ((order.length & 1) == 0)
		    out.writeByte(0); // padding
		for (int i = 0 ; i < order.length; i++)
		    writeBits(out, charfreq[order[i]], 32);
		writeBits(out, uncompressed.length, 32);
		int offset = 0;
		for (int i = 0 ; i < order.length; i++) {
		    out.writeShort(offset);
		    offset += freqlen[order[i]];
		}
		out.writeShort(offset);
		return;
	    }

	    int[] offsets = new int[BLOCKSIZE/SUBBLOCKSIZE];
	    int offset = (BLOCKSIZE/SUBBLOCKSIZE) * 16;
	    for (int i = 0 ; i < order.length; i++)
		offset += freqlen[order[i]];
	    offset = (offset+7) >> 3;
	    for (int i = 0 ; i < BLOCKSIZE/SUBBLOCKSIZE; i++) {
		for (int j = 0; j <= i; j++) {
		    if (j == i) {
			compressed[i] = compressSubblock(i);
			if (i < BLOCKSIZE/SUBBLOCKSIZE-1)
			    offset += suborder[i].length + 1;
			offset += suborder[i].length;
			offsets[i] = offset;
			out.writeShort(offset);
			offset += compressed[i].length;
		    } else if (subblockEqual(j, i)) {
			offsets[i] = offsets[j];
			out.writeShort(offsets[i]);
			break;
		    }
		}
	    }
	    for (int i = 0 ; i < order.length; i++)
		writeBits(out, cf[order[i]], freqlen[order[i]]);
	    flushBits(out);

	    for (int i = 0 ; i < BLOCKSIZE/SUBBLOCKSIZE; i++) {
		if (i*SUBBLOCKSIZE + index*BLOCKSIZE >= indices.length)
		    break;
		if (compressed[i] != null) {
		    if (i < BLOCKSIZE/SUBBLOCKSIZE-1) {
			for (int j = suborder[i].length-1; j >= 0; j--)
			    out.writeByte(subcount[i][j]-1);
			out.writeByte(0xff); /* marker */
		    }
		    for (int j = suborder[i].length-1; j >= 0; j--)
			out.writeByte(suborder[i][j]);
		    for (int j = 0; j < compressed[i].length; j++)
			out.writeByte(compressed[i][j]);
		}
	    }
	}
    }

    static int bitlen(int value) {
	for (int i = 0; ; i++)
	    if (value < (1<<i))
		return i;
    }

    class BWIndex implements Comparable<BWIndex> {
	private int index;

	public BWIndex(int idx) {
	    this.index = idx;
	}
	
	public int compareTo(BWIndex other) {
	    int l = uncompressed.length;
	    char ca, cb;
	    int a = this.index;
	    int b = other.index;
	    if (a == b)
		return 0;
	    while (true) {
		a--;
		b--;
		/* First index is always the smallest! */
		if (a < 0)
		    return -1;
		if (b < 0)
		    return +1;
		ca = (char) (lowercase[a] & 0xff);
		cb = (char) (lowercase[b] & 0xff);
		if (ca != cb)
		    return ca - cb;
		/* Sort zeros by record position */
		if (ca == 0)
		    return a - b;
	    }
	}
    }

    public BurrowsWheeler(List<CatalogRecord> catEntries) {
	ByteArrayOutputStream bas = new ByteArrayOutputStream();

	for (CatalogRecord cr : catEntries) {
	    try {
		bas.write(cr.description.getBytes());
		bas.write(0);
	    } catch (IOException ex) {
		throw (InternalError) new InternalError().initCause(ex);
	    }
	}
	uncompressed = bas.toByteArray();
	int l = uncompressed.length;
	lowercase = new byte[uncompressed.length];

	for (int i = 0; i < l; i++)
	    lowercase[i] = (byte) toLower(uncompressed[i]);
	/* build indices array and sort it. */
	indices = new BWIndex[l];
	for (int i = 0; i < l; i++) {
	    indices[i] = new BWIndex(i);
	}
	Arrays.sort(indices);

	charfreq = new int[256];
	int[][] blockCharFreq = new int[(l-1)/BLOCKSIZE + 1][];
	for (int b = 0; b < (l-1)/BLOCKSIZE + 1; b++) {
	    blockCharFreq[b] = (int[]) charfreq.clone();
	    for (int i = BLOCKSIZE*b; i < l && i < BLOCKSIZE*(b+1); i++) {
		charfreq[lowercase[indices[i].index]]++;
	    }
	}
	freqlen = new byte[256];
	for (int i = 0; i < 256; i++) {
	    freqlen[i] = (byte) bitlen(charfreq[i]);
	}
	int orderlen = 0;
	for (int i = 0; i < 256; i++) {
	    if (charfreq[i] > 0)
		orderlen++;
	}
	order = new byte[orderlen];
	orderlen = 0;
	for (int i = 0; i < 256; i++) {
	    if (charfreq[i] > 0)
		order[orderlen++] = (byte)i;
	}
	int[] cf = new int[256];
	cf[0] = 0;
	for (int i = 1; i < 256; i++)
	    cf[i] = cf[i-1] + charfreq[i-1];
	charfreq = cf;

	entries = new ArrayList<Record>();
	entries.add(new Record(-2));
	for (int i = 0; i * BLOCKSIZE < l; i ++) {
	    entries.add(new Record(i, blockCharFreq[i]));
	}
    }
}
