/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: Catalog.java 976 2010-04-14 23:17:51Z hoenicke $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.sf.astroinfo;
import net.sf.astroinfo.pdb.*;
import java.io.*;
import java.util.*;


public class Catalog implements AstroInfoConstants {
    Database  pdb;
    String    name;
    String    description;
    static int inputVersion;
    List<CatalogRecord> entries;
    static int version = 2;
    static int qTreeObjectSize = V2_OBJECT_SIZE;
    static int qTreeNodeSize   = V2_QTREE_NODE_SIZE;
    static int maxQTreeNodeSize = V2_MAX_QTREE_NODE_SIZE;
    static int qTreeNodeMultiplier = maxQTreeNodeSize;
    static boolean noSeconds = true;
    static boolean distance = false;
    static int maxRecordSize = 64720;
    static boolean unsorted = false;
    static boolean isstar = false;

    int magHigh = 0;
    int numStars;
    int bytesPerEntry;
    int entriesPerIndex;
    int numberOfIndexNodes;

    public Catalog() {
    }
    
    class CatalogInfo extends AppInfoBlock {
	int numberOfBWNodes;
	int numberOfQtreeNodes;

	public CatalogInfo(int bwNodes, int qtreeNodes) {
	    numberOfBWNodes    = bwNodes;
	    numberOfQtreeNodes = qtreeNodes;
	}

	public void read(PalmDataInputStream in) throws IOException
	{
	    throw new IOException("Not Implemented!");
	}

	public void write(PalmDataOutputStream out) throws IOException
	{
	    out.writeInt(numStars);
	    out.writeShort(numberOfIndexNodes);
	    out.writeShort(numberOfBWNodes);
	    out.writeShort(numberOfQtreeNodes);
	    out.writeShort(qTreeObjectSize);
	    out.writeShort(qTreeNodeMultiplier);
	    out.writeShort(entriesPerIndex);
	    out.writeByte(bytesPerEntry);

	    out.writeByte((distance?0x80:0)+(isstar?0x40:0));
	    out.writeByte(magHigh);
	}
    }

    class IndexNode extends PDBRecord {
	int   nr;
	int[] indices;

	public IndexNode(int nr, int[] indices) {
	    this.nr = nr;
	    this.indices = indices;
	}

	public void read(PalmDataInputStream in) throws IOException
	{
	    throw new IOException("Not Implemented!");
	}

	public void write(PalmDataOutputStream out) throws IOException
	{
	    int start = nr * entriesPerIndex;
	    int end = Math.min(indices.length, start + entriesPerIndex);
	    for (int i = start; i < end; i++) {
		int number = indices[i];
		int shift = 8*bytesPerEntry;
		while (shift > 0) {
		    shift -= 8;
		    out.writeByte((number >> shift) & 0xff);
		}
	    }
	    /* support for version 3 */
	    if (version == 3 && nr == 0) {
		out.writeShort(qTreeNodeMultiplier);
		out.writeShort(qTreeObjectSize);
		out.writeByte(magHigh);
		out.writeByte((distance?0x80:0));
		out.writeShort(8);
	    }
	}
    }

    public void read(InputStream is) throws IOException {
	throw new IOException("Not Implemented!");
    }

    public void write(OutputStream os) throws IOException {
	numStars = entries.size();
	QuadTree quad = new QuadTree(name, entries, 
				     version, 
				     maxQTreeNodeSize,
				     qTreeNodeSize,
				     qTreeObjectSize,
				     distance);

	if (unsorted) {
	    /* reorder entries in qtree order */
	    CatalogRecord[] newEntries = new CatalogRecord[numStars];
	    int entryCtr = 0;
	    for (QuadTree.Node node : quad.allNodes) {
		for (CatalogRecord rec : node.records) {
		    newEntries[entryCtr] = rec;
		    rec.index = entryCtr;
		    entryCtr++;
		}
	    }
	    entries = Arrays.asList(newEntries);
	}

	BurrowsWheeler bw = new BurrowsWheeler(entries);

	if (qTreeObjectSize <= 10)
	    magHigh = quad.magHigh;

	CatalogInfo catInfo = 
	    new CatalogInfo(bw.entries.size(), quad.allNodes.size());
	List<PDBRecord> allRecords = new LinkedList<PDBRecord>();

	if (!unsorted) {	
	    int numNodes = (numStars-1)/(maxRecordSize/2) + 1 + 
		bw.entries.size() + quad.allNodes.size();
	    if (numNodes * maxQTreeNodeSize > 65536) {
		bytesPerEntry = 3;
		entriesPerIndex = (maxRecordSize / 3) & ~1;
		qTreeNodeMultiplier = 256;
	    } else {
		bytesPerEntry = 2;
		entriesPerIndex = (maxRecordSize / 2);
		qTreeNodeMultiplier = maxQTreeNodeSize;
	    }
	    numberOfIndexNodes = (numStars-1) / entriesPerIndex + 1;
	    int rootNode = numberOfIndexNodes + bw.entries.size();
	    quad.setRootNodeIndex(rootNode);

	    int[] indices = new int[numStars];
	    for (QuadTree.Node node : quad.allNodes) {
		int offset = 0;
		for (CatalogRecord star : node.records) {
		    indices[star.index] = 
			(rootNode + node.getQtreeIndex()) * qTreeNodeMultiplier
			+ offset;
		    offset++;
		}
	    }
	    for (int i = 0; i < numberOfIndexNodes; i++)
		allRecords.add(new IndexNode(i, indices));

	    if (version < 4 && numberOfIndexNodes != 1
		|| bytesPerEntry != 2)
		throw new IOException("Catalog too big, use version 4.");

	} else {
	    qTreeNodeMultiplier = 0;
	    bytesPerEntry = 4;
	    entriesPerIndex = (maxRecordSize / bytesPerEntry);
	    numberOfIndexNodes = quad.allNodes.size() / entriesPerIndex + 1;

	    int rootNode = numberOfIndexNodes + bw.entries.size();
	    quad.setRootNodeIndex(rootNode);
	    int offset = 0;
	    int[] qTreeOffsets = new int[quad.allNodes.size()];
	    for (QuadTree.Node node : quad.allNodes) {
		qTreeOffsets[node.getQtreeIndex()] = offset;
		for (CatalogRecord star : node.records) {
		    if (star.index != offset++)
			throw new InternalError("reordering is broken");
		}
	    }
	    for (int i = 0; i < numberOfIndexNodes; i++)
		allRecords.add(new IndexNode(i, qTreeOffsets));
	}

	allRecords.addAll(bw.entries);
	allRecords.addAll(quad.allNodes);

	Database pdb = new Database();
	pdb.setName("AstroInfo "+name);
	pdb.setVersion(version);
	pdb.setAttributes(Database.DMHDRATTRBUNDLE+Database.DMHDRATTRBACKUP);
	pdb.setCreator(creatorID);
	pdb.setType(catID);
	PDBRecord[] recs = allRecords.toArray(new PDBRecord[0]);
	pdb.setEntries(recs);
	if (version >= 4)
	    pdb.setAppInfo(catInfo);
	pdb.write(os);
    }

    public void parseV2(Reader r) throws IOException {
	BufferedReader br = new BufferedReader(r);
	String line = br.readLine();
	name = line;
	int linenr = 1;
	try {
	    while (br.ready()) {
		line = br.readLine().trim();
		linenr++;
		if (line.length() == 0 || line.startsWith("#"))
		    continue;
		CatalogRecord rec = new CatalogRecord(entries.size());
		rec.parseV2(line);
		entries.add(rec);
	    }
	} catch (RuntimeException ex) {
	    System.err.println("line "+linenr+": parse error");
	    throw ex;
	}
    }
    
    public void parse(Reader r) throws IOException {
	entries = new ArrayList<CatalogRecord>();
	if (inputVersion <= 2) {
	    parseV2(r);
	    return;
	}

	String prefix ="";
	int layer;
	int objects = 0;
	BufferedReader br = new BufferedReader(r);
	int linenr = 0;
	try {
	    while (br.ready()) {
		String line = br.readLine().trim();
		linenr++;
		if (line.startsWith("#"))
		    continue;
		StringTokenizer st = new StringTokenizer(line);
		if (!st.hasMoreTokens())
		    continue;
		String token = st.nextToken();
		if (token.equals("Version:")) {
		    inputVersion = Integer.parseInt(st.nextToken());
		} else if (token.equals("Catalog:")) {
		    StringBuffer sb = new StringBuffer();
		    sb.append(st.nextToken());
		    while (st.hasMoreTokens()) {
			sb.append(' ').append(st.nextToken());
		    }
		    name = sb.toString();
		} else if (token.equals("CatDesc:")) {
		    StringBuffer sb = new StringBuffer();
		    sb.append(st.nextToken());
		    while (st.hasMoreTokens()) {
			sb.append(' ').append(st.nextToken());
		    }
		    description = sb.toString();
		} else if (token.equals("Prefix:")) {
		    StringBuffer sb = new StringBuffer();
		    sb.append(st.nextToken());
		    while (st.hasMoreTokens()) {
			sb.append(' ').append(st.nextToken());
		    }
		    prefix = sb.toString();
		} else if (token.equals("Layer:")) {
		    layer = Integer.parseInt(st.nextToken());
		} else if (token.equals("Objects:")) {
		    objects = Integer.parseInt(st.nextToken());
		    break;
		} else {
		    System.err.println("line "+linenr+": Unknown entry "+token);
		}
	    }
	    
	    while (br.ready()) {
		String line = br.readLine().trim();
		linenr++;
		if (line.length() > 0 && !line.startsWith("#")) {
		    CatalogRecord rec = new CatalogRecord(entries.size());
		    rec.parse(prefix, line);
		    entries.add(rec);
		    objects--;
		}
	    }
	} catch (RuntimeException ex) {
	    System.err.println("line "+linenr+": parse error");
	    throw ex;
	}
	if (objects != 0)
	    throw new IllegalArgumentException("Number of objects doesn't match.");
    }
    
    public void dump(Writer w) throws IOException {
	PrintWriter pw = new PrintWriter(w);
	pw.println(name);
	for (CatalogRecord entry : entries) {
	    pw.println(entry.toString());
	}
	pw.flush();
    }

    public static void usage() {
	System.err.println("SYNTAX: java net.sf.astroinfo.Catalog"+
			   " {-v/o/m/a/u [arg]} <infile> <outfile>");
	// unsupported: System.err.println("\t-r:\tfrom binary to text (default is reverse)");
	System.err.println("\t-v <num>:\tset version number");
	System.err.println("\t-o <num>:\tset size of qtree object (14,16,18,20 or 22))");
	System.err.println("\t-m <num>:\tmax stars per qtree record");
	System.err.println("\t-a <num>:\tavg stars per qtree record");
	System.err.println("\t-u:\tdo not preserve the ordering (unsorted)");
    }

    public static void main(String[] args) throws IOException {
	boolean reverse = false;
	int argsOffset = 0;
	noSeconds = true;
	inputVersion = 3;
        while (argsOffset < args.length && args[argsOffset].charAt(0)=='-') {
            switch (args[argsOffset].charAt(1)) {
            case 'd':
                distance = true;
                if (qTreeObjectSize < 22)
                    qTreeObjectSize = 22;
                break;
	    case 'u':
		unsorted = true;
		break;
	    case 'r':
		reverse = true;
		break;
	    case 's':
		noSeconds = false;
		break;
	    case 'v':
		version = Integer.parseInt(args[++argsOffset]);
		if (version >= 3) {
		    qTreeNodeSize = V3_QTREE_NODE_SIZE;
		    maxQTreeNodeSize = V3_MAX_QTREE_NODE_SIZE;
		}
		break;
	    case 'm':
		maxQTreeNodeSize = Integer.parseInt(args[++argsOffset]);
		break;
	    case 'a':
		qTreeNodeSize = Integer.parseInt(args[++argsOffset]);
		break;
	    case 'o':
		qTreeObjectSize = Integer.parseInt(args[++argsOffset]);
		if (distance && qTreeObjectSize < 22)
		    qTreeObjectSize = 22;
		break;
	    case '2':
		inputVersion = 2;
		break;
	    default:
		System.err.println("Invalid option: "+args[argsOffset]);
		usage();
		return;
            }
            argsOffset++;
        }
	if (args.length - argsOffset != 2) {
	    usage();
	    return;
	}

	if (version < 4 && unsorted) {
	    System.err.println("Unsorted catalogs require version 4.");
	    usage();
	    return;
	}

	if (version < 3 && qTreeObjectSize != V2_OBJECT_SIZE) {
	    System.err.println("Changing the object size requires at least version 3 output");
	    usage();
	    return;
	}

	Catalog db = new Catalog();
	if (reverse) {
	    db.read(new FileInputStream(args[argsOffset]));
	    if (args[argsOffset+1].equals("-"))
		db.dump(new OutputStreamWriter(System.out));
	    else
		db.dump(new FileWriter(args[argsOffset+1]));
	} else {
	    if (args[argsOffset].equals("-"))
		db.parse(new InputStreamReader(System.in));
	    else
		db.parse(new FileReader(args[argsOffset]));
	    db.write(new FileOutputStream(args[argsOffset+1]));
	}
    }
}
