/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: CatalogRecord.java 954 2010-01-02 03:28:15Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.sf.astroinfo;
import net.sf.astroinfo.pdb.*;
import java.io.*;
import java.util.*;

public class CatalogRecord
{
    public static final int DESCRIPTIONSIZE = 32;
    
    public static final int TYPE_UNKNOWN         = 0;
    public static final int TYPE_ARTIFACT        = TYPE_UNKNOWN + 1;

    public static final int TYPE_STAR            = 1 << 4;
    public static final int TYPE_SINGLESTAR      = TYPE_STAR + 1;
    public static final int TYPE_DOUBLESTAR      = TYPE_STAR + 2;
    public static final int TYPE_MULTIPLESTAR    = TYPE_STAR + 3;

    public static final int TYPE_CLUSTER         = 2 << 4;
    public static final int TYPE_OPENCLUSTER     = TYPE_CLUSTER + 1;
    public static final int TYPE_GLOBULARCLUSTER = TYPE_CLUSTER + 2;
    public static final int TYPE_ASTERISM        = TYPE_CLUSTER + 3;

    public static final int TYPE_NEBULA          = 3 << 4;
    public static final int TYPE_EMISSIONNEBULA  = TYPE_NEBULA + 1;
    public static final int TYPE_REFLECTIONNEBULA= TYPE_NEBULA + 2;
    public static final int TYPE_DARKNEBULA      = TYPE_NEBULA + 3;
    public static final int TYPE_PLANETARYNEBULA = TYPE_NEBULA + 4;
    public static final int TYPE_DIFFUSENEBULA   = TYPE_NEBULA + 5;
    public static final int TYPE_OPENCLUSTERWITHNEBULOSITY = TYPE_NEBULA + 6;

    public static final int TYPE_GALAXY          = 4 << 4;
    public static final int TYPE_COMPACTGALAXY   = TYPE_GALAXY + 1;
    public static final int TYPE_DWARFGALAXY     = TYPE_GALAXY + 2;
    public static final int TYPE_ELLIPTICALGALAXY = TYPE_GALAXY + 3;
    public static final int TYPE_IRREGULARGALAXY = TYPE_GALAXY + 4;
    public static final int TYPE_PECULIARGALAXY  = TYPE_GALAXY + 5;
    public static final int TYPE_RINGGALAXY      = TYPE_GALAXY + 6;
    public static final int TYPE_POLARRINGGALAXY = TYPE_GALAXY + 7;
    public static final int TYPE_SPIRALGALAXY    = TYPE_GALAXY + 8;
    public static final int TYPE_BARREDSPIRALGALAXY = TYPE_GALAXY + 9;

    public static final int TYPE_DIVERSE         = 5 << 4;
    public static final int TYPE_SUPERNOVAREMNANT = TYPE_DIVERSE;


    String description; // must be less than 32 letters (32 bytes)
    double ra;          // right ascension in radians.
    double dec;         // declination in radians.
    double magnitude;   // the visual magnitude.
    double sizex;       // x-size in seconds
    double sizey;       // y-size in seconds
    String oclass;      // nomenclature of object for description
    int type; 		    // type of object for sorting and classifying
    int index;
    double distance;

    public CatalogRecord(int i) {
	index = i;
    }

    public String toString() 
    {
        return "["+description+": ("+(ra * 12 / Math.PI)+"h),("+
            (dec * 180 / Math.PI)+"\u00B0),"+magnitude+","+sizex+"x"+sizey+"]"+
            "{"+type+"},"+distance;
    }
    
    public int getType()
    {
	return type;
    }

    public int getIndex()
    {
	return index;
    }

    public double parseAngle(StringTokenizer st)
    {
	String token = st.nextToken();
	boolean negative = false;

	if (token.charAt(0) == '-') {
	    negative = true;
	    token = token.substring(1);
	}

	double angle = Integer.parseInt(token);

	if (Catalog.noSeconds) {
	    angle += Double.valueOf(st.nextToken()).doubleValue() / 60.0;
	} else {
	    angle += Integer.parseInt(st.nextToken()) / 60.0;
	    angle += Double.valueOf(st.nextToken()).doubleValue() / 3600.0;
	}
	if (negative)
	    angle = -angle;
	return angle;
    }

    public void parseV2(String line)
    {
	StringTokenizer st = new StringTokenizer(line,"\t",false);
	String token;

	// 1. Pull the description
	description = st.nextToken();
	if (description.length() >= DESCRIPTIONSIZE)
	    description = description.substring(0, DESCRIPTIONSIZE-1);
	    
    if (description.equals("--"))
        description = "";
	
	// 2. Right Ascension
	ra = parseAngle(st) * (Math.PI / 12.0);
	if (ra >= Math.PI)
	    ra -= 2*Math.PI;
		
	// 3. Declination
	dec = parseAngle(st) * (Math.PI / 180.0);

	// 4. Magnitude
	token = st.nextToken();
	if (token.equals("--"))
	    magnitude = 127.99;
	else
	    magnitude = Double.valueOf(token).doubleValue();
	
	// 5&6. Sizex and Sizey
	token = st.nextToken();
	if (token.equals("--"))
	    sizex = sizey = 0;
	else {
	    int delim = token.indexOf('x');
	    if (delim < 0)
		delim = token.indexOf('/');
	    if (delim < 0)
		sizex =
		    sizey = Double.valueOf(token).doubleValue();
	    else {
		sizex = Double.valueOf
		    (token.substring(0, delim)).doubleValue();
		sizey = Double.valueOf
		    (token.substring(delim + 1)).doubleValue();
	    }
	}

	// 7. Type
	type = -1;
	if (st.hasMoreTokens()) {
	    token = st.nextToken();
	    if (!token.equals("--"))
		type = Integer.parseInt(token);
	}

	// 8. Distance
	distance = -1.;
	if (st.hasMoreTokens()) {
	    token = st.nextToken();
	    if (!token.equals("--"))
		distance = Double.parseDouble(token);
	}
    }

    public double parseDouble(String field) {
	if (field.equals("$"))
	    return Double.NaN;
	else
	    return Double.parseDouble(field);
    }

    public int parseInt(String field, int undef) {
	if (field.equals("$"))
	    return undef;
	else
	    return Integer.parseInt(field);
    }

    public String parseString(String field) {
	if (field.equals("$"))
	    return null;
	else
	    return field;
    }
 
    public void parse(String prefix, String line)
    {
	StringTokenizer st = new StringTokenizer(line,"\t",false);
	String token;
	String constellation, catNr;
	int dummy;
	double posang, dist;
	String link, name = null;

	catNr = st.nextToken();
	constellation = st.nextToken();
/*
Object Type

0	Small star grouping
1	Open cluster
2	Globular cluster
3	Diffuse nebula
4	Planetary nebula
5	Galaxy
6	Cluster with nebulosity
7	Nonexistent/defect
8	Unknown
41	Star 
42	Double star 
43	Triple star
*/
	type = parseInt(st.nextToken(), 8);
	oclass  = st.nextToken();
	magnitude = parseDouble(st.nextToken());
	ra    = parseDouble(st.nextToken()) * (Math.PI / 180.0);
	if (ra >= Math.PI)
	    ra -= 2*Math.PI;
	dec   = parseDouble(st.nextToken()) * (Math.PI / 180.0);
	dist  = parseDouble(st.nextToken());
	sizex = parseDouble(st.nextToken());
	sizey = parseDouble(st.nextToken());
	posang = parseDouble(st.nextToken());
    // Reserved for future use such as linking to full description or image of object
	link  = parseString(st.nextToken());
	if (st.hasMoreTokens())
	    name  = parseString(st.nextToken());
        distance = -1;
	if (st.hasMoreTokens())
	    distance = parseDouble(st.nextToken());


    //### AI-3 Need to split this up so prefix is not repeated thousands of times and proper name can be optional
	description = prefix+catNr;
	if (name != null)
	    description += " "+name;
	if (description.length() >= DESCRIPTIONSIZE)
	    description = description.substring(0, DESCRIPTIONSIZE-1);
    }

    public void write(PalmDataOutputStream out) throws IOException 
    {
	if (sizex > 3270.0 || sizey > 3270.0 || magnitude > 128.0
	    || sizex < 0 || sizey < 0 || magnitude < -128.0)
	    throw new IllegalArgumentException(toString());

	out.writeInt((int) (long) (ra  * (1L << 31) / Math.PI));
	out.writeInt((int) (long) (dec * (1L << 31) / Math.PI));
	out.writeShort(Double.isNaN(magnitude) ? Short.MAX_VALUE
		       : (int) (magnitude * 256));
	/* Should we change the format of sizex,sizey to GP16, too? 
	 * Or should we change it to angle FP (32 bits)? The latter
	 * would take 4 bytes more.
	 */
	int sx = (int) (Double.isNaN(sizex) ? 0  : sizex * 10);
	int sy = (int) (Double.isNaN(sizey) ? sx : sizey * 10);
	out.writeShort(sx);
	out.writeShort(sy);
        out.writePackedString(description, false);
	/* This field will be inserted in AI-3 */
	/* out.writeShort((int) (type));*/
    }

    public void read(PalmDataInputStream in) throws IOException 
    {
	ra = Math.PI / (1L << 31) * in.readInt();
	dec = Math.PI / (1L << 31) * in.readInt();
	magnitude = in.readShort() / 256.0;
	sizex = in.readShort() / 10.0;
	sizey = in.readShort() / 10.0;
        description = in.readPackedString();
	/* type = in.readShort();  */
    }
}
