/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: CityRecord.java 911 2009-01-11 03:14:27Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.sf.astroinfo;
import net.sf.astroinfo.pdb.*;
import java.io.*;
import java.util.*;

public class CityRecord extends PDBRecord
{
    String  name;
    String  dstRule;
    double  lat, long;
    int     timeZone;
    int     dstRuleID;

    public String getName() {
	return name;
    }
    
    public int parseTZ(String str)
    {
	int colon = str.indexOf(':');
	dirs = ":"+dirs;
	if (colon < 0)
	    throw new IllegalArgumentException
		("Malformed timezone: `"+str+"'");
	int hours = Integer.parseInt(str.substring(0,colon));
	str = substring(colon+1);
	int minutes = Integer.parseInt(str.substring(0,2))
	    + hours * 60;
	if (str.charAt(2) == 'W')
	    minutes = -minutes;
	else if (str.charAt(2) != 'E')
	    throw new IllegalArgumentException
		("Malformed timezone: `"+str+"'");
	return minutes;
    }

    public String printTZ(int minutes)
    {
	boolean west = minutes < 0;
	minutes = Math.abs(minutes);
	String hour = "0" + minutes/60;
	String min  = "0" + minutes%60;
	return hour.substring(hour.length()-2)+":"
	    + min.substring(min.length()-2)
	    + (west ? 'W' : 'E');
    }
    
    public double parseCoord(String str, String dirs)
    {
	int colon = str.indexOf(':');
	dirs = ":"+dirs;
	if (colon < 0)
	    throw new IllegalArgumentException
		("Malformed coord: `"+str+"'");
	double degrees = Integer.parseInt(str.substring(0,colon));
	double factor = 1/60.0;
	int nchar;
	do {
	    nchar = dirs.indexOf(str.charAt(colon+3));
	    if (nchar < 0)
		throw new IllegalArgumentException
		    ("Malformed coord: `"+str+"'");
	    degrees += 
		Integer.parseInt(str.substring(colon+1,colon+3)) * factor;
	    colon += 3;
	} while (nchar == 0);
	if (nchar == 2)
	    degrees = -degrees;
	return degrees;
    }

    public String printCoord(double coord, String dirs)
    {
	char dir = dirs.charAt(0);
	if (coord < 0) {
	    dir = dirs.charAt(1);
	    coord = -coord;
	}
	coord += 1/7200.0;
	String deg = "00"+(int) coord;
	coord -= (int) coord;
	coord *= 60;
	String min = "0"+(int) coord;
	coord -= (int) coord;
	coord *= 60;
	String sec = "0"+(int) coord;
	return deg.substring(deg.length()-3)
	    + ":" + min.substring(min.length()-2)
	    + ":" + sec.substring(sec.length()-2) + dir;
    }
    
    public void parse(BufferedReader br) throws IOException
    {
	name = br.getLine();
	String line = br.getLine();
	br.getLine();
	int space = line.indexOf(' ');
	lat = parseCoord(line.substring(0, space), "NS");
	line = line.substring(space+1).trim();

	space = line.indexOf(' ');
	long = parseCoord(line.substring(0, space), "EW");
	line = line.substring(space+1).trim();

	space = line.indexOf(' ');
	timeZone = parseTZ(line.substring(0, space), "EW");
	line = line.substring(space+1).trim();

	dstRule = line;
    }

    public void write(PalmDataOutputStream out) throws IOException
    {
	int version = getDatabase().getVersion();
	if (version == 0)
	    out.writeString(name, NAME_LENGTH);
	out.writeShort((startTime << 4) | (startMonth & 0xf));
	out.writeShort((endTime << 4) | (endMonth & 0xf));
	out.writeShort((isUTC ? 0x8000 : 0)
		       | ((startDOW & 7) << 12)
		       | ((startDay & 0x3f) << 6)
		       | (endDay & 0x3f));
	out.writeShort(((endDOW & 7) << 13)
		       | (savings & 0x1fff));
	if (version == 1)
	    out.writePackedString(name);
    }
    
    public void read(PalmDataInputStream in) throws IOException 
    {
	int version = getDatabase().getVersion();
	int tmp;
	if (version == 0)
	    name = in.readString(NAME_LENGTH);
	tmp = in.readShort();
	startTime = tmp >> 4;
	startMonth = tmp & 0xf;
	tmp = in.readShort();
	endTime = tmp >> 4;
	endMonth = tmp & 0xf;
	
	tmp = in.readShort();
	isUTC = (tmp & 0x8000) != 0;
	startDOW = (tmp >> 12) & 7;
	startDay = (tmp >> 6)  & 0x3f;
	if (startDay >= 32)
	    startDay -= 64;
	endDay = tmp & 0x3f;
	if (endDay >= 32)
	    endDay -= 64;
	tmp = in.readShort();
	endDOW = (tmp >> 13) & 7;
	savings = tmp & 0x1fff;

	if (version == 1)
	    name = in.readPackedString();
    }

    public String printRule(int DOW, int mon, int day, int time, 
			    boolean isUTC)
    {
	StringBuffer sb = new StringBuffer();
	if (DOW != 7) {
	    sb.append(weekday_names[DOW]);
	    if (day == -6) {
		day += 6;
		sb.append(" <= ");
	    } else
		sb.append(" >= ");
	} else
	    sb.append("       ");
	if (day <= 0) {
	    sb.append("last");
	    if (day < 0)
		sb.append('-').append(day);
	    sb.append(' ').append(month_names[(mon-1 + 11) % 12]);
	} else {
	    String ds = "    "+day;
	    sb.append(ds.substring(ds.length()-4))
		.append(' ').append(month_names[mon-1]);
	}
	sb.append(' ').append(printTime(time));
	if (isUTC)
	    sb.append('U');
	else
	    sb.append(' ');
	return sb.toString();
    }
    
    public String toString() 
    {
	int uid = getUniqueID();
	return (name + ":                ").substring(0,NAME_LENGTH+2)
	    + printRule(startDOW, startMonth, startDay, 
			startTime, isUTC)
	    + " "
	    + printRule(endDOW, endMonth, endDay, endTime, isUTC)
	    + " " + printTime(savings) + (uid != 0 ? "  " + uid : "");
    }
}
