/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: DSTRecord.java 613 2006-08-21 18:24:39Z hoenicke $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.sf.astroinfo;
import net.sf.astroinfo.pdb.*;
import java.io.*;
import java.util.*;

public class DSTRecord extends PDBRecord
{
    final static String weekday_names[] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
    };
    final static String month_names[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
    };
    
    final static int NAME_LENGTH = 16;

    String  name;
    int     startTime;
    int     startDay;
    int     startMonth;
    int     startDOW;
    int     endTime;
    int     endDay;
    int     endMonth;
    int     endDOW;
    int     savings;
    boolean isUTC;

    public String getName() {
	return name;
    }
    
    public String printTime(int minutes)
    {
	String hour = "0" + minutes/60;
	String min  = "0" + minutes%60;
	return hour.substring(hour.length()-2)+":"
	    + min.substring(min.length()-2);
    }
    
    public int parseTime(String str)
    {
	if (str.charAt(2) != ':')
	    throw new IllegalArgumentException
		("Malformed time: `"+str+"'");
	return Integer.parseInt(str.substring(0,2)) * 60 
	    + Integer.parseInt(str.substring(3,5));
    }
    
    
    public String parseRule(boolean isStart, String line)
    {
	int dow;
	int day = 0;
	int month;
	int time;
	boolean ge = true;
	boolean isLast = false;
	boolean parseDay = true;
	boolean utc = false;
	for (dow=0; dow < 7; dow++) {
	    if (line.toLowerCase()
		.startsWith(weekday_names[dow].toLowerCase())) {
		line = line.substring(3).trim();
		if (line.startsWith("<="))
		    ge = false;
		else if (line.startsWith(">="))
			ge = true;
		else
		    throw new IllegalArgumentException
			("Weekday without <=/>=");
		line = line.substring(2).trim();
		break;
	    }
	}
	if (line.startsWith("last")) {
	    isLast = true;
	    line = line.substring(4).trim();
	    if (!line.startsWith("-"))
		parseDay = false;
	    else
		line = line.substring(1).trim();
	}
	if (parseDay) {
	    int i;
	    for (i = 0; 
		 i < line.length() && Character.isDigit(line.charAt(i));
		 i++);
	    day = Integer.parseInt(line.substring(0,i));
	    line = line.substring(i).trim();
	}
	for (month = 0; month < 12; month++) {
	    if (line.toLowerCase()
		.startsWith(month_names[month].toLowerCase())) {
		line = line.substring(3).trim();
		break;
	    }
	}
	if (month == 12)
	    throw new IllegalArgumentException("Unknown Month");
	time = parseTime(line.substring(0, 5));
	line = line.substring(5);
	if (line.charAt(0) == 'U') {
	    utc = true;
	    line =line.substring(1);
	}
	line = line.trim();
	month++;
	if (isLast) {
	    day = -day;
	    month = (month % 12) + 1;
	}
	if (!ge)
	    day -= 6;
	
	if (isStart) {
	    startDOW = dow;
	    startDay = day;
	    startMonth = month;
	    startTime = time;
	    isUTC = utc;
	} else {
	    endDOW = dow;
	    endDay = day;
	    endMonth = month;
	    endTime = time;
	    if (isUTC != utc)
		throw new IllegalArgumentException("UTC inconsistent");
	}
	return line;
    }

    public void parse(String line)
    {
	int colon = line.indexOf(':');
	if (colon < 0 || colon > NAME_LENGTH)
	    throw new IllegalArgumentException("Colon not found");
	name = line.substring(0, colon);
	line = line.substring(colon+1).trim();
	line = parseRule(true, line);
	line = parseRule(false, line);
	savings = parseTime(line.substring(0,5));
	line = line.substring(5).trim();
	if (line.length() > 0) {
	    int idx = line.indexOf(' ');
	    if (idx < 0)
		idx = line.length();
	    setUniqueID(Integer.parseInt(line.substring(0, idx)));
	    line = line.substring(idx).trim();
	}
	if (line.length() > 0)
	    throw new IllegalArgumentException("Trailing garbage: `"
					       +line+"'");
    }
    
    public void write(PalmDataOutputStream out) throws IOException 
    {
	int version = getDatabase().getVersion();
	if (version == 0)
	    out.writeString(name, NAME_LENGTH);
	out.writeShort((startTime << 4) | (startMonth & 0xf));
	out.writeShort((endTime << 4) | (endMonth & 0xf));
	out.writeShort((isUTC ? 0x8000 : 0)
		       | ((startDOW & 7) << 12)
		       | ((startDay & 0x3f) << 6)
		       | (endDay & 0x3f));
	out.writeShort(((endDOW & 7) << 13)
		       | (savings & 0x1fff));
	if (version == 1)
	    out.writePackedString(name);
    }
    
    public void read(PalmDataInputStream in) throws IOException 
    {
	int version = getDatabase().getVersion();
	int tmp;
	if (version == 0)
	    name = in.readString(NAME_LENGTH);
	tmp = in.readShort();
	startTime = tmp >> 4;
	startMonth = tmp & 0xf;
	tmp = in.readShort();
	endTime = tmp >> 4;
	endMonth = tmp & 0xf;
	
	tmp = in.readShort();
	isUTC = (tmp & 0x8000) != 0;
	startDOW = (tmp >> 12) & 7;
	startDay = (tmp >> 6)  & 0x3f;
	if (startDay >= 32)
	    startDay -= 64;
	endDay = tmp & 0x3f;
	if (endDay >= 32)
	    endDay -= 64;
	tmp = in.readShort();
	endDOW = (tmp >> 13) & 7;
	savings = tmp & 0x1fff;

	if (version == 1)
	    name = in.readPackedString();
    }

    public String printRule(int DOW, int mon, int day, int time, 
			    boolean isUTC)
    {
	StringBuffer sb = new StringBuffer();
	if (DOW != 7) {
	    sb.append(weekday_names[DOW]);
	    if (day == -6) {
		day += 6;
		sb.append(" <= ");
	    } else
		sb.append(" >= ");
	} else
	    sb.append("       ");
	if (day <= 0) {
	    sb.append("last");
	    if (day < 0)
		sb.append('-').append(day);
	    sb.append(' ').append(month_names[(mon-1 + 11) % 12]);
	} else {
	    String ds = "    "+day;
	    sb.append(ds.substring(ds.length()-4))
		.append(' ').append(month_names[mon-1]);
	}
	sb.append(' ').append(printTime(time));
	if (isUTC)
	    sb.append('U');
	else
	    sb.append(' ');
	return sb.toString();
    }
    
    public String toString() 
    {
	int uid = getUniqueID();
	return (name + ":                ").substring(0,NAME_LENGTH+2)
	    + printRule(startDOW, startMonth, startDay, 
			startTime, isUTC)
	    + " "
	    + printRule(endDOW, endMonth, endDay, endTime, isUTC)
	    + " " + printTime(savings) + (uid != 0 ? "  " + uid : "");
    }
}
