/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: QuadTree.java 980 2010-04-16 02:09:05Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package net.sf.astroinfo;
import net.sf.astroinfo.pdb.*;
import java.io.*;
import java.util.*;

public class QuadTree implements AstroInfoConstants {
    private static final boolean VERBOSE = false;

    private int version;
    private int qTreeNodeSize;
    private int maxQTreeNodeSize;
    private int qTreeObjectSize;
    private boolean distance;
    static int magHigh;
    List<Node> allNodes = new ArrayList<Node>();
    Node root;
    String qname;
    int rootNodeIdx;

    Comparator<CatalogRecord> magnitudeComparator = 
	new Comparator<CatalogRecord>() {
	    public int compare(CatalogRecord r1, CatalogRecord r2) {
		int cmp;
		if (r1 == r2)
		    return 0;

		double mag1 = r1.magnitude;
		double mag2 = r2.magnitude;
		if (mag2 != mag2) {
		    cmp = mag1 != mag1 ? 0 : -1;
		}
		cmp = mag1 < mag2 ? -1 : mag1 == mag2 ? 0 : 1;
		if (cmp == 0)
		    return r1.getIndex() - r2.getIndex();
		return cmp;
	    }
	};

    public class Node extends PDBRecord {
	boolean raSplit;
	double  coord;
	double  limMagnitude;
	Node    low, high;

	Collection<CatalogRecord> records;

	private double  minRa;
	private double  maxRa;
	private double  minDec;
	private double  maxDec;
	private int index;


	public Node(double minRa, double maxRa, double minDec, double maxDec,
		    Collection<CatalogRecord> recs)
	{
	    records = recs;
	    this.minRa  = minRa;
	    this.maxRa  = maxRa;
	    this.minDec = minDec;
	    this.maxDec = maxDec;
	    this.index = allNodes.size();
	    this.limMagnitude = Double.NaN;
	    allNodes.add(this);
	}

	public int getQtreeIndex() {
	    return index;
	}

	public void split()
	{
	    /* No split necessary */
	    if (records.size() < maxQTreeNodeSize) {
		if (VERBOSE)
		    System.err.println("Build leaf: ["
				       +(minRa*180/Math.PI)+","
				       +(minDec*180/Math.PI)+"]-["
				       +(maxRa*180/Math.PI)+","
				       +(maxDec*180/Math.PI)+"]");
		return;
	    }

	    /* First extract the brightest objects */
	    List<CatalogRecord> bright = 
		new ArrayList<CatalogRecord>(maxQTreeNodeSize);
	    CatalogRecord cr = null;
	    Iterator<CatalogRecord> i = records.iterator();
	    for (int k = 0; k < qTreeNodeSize; k++) {
		cr = i.next();
		i.remove();
		bright.add(cr);
	    }

	    cr = i.next();
	    limMagnitude = cr.magnitude;
	    if(qTreeObjectSize==10) {
		if(Double.isNaN(cr.magnitude) ||
		   (magHigh != -10000 && (int)cr.magnitude != magHigh)) {
      		    throw new IllegalArgumentException("This format needs the integer part of "+
						       "all magnitudes to be the same, "+
						       "but it is not so for "+cr.magnitude);
		} else {
		    magHigh = (int)cr.magnitude;
		}
	    }
	    
	    while (cr.magnitude == limMagnitude 
		   && bright.size() < maxQTreeNodeSize) {
		i.remove();
		bright.add(cr);
		if (!i.hasNext()) {
		    records = bright;
		    if (VERBOSE)
			System.err.println("early done.");
		    return;
		}
		cr = i.next();
	    }

	    /* Now determine in which direction to split */
	    if (Math.min(Math.cos(maxDec), Math.cos(minDec))
		* (maxRa - minRa) > (maxDec - minDec)) {
		raSplit = true;
		coord = (maxRa + minRa) / 2.0;
	    } else {
		raSplit = false;
		coord = (maxDec + minDec) / 2.0;
	    }

	    if (VERBOSE)
		System.err.println("splitting " + (raSplit? "ra" : "dec")
				   + ": " +(coord*180 / Math.PI)+".");

	    Collection<CatalogRecord> lowRecs = new ArrayList<CatalogRecord>();
	    Collection<CatalogRecord> highRecs = new ArrayList<CatalogRecord>();
	    while (true) {
		if (raSplit) {
		    if (cr.ra <= coord)
			lowRecs.add(cr);
		    else
			highRecs.add(cr);
		} else {
		    if (cr.dec <= coord)
			lowRecs.add(cr);
		    else
			highRecs.add(cr);
		}
		if (!i.hasNext())
		    break;
		cr = i.next();
	    }
	    
	    /* Now set records to the bright ones and create the
	     * sub nodes.
	     */
 	    records = bright;
	    if (raSplit) {
		if (lowRecs.size() > 0)
		    low  = new Node(minRa, coord, minDec, maxDec, lowRecs);
		if (highRecs.size() > 0)
		    high = new Node(coord, maxRa, minDec, maxDec, highRecs);
	    } else {
		if (lowRecs.size() > 0)
		    low  = new Node(minRa, maxRa, minDec, coord, lowRecs);
		if (highRecs.size() > 0)
		    high = new Node(minRa, maxRa, coord, maxDec, highRecs);
	    }

	    if (low != null)
		low.split();
	    if (high != null)
		high.split();
	}

	public void read(PalmDataInputStream out) throws IOException
	{
	    throw new IOException("Not Implemented!");
	}

	private void writeInt24(PalmDataOutputStream out, int value) 
	    throws IOException
	{
	    out.writeByte((value >> 24)&0xFF);
	    out.writeByte((value >> 16)&0xFF);
	    out.writeByte((value >> 8)&0xFF);
	}

	public void write(PalmDataOutputStream out) throws IOException
	{
	    out.writeByte(raSplit ? 1 : 0);
	    out.writeByte(records.size());
	    out.writeShort(Double.isNaN(limMagnitude) ? Short.MAX_VALUE
			   : (int) (limMagnitude * 256));
	    out.writeInt((int) (long) (coord  * (1L << 31) / Math.PI));
	    out.writeShort(low == null ? 0 : rootNodeIdx + low.index);
	    out.writeShort(high == null ? 0 : rootNodeIdx + high.index);
	    for (CatalogRecord cr : records) {
		/*
		 *   X = cos(ra)*cos(dec)
		 *   Y = sin(ra)*cos(dec)
		 *   Z = sin(dec)
		 */
		double cosdec = Math.cos(cr.dec);
		int radec0 = (int) (Math.cos(cr.ra) * cosdec * (1L << 30));
		int radec1 = (int) (Math.sin(cr.ra) * cosdec * (1L << 30));
		int radec2 = (int) (Math.sin(cr.dec) * (1L << 30));

		if ( 14 <= qTreeObjectSize) {
		    out.writeInt(radec0);
		    out.writeInt(radec1);
		    out.writeInt(radec2);
		} else {
		    writeInt24(out, radec0);
		    writeInt24(out, radec1);
		    writeInt24(out, radec2);
		}
		
		if ( 11 <= qTreeObjectSize) {
		    out.writeShort(Double.isNaN(cr.magnitude) ? Short.MAX_VALUE
				   : (int) (cr.magnitude * 256));
		}
		else {
		    out.writeByte((int)(cr.magnitude * 256));
		}
		
		/* sizex and sizey are now in Fixedpoint angles shifted right
		 * by 10 bits. This way angles upto 5.6 degrees can be
		 * represented with a precision of 0.31 arcseconds.
		 */
		double sx = Double.isNaN(cr.sizex) ? 0  : cr.sizex;
		double sy = Double.isNaN(cr.sizey) ? sx : cr.sizey;
		if ( 16 <= qTreeObjectSize ) {
		    out.writeShort((int) (sx * (1L << 20) / (90*60)));
		    if ( 18 <= qTreeObjectSize ) {
          		out.writeShort((int) (sy * (1L << 20) / (90*60)));
          		if ( 20 <= qTreeObjectSize ) {
			    if (!distance) {
                  		out.writeShort(cr.type);
                  		if ( 22 <= qTreeObjectSize ) {
                  		    out.writeShort(cr.getIndex());
                                }
			    } else {
				out.writeFloat((float)cr.distance);
			    }
			}
		    }
        	}
	    }
	}


	public void dump(String prefix) {
	    System.err.println(prefix + prefix.length()/2 + ": Node["+
			       + ((int) (minRa * 18000/Math.PI)) / 100.0 + ":"
			       + ((int) (maxRa * 18000/Math.PI)) / 100.0 + ","
			       + ((int) (minDec * 18000/Math.PI)) / 100.0 + ":"
			       + ((int) (maxDec * 18000/Math.PI)) / 100.0
			       + "] " + limMagnitude
			       + " ("+records.size()+")");
	    prefix += "  ";
	    for (CatalogRecord cr : records) {
		System.err.println(prefix + cr);
	    }
	    if (low != null)
		low.dump(prefix);
	    if (high != null)
		high.dump(prefix);
	}
    }

    public QuadTree(String name, Collection<CatalogRecord> records, 
        int v, int maxNodeSize, int nodeSize, int objectSize,
        boolean doDistance) {
	version           = v;
	maxQTreeNodeSize  = maxNodeSize;
	qTreeNodeSize     = nodeSize;
	qTreeObjectSize   = objectSize;
	distance          = doDistance;
	magHigh           = -10000;

	TreeSet<CatalogRecord> sortedrecs =
	    new TreeSet<CatalogRecord>(magnitudeComparator);
	sortedrecs.addAll(records);
	root = new Node(-Math.PI, Math.PI, -Math.PI/2.0, Math.PI/2.0,
			sortedrecs);
	root.split();
	qname = "Q"+name;
    }

    public void setRootNodeIndex(int idx) {
	rootNodeIdx = idx;
    }

    public void write(OutputStream os) throws IOException {
	Database pdb = new Database();
	pdb.setName(qname);
	pdb.setVersion(version);
	pdb.setAttributes(Database.DMHDRATTRBUNDLE);
	pdb.setCreator(creatorID);
	pdb.setType(qtreeID);
	Node[] recs = (Node[]) allNodes.toArray(new Node[0]);
	pdb.setEntries(recs);
	pdb.write(os);
    }

    public void dump() {
	root.dump("");
    }
}

