/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: Astro.h 992 2010-06-22 03:39:39Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#ifndef Astro_h
#define Astro_h

#include <VFSMgr.h>

#include "PalmUtil.h"
#include "AstroDST.h"
#include "AstroRocker.h"
#include "AstroMatrix.h"

#if ERROR_CHECK_LEVEL != ERROR_CHECK_FULL
#define HostTraceInit()
#define HostTraceOutputTL(...)
#define HostTraceClose()
#endif

#define ASTROSECTION __attribute__ ((section ("code3")))

typedef struct {
    UInt16 projection         : 2;
    UInt16 showRA             : 1;
    UInt16 showHorizon        : 1;
    UInt16 showConstellation  : 1;
    UInt16 brightness         : 3;

    UInt16 displayPosText     : 1;  /* Show coords in map */
    UInt16 displayPosGraph    : 1;  /* Show position circles in map */
    UInt16 displayZoom        : 1;
    UInt16 sizePosCircles     : 2;  /* Size of position circles */
                                    /* 0(small)..2(large) */
    UInt16 noDisplayTimeout   : 1;
    UInt16 measureMode        : 1;
    UInt16 showAllNonStars    : 1;
} AstroSkyFlags;


/* Function of the pageBtn. */
#define PAGEBTNROCKER  0
#define PAGEBTNOBJECT  1
#define PAGEBTNDATE    2
#define PAGEBTNZOOM    3

#define MAXSSDETAILROWS  7
#define MAXOBJDETAILROWS  5

#define NO_OBJECT ((UInt32) -1)

/* Application Preferences
 *
 * Be careful when changing this, as the user doesn't like loosing his
 * preferences.  Best ask me <hoenicke@users.sourceforge.net> if the
 * change is okay.
 *
 * You may safely add new fields to the end, but think twice if its
 * really a preference before adding a field as this structure should
 * remain clean.
 *
 * If you change something early in the structure, user looses
 * preferences when upgrading so don't do it.
 *
 * If you change something at the end, user will loose preferences on
 * downgrading, which is not as bad.
 */

typedef enum {
    COLOR_PLANETS = 0,
    COLOR_MOON,
    COLOR_SUN,
    COLOR_LINES,
    COLOR_EXTRAS,
    COLOR_TEXT,
    COLOR_SHADOW,
    COLOR_CIRCLES,
    COLOR_CONST_NAMES
} ColorTypeType;

/* If you want to add additional color prefs,
change the following line, making sure that
it does not exceed 16 */
#define NUM_COLOR_TYPES        9
#define RESERVED_COLOR_TYPES   (16-NUM_COLOR_TYPES)

typedef enum {
    MODE_ONBLACK = 0,
    MODE_ONWHITE,
    MODE_NIGHT
} ColorModeType;
#define NUM_COLOR_MODES        3

typedef enum {
    ORIENTATION_CORRECT_IMAGE = 0,
    ORIENTATION_HORIZONTAL_FLIP,
    ORIENTATION_UPSIDE_DOWN,
    ORIENTATION_VERTICAL_FLIP
} OrientationType;

#define XFLIP ((userPrefs.optFlags.orientation==ORIENTATION_HORIZONTAL_FLIP || \
                userPrefs.optFlags.orientation==ORIENTATION_UPSIDE_DOWN) ? -1 : 1)

#define YFLIP ((userPrefs.optFlags.orientation==ORIENTATION_UPSIDE_DOWN || \
                userPrefs.optFlags.orientation==ORIENTATION_VERTICAL_FLIP) ? -1 : 1)


#define FIRST_CUSTOM_CIRCLES   6
#define NUM_CUSTOM_CIRCLES     7

typedef struct {
    UInt16              version;     /* validates the prefs structure */

    struct {
        Int16               dstRuleNr;   /* Number for current DST Rule. */
        Int32               tz;          /* timezone in minutes relative to UTC. */
        Int32               lat, lng;    /* lat/lng  */
        DSTRule             dstRule;
        Char                name[32];
    } Location[5];
    UInt8               currentLoc;      /* defaults to zero */
    UInt8               padding;

    /* If you change anything above increase version number to 1024.
     * The user will loose all his preferences when up- or downgrading.
     * You also need to adapt ASTROCOMPSIZE() below.
     */

    Int32               ra, dec;     /* last visited ra/dec */

    /* The last form we were on that we would like to remember.
     * e.g. Doesn't include option forms.
     */
    UInt16  activeFormId;

    Int32   timeInc;
    Int32   timeInt;
    Int32   timeIncRocker;

    UInt8   radec2000:1;
    UInt8   parsec:1;
    UInt8   slowScroll:1;
    UInt8   unused:3;
    UInt8   pageBtnFunction:2;      /* Specifies what hard page buttons do */
    UInt8   showDetailRows   :4;    /* Show this many rows of detail info */
    UInt8   spacingDetailRows:4;    /* How tight to layout the rows */

    AstroSkyFlags       skyFlags;

    struct {
        /* first 16 bit */
        UInt16 time24Hour      : 1;
        UInt16 moreAccurate    : 1;  /* provide more details and accuracy */

        UInt16 nightMode       : 1;
        UInt16 blackBackground : 1;

        /* These flags are for the Catalog Options form */
        UInt16 showAbove       : 1;  /* Show catalog objects above horizon */
        UInt16 showBelow       : 1;  /* Show catalog objects below horizon */
        UInt16 showNever       : 1;  /* Show catalog objects never visible */

        UInt16 timeStopped     : 1;  /* Allows clock to be frozen */
        UInt16 isNow           : 1;  /* Set time to now. */

        UInt16 fontSize        : 2;  /* A baseline for laying out the display*/
        UInt16 omitStars       : 1;  /* omit stars from catalog */
        UInt16 orientation     : 2;
        UInt16 fadeBelowHorizon : 1;  /* fade objects below horizon? */
        UInt16 showConstNames   : 1;  /* show constellation names */

        /* second 16 bit */
        UInt16 reserved2       : 3;  /* reserved default to zero */

        UInt16 reserved5       :12;  /* reserved default to one */
    } optFlags;

    Int32              jday;          /* currently selected julian day. */
    Int32              jtime;         /* currently selected time. */

    /* Remember which object we had selected when returning to detail
     * screen.
     */
    UInt32  currentObj;

    Int16   zoomValue;
    Int16   detailPage;

    UInt8   angleFormat : 4;
    UInt16  lastConstellation : 8;
    UInt16  pad         : 4;     /* Padding to get to even address */

    /* Add new fields always to the end.  Don't forget to change the
     * defaultPrefs in Astro.c.  The prefs mechanism will always fill
     * new fields with their default value so unless you changed anything
     * above you don't need to change the ASTROVERSION.
     *
     * You do not need to increase ASTROVERSION if you just changed a
     * reserved field, but make sure that the old default is useful.
     */

    RGBColorType    colors[NUM_COLOR_MODES][NUM_COLOR_TYPES+RESERVED_COLOR_TYPES];
    ColorModeType   prevColorMode; /* this is the color mode before
                                      the last nightMode */

    UInt32          circles;
    Int32           customCircles[FIRST_CUSTOM_CIRCLES];
    Char            survey[FIELDLENGTH];
    Int32           moreCustomCircles[NUM_CUSTOM_CIRCLES-FIRST_CUSTOM_CIRCLES];
    Int32           reservedForMoreCustomCircles[9];
} UserPrefsType;

#define CUSTOM_CIRCLES(i) (((i)<FIRST_CUSTOM_CIRCLES) ? userPrefs.customCircles[(i)] : \
                        userPrefs.moreCustomCircles[(i)-FIRST_CUSTOM_CIRCLES])

                        /* The following constant needs to be incremented everytime something
 * changes in the format of the userPrefs data.  This constant is written
 * to the saved preferences, allowing us to verify that they are not
 * obsolete when we start.  (see the UserPrefsType, above.)
 *
 * It should also be incremented if the number of detail pages for an
 * object DECREASES to ensure current page is reset below max.
 */

#define ASTROVERSION   157

/* This is the version number when the core options changed.  These
 * are the most important options such as location and time zone and
 * shouldn't be changed as often as the other parts.
 *
 * Everytime it changes the user has to reconfigure AstroInfo for their
 * location again, so don't change it unnecessarily.
 */
#define SIZEUPTO(type,field) ((UInt16) (&((type*)NULL)->field + 1))
#define ASTROCOMPSIZE(v) \
    v < 154        ? 0 : \
    v < 156        ? SIZEUPTO(UserPrefsType, Location[0].dstRule) : \
    v < 1024       ? SIZEUPTO(UserPrefsType, currentLoc) : 0;



typedef struct {
    VectorType  radec;
    float  distance;
    Int16  magnitude;
    UInt16 width;
    UInt16 minoraxis;
    Int16  posangle;
    Int16  objclass;
    Char   description[32];
} AstroRecordType;

typedef struct _DBRecord
{
    Char             name[32];
    UInt32           numRecords;
    UInt32           filterStart;
    UInt32           filterEnd;
    union {
        struct {
            DmOpenRef db;
            UInt16    cardNo;
        } ram;
        struct {
            Char*      pathName;
            FileRef    fileRef;
            UInt16     volRef;
        } vfs;
    } dbInfo;
    struct {
    	UInt16           numBWBlocks;
    	UInt8            *superBlock;
    	MemHandle        superBlockHandle;
    	UInt16           *records;
    	UInt16           **data;
    	MemHandle        *dataHandle;
    } bwdata;
    UInt16           catNum;
    UInt16           numIndexNodes;
    UInt16           firstBWBlock;
    UInt16           qtreeRootBlock;
    UInt16           maxQTreeNodeSize;
    UInt16           qTreeObjectSize;
    Int16            minMag;
    UInt16           entriesPerIndex;
    Boolean          vfs;
    Boolean          isVisible;
    Boolean          isStars;
    Boolean          isUnsorted;
    Boolean          pad;
    UInt8            version;
    UInt8            bytesPerEntry;
    UInt8            magHigh;
    UInt8            distance:1;
    UInt8            padding:7;
    RGBColorType     color[NUM_COLOR_MODES];
} DBRecord;



#define ASTRODBTYPE    'DATA'

/* Flags used to send events */

#define updateRedraw            0
#define updateDateChanged       1
#define updateDetailChanged     2
#define updateZoomChanged       3
#define updateObjectListChanged 4
#define updatePosTimeout        5
#define updateDetailsScrolled   6
#define updateDateChangedAutomatic  7
#define updateListItem              8

/*
 * Needed by other code sections
 */
extern UInt16 mainMenu;

/* Font sizes can change with screen resolution */
extern FontID fontSmall;
extern FontID fontNormal;
extern FontID fontBold;

extern UserPrefsType defaultPrefs;
extern UserPrefsType userPrefs;
extern SystemPreferencesType sysPrefs;

extern RectangleType skyRect, screenRect;

extern Char dstRuleName[FIELDLENGTH];
extern Char* astroInfoArgument;

/* A catalog may consist of several databases */
/* The databases in a catalog all start with the same name,
   followed by "|" and then something distinguishing them, such
   a part number, or a magnitude range */
#define MAX_DBS         128
extern DBRecord* catDBs[MAX_DBS];
extern Int16       numDBs;

void GetUserSettings(Boolean reset) ASTROSECTION;
Boolean FormChangeHandleEvent(EventPtr event) ASTROSECTION;
void SetMenu(FormPtr formPtr) ASTROSECTION;

/* Allow quick remote access to current object */
#define GetCurrentObject()    userPrefs.currentObj
#define SetCurrentObject(num) userPrefs.currentObj = num

extern UInt8 CurrentDetailRow;

/*
 * Converts hms to floating point hours.
 */
#define doubletime(h, m, s) (h + m / 60.0 + s / 3600.0)
#define inttime(h, m) (h * 60 + m)

/* IDs for controls dynamically created at runtime */
#define f_SkyGadget     9000
#define p_resizeSky     f_SkyGadget+1

#ifndef vchrRockerLeft
# define vchrRockerLeft 0x134
#endif

#ifndef vchrRockerRight
# define vchrRockerRight 0x135
#endif

#ifndef vchrRockerCenter
# define vchrRockerCenter 0x136
#endif

#define gotoRaDecEvent firstUserEvent
#define gotoAlignEvent (firstUserEvent+1)
#define gotoConstellationEvent (firstUserEvent+2)
#define listAddEvent (firstUserEvent+3)
#define catalogSelectEvent (firstUserEvent+4)
#define dssLookupEvent (firstUserEvent+5)
#define zoomOutEvent (firstUserEvent+6)
#define chooseCirclesEvent (firstUserEvent+7)
#define findConstellationEvent (firstUserEvent+8)

#endif

