/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroAlign.c 990 2010-06-10 02:11:39Z arpruss $
 * Copyright (C) 2000, Michael Heinz
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "Astro.h"
#include "AstroSky.h"
#include "AstroCat.h"
#include "resource.h"
#include "PalmUtil.h"
#include "AstroAlign.h"
#include "AstroDetails.h"

/* cannot exceed 127 */
#define MAX_ALIGNMENT_OBJECTS  32

#ifndef M_PI
# define M_PI 3.141592653589793
#endif

//#define DEG(x) ( 180. / M_PI * (x) )
//#define H(x) ( 12. / M_PI * (x) )

typedef struct {
    Int32 alt;
    Int32 az;
} Angles32;

#define OBJECT_MODE_NONE   0
#define OBJECT_MODE_ALTAZ  1
#define OBJECT_MODE_EQ     2

typedef struct {
    Char              name[FIELDLENGTH];
    Angles32          apparentTelescopeAngles;
    VectorType        skyPosition;
    UInt8             mode;
    UInt8             padding[3];
} AlignmentObjectType;

typedef struct {
    UInt8     eqMode:1;
    UInt8     reserved1:7;
    UInt8     numObjects;
    UInt8     reserved2[2];
    AlignmentObjectType objects[0];
} AlignmentDataType;

static VectorType   currentAltAz;
static Char         currentName[FIELDLENGTH];
static Char         estAlt[32];
static Char         estAz[32];

static MatrixType T;
static Boolean   validT = false;
static Boolean   validAltDelta = false;
static AlignmentDataType* align = NULL;
static Int16 lastItem = -1;
// The following compensates in case alt circle 0 is not
// parallel to az circle.
static Int32 altDelta = 0;

ALIGNSECTION static
Boolean MakeMatrix(MatrixType m, VectorType col1, VectorType col2)
{
    Int32  f;
    Int32 a[3];
    UInt16 i;

    /* Divide by two to prevent possible overflows */
    a[0] = fpmul(col1[1]/2, col2[2]) - fpmul(col1[2]/2, col2[1]);
    a[1] = fpmul(col1[2]/2, col2[0]) - fpmul(col1[0]/2, col2[2]);
    a[2] = fpmul(col1[0]/2, col2[1]) - fpmul(col1[1]/2, col2[0]);

    f = fpsq(a[0]) + fpsq(a[1]) + fpsq(a[2]);

    if (f <= 0)
        return false;

    f = fpsqrt(f);

    for (i=0 ; i<3; i++) {
        m[i][0] = col1[i];
        m[i][1] = col2[i];
        m[i][2] = fpdiv(a[i], f);
    }

    return true;
}

ALIGNSECTION static
void GetVectorDouble(VectorType c, Angles32* a, Int32 altDelta )
{
    Int32 sinAlt, cosAlt, sinAz, cosAz;

    fpsincos(a->alt - altDelta, &sinAlt, &cosAlt);
    fpsincos(a->az, &sinAz, &cosAz);

    c[0] = fpmul(cosAlt, sinAz);
    c[1] = fpmul(cosAlt, cosAz);
    c[2] = sinAlt;
}


/* This does not require normalization */
ALIGNSECTION static
void GetAngles( Angles32* a, VectorType c, Int32 altDelta )
{
    a->az = fpatan2(c[0], c[1]);
    a->alt = fpatan2(c[2], fpsqrt(fpsq(c[0])+fpsq(c[1]))) + altDelta;
}

// Since all we care about is the angles, we don't need to
// divide by the determinant.

ALIGNSECTION
static void InvertMatrixDontNormalize(MatrixType I, MatrixType A)
{
#define a11 A[0][0]
#define a12 A[0][1]
#define a13 A[0][2]
#define a21 A[1][0]
#define a22 A[1][1]
#define a23 A[1][2]
#define a31 A[2][0]
#define a32 A[2][1]
#define a33 A[2][2]
    I[0][0] = fpmul(a33/6,a22)-fpmul(a32/6,a23);
    I[0][1] = fpmul(a32/6,a13)-fpmul(a33/6,a12);
    I[0][2] = fpmul(a23/6,a12)-fpmul(a22/6,a13);

    I[1][0] = fpmul(a31/6,a23)-fpmul(a33/6,a21);
    I[1][1] = fpmul(a33/6,a11)-fpmul(a31/6,a13);
    I[1][2] = fpmul(a21/6,a13)-fpmul(a23/6,a11);

    I[2][0] = fpmul(a32/6,a21)-fpmul(a31/6,a22);
    I[2][1] = fpmul(a31/6,a12)-fpmul(a32/6,a11);
    I[2][2] = fpmul(a22/6,a11)-fpmul(a21/6,a12);

#undef a11
#undef a12
#undef a13
#undef a21
#undef a22
#undef a23
#undef a31
#undef a32
#undef a33
}



ALIGNSECTION
static UInt16 GetMode(void)
{
    return (align->eqMode) ? OBJECT_MODE_EQ : OBJECT_MODE_ALTAZ;
}



ALIGNSECTION
static UInt16 CountAlignmentObjects(void)
{
    UInt16 i;
    UInt16 count = 0;
    UInt8  mode = GetMode();

    for(i=0; i<align->numObjects; i++)
        if (align->objects[i].mode == mode)
            count++;

    return count;
}


ALIGNSECTION
static Boolean Transform1(Angles32* atel1, VectorType star1,
                          Angles32* out, VectorType in)
{
    Angles32 anglesStar1;

    GetAngles(&anglesStar1, star1, 0);
    GetAngles(out, in, 0);
    out->az  += atel1->az  - anglesStar1.az;
    out->alt += atel1->alt - anglesStar1.alt;

    if (fpabs(out->alt) > FP_ONE) {
        out->alt = -2*FP_ONE - out->alt;
        out->az += -2*FP_ONE;
    }
    return true;
}


ALIGNSECTION
static void ResetValidity(void)
{
    validAltDelta = false;
    validT     = false;
}


ALIGNSECTION
static Boolean Transform2(Angles32* atel1, VectorType star1,
                          Angles32* atel2, VectorType star2,
                          Angles32* atel3, VectorType tel3, VectorType star3,
                          Boolean useCaching)
{
    if (!validT || !useCaching) {
        VectorType tel1;
        VectorType tel2;
        MatrixType M1;
        MatrixType M2;
        MatrixType InvM2;

        GetVectorDouble(tel1, atel1, altDelta);
        GetVectorDouble(tel2, atel2, altDelta);

        if (!MakeMatrix(M1, tel1, tel2) ||
            !MakeMatrix(M2, star1, star2))
            return false;

        InvertMatrixDontNormalize(InvM2, M2);
        MatMatMul(T, M1, InvM2);
        validT = true;
    }

    MatVecMul(tel3, T, star3);

    if (atel3 != NULL)
        GetAngles(atel3, tel3, altDelta);

    return true;
}


ALIGNSECTION static
AlignmentObjectType* GetAlignmentObject(UInt16 i)
{
   UInt16 j;
   UInt8  mode = GetMode();

   for (j=0; j<align->numObjects; j++) {
        if (align->objects[j].mode == mode) {
            if (i==0)
                return &(align->objects[j]);
            i--;
        }
   }

   return NULL;
}


ALIGNSECTION static
UInt32 GetWeight(VectorType current, VectorType target)
{
    Int32 dot;

    dot = fpmul(current[0]/2, target[0])
        + fpmul(current[1]/2, target[1])
        + fpmul(current[2]/2, target[2]);

    if (dot < -FP_ONE) {
        dot = -FP_ONE;
    }
    else if (1 < dot) {
        dot = FP_ONE;
    }

    return ((UInt32)2*FP_ONE - fpatan2(fpsqrt(1-fpsq(dot)), dot)) / 2;
}



ALIGNSECTION static
Int32 HalfDistance(VectorType v1, VectorType v2)
{
    return fpsq(v1[0]/2-v2[0]/2)+fpsq(v1[1]/2-v2[1]/2)+fpsq(v1[2]/2-v2[2]/2);
}


ALIGNSECTION static
Int32 GetPairSeparationError(AlignmentObjectType* o1,
                             AlignmentObjectType* o2, Int32 delta)
{
    VectorType v1;
    VectorType v2;

    GetVectorDouble(v1, &(o1->apparentTelescopeAngles), delta);
    GetVectorDouble(v2, &(o2->apparentTelescopeAngles), delta);

    return fpabs(HalfDistance(v1, v2) - HalfDistance(o1->skyPosition, o2->skyPosition));
}


ALIGNSECTION static
Int32 GetSeparationError(Int32 delta, UInt8* furthest)
{
    UInt16 count;
    UInt16 i;
    Int32 error;

    error = 0;

    count = CountAlignmentObjects();

    for (i=0; i<count; i++) {
        if (furthest[i] != i)
            error += fpdiv(GetPairSeparationError(GetAlignmentObject(i),
                GetAlignmentObject(furthest[i]), delta), count);
    }

    return error;
}


ALIGNSECTION static
void CalculateAltDelta(void)
{
    Int32 bestError;
    Int32 delta;
    Int32 lastDelta;
    UInt8* furthest;
    UInt16 count;
    UInt16 i;

    if (validAltDelta)
        return;

    validAltDelta = true;
    altDelta      = 0;
return;

    count = CountAlignmentObjects();
    if (count < 2)
        return;

    furthest = MemPtrNew(sizeof(UInt8)*count);

    if (furthest==NULL)
        return;

    for(i=0; i<count; i++) {
        UInt16 j;
        AlignmentObjectType* o1;
        UInt16 bestJ;
        Int32 bestCos;

        bestJ   = i;
        bestCos = 1;

        o1 = GetAlignmentObject(i);

        for (j=0; j<count; j++) {
            Int32 cosDistance;

            if (j==i)
                continue;

            cosDistance = fpcos(o1->apparentTelescopeAngles.az -
                            GetAlignmentObject(j)->apparentTelescopeAngles.az);

            if (cosDistance < bestCos) {
                bestCos = cosDistance;
                bestJ   = j;
            }
        }

        if (bestJ < i && furthest[bestJ] == i)
            furthest[i] = i;
        else
            furthest[i] = bestJ;
    }

    bestError = GetSeparationError(altDelta, furthest);

    for (delta = -FP_DMS(15,0,0); delta <= FP_DMS(15,0,0); delta += FP_DMS(3,0,0)) {
         Int32 error;

         error = GetSeparationError(delta, furthest);

         if (error < bestError) {
             altDelta = delta;
             bestError = error;
         }
    }

    lastDelta = altDelta;
    for (delta = lastDelta - FP_DMS(3,0,0); delta <= lastDelta + FP_DMS(3,0,0); delta += FP_DMS(0,30,0)) {
         Int32 error;

         error = GetSeparationError(delta, furthest);

         if (error < bestError) {
             altDelta = delta;
             bestError = error;
         }
    }

    lastDelta = altDelta;
    for (delta = lastDelta - FP_DMS(0,30,0); delta <= lastDelta + FP_DMS(0,30,0); delta += FP_DMS(0,3,0)) {
         Int32 error;

         error = GetSeparationError(delta, furthest);

         if (error < bestError) {
             altDelta = delta;
             bestError = error;
         }
    }

    MemPtrFree(furthest);
}


ALIGNSECTION static
Boolean Transform(Angles32* out, VectorType in)
{
    UInt16 count;
    UInt16 i;
    VectorType transformed;
    VectorType transformedAvg;
    UInt32     totalWeight;

    count = CountAlignmentObjects();

    CalculateAltDelta();

    if (count==0)  {
        return false;
    }
    else if (count==1) {
        if (align->eqMode) {
            return false;
        }
        else {
            AlignmentObjectType* o;
            o = GetAlignmentObject(0);
            return Transform1(&(o->apparentTelescopeAngles),o->skyPosition,out,in);
        }
    }
    else if (count==2) {
        AlignmentObjectType* o1;
        AlignmentObjectType* o2;

        o1 = GetAlignmentObject(0);
        o2 = GetAlignmentObject(1);
        return Transform2(&(o1->apparentTelescopeAngles),
                    (o1->skyPosition),
                    &(o2->apparentTelescopeAngles),
                    (o2->skyPosition),out,transformed,in,true);
    }

    /* Now we do a weighted average for more than two alignment objects */

    transformedAvg[0] = 0;
    transformedAvg[1] = 0;
    transformedAvg[2] = 0;

    totalWeight = 0;
    
    for(i=0; i<count; i++) {
        AlignmentObjectType* o1;
        AlignmentObjectType* o2;
        UInt16               j;
        Int32                bestDot;

        o1 = GetAlignmentObject(i);
        o2 = NULL;

        /* Find the best object to pair o1 with.  Heuristic: The best one is
           the one which is closest to orthogonality with o1. */
        bestDot = FP_ONE;

        for (j=0; j<count; j++) {
            Int32      absDot;

            if (i==j)
                continue;

            o2 = GetAlignmentObject(j);

            absDot = fpabs(fpmul(o2->skyPosition[0], o1->skyPosition[0]) +
                           fpmul(o2->skyPosition[1], o1->skyPosition[1]) +
                           fpmul(o2->skyPosition[2], o1->skyPosition[2]));

            if (absDot < bestDot) {
                bestDot = absDot;
            }
        }

        if (o2 != NULL && Transform2(&(o1->apparentTelescopeAngles),
                              (o1->skyPosition),
                              &(o2->apparentTelescopeAngles),
                              (o2->skyPosition),NULL,transformed,in,false)) {
            UInt32 weight;

            weight = GetWeight(o1->skyPosition, in) / count;

            transformedAvg[0] += fpmul(transformed[0], weight);
            transformedAvg[1] += fpmul(transformed[1], weight);
            transformedAvg[2] += fpmul(transformed[2], weight);

            totalWeight += weight;
        }
    }

    if (totalWeight == 0)
        return false;

    transformedAvg[0] = fpdiv(transformedAvg[0], totalWeight);
    transformedAvg[1] = fpdiv(transformedAvg[1], totalWeight);
    transformedAvg[2] = fpdiv(transformedAvg[2], totalWeight);

    GetAngles(out, transformedAvg, altDelta);

    return true;
}


static ALIGNSECTION
void DrawObjectName(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    AlignmentObjectType* o =
        GetAlignmentObject(itemNum);

    if (o!=NULL)
        WinDrawChars(o->name,StrLen(o->name),bounds->topLeft.x,bounds->topLeft.y);
}


static ALIGNSECTION
void AlignDraw(void)
{
    FormPtr formPtr = FrmGetActiveForm();
    char format[FIELDLENGTH];
    Angles32 currentEstimate;
    ListType*    list;
    UInt8 mode;
    UInt16 count;

    if (align->eqMode) {
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_EstAlt));
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_EstAz));
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_ActAlt));
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_ActAz));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_EstRA));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_EstDec));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_ActRA));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_ActDec));
    }
    else {
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_EstRA));
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_EstDec));
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_ActRA));
        FrmHideObject(formPtr, FrmGetObjectIndex(formPtr, l_ActDec));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_EstAlt));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_EstAz));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_ActAlt));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, l_ActAz));
    }
    
    CtlSetValue(FrmGetObjectPtrByID(formPtr, c_RaDecMode), align->eqMode);

    if (Transform(&currentEstimate, currentAltAz)) {
	SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			 userPrefs.angleFormat, format, FIELDLENGTH);
        FormatAngle(estAlt, currentEstimate.alt, format, true);
        if (align->eqMode)
            SysStringByIndex(stbl_AngleFormatsHMS_sprintf,
                             userPrefs.angleFormat, format, FIELDLENGTH);
        FormatAngle(estAz, currentEstimate.az, format, false);
    }
    else {
        estAlt[0] = 0;
        estAz[0]  = 0;
    }

    FldSetTextPtr(FrmGetObjectPtrByID(formPtr, f_CurrentName), currentName);
    FldSetTextPtr(FrmGetObjectPtrByID(formPtr, f_EstAz), estAz);
    FldSetTextPtr(FrmGetObjectPtrByID(formPtr, f_EstAlt), estAlt);

    mode = align->eqMode ? OBJECT_MODE_EQ : OBJECT_MODE_ALTAZ;

    count = CountAlignmentObjects();

    list = FrmGetObjectPtrByID(formPtr, l_Objects);
    LstSetListChoices(list, NULL, count);
    LstSetDrawFunction(list, DrawObjectName);
    
    if (count == 0)
        lastItem = -1;
    else if ((Int16)count <= lastItem)
        lastItem = count - 1;

    LstSetSelection(list, lastItem);

    if (0 <= lastItem)
        LstMakeItemVisible(list, lastItem);
    else if (0<count)
        LstMakeItemVisible(list, count-1);

    FrmDrawForm(formPtr);
}


ALIGNSECTION static
void ShowAlignAdjustment(void)
{
    Char format[FIELDLENGTH];
    Char altDeltaString[32];

    CalculateAltDelta();
    SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
                     userPrefs.angleFormat, format, FIELDLENGTH);
    FormatAngle(altDeltaString, altDelta, format, true);
    FrmCustomAlert(a_AltAdjustment, altDeltaString, "", "");
}


ALIGNSECTION static
Int32 Wrap32(double x)
{
    while (2.*FP_ONE <= x)
        x -= 4.*FP_ONE;
    while (x < -2.*FP_ONE)
        x += 4.*FP_ONE;
    return (Int32)x;
}


ALIGNSECTION static
void AlignAddObjectData(void)
{
    Char* azStr;
    Char* altStr;
    FlpCompDouble az;
    FlpCompDouble alt;
    AlignmentDataType* align2;
    UInt16             index;
    FormPtr formPtr = FrmGetActiveForm();
    
    if (MAX_ALIGNMENT_OBJECTS <= CountAlignmentObjects()) {
        FrmAlert(a_TooManyAlignmentObjects);
        return;
    }

    ResetValidity();

    azStr  = FldGetTextPtr(FrmGetObjectPtrByID(formPtr, f_ActAz));

    if (azStr == NULL || (azStr[0] != '-' && azStr[0] != '.' &&
        ( azStr[0] < '0' || azStr[0] > '9') ) )
        return;

    altStr = FldGetTextPtr(FrmGetObjectPtrByID(formPtr, f_ActAlt));

    if (altStr == NULL || (altStr[0] != '-' && altStr[0] != '.' &&
        ( altStr[0] < '0' || altStr[0] > '9') ) )
        return;

    FlpBufferAToF(&az.fd,azStr);
    FlpBufferAToF(&alt.fd,altStr);

    index = align->numObjects;

    align2 = MemPtrNew(sizeof(AlignmentDataType) +
                 (index+1)*sizeof(AlignmentObjectType));

    if(align2 == NULL)
        return;

    MemMove(align2, align, sizeof(AlignmentDataType) +
                 index*sizeof(AlignmentObjectType));
    MemPtrFree(align);
    align = align2;

    align->objects[index].apparentTelescopeAngles.az  =
        Wrap32(align->eqMode ? ( az.d * (FP_ONE/6.) ) : ( az.d * (FP_ONE/90.) ));
    align->objects[index].apparentTelescopeAngles.alt = Wrap32(alt.d * (FP_ONE/90.0));
    MemMove(align->objects[index].skyPosition, currentAltAz, sizeof(currentAltAz));
    StrNCopy(align->objects[index].name,currentName,FIELDLENGTH);
    align->objects[index].mode = align->eqMode ? OBJECT_MODE_EQ : OBJECT_MODE_ALTAZ;

    align->numObjects++;

    lastItem = index;

    AlignDraw();
}


ALIGNSECTION static
void DeleteAlignmentObject(Int16 index)
{
    UInt8  mode;
    UInt16 i;

    if (index<0)
        return;

    mode = GetMode();

    for (i=0; i<align->numObjects; i++) {
        if(align->objects[i].mode == mode) {
            if(index == 0) {
                ResetValidity();
                if (i+1<align->numObjects) {
                    MemMove(&(align->objects[i]),&(align->objects[i+1]),
                        sizeof(AlignmentObjectType)*(align->numObjects-(i+1)));
                }
                align->numObjects--;
                lastItem = -1;
                return;
            }
            else  {
                index--;
            }
        }
    }
}


ALIGNSECTION
Boolean AlignFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr formPtr = FrmGetActiveForm();

    switch (event->eType) {
        case menuEvent:
            switch (event->data.menu.itemID) {
                case mi_AlignHelp:
                    FrmHelp(s_AlignHelp);
                    break;

                case mi_AdjAlignment:
                    ShowAlignAdjustment();
                    break;
            }

        case frmOpenEvent:
            lastItem = -1;
            AlignDraw();
            handled = true;
            break;
        case frmCloseEvent:
            break;
        case ctlSelectEvent:
            switch (event->data.ctlSelect.controlID) {
                case c_RaDecMode:
                    align->eqMode = CtlGetValue(
                        FrmGetObjectPtrByID(formPtr, c_RaDecMode));
                    ResetValidity();
                    lastItem = -1;
                    AlignDraw();
                    handled = true;
                    break;
                case b_OK:
                    FrmReturnToForm(0);
                    handled = true;
                    break;
                case b_DeleteAll: {
                    UInt16 count;

                    count = CountAlignmentObjects();

                    while(count--) {
                        DeleteAlignmentObject(0);
                    }
                    AlignDraw();
                    handled = true;
                    break;
                }
                case b_Delete:
                    DeleteAlignmentObject(LstGetSelection(
                        FrmGetObjectPtrByID(formPtr, l_Objects)));
                    AlignDraw();
                    handled = true;
                    break;
                case b_AddObject:
                    AlignAddObjectData();
                    handled = true;
                    break;
                default:
                    break;
        }
        default:
            break;
    }

    return handled;
}


ALIGNSECTION
void AlignStart (void)
{
    UInt16 size;

    size = 0;
    if (noPreferenceFound == PrefGetAppPreferences(APPID, 'Al', NULL, &size, true)) {
        align = MemPtrNew(sizeof(AlignmentDataType));
        if (align == NULL)
            return;
        MemSet(align, sizeof(AlignmentDataType), 0);
        return;
    }

    align = MemPtrNew(size);
    if (align == NULL)
        return;

    MemSet(align, size, 0);
    PrefGetAppPreferences(APPID, 'Al',align,&size,true);
}


ALIGNSECTION
void AlignClose (void)
{
    if (align != NULL) {
        PrefSetAppPreferences(APPID, 'Al', 1, align, 
            sizeof(*align)+sizeof(AlignmentObjectType)*align->numObjects, true);
        MemPtrFree(align);
        align = NULL;
    }
}


ALIGNSECTION void Align(UInt32 currentObj)
{
    MatrixType rotate;
    
    if (align == NULL)
        return;

    if (currentObj == NO_OBJECT) {
        FrmHelp(s_AlignHelp);
        return;
    }
    
    lastItem = -1;

    SkyGetRaDec2AltAz(rotate);

    if (currentObj < NUMPLANETS) {
        PlanetInfoType p;
        SkyGetPlanetInfo(&p, currentObj);
        MatVecMul(currentAltAz, rotate, p.pos);
        SysStringByIndex(stbl_SolarObjects, currentObj, currentName, FIELDLENGTH);
    } else {
        AstroRecordType art;
        if (!GetCatalogObject(currentObj - NUMPLANETS, &art, NULL))
            return;
        MatVecMul(currentAltAz, rotate, art.radec);
        StrNCopy(currentName, art.description, FIELDLENGTH);
    }

    FrmPopupForm(f_Align);
}



ALIGNSECTION Boolean AlignValid(void)
{
    UInt16 c;
    
    if (align == NULL)
        return false;

    c = CountAlignmentObjects();
    
    return 2 <= c || (1 == c && !align->eqMode);
}



ALIGNSECTION
Boolean AlignGetStrings(UInt16* axis1LabelP, Char* axis1,
    UInt16* axis2LabelP, Char* axis2, VectorType altAz)
{
    Angles32 adjusted;
    char format[FIELDLENGTH];

    if (align==NULL)
        return false;

    if (!Transform(&adjusted, altAz))
        return false;

    if (align->eqMode) {
	SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			 userPrefs.angleFormat, format, FIELDLENGTH);
        FormatAngle(axis1, adjusted.az, format, true);
        SysStringByIndex(stbl_AngleFormatsHMS_sprintf,
                         userPrefs.angleFormat, format, FIELDLENGTH);
        FormatAngle(axis2, adjusted.alt, format, false);
        *axis1LabelP = s_AdjRa;
        *axis2LabelP = s_AdjDec;
    }
    else {
	SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			 userPrefs.angleFormat, format, FIELDLENGTH);
        FormatAngle(axis1, adjusted.alt, format, true);
        FormatAngle(axis2, adjusted.az, format, false);
        *axis1LabelP = s_AdjAlt;
        *axis2LabelP = s_AdjAz;
    }

    return true;
}
