/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroBW.c 970 2010-04-12 22:19:21Z hoenicke $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <PalmOS.h>
#include "PalmUtil.h"
#include "resource.h"
#include "Astro.h"
#include "AstroCat.h"
#include "AstroBW.h"
#include "AstroQTree.h"

/*
 * The catalog names are compressed by the Burrows-Wheeler algorithm,
 * which also enables fast searching.  This descriptions assumes that
 * you're familiar with Burrows-Wheeler.  However we use small
 * variations marked with (!) in AstroInfo.
 *
 * Data Format:
 *
 * First all catalog strings are concatenated and reversed (!),
 * separated by NUL characters.  Then block sorting is applied,
 * however the NUL characters are never equal but compare according to
 * the record index they belong to (!).  NUL characters are still
 * smaller than any other character.
 *
 * The resulting sequence is first separated in subblocks (256 bytes).
 * On each block the standard last-n encoding is used: 0 encodes the
 * last character encountered, 1 the character before and so on.  To
 * initialize the table, there is a list of the characters in the
 * order they appear in the sub-block together with their frequency.
 *
 * A database record encodes a block consisting of several (currently
 * 16) subblocks. Its encoding is as follows:
 *
 * 16 unsigned shorts: start of subblock in current block.
 *
 * frequency table:  contains the number of occurrences of each character
 * *before* this block.  The layout of the frequency table is given by
 * the super block.
 *
 * subblocks:  Encoding of a subblock is as follows:
 * n bytes: frequency of each character in the subblock - 1.
 * 1 byte : 0xff marker
 * n bytes: characters in reverse order of appearance.
 * -------  <- this is where offset points to
 * compressed subblock.
 *
 * The compression is nibble-based (a nibble is 4 bits).  The codes are:
 * 0 - D: encodes index into last-n table.
 *     E: next two nibbles encode index.
 *     F: next two nibbles encode rep.  The index 0 is repeated rep+4 times.
 *
 */

#define BLOCKSIZE 2048
#define SUBBLOCKSIZE 256

static BWSECTION UInt32
BWReadBits(UInt16* data, UInt16 offset, UInt16 end)
{
    UInt32 result = 0;
    UInt16 bits;
    Int16 nb;

    data += (offset >> 4); 
    nb = 16 - (offset & 15);
    bits = (*data++) & ((1 << nb) - 1);
    end -= offset;
    while (end > nb) {
	result += (UInt32) bits << (end - nb);
	end -= nb;
	bits = *data++;
	nb = 16;
    }
    result += bits >> (nb - end);
    return result;
}

static BWSECTION Int16
BWDecodeValues(UInt8* data, UInt16** values, Int16 offset)
{
    Int16  pos;
    UInt8  raw;
    UInt16 value;
    UInt16 *valptr;
	pos = 0;
	valptr = *values;
	while (pos <= offset) {
	    raw = *data++;
	    /* data in high nibble */
	    if (!((raw ^ 0xf0) & 0xf0)) {
		/* repeated zeros */
		value = (raw & 0xf) << 4;
		raw = *data++;
		value += (raw >> 4) + 4;
		pos += value;
		value |= 0x8000;
	    } else if (!((raw ^ 0xe0) & 0xf0)) {
		/* large value */
		value = (raw & 0xf) << 4;
		raw = *data++;
		value |= raw >> 4;
		pos++;
	    } else {
		value = raw >> 4;
		pos++;
	    }
	    *valptr++ = value;
	    if (pos > offset)
		break;

	    /* data in low nibble */
	    if (!((raw ^ 0xf) & 0xf)) {
		/* repeated zeros */
		value = (UInt16) *data++ + 4;
		pos += value;
		value |= 0x8000;
	    } else if (!((raw ^ 0xe) & 0xf)) {
		/* large value */
		value = *data++;
		pos++;
	    } else {
		value = raw & 0xf;
		pos++;
	    }
	    *valptr++ = value;
	}
	*values = valptr;
	return pos;
}

static BWSECTION Int16
BWGetCount(UInt8* data, UInt8 c)
{
	    UInt8 found = 0, idx = 0;
	    while (*--data != 0xff) {
		idx++;
		if ((*data & 0x7f) == c)
		    found = idx;
	    }
	    if (found) {
		return (Int16) *(data - found) + 1;
	    }
	    return 0;
}

				    
static BWSECTION Int32
BWFollowPosition(DBRecord *db, UInt32 index, Char *dest)
{
    UInt16 rec, subblock;
    Int16  offset;
    UInt16 *suboffsets;
    UInt8  c;
    UInt8  *data;
    UInt16 value;
    UInt32 *freqs;
    UInt8  *order;
    UInt16 *freqoffset;
    Int16    ordlen, datastart;
    Int16    upcase;
    UInt16 values[SUBBLOCKSIZE];
    UInt16 *valptr;

    ordlen = *db->bwdata.superBlock;
    order  = db->bwdata.superBlock + 1;
    freqs      = (UInt32*) (order + (ordlen|1));
    freqoffset = (UInt16*) (freqs + ordlen + 1);
    datastart  = (freqoffset[ordlen]+15) >> 4;

    do {
	if (!dest && index >= db->filterStart && index < db->filterEnd)
	    return index - db->filterStart;

	rec      = index / BLOCKSIZE;
	subblock = (index % BLOCKSIZE) / SUBBLOCKSIZE;
	offset   = index % SUBBLOCKSIZE;
	suboffsets = db->bwdata.data[rec];
	data = (UInt8*) suboffsets + suboffsets[subblock];

	valptr = values;
	index = offset - BWDecodeValues(data, &valptr, offset);
	upcase = 0;
	c = 0;
	while (valptr > values) {
	    value = *--valptr;
	    if (value & 0x8000) {
		/* zeros encoded */
		index += (value & 0x1ff);
	    } else {
		upcase ^= value;
		c = (value & 0x7f);
		index++;
		if (c) {
		    while (valptr > values) {
			value = *--valptr;
			if (!(value & 0x8000)
			    && (value & 0x7f) >= c) {
			    if (!--c) {
				break;
			    }
			}
		    }
		}
	    }
	}
	c    = *(data - c - 1);
	upcase ^= c;
	c &= 0x7f;
	if (dest)
	    *dest++ = upcase & 0x80 ? order[c] ^ 0x20 : order[c];

	while (subblock-- > 0) {
	    index += BWGetCount((UInt8*) suboffsets + suboffsets[subblock], c);
	}
	index += freqs[c] +
	    BWReadBits(suboffsets + BLOCKSIZE/SUBBLOCKSIZE,
		       freqoffset[c], freqoffset[c+1]);
    } while (order[c] != 0);
    return -1;
}

void BWGetString(DBRecord *db, UInt32 index, Char *dest) {
    BWFollowPosition(db, index, dest);
}

static BWSECTION UInt32 
BWSearchOneStep(DBRecord *db, UInt32 *freqs, UInt16 *freqoffset,
		UInt8 c, UInt32 index) {
    UInt16 rec;
    Int16  pos, subblock, offset;
    UInt16 *suboffsets;
    UInt8  *data;
    UInt8  raw;
    UInt16 value;
    Int16    i;

    rec      = index / BLOCKSIZE;
    subblock = (index % BLOCKSIZE) / SUBBLOCKSIZE;
    offset   = index % SUBBLOCKSIZE;
    suboffsets = db->bwdata.data[rec];
    
    index = freqs[c] + BWReadBits(suboffsets + BLOCKSIZE/SUBBLOCKSIZE,
				  freqoffset[c], freqoffset[c+1]);
    for (i = 0; i < subblock; i++) {
	index += BWGetCount((UInt8*) suboffsets + suboffsets[i], c);
    }
    
    data = (UInt8*) suboffsets + suboffsets[subblock];
    i = 0;
    while ((data[--i] & 0x7f) != c) {
	if (data[i] == 0xff) {
	    /* found end marker */
	    return index;
	}
    }
    c = -i - 1;
    pos = 0;
    while (pos < offset) {
	raw = *data++;
	/* data in high nibble */
	if (!((raw ^ 0xf0) & 0xf0)) {
	    /* repeated zeros */
	    value = (raw & 0xf) << 4;
	    raw = *data++;
	    value += (raw >> 4) + 4;
	    pos += value;
	    if (c == 0)
		index += value;
	} else {
	    if (!((raw ^ 0xe0) & 0xf0)) {
		/* large value */
		value = (raw & 0x7) << 4;
		raw = *data++;
		value |= raw >> 4;
	    } else {
		value = raw >> 4;
	    }
	    pos++;
	    if (c == value) {
		c = 0;
		index++;
	    } else if (value > c)
		c++;
	}
	if (pos >= offset)
	    break;
	
	/* data in low nibble */
	if (!((raw ^ 0xf) & 0xf)) {
	    /* repeated zeros */
	    value = *data++ + 4;
	    pos += value;
	    if (c == 0)
		index += value;
	} else {
	    if (!((raw ^ 0xe) & 0xf)) {
		/* large value */
		value = (*data++) & 0x7f;
	    } else {
		value = raw & 0xf;
	    }
	    pos++;
	    if (c == value) {
		c = 0;
		index++;
	    } else if (value > c)
		c++;
	}
    }
    if (c == 0)
	index -= pos - offset;

    return index;
}


static BWSECTION UInt32
BWGetRecord(DBRecord *db, UInt32 index)
{
    Int16  offset;
    UInt16 *suboffsets;
    UInt8  c;
    UInt8  *data;
    UInt16 value;
    UInt8  raw;
    Int16    start, mid, end;
    UInt32 *freqs;
    UInt16 *freqoffset;
    UInt8  *order;
    Int16    ordlen;

    order  = db->bwdata.superBlock;
    ordlen = *order++;
    freqs      = (UInt32*) (order + (ordlen|1));
    freqoffset = (UInt16*) (freqs + ordlen + 1);

    while (index >= freqs[1]) {
	/* Reverse step: Find the predecessor of index */

	/* Search of character */
	start = 1;
	end   = ordlen - 1;
	while (end > start) {
	    mid = (start+end+1) / 2;
	    if (freqs[mid] > index)
		end = mid - 1;
	    else
		start = mid;
	}
	c = start;
	index -= freqs[c];

	/* Search of record */
	start = 0;
	end   = db->bwdata.numBWBlocks - 1;
	while (end > start) {
	    mid = (start+end+1) / 2;
	    if (BWReadBits(db->bwdata.data[mid] + BLOCKSIZE/SUBBLOCKSIZE,
			   freqoffset[c], freqoffset[c+1]) > index)
		end = mid - 1;
	    else
		start = mid;
	}
	suboffsets = db->bwdata.data[start];
	offset = index - BWReadBits(suboffsets + BLOCKSIZE/SUBBLOCKSIZE,
				    freqoffset[c], freqoffset[c+1]);
	index = (UInt32)BLOCKSIZE * start;

	/* search subblock */
	for (start = 0; start < BLOCKSIZE/SUBBLOCKSIZE - 1; start++) {
	    Int16 count = BWGetCount((UInt8*) suboffsets + suboffsets[start], c);
	    if (count > offset)
		break;
	    offset -= count;
	    index += SUBBLOCKSIZE;
	}
	data = (UInt8*) suboffsets + suboffsets[start];

	/* search character index */
	start = 0;
	while ((data[--start] & 0x7f) != c)
	    ;
	c = -start - 1;

	for (;;) {
	    raw = *data++;
	    /* data in high nibble */
	    if (!((raw ^ 0xf0) & 0xf0)) {
		/* repeated zeros */
		value = (raw & 0xf) << 4;
		raw = *data++;
		value += (raw >> 4) + 4;
		index += value;
		if (c == 0) {
		    offset -= value;
		    if (offset < 0) {
			index += offset;
			break;
		    }
		}
	    } else {
		if (!((raw ^ 0xe0) & 0xf0)) {
		    /* large value */
		    value = (raw & 0x7) << 4;
		    raw = *data++;
		    value |= raw >> 4;
		} else {
		    value = raw >> 4;
		}
		if (c == value) {
		    c = 0;
		    if (--offset < 0)
			break;
		} else if (value > c)
		    c++;
		index++;
	    }
	
	    /* data in low nibble */
	    if (!((raw ^ 0xf) & 0xf)) {
		/* repeated zeros */
		value = *data++ + 4;
		index += value;
		if (c == 0) {
		    offset -= value;
		    if (offset < 0) {
			index += offset;
			break;
		    }
		}
	    } else {
		if (!((raw ^ 0xe) & 0xf)) {
		    /* large value */
		    value = (*data++) & 0x7f;
		} else {
		    value = raw & 0xf;
		}
		if (c == value) {
		    c = 0;
		    if (--offset < 0)
			break;
		} else if (value > c)
		    c++;
		index++;
	    }
	}
    }
    
    return index;
}

static BWSECTION UInt32
BWFilterObjectsInCatalog(DBRecord *db, Char *search, Boolean skipAll)
{
    UInt8  c, sc;
    UInt32 *freqs;
    UInt8  *order;
    UInt16 *freqoffset;
    Int16    ordlen, datastart;
    UInt32 start, end;

    order  = db->bwdata.superBlock;
    ordlen = *order++;
    freqs      = (UInt32*) (order + (ordlen|1));
    freqoffset = (UInt16*) (freqs + ordlen + 1);
    datastart  = (freqoffset[ordlen]+15) >> 4;
    start = 0;
    end = freqs[ordlen];

    if (skipAll) {
        start = end;
    } else if (! *search) {
        end   = db->numRecords;
    } else {
        while ((sc = *search++) != 0) {
            for (c = 0; sc != order[c]; c++) {
                if (c == ordlen) {
                    start = end; /* Character not in whole database */
                    break;
                }
            }
            
            start = BWSearchOneStep(db, freqs, freqoffset, c, start);
            end   = BWSearchOneStep(db, freqs, freqoffset, c, end);
            if (start >= end) {
                start = end; /* Character not in whole database */
                break;
            }
        }
    }    
    db->filterStart = start;
    db->filterEnd = end;
    return end - start;
}

UInt32 BWFilterObjects(Char *search) {
    UInt32 count = 0;
    Int16 i;
    for (i = 0; i < numDBs; i++) {
	if (catDBs[i]->isVisible) {
	    count += BWFilterObjectsInCatalog(catDBs[i], search,
                     catDBs[i]->isStars && userPrefs.optFlags.omitStars);
	}
    }
    return count;
}

UInt32 BWGetFilteredPosition(UInt32 record) {
    Int16 i;
    UInt32 offset = 0;
    for (i = 0; i < numDBs; i++) {
	if (catDBs[i]->isVisible) {
	    if (record < catDBs[i]->numRecords) {
		Int32 pos = BWFollowPosition(catDBs[i], record, NULL);
		return pos < 0 ? NO_OBJECT : offset + pos;
	    } else {
		record -= catDBs[i]->numRecords;
		offset += catDBs[i]->filterEnd - catDBs[i]->filterStart;
	    }
	}
    }
    return NO_OBJECT;
}

UInt32 BWGetFilteredObject(UInt32 position) {
    Int16 i;
    UInt32 catOffset = 0;
    for (i = 0; i < numDBs; i++) {
	if (catDBs[i]->isVisible) {
	    if (catDBs[i]->filterStart + position < catDBs[i]->filterEnd) {
		return catOffset + 
		    BWGetRecord(catDBs[i], catDBs[i]->filterStart + position);
	    } else {
		position -= catDBs[i]->filterEnd - catDBs[i]->filterStart;
		catOffset += catDBs[i]->numRecords;
	    }
	}
    }
    return NO_OBJECT;
}

void BWOpenCatalogs(void)
{
    MemHandle handle;
    Int16 i,j,size, ordlen;
    UInt32 *freqs;
    for (i = 0; i < numDBs; i++) {
        QTreeHeader *header;

      	handle = CatalogGetHandle(catDBs[i], catDBs[i]->firstBWBlock);
      	catDBs[i]->bwdata.superBlockHandle = handle;
      	catDBs[i]->bwdata.superBlock = MemHandleLock(handle);

      	ordlen = *(catDBs[i]->bwdata.superBlock);
      	freqs  = (UInt32*) (catDBs[i]->bwdata.superBlock + 1 + (ordlen|1));
      	size   = (freqs[ordlen]+BLOCKSIZE-1) / BLOCKSIZE;
      	catDBs[i]->bwdata.numBWBlocks = size;
      	catDBs[i]->bwdata.data = MemPtrNew(size*sizeof(UInt16*));
      	catDBs[i]->bwdata.dataHandle = MemPtrNew(size*sizeof(MemHandle));
        if (catDBs[i]->qtreeRootBlock == 0)
            catDBs[i]->qtreeRootBlock = catDBs[i]->firstBWBlock + 1 + size;

      	if (catDBs[i]->bwdata.data == NULL || catDBs[i]->bwdata.dataHandle == NULL) {
            SysFatalAlert("Out of memory");
        }

      	for (j = 0; j < size; j++) {
      	    handle = CatalogGetHandle(catDBs[i], catDBs[i]->firstBWBlock + 1 + j);
      	    catDBs[i]->bwdata.dataHandle[j] = handle;
      	    catDBs[i]->bwdata.data[j] = MemHandleLock(handle);
      	}

        handle = CatalogGetHandle(catDBs[i], catDBs[i]->qtreeRootBlock);
        header = MemHandleLock(handle);
        catDBs[i]->minMag = MAGNITUDE(catDBs[i]->qTreeObjectSize,
            (QTreeObject*) (header+1), ((UInt16)catDBs[i]->magHigh << 8));
        MemHandleUnlock(handle);
        CatalogFreeHandle(catDBs[i], handle);
    }

}

void BWCloseCatalogs(void)
{
    Int16 i, j, size;

    for (i = 0; i < numDBs; i++) {
    	MemHandleUnlock(catDBs[i]->bwdata.superBlockHandle);
        CatalogFreeHandle(catDBs[i], catDBs[i]->bwdata.superBlockHandle);
    	size = catDBs[i]->bwdata.numBWBlocks;
    	for (j = 0; j < size; j++) {
            MemHandleUnlock(catDBs[i]->bwdata.dataHandle[j]);
            CatalogFreeHandle(catDBs[i], catDBs[i]->bwdata.dataHandle[j]);
    	}
    	MemPtrFree(catDBs[i]->bwdata.data);
    	MemPtrFree(catDBs[i]->bwdata.dataHandle);
    }
}
