/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroCat.c 1006 2010-10-02 02:54:19Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <PalmOS.h>
#include <HostControl.h>
#include "PalmUtil.h"
#include "resource.h"
#include "Astro.h"
#include "AstroLib.h"
#include "AstroCat.h"
#include "AstroQTree.h"
#include "AstroSky.h"
#include "FixedPoint.h"
#include "AstroTime.h"
#include "AstroBW.h"

// Prior to 3.20, this was 'DB', but 3.20 increases the number
// of possible catalogs to 128, forcing this to change.
#define CAT_PREF_ID  'CA'

DBRecord* catDBs[MAX_DBS];
Int16       numDBs;

/*
 * Pointers to catalog objects
 */
static ListPtr         catNamePtr;
static TablePtr        catalogPtr;
static ScrollBarPtr    catScrollPtr;
static UInt8           catScrollShift;
static Int16           visibleRows;
static UInt32          firstIndex;
UInt32                 totalRecords;
static const RGBColorType defaultOnBlackColor = { 0, 255, 255, 255 };
static const RGBColorType defaultOnWhiteColor = { 0, 0, 0, 0 };
static const RGBColorType defaultNightModeColor = { 0, 255, 255, 255 };
Char dbPrefix[] = "AstroInfo ";

// Global used to hold last find text.
Char filterString[FIELDLENGTH] = { "\0" };


static void DBSECTION CatalogStart(EventPtr event);
static Boolean DBSECTION IsStars(Char* name);


UInt32 GetTotalRecords() {
    UInt32 total = 0;
    Int16 i;
    for (i = 0; i < numDBs; i++) {
	if (catDBs[i]->isVisible)
	    total += catDBs[i]->numRecords;
    }
    return total;
}

/*
 * Display the list of catalogs of objects.
 */
static void DBSECTION
CatNamesListDrawData(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    FontID oldfont;
    DBRecord **items = (DBRecord**) itemsText;
    Char *name = items[itemNum]->name;
    Coord x;

    /* Skip AstroInfo prefix if present */
    name = TrimName(name, dbPrefix);

    /* Set to symbol font for check box */
    oldfont = FntSetFont(symbol11Font);

    /* Show Visible state */
    WinDrawChars((Char*)&items[itemNum]->isVisible, 1,
		 bounds->topLeft.x, bounds->topLeft.y);
    x = 20;

    /* Show catalog name */
    FntSetFont(fontNormal);
    WinDrawChars(name, StrLen(name),
		 bounds->topLeft.x + x, bounds->topLeft.y);

    /* Restore entry font */
    FntSetFont(oldfont);
}



static DBSECTION 
void CatalogUpdateSelection(void)
{
    Int32 currRecord = GetCurrentObject() - NUMPLANETS;
    if (filterString[0] || userPrefs.optFlags.omitStars)
	    currRecord = BWGetFilteredPosition(currRecord);
    if (currRecord >= firstIndex
        && currRecord < firstIndex + visibleRows)
        TblSelectItem(catalogPtr, currRecord - firstIndex, 0);
    else
        TblUnhighlightSelection(catalogPtr);
}



/* Get the amount of shift needed to transform value into a 16-bit value */
static DBSECTION
UInt8 GetShift(Int32 value)
{  
    UInt8 s;

    if ( value < 0 )
        value = -value;

    s = 0;
    while ( value & 0xFFFF8000ul ) {
        s++;
        value >>= 1;
    }
    
    return s;
}

/* The following routines provide glue for 32-bit scrollbars, allowing
   more than 32767 objects */
static DBSECTION
void CatSetScrollBar(Int32 value, Int32 min, Int32 max, Int32 pageSize)
{
    UInt8 shift1;
    UInt8 shift2;

    shift1 = GetShift(min);
    shift2 = GetShift(max);
    if (shift1 <= shift2)
        catScrollShift = shift2;
    else
        catScrollShift = shift1;

    SclSetScrollBar(catScrollPtr, value >> catScrollShift,
        min >> catScrollShift, max >> catScrollShift,
        pageSize >> catScrollShift);
}

static DBSECTION
void CatalogUpdateScrollBar(void)
{
    /*
     * Connect the scrollbar to the list.  The list cannot change.
     */
    UInt32 max = 0;
    max = totalRecords;
    max = max <= visibleRows ? 0 : max - visibleRows;
    if (firstIndex > max)
        firstIndex = max;

    CatSetScrollBar(firstIndex, 0, max, visibleRows);
}

/*
 * Scroll to the record.  Update table and scrollbar.
 *
 * It can't be less than zero of course, but also we don't allow
 * whitespace at the bottom if there are enough rows to fill the
 * display.  However we leave the display as close as possible to
 * where the user put it last.
 */
static DBSECTION
void CatalogScroll(UInt32 newPos, Boolean drawTable)
{
    firstIndex = newPos;
    CatalogUpdateScrollBar();
    TblUnhighlightSelection(catalogPtr);
    TblMarkTableInvalid(catalogPtr);
    if (drawTable)
        TblRedrawTable(catalogPtr);
    CatalogUpdateSelection();
}


DBSECTION void
CatalogFreeHandle(DBRecord* db, MemHandle h)
{
    if (h != NULL && db->vfs)
        MemHandleFree(h);
}


DBSECTION MemHandle
CatalogGetHandle(DBRecord* db, UInt16 index)
{
   if (db->vfs) {
       MemHandle h;
       
       if (errNone != VFSFileDBGetRecord(db->dbInfo.vfs.fileRef, index, &h, NULL, NULL))
           return NULL;
       
       return h;
   } else {
       return DmQueryRecord(db->dbInfo.ram.db, index);
   }
}

static UInt32 DBSECTION
CatalogReadIndexEntry(DBRecord *db, UInt16 *indexPtr, UInt16 blkindex) {
    if (db->bytesPerEntry == 2) {
        return indexPtr[blkindex];
    } else if (db->bytesPerEntry == 3) {
        indexPtr += 3 * (blkindex / 2);
        if ((blkindex & 1) == 0) {
            return *(UInt32*)indexPtr >> 8;
        } else {
            return *(UInt32*)(indexPtr+1) & 0xffffff;
        }
    } else {
        return ((UInt32*)indexPtr)[blkindex];
    }
}

static UInt32 DBSECTION
CatalogReadIndex(DBRecord *db, UInt32 index) {
    UInt32 value;
    MemHandle handle;
    UInt16 *indexPtr;
    UInt16 block = index / db->entriesPerIndex;
    UInt16 blkindex = index % db->entriesPerIndex;

    handle = CatalogGetHandle(db, block);
    indexPtr = MemHandleLock(handle);
    value = CatalogReadIndexEntry(db, indexPtr, blkindex);
    MemHandleUnlock(handle);
    CatalogFreeHandle(db, handle);
    return value;
}

static UInt32 DBSECTION
CatalogGetQTreeIndex(DBRecord *db, UInt32 index) {
    if (db->isUnsorted) {
        UInt16 start, end, block, offset;
        MemHandle handle;
        UInt16 *indexPtr;

        start = 0;
        end = db->numIndexNodes;
        while (start < end-1) {
            UInt16 mid = (start+end)/2;
            if (index >= CatalogReadIndex(db, mid*db->entriesPerIndex)) {
                start = mid;
            } else {
                end = mid;
            }
        }
        block = start;

        handle = CatalogGetHandle(db, block);
        indexPtr = MemHandleLock(handle);
        start = 0;
        end   = MemHandleSize(handle) / db->bytesPerEntry;
        while (start < end-1) {
            UInt16 mid = (start+end)/2;
            if (index >= CatalogReadIndexEntry(db, indexPtr, mid)) {
                start = mid;
            } else {
                end = mid;
            }
        }
        offset = index - CatalogReadIndexEntry(db, indexPtr, start);
        MemHandleUnlock(handle);
        CatalogFreeHandle(db, handle);
        start += block * db->entriesPerIndex + db->qtreeRootBlock;
        return (UInt32) start * db->maxQTreeNodeSize + offset;
    } else {
        return CatalogReadIndex(db, index);
    }
}

UInt32 DBSECTION
CatalogGetIndexFromQTree(DBRecord *db, UInt32 qtreeIndex) {
    int i, j;
    UInt32 index = NO_OBJECT;
    MemHandle handle;
    UInt16 *indexPtr;

    if (db->isUnsorted) {
        return CatalogReadIndex(db, qtreeIndex / db->maxQTreeNodeSize
                                - db->qtreeRootBlock)
            + qtreeIndex % db->maxQTreeNodeSize;
    }

    for (i = 0; i < db->numIndexNodes; i++) {
        handle = CatalogGetHandle(db, i);
        ErrFatalDisplayIf(handle == NULL,"Could not retrieve record!");
        indexPtr = MemHandleLock(handle);

        if (db->bytesPerEntry == 2) {
            for (j = 0; j < db->numRecords; j++) {
                if (*indexPtr == qtreeIndex) {
                    index = j;
                    break;
                }
                indexPtr++;
            }
        } else {
            for (j = 0; j < ENTRIES_PER_INDEX; j += 2) {
                if ((*(UInt32*)indexPtr >> 8) == qtreeIndex) {
                    index = ENTRIES_PER_INDEX*i + j;
                    break;
                }
                if ((*(UInt32*)(indexPtr+1) & 0xffffff) == qtreeIndex) {
                    index = ENTRIES_PER_INDEX*i + j + 1;
                    break;
                }
                indexPtr += 3;
            }
        }
        MemHandleUnlock(handle);
        CatalogFreeHandle(db, handle);
        if (index != NO_OBJECT)
            break;
    }
    return index;
}        

/* Get R2 catalogue object */
Boolean DBSECTION
GetCatalogObject(UInt32 index, AstroRecordType *art, UInt16* dbNumberP) {
    static UInt32 lastIndex = -1;
    static AstroRecordType lastART;
    Int16 i;
    UInt32 qtreeIndex;

    if (index != lastIndex) {
	lastIndex = index;

	/*
	 * Get the cat string
	 */
	for (i = 0; i < numDBs; i++) {
	    if (catDBs[i]->isVisible) {
    		if (index >= catDBs[i]->numRecords) {
    		    index -= catDBs[i]->numRecords;
    		} else {
    		    break;
    		}
	    }
	}
	if (i >= numDBs) {
	    lastIndex = -1;
	    return false;
	}

        qtreeIndex = CatalogGetQTreeIndex(catDBs[i], index);

	QTreeGetObject(catDBs[i], qtreeIndex, &lastART);
	BWGetString(catDBs[i], index, lastART.description);
	lastART.description[31] = 0; // force null termination
	if (lastART.description[0] == 0) {
        Char* p;
        UInt16  len;

        lastART.description[0] = '(';

        p = TrimName(catDBs[i]->name, dbPrefix);

        if ( 29 < StrLen(p)) {
            StrNCopy(lastART.description+1,p,30);
            lastART.description[1+29]=0;
        }
        else {
            StrCopy(lastART.description+1,p);
        }
        len = StrLen(lastART.description);
        lastART.description[len]=')';
        lastART.description[len+1]=0;
    }
    if (dbNumberP != NULL)
        *dbNumberP = i;
    }
    *art = lastART;
    return true;
}

/*
 * Display the current catalog of objects.
 */
static void DBSECTION
CatalogTableDrawData(TablePtr table, Int16 row, Int16 col, RectanglePtr bounds)
{
    AstroRecordType art;
    FontID oldfont;
    UInt32 index;
    Coord  x;

    /* Set to normal font by default */
    oldfont = FntSetFont(fontNormal);

    index = firstIndex + row;
    if (filterString[0] || userPrefs.optFlags.omitStars)
	    index = BWGetFilteredObject(index);

    if (index != NO_OBJECT && GetCatalogObject(index, &art, NULL)) {

        x = bounds->topLeft.x;

        /* Optionally show flag indicating visibility of object */
        if (userPrefs.optFlags.moreAccurate
            && (userPrefs.optFlags.showAbove
                || userPrefs.optFlags.showBelow
                || userPrefs.optFlags.showNever)) {

            Int32  sinlat, coslat;
            Int32  UTr, UTs;
            Int16    flags;
            Int16    upnow = 0;

            x += 9;

            /* XXX This could be made 100 times faster:
             *
             * object is circumpolar if
             *    |art->dec + lat| > FP_ONE
             * object is never visible if
             *    |art->dec - lat| > FP_ONE
             * object is up if
             *    RaDec2AltAz gives alt > 0
             *
             * I fix it later when I optimize RaDec2AltAz calculations.
             */

            fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

            /* TODO: Precession! */
            flags = ObjectRiseSet(userPrefs.jday, userPrefs.jtime,
                                  art.radec,
                                  sinlat, coslat, userPrefs.Location[userPrefs.currentLoc].lng,
                                  &UTr, &UTs);
            
            /* Prepend a special symbol if object is up or circumpolar */
            upnow = UpNow(flags, UTr, UTs, userPrefs.jtime);
            FntSetFont(fontSmall);
            if (upnow && userPrefs.optFlags.showAbove) {
                WinDrawChars("\012", 1,
                             bounds->topLeft.x, bounds->topLeft.y + 2);
            } else if ((flags & (ASTROLIB_DONTRISESET) && userPrefs.optFlags.showNever)) {
                WinDrawChars("\026", 1,
                             bounds->topLeft.x, bounds->topLeft.y + 2);
            } else if (!upnow && userPrefs.optFlags.showBelow) {
                WinDrawChars("\013", 1,
                             bounds->topLeft.x, bounds->topLeft.y + 2);
            }
        }

        /* Show object name */
        FntSetFont(fontNormal);

        WinDrawChars(art.description, StrLen(art.description),
            x, bounds->topLeft.y);

        /* Restore entry font */
        FntSetFont(oldfont);
    }
}

static void DBSECTION
SaveCurrEntry()
{
    Int16 i;
    CatalogPrefType* catalogPrefsP;
    Int32 index = userPrefs.currentObj - NUMPLANETS;

    catalogPrefsP = MemPtrNew(sizeof(CatalogPrefType));

    if (catalogPrefsP == NULL)
        return;

    catalogPrefsP->selectedCatalog = -1;
    catalogPrefsP->selectedIndex = -1;
    catalogPrefsP->visible[ 0 ] = 0;
    catalogPrefsP->visible[ 1 ] = 0;
    catalogPrefsP->visible[ 2 ] = 0;
    catalogPrefsP->visible[ 3 ] = 0;
    StrCopy(catalogPrefsP->filterString, filterString);

    for (i = 0; i < numDBs; i++) {
        UInt16 j;
    	MemMove(catalogPrefsP->known[i], catDBs[i]->name, 32);
    	if (catDBs[i]->isVisible) {
    	    if (index >= 0 && index < catDBs[i]->numRecords) {
    		catalogPrefsP->selectedCatalog = i;
    		catalogPrefsP->selectedIndex = index;
    	    }
    	    catalogPrefsP->visible[ i / 32 ] |= 1L << ( i % 32 );
    	    index -= catDBs[i]->numRecords;
    	}

        for (j=0; j < NUM_COLOR_MODES; j++)
            catalogPrefsP->color[j][i] = catDBs[i]->color[j];
    }
    
    if (index<0 && userPrefs.currentObj != NO_OBJECT) {
        catalogPrefsP->selectedCatalog = -1;
        catalogPrefsP->selectedIndex   = userPrefs.currentObj;
    }

    for (; i < MAX_DBS; i++)
         catalogPrefsP->known[i][0] = 0;

    PrefSetAppPreferences(APPID, CAT_PREF_ID, 1,
                          catalogPrefsP,
                          sizeof(*catalogPrefsP),
                          true);

    MemPtrFree(catalogPrefsP);
}

static void DBSECTION
RestoreCurrEntry(Boolean readVisible)
{
    UInt16 size;
    UInt16 i;
    CatalogPrefType* catalogPrefsP;

    filterString[0] = 0;
    catalogPrefsP = MemPtrNew(sizeof(CatalogPrefType));

    SetCurrentObject(NO_OBJECT);

    size = 0;

    if (catalogPrefsP == NULL ||
        noPreferenceFound == PrefGetAppPreferences(APPID, CAT_PREF_ID,
                              NULL, &size, true) ||
        sizeof(*catalogPrefsP) < size) {

        for (i = 0; i < MAX_DBS; i++) {
             catalogPrefsP->color[MODE_ONBLACK][i] = defaultOnBlackColor;
             catalogPrefsP->color[MODE_ONWHITE][i] = defaultOnWhiteColor;
             catalogPrefsP->color[MODE_NIGHT][i] = defaultNightModeColor;
             if ( i < numDBs ) {
                catDBs[i]->color[MODE_ONBLACK] = defaultOnBlackColor;
                catDBs[i]->color[MODE_ONWHITE] = defaultOnWhiteColor;
                catDBs[i]->color[MODE_NIGHT]   = defaultNightModeColor;
             }
        }

        if (NULL!=catalogPrefsP)
            MemPtrFree(catalogPrefsP);
        return;
    }

    MemSet(catalogPrefsP, sizeof(CatalogPrefType), 0 );

    if (readVisible)
        for (i = 0; i < MAX_DBS; i++) {
             catalogPrefsP->color[MODE_ONBLACK][i] = defaultOnBlackColor;
             catalogPrefsP->color[MODE_ONWHITE][i] = defaultOnWhiteColor;
             catalogPrefsP->color[MODE_NIGHT][i]   = defaultNightModeColor;
        }

    if (PrefGetAppPreferences(APPID, CAT_PREF_ID,
			      catalogPrefsP, &size,
			      true) != noPreferenceFound) {
         Int16 i, j;
	     UInt32 offset = 0;

	for (i = 0; i < numDBs; i++) {
        if (readVisible) {
            catDBs[i]->color[MODE_ONBLACK] = defaultOnBlackColor;
            catDBs[i]->color[MODE_ONWHITE] = defaultOnWhiteColor;
            catDBs[i]->color[MODE_NIGHT]   = defaultNightModeColor;
        }

	    for (j = 0; j < MAX_DBS && catalogPrefsP->known[j][0]; j++) {
		if (StrCompare(catDBs[i]->name,
			       catalogPrefsP->known[j]) == 0) {
		    if (readVisible) {
                        UInt16 k;
  			catDBs[i]->isVisible =
  			    (catalogPrefsP->visible [ j / 32 ] & (1L << ( j % 32 ))) ? true : false;
  			            for (k=0; k<NUM_COLOR_MODES; k++) {
                             catDBs[i]->color[k] = catalogPrefsP->color[k][j];
                        }
                    }
		    if (j == catalogPrefsP->selectedCatalog) {
			    SetCurrentObject(offset +
					 catalogPrefsP->selectedIndex +
					 NUMPLANETS);
		    }
		    break;
		}
	    }
	    if (catDBs[i]->isVisible)
		offset += catDBs[i]->numRecords;
	}
    }

    if (catalogPrefsP->selectedCatalog == -1 &&
        catalogPrefsP->selectedIndex != (UInt16)-1) {
           /* Planet */
           SetCurrentObject(catalogPrefsP->selectedIndex);
    }

    if (astroInfoArgument[0] == 'f') {
        StrNCopy(filterString, astroInfoArgument+1, FIELDLENGTH - 1);
        filterString[FIELDLENGTH-1] = 0;
    }
    else
        MemMove(filterString, catalogPrefsP->filterString, FIELDLENGTH);
    MemPtrFree(catalogPrefsP);
}


static Boolean DBSECTION
IsStars(Char* name)
{
    return NULL != StrStr(name, "Stars") ||
           NULL != StrStr(name, "HYG") ||
           NULL != StrStr(name, "UCAC") ||
           NULL != StrStr(name, "Tycho");
}


static Int16 DBSECTION
CompareDBNames(void* a, void* b, Int32 other)
{
    return StrCaselessCompare((*(DBRecord**)a)->name, (*(DBRecord**)b)->name);
}



static void DBSECTION
SetupCatalogQTreeData(DBRecord* db, MemHandle appInfo, UInt16 version)
{
    db->isUnsorted = false;
    if (appInfo) {
        CatalogAppInfo*      appInfoPtr;

        appInfoPtr = MemHandleLock(appInfo);

        db->maxQTreeNodeSize = appInfoPtr->maxQTreeNodeSize;
        db->bytesPerEntry = appInfoPtr->bytesPerEntry;
        db->entriesPerIndex = appInfoPtr->entriesPerIndex;
        db->qTreeObjectSize  = appInfoPtr->qTreeObjectSize;
        db->distance = appInfoPtr->distance;
        db->numRecords = appInfoPtr->numStars;
        db->numIndexNodes = appInfoPtr->numIndexNodes;
        db->firstBWBlock = appInfoPtr->numIndexNodes;
        db->qtreeRootBlock =
            appInfoPtr->numIndexNodes + appInfoPtr->numBWNodes;
        db->magHigh = appInfoPtr->magHigh;

        if (db->maxQTreeNodeSize == 0) {
            db->isUnsorted = true;
            db->maxQTreeNodeSize = 256;
        }

        MemHandleUnlock(appInfo);
    } else {
        MemHandle record0;
        UInt16 recordSize;
        
        record0 = CatalogGetHandle(db, 0);
        recordSize = MemHandleSize(record0);

        if (version == 3) {
            UInt8 *p;
            UInt16 extrasSize;
            CatalogIndexV3Extras* extras;
            p = MemHandleLock(record0);
            extrasSize = *(UInt16*)(p + recordSize - sizeof(UInt16));
            extras = (CatalogIndexV3Extras*) (p + recordSize - extrasSize);
            db->numRecords = (recordSize - extrasSize) / sizeof(UInt16);
            db->maxQTreeNodeSize = extras->maxQTreeNodeSize;
            db->qTreeObjectSize  = extras->qTreeObjectSize;
            if (extrasSize > 6) {
                db->magHigh  = extras->magHigh;
                db->distance = extras->distance;
            } else {
                db->magHigh  = 0;
                db->distance = 0;
            }
            MemHandleUnlock(record0);

        } else {
            db->numRecords = recordSize / sizeof(UInt16);
            db->maxQTreeNodeSize = V2_QTREE_MAX_NODE_SIZE;
            db->qTreeObjectSize  = V2_SIZE_QTreeObject;
            db->magHigh = 0;
            db->distance = 0;
        }
        CatalogFreeHandle(db, record0);
        db->firstBWBlock = 1;
        db->bytesPerEntry = 2;
        db->entriesPerIndex = 65535;
        db->qtreeRootBlock = 0; // set in BWOpenCatalogs
        db->numIndexNodes = 1;
    }
}


static void DBSECTION
ScanVFSCatalogs(UInt16 volRef, Char* dir)
{
    FileRef dirRef;
    UInt32 dirEnumerator;
    FileInfoType info;

    if (errNone != VFSFileOpen(volRef, dir, vfsModeRead, &dirRef))
        return;

    info.nameP = MemPtrNew(128);
    if (info.nameP == NULL) {
        VFSFileClose(dirRef);
        return;
    }

    info.nameBufLen = 128;

    dirEnumerator = vfsIteratorStart;

    while (numDBs < MAX_DBS &&
           errNone == VFSDirEntryEnumerate(dirRef, &dirEnumerator, &info)) {
        FileRef fileRef;
        UInt16  version;
        Char* pathName;

        pathName = MemPtrNew(StrLen(dir) + 1 + StrLen(info.nameP) + 1);
        if (pathName == NULL)
            continue;

        StrPrintF(pathName, "%s/%s", dir, info.nameP);

        if (errNone == VFSFileOpen(volRef, pathName, vfsModeRead, &fileRef)) {
            UInt32 type;
            UInt32 creator;
            Char   name[dmDBNameLength];
            MemHandle appInfo;

            if (errNone == VFSFileDBInfo(fileRef, name, NULL, &version,
                                         NULL, NULL, NULL, NULL, 
                                         &appInfo, NULL, 
                                         &type, &creator, NULL) &&
                2 <= version && type == ASTRODBTYPE && creator == APPID ) {

                catDBs[numDBs] = MemPtrNew(sizeof(DBRecord));
                if (catDBs[numDBs] == NULL)
                    SysFatalAlert("Out of memory");

                catDBs[numDBs]->vfs = true;
                catDBs[numDBs]->isStars = IsStars(name);
                catDBs[numDBs]->dbInfo.vfs.fileRef = fileRef;
                catDBs[numDBs]->dbInfo.vfs.volRef  = volRef;
                StrCopy(catDBs[numDBs]->name, name);

                SetupCatalogQTreeData(catDBs[numDBs], appInfo, version);
                catDBs[numDBs]->dbInfo.vfs.pathName = pathName;
                catDBs[numDBs]->isVisible = true;
                catDBs[numDBs]->version = version;
                numDBs++;
                if (appInfo != NULL) {
                    MemHandleFree(appInfo);
                }
            } else {
                VFSFileClose(fileRef);
                MemPtrFree(pathName);
            }
        } else {
            MemPtrFree(pathName);
        }
    }

    VFSFileClose(dirRef);
    MemPtrFree(info.nameP);
}


void DBSECTION
OpenCatalogs(void)
{
    UInt16 cardNo, version;
    LocalID dbID;
    static LocalID oldDBs[MAX_DBS];
    UInt16  numOldDBs = 0;
    Err err;
    DmSearchStateType state;
    Boolean newSearch = true;
    LocalID appInfoId;
    MemHandle appInfo;
    UInt32 value;
    UInt16 i;

    HostTraceOutputTL(appErrorClass, "OpenCatalog");

    numDBs = 0;
    while (numDBs < MAX_DBS)
    {
        err = DmGetNextDatabaseByTypeCreator(newSearch, &state,
                                             ASTRODBTYPE, APPID,
                                             false, &cardNo, &dbID);
        if (err)
	    break;
	newSearch = false;

	catDBs[numDBs] = MemPtrNew(sizeof(DBRecord));

        err = DmDatabaseInfo(cardNo, dbID, catDBs[numDBs]->name,
			     NULL, &version,      /* attr, vers */
			     NULL, NULL, NULL,    /* time stamps */
			     NULL, &appInfoId, NULL,    /* modnum, app, sort*/
			     NULL, NULL);         /* type, creator */
        if (err) {
	    MemPtrFree(catDBs[numDBs]);
	    continue;
	}


	if (version < 2) {
	    if (numOldDBs < MAX_DBS) {
    		oldDBs[numOldDBs++] = dbID;
    		MemPtrFree(catDBs[numDBs]);
	    }
	} else {
	    catDBs[numDBs]->vfs = false;
	    catDBs[numDBs]->isStars = IsStars(catDBs[numDBs]->name);
	    catDBs[numDBs]->dbInfo.ram.db = DmOpenDatabase(cardNo, dbID, dmModeReadOnly);
	    catDBs[numDBs]->dbInfo.ram.cardNo = cardNo;

        if (appInfoId != NULL)
            appInfo = MemLocalIDToGlobal(appInfoId,cardNo);
        else
            appInfo = NULL;

	    SetupCatalogQTreeData(catDBs[numDBs], appInfo, version);

	    catDBs[numDBs]->isVisible = true;
	    catDBs[numDBs]->version = version;
	    numDBs++;
	}
    }

    /* We can only handle version 2 catalogs.  Since there can
     * be only one AstroInfo installed at a time it makes sense
     * to remove all other catalogs.  Otherwise, the user may
     * not notice that they are still there and take up memory.
     */
    while (numOldDBs-- > 0)
    	DmDeleteDatabase(cardNo, oldDBs[numOldDBs]);

    if (errNone == FtrGet(sysFileCVFSMgr, vfsFtrIDVersion, &value)) {
        UInt16 volRef;
        UInt32 volIterator = vfsIteratorStart;

        while (errNone == VFSVolumeEnumerate(&volRef, &volIterator) &&
               numDBs < MAX_DBS) {
            ScanVFSCatalogs(volRef, "/PALM/Launcher");
            ScanVFSCatalogs(volRef, "/PALM/Programs/AstroInfo");
        }
    }

	if (0<numDBs)
	    SysInsertionSort(catDBs, numDBs, sizeof(catDBs[0]), CompareDBNames, 0);
	    
    for (i=0; i<numDBs; i++)
        catDBs[i]->catNum = i;

    RestoreCurrEntry(true);

    BWOpenCatalogs();

    if (filterString[0] || userPrefs.optFlags.omitStars) {
	    totalRecords = BWFilterObjects(filterString);
    }

    /* Call this even if catalog could not be loaded */
    SkyMapLoadCatalog();
}



/*
 * Release the open catalog
 */
void DBSECTION
FreeCatalogs()
{
    Int16 i;

    HostTraceOutputTL(appErrorClass, "FreeCatalogs");

    BWCloseCatalogs();
    SaveCurrEntry();
    for (i = 0; i < numDBs; i++) {
    	if (catDBs[i]->vfs && catDBs[i]->dbInfo.vfs.fileRef) {
            if(catDBs[i]->dbInfo.vfs.pathName) {
                MemPtrFree(catDBs[i]->dbInfo.vfs.pathName);
            }
            VFSFileClose(catDBs[i]->dbInfo.vfs.fileRef);
        }
    	else if (! catDBs[i]->vfs && catDBs[i]->dbInfo.ram.db) {
    	    DmCloseDatabase(catDBs[i]->dbInfo.ram.db);
    	}
    	MemPtrFree(catDBs[i]);
    }
    numDBs = 0;

    SkyMapFreeCatalog();
}


static void DBSECTION
FindObject(Char *newString, Boolean drawTable)
{
    Int32 currRecord = GetCurrentObject() - NUMPLANETS;
    StrToLower(filterString, newString);
    if (filterString[0] || userPrefs.optFlags.omitStars) {
	    totalRecords = BWFilterObjects(filterString);
	    currRecord = BWGetFilteredPosition(currRecord);
    } else {
	    totalRecords = GetTotalRecords();
    }
    CatalogScroll(currRecord+1 > 5 ? currRecord - 5 : 0, drawTable);
}

/*
 * Events for the SelectCatalog form
 */
Boolean DBSECTION
CatalogListFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();

    switch (event->eType) {
    case keyDownEvent:
        if ( event->data.keyDown.modifiers & commandKeyMask ) {
            ListType* list = (ListPtr)FrmGetObjectPtrByID(formPtr, ls_CatName);

            switch (event->data.keyDown.chr) {
                case vchrRockerUp:
                case chrPageUp:
                    if (!InteractionMode())
                        LstScrollList( list, winUp, LstGetVisibleItems(list) - 1);
                    handled = true;
                    break;
                case vchrRockerDown:
                case chrPageDown:
                    if (!InteractionMode())
                        LstScrollList(list, winDown, LstGetVisibleItems(list) - 1);
                    handled = true;
                    break;
                default:
                    break;
            }
        }
        break;

    case lstSelectEvent:
        if (event->data.lstSelect.listID == ls_CatName) {
            Int16 i = event->data.lstSelect.selection;

            if ( i != noListSelection) {
                catDBs[i]->isVisible = !catDBs[i]->isVisible;
                LstDrawList(catNamePtr);
            }
            handled = true;
        }
        break;

    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
        case b_OK:
	        RestoreCurrEntry(false);
            FrmReturnToForm(0);
            FrmUpdateForm(f_Catalog, updateObjectListChanged);
            handled = true;
            break;

        case b_Delete: {
            // Get the database that was selected for deletion.
            Int16 i = LstGetSelection(catNamePtr);

            // Confirm that the user really wants to delete a db.
            if (i != noListSelection && FrmAlert(a_AreYouSure) == 0)
            {
              Char* pathName = NULL;
      		Int16     cardNo  = 0;
      		LocalID dbID = 0;
      		UInt16  volRef = 0;

      		Boolean vfs = false;

            if(catDBs[i]->vfs) {
                if(catDBs[i]->dbInfo.vfs.pathName != NULL) {
                    pathName = MemPtrNew(StrLen(catDBs[i]->dbInfo.vfs.pathName));
                    if(pathName == NULL ) {
                        handled = true;
                        break;
                    }
                    StrCopy(pathName, catDBs[i]->dbInfo.vfs.pathName);
                    volRef = catDBs[i]->dbInfo.vfs.volRef;
                    vfs = true;
                }
            }
            else {
        		dbID = DmFindDatabase(cardNo, catDBs[i]->name);
        		cardNo = catDBs[i]->dbInfo.ram.cardNo;
      		}

    		FreeCatalogs();
    		if (vfs) {
                VFSFileDelete(volRef, pathName);
                MemPtrFree(pathName);
            }
            else {
                DmDeleteDatabase(cardNo, dbID);
            }
    		OpenCatalogs();
    		LstSetSelection(catNamePtr,-1);
    		LstSetListChoices(catNamePtr, (Char**)catDBs, numDBs);
    		LstSetTopItem(catNamePtr,0);
    		LstDrawList(catNamePtr);
            }
            handled = true;
            break;
        }
            
        default:
        }
        break;
        
    case frmOpenEvent:
	    SaveCurrEntry();
        catNamePtr = (ListPtr)FrmGetObjectPtrByID(formPtr, ls_CatName);
	    LstSetSelection(catNamePtr,-1);
	    LstSetListChoices(catNamePtr, (Char**)catDBs, numDBs);
	    LstSetTopItem(catNamePtr,0);
	    LstSetDrawFunction(catNamePtr, CatNamesListDrawData);
        FrmDrawForm(formPtr);
        
        handled = true;
        break;

    case frmCloseEvent:
        handled = false;
        break;
        
    default:
        handled = false;
    }

    return handled;
}



static UInt16 DBSECTION
GetNumberOfVisibleRows(TableType* tableP)
{
    RectangleType bounds;
    TblGetBounds(tableP, &bounds);

    return bounds.extent.y / TblGetRowHeight(tableP,0);
}



static void DBSECTION
CatalogStart(EventPtr event)
{
    UInt16 findIdx;
    UInt16 row;
    FormType* formPtr;
    FieldPtr findPtr;
    MemHandle handle;

    formPtr = FrmGetActiveForm();

    SetMenu(formPtr);

    catalogPtr = FrmGetObjectPtrByID(formPtr, tb_Catalog);
    catScrollPtr = FrmGetObjectPtrByID(formPtr, sc_Catalog);

    /* Get the total number of rows allocated.  */
    visibleRows = GetNumberOfVisibleRows(catalogPtr);

    /* Mark all rows as custom
     */
    for (row = 0; row < visibleRows; row++) {
	TblSetItemStyle(catalogPtr, row, 0, customTableItem);
    }
    TblSetColumnUsable(catalogPtr, 0, true);

    TblHasScrollBar(catalogPtr, true);

    DrawDatePicker();
    DrawTimePicker();

    findIdx = FrmGetObjectIndex(formPtr,f_Find);

    if (event->eType==frmOpenEvent) {
        /* Fill find field */
        findPtr = (FieldPtr) FrmGetObjectPtr(formPtr, findIdx);
        handle = MemHandleNew(FIELDLENGTH);
        StrCopy(MemHandleLock(handle), filterString);
        MemHandleUnlock(handle);
        FldSetText(findPtr, handle, 0, FIELDLENGTH);
    }

    TblSetCustomDrawProcedure(catalogPtr, 0,
         (TableDrawItemFuncPtr) CatalogTableDrawData);

	    /* Filter objects */
    FindObject(filterString, false);
    if (event->eType==frmOpenEvent)
         FrmSetFocus(formPtr, findIdx);

    FrmDrawForm(formPtr);

    CatalogUpdateSelection();
}


/*
 * Events for the main catalog form showing list of catalogue objects
 */
Boolean DBSECTION
CatalogFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr formPtr = FrmGetActiveForm();
    static  Coord lastExtentX;
    static  Coord lastExtentY;

    if (DateChangeHandleEvent(event)
        || FormChangeHandleEvent(event))
        return true;
        
    switch (event->eType) {
        case keyDownEvent:

            // Use Page Up/Down buttons to scroll object list.

            if (event->data.keyDown.chr == pageUpChr)
            {
                CatalogScroll(firstIndex <= visibleRows ? 0
			      : firstIndex - visibleRows, true);
                handled = true;
            }
            else if (event->data.keyDown.chr == pageDownChr)
            {
                CatalogScroll(firstIndex + visibleRows, true);
                handled = true;
            }
	    else
	    {
		FieldPtr findField = FrmGetObjectPtrByID(formPtr,f_Find);
		if (FldHandleEvent(findField, event))
		    FindObject(FldGetTextPtr(findField), true);
		    handled = true;
	    }
            break;

        case fldChangedEvent:
            if (event->data.fldChanged.fieldID == f_Find) {
		        FindObject(FldGetTextPtr(FrmGetObjectPtrByID(formPtr,f_Find)), 
                    true);
                handled = true;
            }
            break;

        case ctlSelectEvent:
            // Process various button clicks.

            switch (event->data.ctlSelect.controlID) {
            case b_CatalogSelect:
                FrmPopupForm(f_SelectCatalog);
                break;
            case b_Clear: {
          		FieldPtr findField = FrmGetObjectPtrByID(formPtr,f_Find);
          		FldDelete(findField, 0, 0xFFFF);
      		    FindObject(FldGetTextPtr(findField), true);
            }
            handled = true;
            break;
            }

        case tblSelectEvent:
            if (event->data.tblSelect.tableID == tb_Catalog
                && event->data.tblSelect.row >= 0)
            {
                UInt32 i = firstIndex + event->data.tblSelect.row;
                if (i < totalRecords) {
		    if (filterString[0] || userPrefs.optFlags.omitStars)
			i = BWGetFilteredObject(i);
                    SetCurrentObject(NUMPLANETS + i);
                    SkyCenterObject(NUMPLANETS + i);
                    FrmGotoForm(f_ObjectDetail);
                }
            }
            break;

    	case sclRepeatEvent: {
            Int16 value;
            Int16 min;
            Int16 max;
            Int16 pageSize;
            Int16 delta;
            
            /* This is done in such a way as to allow more than 32767 items,
               as well as single-line scrolling. */

            SclGetScrollBar(catScrollPtr, &value, &min, &max, &pageSize);
            delta = (Int16)event->data.sclRepeat.newValue
                        -(firstIndex>>catScrollShift);
            if (delta == -1 || delta == 1) {
                CatalogScroll(firstIndex+delta, true);
            }
            else {
    	        CatalogScroll(( Int32 )event->data.sclRepeat.newValue << catScrollShift, true);
            }
    	    break;
    	}

        case winDisplayChangedEvent: {
            handled = true;
            break;
        }

        case frmOpenEvent: {
            WinGetDisplayExtent( &lastExtentX, &lastExtentY );
            CatalogStart(event);

            handled = true;
            break;
	}

        case frmUpdateEvent:
	    if (event->data.frmUpdate.updateCode == updateObjectListChanged) {
		FindObject(filterString, false);
		FrmDrawForm(formPtr);
	    } else if (event->data.frmUpdate.updateCode == updateDateChanged) {
		/* FIXME */
	    } else {
                Coord extentX;
                Coord extentY;
                
                WinGetDisplayExtent(&extentX,&extentY);
                
                if (extentX==lastExtentX && extentY==lastExtentY) {
                    TblUnhighlightSelection(catalogPtr);
                    FrmDrawForm(formPtr);
                    CatalogUpdateSelection();
                }
                else {
                    lastExtentX = extentX;
                    lastExtentY = extentY;

                    CatalogStart(event);
                }
	    }
            handled = true;
            break;
            
        case catalogSelectEvent:
            FrmPopupForm(f_SelectCatalog);
            handled = true;
            break;

        case frmCloseEvent:
	{
            handled = false;
            break;
	}

        default:
            handled = false;
    }

    return handled;
}
