/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroDetails.c 997 2010-07-18 20:24:43Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <TxtGlue.h>
#include "AstroDetails.h"
#include "PalmUtil.h"
#include "Astro.h"
#include "resource.h"
#include "FixedPoint.h"
#include "AstroLib.h"
#include "AstroMoon.h"
#include "AstroEphem.h"
#include "AstroTime.h"
#include "AstroTable.h"
#include "AstroCat.h"
#include "AstroSky.h"
#include "PalmUtils.h"
#include "AstroGlue.h"
#include "AstroAlign.h"
#include "constants.h"

#include <HostControl.h>

#define IS_FULLSCREEN (FrmGetActiveFormID() == f_FullScreen)

static const Char xlatarr[] =
        { 0, 1,1,1,1, 2, 3,3,3,3, 4, 5,5,5,5, 6, 7,7,7,7, 0 };

/* These are in English because they're for English Wikipedia lookup */
static const Char* planetNames[NUMPLANETS] = {
    "Mercury",
    "Venus",
    "Moon",
    "Mars",
    "Jupiter",
    "Saturn",
    "Uranus",
    "Neptune",
    "Pluto",
    "Sun"
};

#if 0
/* mass in kilograms divided by 1e19
 * Would be nice to actually use these real numbers to allow conversion between
 * units, rather than the simple string lookup further down.
 */
static const UInt32 planetmass[] = {
           33000,   //3.30e23, 
          486900,   //4.869e24,
          597200,   //5.972e24,Earth
            7350,   //7.35e22,Moon
           64219,   //6.4219e23
       190000000,   //1.900e27,
        56800000,   //5.68e26,
        86830000,   //8.683e25,
        10247000,   //1.0247e26,
            1270,   //1.27e22,
    198900000000    //1.989e30
};
#endif

RectangleType detailRect, descRect;

static void ShowDetailCell(Int8 y, Int8 x,
                           UInt16 fieldnameid,
                           Char *contents) ADETAILSECTION;


static void ADETAILSECTION
FormatDistance(Char* out, double d)
{
    Char* units;
    Char* unitModifier;
    
    if(d<0) {
        *out=0;
        return;
    }
    
    if (userPrefs.parsec) {
        d /= 3.26167;
        units = "pc";
    }
    else {
        units = "ly";
    }

    if(d>=1000000000.) {
        d /= 1000000000.;
        unitModifier = "G";
    }
    else if (d>=1000000.) {
        d /= 1000000.;
        unitModifier = "M";
    }
    else if (d>=1000.) {
        d /= 1000.;
        unitModifier = "k";
    }
    else {
        unitModifier = "";
    }

    if (d>=10)
        StrPrintF(out, "%d %s%s", (Int16)d, unitModifier, units);
    else if (d>=1)
        StrPrintF(out, "%d.%d %s%s", (Int16)d, (Int16)((d-(Int16)d)*10), 
            unitModifier, units);
    else
        StrPrintF(out, "0.%03d %s", (Int16)(d*1000), unitModifier, units);
}


void ADETAILSECTION
ClearObjectDetail()
{
    if (userPrefs.showDetailRows == 0
        || IS_FULLSCREEN)
        return;

    WinEraseRectangle(&detailRect, 0);
    WinEraseRectangle(&descRect, 0);
}


LocalID ADETAILSECTION
GetWebBrowser(void)
{
    static Boolean searched = false;
    static LocalID id;

    if (!searched) {
        id = DmFindDatabase(0, "Blazer");
        searched = true;
    }

    return id;
}

LocalID ADETAILSECTION
GetDictionary(void)
{
    static Boolean searched = false;
    static LocalID id;

    if (!searched) {
        id = DmFindDatabase(0, "AstroInfo:Dictionary Interface");
        searched = true;
    }

    return id;
}

/* Get name for use with seds.org */
/* May overwrite sedName even if it returns false. */
static Boolean ADETAILSECTION
GetObjectID(Char* id, Char* description, LookupMethodType method)
{
    UInt16 digits;
    Char*  p;
    Char*  ngcPrefix;
    Char*  id0;
    Char*  description0;

    id0 = id;
    description0 = description;

    *id = 0;

    ngcPrefix = "NGC";

    if (description[0]=='M' && method == LOOKUP_WEB) {
        StrCopy(id, "M");
        description++;
    }
    else if (!StrNCompare(description,"NGC",3)) {
        StrCopy(id, ngcPrefix);
        description += 3;
    }
    else if (!StrNCompare(description,"IC",2)) {
        StrCopy(id, "IC");
        description += 2;
    }
    else if (description[0]=='B' && 0<=description[1] &&
         description[1]<='9') {

        if (method == LOOKUP_WEB)
            StrCopy(id, "Barnard");
        else
            StrCopy(id, "B");

        description += 1;
    }
    else if (NULL != (p = StrStr(description, "NGC"))) {
        StrCopy(id, ngcPrefix);
        description = p + 3;
    }
    else if (NULL != (p = StrStr(description, " IC")) ||
             NULL != (p = StrStr(description, "(IC"))) {
        StrCopy(id, "IC");
        id += 2;
        description = p + 3;
    }
    else if ((description[0] == 'C' ||
                  ( method == LOOKUP_DICT && description[0] == 'M')) &&
             TxtGlueCharIsDigit(description[1]) &&
             NULL != (p = StrChr(description, '(')) &&
             TxtGlueCharIsDigit(p[1])) {
        /* Caldwell catalog NGC number */
        StrCopy(id, ngcPrefix);
        description = p + 1;
    }
    else if (description[0]=='M' && method != LOOKUP_WEB) {
        StrCopy(id, "M");
        description++;
    }
    else if (TxtGlueCharIsDigit(description[0]) &&
             StrNCompare(description, "3C ", 3)) {
        StrCopy(id, ngcPrefix);
    }
    else {
        UInt16 i;

        if (method == LOOKUP_WEB)
            return false;

        i = 0;
        while(*description) {
            if (TxtGlueCharIsAlpha(*description) ||
                *description == '3') {
                id[i++] = TxtGlueUpperChar(*description++);
            }
            else {
                break;
            }
        }

        if (i==0) {
            StrCopy(id0, description0);
            return true;
        }

        id[i] = 0;
    }

    id += StrLen( id );

    while (*description == ' ') {
        description++;
    }

    if (!TxtGlueCharIsDigit(*description)) {
        if (method == LOOKUP_WEB)
            return false;
        StrCopy(id0, description0);
        return true;
    }

    digits = 0;

    do {
        *id++ = *description++;
        digits++;
    }
    while((TxtGlueCharIsDigit(*description) ||
               (*description =='.')) && digits<10);

    if (TxtGlueCharIsDigit(*description)) {
        if (method == LOOKUP_WEB)
            return false;
        StrCopy(id0, description0);
        return true;
    }

    if (TxtGlueCharIsAlpha(*description)) {
        /* This may be an alphabetic variant letter */
        *id++ = *description++;
    }
    *id = 0;
    return true;
}

/* URL must be at most 255 characters long */
static Boolean ADETAILSECTION
GetObjectLookupString(Char* url, UInt32 currentObj, 
    Char* description, LookupMethodType method)
{
    AstroRecordType art;
    Char id[FIELDLENGTH];

    if (currentObj < NUMPLANETS) {
        if (method == LOOKUP_DICT) {
            StrCopy(url, planetNames[currentObj]);
        }
        else {
            StrPrintF(url,"http://en.wikipedia.org/wiki/%s",
                planetNames[currentObj]);
        }
        return true;
    }

    if (description == NULL) {
        GetCatalogObject(currentObj - NUMPLANETS, &art, NULL);
        description = art.description;
    }

    if (GetObjectID(id, description, method)) {
      
        if (method == LOOKUP_DICT)
            StrCopy(url, id);
        else if (id[0] != 'B')
            StrPrintF(url,"http://spider.seds.org/ngc/ngc.cgi?%s",id);
        else
            StrPrintF(url,"http://simbad.harvard.edu/simbad/sim-id?Ident=%s",id);

        return true;
    }

    return false;
}


void ADETAILSECTION
LookupObject(UInt32 currentObj, Char* description, LookupMethodType method)
{
    LocalID id;
    Char*   s;
    Boolean success;

    if (method == LOOKUP_WEB)
        id = GetWebBrowser();
    else
        id = GetDictionary();

    if (id == NULL) {
        FrmAlert(method == LOOKUP_WEB ? a_NoBrowser : a_NoDict);
        return;
    }

    if (currentObj == NO_OBJECT && description == NULL)
        return;

    success = false;
    s = MemPtrNew( 256 );
    if (s != NULL) {
        if ( GetObjectLookupString(s, currentObj, description, method) ) {
            MemPtrSetOwner(s, 0);

            success = (errNone == SysUIAppSwitch(0, id,
                               sysAppLaunchCmdGoToURL, s));

            if (!success) {
                MemPtrFree(s);
            }
        }
    }
}

void ADETAILSECTION
ShowObjectDetail(UInt32 currentObj)
{
    static Char str[64];
    static Char format[64];
    Char flstr[3];
    Int32 gst, lst, sinlat, coslat;
    VectorType earth_celestial;
    VectorType radec2000, radec, altaz;
    MatrixType precession, radec2altaz;
    Int32 phi, psi;
    Int16   len;
    Coord x;
    AstroRecordType art;
    Coord y;
    Int32 illf, dist, semid, pa;
    Int16 mag;
    Int32 UTr, UTs;
    Int16 riseFlags;
    RGBColorType oldColor;
    UInt16 showAlign = 0;

    if (IS_FULLSCREEN)
        return;

    WinEraseRectangle(&detailRect, 0);
    WinEraseRectangle(&descRect, 0);

    if (currentObj == NO_OBJECT)
        return;

    if (CheckMinRomVersion(sysVersion35)) {
        RGBColorType back;

        WinGlueSetTextColorRGB(&(userPrefs.colors[GetColorMode()][COLOR_TEXT]),&oldColor);
        if (GetColorMode()==MODE_NIGHT) {
            back.r = 0;
            back.g = 0;    
            back.b = 0;
        }
        else {
            back.r = 255;
            back.g = 255;
            back.b = 255;
        }
        WinGlueSetBackColorRGB(&back,NULL);
    }

    gst = UTCToGST(userPrefs.jtime, currentGSTOffset);
    lst = GSTToLST(gst,userPrefs.Location[userPrefs.currentLoc].lng);
    PrecessionMatrix(userPrefs.jday, userPrefs.jtime, precession);
    MatRotate(radec2altaz,
              FP_ONE-userPrefs.Location[userPrefs.currentLoc].lat,
              FP_ONE + lst);
    fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

    if (currentObj < NUMPLANETS) {
        if (currentObj == SUN) {
	    HelioPlanetPos(EARTH, userPrefs.jday, userPrefs.jtime,
			   NULL, &dist, earth_celestial);
            /* TODO: semid and mag is dependent of distance sun->earth */
	    semid = fpdiv(GP(1919.26 / 2 / 40), dist);
            mag   = GP16(-26.7);
	    SolarPos(userPrefs.jday, userPrefs.jtime, radec2000);
	    GetStringResource(s_OurOwnStar, str);
        } else if (currentObj == MOON) {
	    Int32 EC, centerdist;
	    Int32 sinlst, coslst;
	    VectorType tmp, zenith;
	    LunarPos(userPrefs.jday, userPrefs.jtime, &EC, radec2000,
		     &illf, &mag, &dist, &semid, &pa);
	    /* adjust for topocentric moon.
	     */
	    fpsincos(lst, &sinlst, &coslst);
	    centerdist =
		fpdiv(FP(6378.14/384401),
		      fpmul(EC, FP_ONE + fpmul(fpsq(sinlat), 
					       FP(6378.14/6356.755 - 1))));
	    tmp[0] = fpmul(coslat, coslst);
	    tmp[1] = fpmul(coslat, sinlst);
	    tmp[2] = sinlat;
	    MatTranspose(precession);
	    MatVecMul(zenith, precession, tmp);
	    MatTranspose(precession);
	    radec2000[0] -= fpmul(zenith[0], centerdist);
	    radec2000[1] -= fpmul(zenith[1], centerdist);
	    radec2000[2] -= fpmul(zenith[2], centerdist);
	    GetMoonPhaseDescription(pa, str);
        } else {
	        HelioPlanetPos(EARTH, userPrefs.jday, userPrefs.jtime,
			   NULL, NULL, earth_celestial);
            PlanetPos(userPrefs.currentObj, userPrefs.jday, userPrefs.jtime,
		      earth_celestial,
                      radec2000, &illf, &mag, &dist, &semid, &pa);
            SysStringByIndex(stbl_SolarObjects, currentObj,
                             str, FIELDLENGTH);
        }
    } else {
        GetCatalogObject(currentObj - NUMPLANETS, &art, NULL);
        StrNCopy(str, art.description, FIELDLENGTH);

        mag = art.magnitude;
	    radec2000[0] = art.radec[0];
        radec2000[1] = art.radec[1];
        radec2000[2] = art.radec[2];
    }

    MatVecMul(radec, precession, radec2000);
    MatVecMul(altaz, radec2altaz, radec);

    if (currentObj < NUMPLANETS) {
        riseFlags = SolObjectRiseSet(currentObj,
                                     userPrefs.jday, userPrefs.jtime,
                                     sinlat, coslat,
                                     userPrefs.Location[userPrefs.currentLoc].lng,
                                     &UTr, &UTs);
    } else {
        riseFlags = ObjectRiseSet(userPrefs.jday, userPrefs.jtime,
                                  radec,
                                  sinlat, coslat,
                                  userPrefs.Location[userPrefs.currentLoc].lng,
                                  &UTr, &UTs);
    }


    // Show Object Name/Short Description
    FntSetFont(fontNormal);
    len = StrLen(str);
    x = descRect.topLeft.x + (descRect.extent.x - FntCharsWidth(str,len)) / 2;

    WinDrawChars(str, len, x, descRect.topLeft.y);
    y = detailRect.topLeft.y + detailRect.extent.y;
    WinDrawLine(detailRect.topLeft.x, y,
                detailRect.topLeft.x + detailRect.extent.x, y);

    /* Ra/Dec */
    if (userPrefs.detailPage <= 0
	&& userPrefs.detailPage + userPrefs.showDetailRows > 0) {
    if (userPrefs.radec2000) {
    	phi = fpatan2(radec2000[1], radec2000[0]);
    	psi = fpatan2(radec2000[2], fpsqrt(fpsq(radec2000[0]) + fpsq(radec2000[1])));
    }
    else {
    	phi = fpatan2(radec[1], radec[0]);
    	psi = fpatan2(radec[2], fpsqrt(fpsq(radec[0]) + fpsq(radec[1])));
    }
	SysStringByIndex(stbl_AngleFormatsHMS_sprintf,
			 userPrefs.angleFormat, format, FIELDLENGTH);
	FormatAngle(str, phi, format, false);
	ShowDetailCell(0, 0, s_Ra, str);

	SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			 userPrefs.angleFormat, format, FIELDLENGTH);
	FormatAngle(str, psi, format, true);
	ShowDetailCell(0, 1, s_Dec, str);
    }

    /* Alt/Az */
    if (userPrefs.detailPage <= 1
	&& userPrefs.detailPage + userPrefs.showDetailRows > 1) {

	phi = fpatan2(altaz[0], altaz[1]);
	psi = fpatan2(altaz[2], fpsqrt(FP_ONE - fpsq(altaz[2])));
	SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			 userPrefs.angleFormat, format, FIELDLENGTH);
	FormatAngle(str, psi, format, true);
	ShowDetailCell(1, 0, s_Alt, str);

	FormatAngle(str, phi, format, false);
	ShowDetailCell(1, 1, s_Az, str);
    }

    if (userPrefs.detailPage <= 2
	    && userPrefs.detailPage + userPrefs.showDetailRows > 2) {
         Char axis1[16];
         Char axis2[16];
         UInt16 axis1Label;
         UInt16 axis2Label;

         if (AlignGetStrings(&axis1Label, axis1, &axis2Label, axis2, altaz)) {
             showAlign = 1;
             ShowDetailCell(2, 0, axis1Label, axis1);
             ShowDetailCell(2, 1, axis2Label, axis2);
         }
    }

    if (userPrefs.detailPage <= 2+showAlign
	&& userPrefs.detailPage + userPrefs.showDetailRows > 2+showAlign) {

    if (mag == 0x7FFD)
        StrCopy(str,"--");
    else {
	    FormatGPNumber(GP16TO32(mag), 2, str, FIELDLENGTH);
    }

	ShowDetailCell(2+showAlign, 0, s_Mag, str);

	if (riseFlags & ASTROLIB_DONTRISE) {
	    GetStringResource(riseFlags & ASTROLIB_CIRCUMPOLAR
			      ? s_CirPl: s_NvrUp,
			      str);
	} else {
	    FormatTime(UTr + currentTzDST,  str, FIELDLENGTH, flstr, true);
	    StrCat(str, flstr);
	}
	ShowDetailCell(2+showAlign, 1, s_Rise, str);
    }

    if (userPrefs.detailPage <= 3+showAlign
	&& userPrefs.detailPage + userPrefs.showDetailRows > 3+showAlign) {

        if ( currentObj < NUMPLANETS ) {
  	    FormatGPNumber(FPTOGP(dist) * 40, 2, str, FIELDLENGTH);
            StrCat(str, " AU");
  	    ShowDetailCell(3+showAlign, 0, s_Dist, str);
        }
        else if ( 0 <= art.distance ) {
            FormatDistance(str, art.distance);
  	    ShowDetailCell(3+showAlign, 0, s_Dist, str);
        }

	if ((riseFlags & (ASTROLIB_DONTRISE | ASTROLIB_DONTSET))) {
	    GetStringResource(riseFlags & ASTROLIB_CIRCUMPOLAR ? s_CirPl: s_NvrUp,
			      str);
	} else {
	    FormatTime((UTr + UTs) / 2 + currentTzDST,
		       str, FIELDLENGTH, flstr, true);
	    StrCat(str, flstr);
	}
	ShowDetailCell(3+showAlign, 1, s_Trans, str);
    }

    if (userPrefs.detailPage <= 4+showAlign
	&& userPrefs.detailPage + userPrefs.showDetailRows > 4+showAlign) {
	if (riseFlags & ASTROLIB_DONTSET) {
	    GetStringResource(riseFlags & ASTROLIB_CIRCUMPOLAR ? s_CirPl: s_NvrUp,
			      str);
	} else {
	    FormatTime(UTs + currentTzDST,  str, FIELDLENGTH, flstr, true);
	    StrCat(str, flstr);
	}
	ShowDetailCell(4+showAlign, 1, s_Set, str);
    }

    if (userPrefs.detailPage <= 6+showAlign
	&& userPrefs.detailPage + userPrefs.showDetailRows > 4+showAlign) {
	if (currentObj < NUMPLANETS) {
	    FormatGPNumber(semid, 1, str, FIELDLENGTH);
	    StrCat(str, "\"");
	    ShowDetailCell(4+showAlign, 0, s_Semi, str);

	    if (currentObj == SUN) {
		/* Nothing special */
	    } else {
		GetStringResource(s_PhaseFormat, format);
		FormatAngle(str, pa, format, true);
		ShowDetailCell(5+showAlign, 0, s_Phase, str);

		FormatGPNumber(FPTOGP(illf) * 100, 2, str, FIELDLENGTH);
		StrCat(str, "%");
		ShowDetailCell(5+showAlign, 1, s_IllF, str);
	    }

	    if (currentObj == MOON) {
		/* pa is from -2..2, age should be from 0..100 */
		FormatGPNumber(FPTOGP(pa) * 25 + GP(50), 1, str, FIELDLENGTH);
		StrCat(str, "%");
		ShowDetailCell(6+showAlign, 0, s_Age, str);
	    } else {
		/* No calculation required */
		SysStringByIndex(stbl_DayLength, (UInt16)currentObj,
				 str, sizeof(str));
		StrCat(str," hrs");
		ShowDetailCell(6+showAlign, 0, s_SSDay, str);
	    }

	    SysStringByIndex(stbl_SolarMasses, (UInt16)currentObj,
			     str, sizeof(str));
	    StrCat(str," kg");
	    ShowDetailCell(6+showAlign, 1, s_SSMass, str);

	} else {
	    if (art.width != 0 || art.minoraxis != 0) {
		UInt16 wd = ((UInt32) art.width * (90L*600/16) + 0x8000) >> 16;
		if (art.width == art.minoraxis) {
		    StrPrintF(str, "%d.%d\'", wd / 10, wd % 10);
		} else {
		UInt16 hd = ((UInt32) art.minoraxis * (90L*600/16) + 0x8000) >> 16;
                StrPrintF(str, "%d.%d\'x%d.%d\'", wd / 10, wd % 10,
                          hd / 10, hd % 10);
		}

		ShowDetailCell(4+showAlign, 0, s_Size, str);
	    } else {
		ShowDetailCell(4+showAlign, 0, s_Size, "-");
	    }
	}
    }

    if (CheckMinRomVersion(sysVersion35))
        WinGlueSetTextColorRGB(&oldColor,NULL);
}


UInt16 ADETAILSECTION
GetMoonPhaseNumber(Int32 phase)
{
    return xlatarr[GP_ROUND((FPTOGP(phase) + GP(2)) * 5)];
}

/*
 * Get a textual description for the given moon age in percent.
 */
Char * ADETAILSECTION
GetMoonPhaseDescription(Int32 phase, Char *phaseName)
{
    /* Convert phase to name. */
    SysStringByIndex(stbl_MoonPhases,
                     GetMoonPhaseNumber(phase),
                     phaseName, FIELDLENGTH);
    return phaseName;
}


/* FormatAngle
 *
 * Function takes an angle or hour angle value and returns it as one
 * of a range of formatted strings.  Uses lookup into string table in
 * resource file.
 *
 * There are some smarts to automatically convert between angle, hour
 * angle and percentage.
 */
Char * ADETAILSECTION
FormatAngle(Char *str, Int32 angle, Char *format, Boolean isSigned)
{
    UInt8 fields;
    UInt32 percent;
    Char *sign;
    Char AngleType;
    UInt16 hour,minute,second;

    // Last digit in format string is number of fields in output
    fields = StrAToI(&format[StrLen(format)-1]);

    // Do smart conversion of values
    AngleType = format[StrLen(format)-2];
    
    if (isSigned) {
        sign = angle < 0 ? "-" : "";
        angle = fpabs(angle);
    } else {
        sign = "";
    }

    switch (AngleType) {
    case 'H':
        AngToHMS(angle, &hour,&minute,&second);
        break;        
        
    case 'D':
        AngToDMS(angle, &hour,&minute,&second);
        break;
            
    case 'P':
        percent = angle / (FP_ONE / 10000 * 4);
        /* Since we only use one field for percentage, we can
         * simply put the values here since it isn't manipulated
         * below.  
         */
        hour = percent / 100;
        minute = percent - hour * 100;
        break;
    }
    
    if (fields == 1) {
        if (AngleType == 'P') {
            // No massaging of number base as we already did that above
            StrPrintF(str, format, sign, hour, minute);
        } else {
            StrPrintF(str, format, sign, hour, (minute * 60 + second) * 5 / 18); // * 1000 / 3600
        }
    } else if (fields == 2) {
        StrPrintF(str, format, sign, hour, minute, second * 5 / 3); // * 100 / 60
    } else if (fields == 3) {
        StrPrintF(str, format, sign, hour, minute, second);
    }
    /* Remove the trailing *X1 */
    str[StrLen(str)-3] = '\0';
    HostTraceOutputTL(appErrorClass, 
                      "FormatAngle %hd,%hd,%hd - %s %hu %c = %s, %ud, %ud, %ud, %ud",
                      hour, minute, second, format, fields, AngleType, str, str[1], str[2], str[3], str[4]);

    return str;
}


/*
static Char * ADETAILSECTION
GetObjectType(UInt32 currentObj, Char * desc, Int16 size)
{
    AstroRecordType art;

    GetCatalogObject(currentObj, &art, NULL);
    GetStringTable(s_ObjectType, art.objClass, desc, size);
    return desc;
}
*/



void ADETAILSECTION
ShowDetailCell(Int8 y, Int8 x, UInt16 fieldnameid, Char *contents)
{
    Coord cx, cy, w;
    Char  str[FIELDLENGTH];
    Int16   len;

    /* Only draw it if it is within the display window. */
    if (y < userPrefs.detailPage ||
        y >= userPrefs.detailPage + userPrefs.showDetailRows)
        return;

    FntSetFont(fontSmall);
        
    cy = detailRect.topLeft.y
        + (y - userPrefs.detailPage) * (FntCharHeight() + userPrefs.spacingDetailRows);
    cx = detailRect.topLeft.x + x * screenRect.extent.x / 2 + 30;
    
    if (fieldnameid) {
        GetStringResource(fieldnameid, str);
        len = StrLen(str);
        w = FntCharsWidth(str, len);
        WinDrawChars(str, len, cx - w, cy);
    }
    
    len = StrLen(contents);
    WinDrawChars(contents, len, cx, cy);
}

