/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroEphem.c 935 2009-02-04 01:02:27Z arpruss $
 * Copyright (C) 2000, Michael Heinz
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "Astro.h"
#include "AstroEphem.h"
#include "AstroLib.h"
#include "AstroMatrix.h"
#include "FixedPoint.h"


/*
 * These formulas are all taken from:
 *
 * J. Simon et al.: Precession formulae and mean elements for the Moon
 * and the planets. Astronomy & Astrophysics 282, 663-683 (1994)
 *
 * Available at http://adc.gsfc.nasa.gov/ as document 6060.
 *
 * We implement formulas (18) section 6.3 with the mean elements referred
 * to the mean dynamical ecliptic and equinox J2000 section 5.8.
 *
 * The mean elements for Pluto are from the Explanatory Supplement to
 * the Astronomical Almanac, 1992, table 5.8.1.
 */

#define MILLENIUM 365250.0


/* Various scaling macros */
#define FP_DEG(x)    FP((x)/90.0)
#define FP_DEGSEC(x) FP_DEG((x)/3600.0)
#define FP_AU(x)     FP((x)/40.0)
#define FP_RAD(x)    FP((x)*2/3.14159265358979)

/* The periodic term constant */
#define MU FP_RAD(0.35953620 / MILLENIUM * (1L << 21))

/* longitude delta has two components.  This is a bit complicated*/
#define DELTALONG(x) \
   (Int32) (FP_DEGSEC(x)), \
   FP((x / 90.0 / 3600.0 * FP_ONE) - ((Int32) (FP_DEGSEC(x)))), \
   (FP(x / 90.0 / 3600.0 * FP_ONE / 86400000))


struct PlanetPeriodics {
    UInt32 p;   /* frequency (in 0.35953620 / Millenia) */
    Int32  C;   /* cosine factor */
    Int32  S;   /* sine factor */
};

struct PlanetOrbitals {
    Int32 a;         /* semi major axis */
    Int32 l;         /* mean longitude + delta */
    Int32 e;         /* eccentricity */
    Int32 p;         /* perihelion */
    Int32 i;         /* inclination */
    Int32 O;         /* ascending node */
};

#define PLANET_DATA_COUNT 9

struct PlanetConstants {
    Int32 a, da;         /* semi major axis */
    Int32 l, dl_hi, dl, dl_msec;  /* mean longitude */
    Int32 e, de;         /* eccentricity */
    Int32 p, dp;         /* perihelion */
    Int32 i, di;         /* inclination */
    Int32 O, dO;         /* ascending node */
    UInt8 de_shift, dp_shift, di_shift, dO_shift;
    struct PlanetPeriodics pa[8];  /* periodic pertubations axis */
    struct PlanetPeriodics pl[8];  /* periodic pertubations longitude */
};

const static struct PlanetConstants planetConsts[PLANET_DATA_COUNT] = {
  /* MERCURY */
  { FP_AU(0.3870983098),  FP_AU(0),
    FP_DEG(252.25090552), DELTALONG(5381016286.88982 / MILLENIUM),
    FP(0.2056317526),     FP(0.0002040653 / MILLENIUM * (1L << 30)),
    FP_DEG( 77.45611904), FP_DEGSEC( 5719.11590 / MILLENIUM * (1L << 25)),
    FP_DEG(  7.00498625), FP_DEGSEC( -214.25629 / MILLENIUM * (1L << 30)),
    FP_DEG( 48.33089304), FP_DEGSEC(-4515.21727 / MILLENIUM * (1L << 25)),
    0,5,0,5,
    { {  69613, FP_AU(     4e-7), FP_AU(   -29e-7) },
      {  75645, FP_AU(   -13e-7), FP_AU(    -1e-7) },
      {  88306, FP_AU(    11e-7), FP_AU(     9e-7) },
      {  59899, FP_AU(    -9e-7), FP_AU(     6e-7) },
      {  15746, FP_AU(    -9e-7), FP_AU(    -6e-7) },
      {  71087, FP_AU(    -3e-7), FP_AU(     5e-7) },
      { 142173, FP_AU(    -1e-7), FP_AU(     4e-7) },
      {   3086, FP_AU(     4e-7), FP_AU(     0e-7) }},
    { {   3086, FP_RAD(    21e-7), FP_RAD(  -342e-7) },
      {  15746, FP_RAD(   -95e-7), FP_RAD(   136e-7) },
      {  69613, FP_RAD(  -157e-7), FP_RAD(   -23e-7) },
      {  59899, FP_RAD(    41e-7), FP_RAD(    62e-7) },
      {  75645, FP_RAD(    -5e-7), FP_RAD(    66e-7) },
      {  88306, FP_RAD(    42e-7), FP_RAD(   -52e-7) },
      {  12661, FP_RAD(    23e-7), FP_RAD(   -33e-7) },
      {   2658, FP_RAD(    30e-7), FP_RAD(    17e-7) }},
  },
  /* VENUS */
  { FP_AU(0.7233298200),  FP_AU(0),
    FP_DEG(181.97980085), DELTALONG(2106641364.33548 / MILLENIUM),
    FP(0.0067719164),     FP(-0.0004776521 / MILLENIUM * (1L << 30)),
    FP_DEG(131.56370300), FP_DEGSEC(   175.48640 / MILLENIUM * (1L << 30)),
    FP_DEG(  3.39466189), FP_DEGSEC(   -30.84437 / MILLENIUM * (1L << 30)),
    FP_DEG( 76.67992019), FP_DEGSEC(-10008.48154 / MILLENIUM * (1L << 24)),
    0,0,0,6,
    { {  21863, FP_AU(  -156e-7), FP_AU(   -48e-7) },
      {  32794, FP_AU(    59e-7), FP_AU(  -125e-7) },
      {  26934, FP_AU(   -42e-7), FP_AU(   -26e-7) },
      {  10931, FP_AU(     6e-7), FP_AU(   -37e-7) },
      {  26250, FP_AU(    19e-7), FP_AU(    18e-7) },
      {  43725, FP_AU(   -20e-7), FP_AU(   -13e-7) },
      {  53867, FP_AU(   -10e-7), FP_AU(   -20e-7) },
      {  28939, FP_AU(   -12e-7), FP_AU(    -2e-7) }},
    { {  21863, FP_RAD(  -160e-7), FP_RAD(   524e-7) },
      {  32794, FP_RAD(  -313e-7), FP_RAD(  -149e-7) },
      {  10931, FP_RAD(  -235e-7), FP_RAD(   -35e-7) },
      {     73, FP_RAD(    60e-7), FP_RAD(   117e-7) },
      {   4387, FP_RAD(   -74e-7), FP_RAD(   151e-7) },
      {  26934, FP_RAD(   -76e-7), FP_RAD(   122e-7) },
      {   1473, FP_RAD(   -27e-7), FP_RAD(   -71e-7) },
      {   2157, FP_RAD(    34e-7), FP_RAD(   -62e-7) }},
  },
  /* EARTH-MOON */
  { FP_AU(1.0000010178),  FP_AU(0),
    FP_DEG(100.46645683), DELTALONG(1295977422.83429 / MILLENIUM),
    FP(0.0167086342),     FP(-0.0004203654 / MILLENIUM * (1L << 30)),
    FP_DEG(102.93734808), FP_DEGSEC(11612.35290 / MILLENIUM * (1L << 24)),
    FP_DEG(  0         ), FP_DEGSEC(  469.97289 / MILLENIUM * (1L << 28)),
    FP_DEG(174.87317577), FP_DEGSEC(-8679.27034 / MILLENIUM * (1L << 24)),
    0,6,2,6,
    { {  16002, FP_AU(    64e-7), FP_AU(  -150e-7) },
      {  21863, FP_AU(  -152e-7), FP_AU(   -46e-7) },
      {  32004, FP_AU(    62e-7), FP_AU(    68e-7) },
      {  10931, FP_AU(    -8e-7), FP_AU(    54e-7) },
      {  14529, FP_AU(    32e-7), FP_AU(    14e-7) },
      {  16368, FP_AU(   -41e-7), FP_AU(    24e-7) },
      {  15318, FP_AU(    19e-7), FP_AU(   -28e-7) },
      {  32794, FP_AU(   -11e-7), FP_AU(    22e-7) }},
    { {     10, FP_RAD(  -325e-7), FP_RAD(  -105e-7) },
      {  16002, FP_RAD(  -322e-7), FP_RAD(  -137e-7) },
      {  21863, FP_RAD(   -79e-7), FP_RAD(   258e-7) },
      {  10931, FP_RAD(   232e-7), FP_RAD(    35e-7) },
      {   1473, FP_RAD(   -52e-7), FP_RAD(  -116e-7) },
      {  32004, FP_RAD(    97e-7), FP_RAD(   -88e-7) },
      {   4387, FP_RAD(    55e-7), FP_RAD(  -112e-7) },
      {     73, FP_RAD(   -41e-7), FP_RAD(   -80e-7) }},
  },
  /* MARS */
  { FP_AU(1.5236793419),  FP_AU(3e-10 / MILLENIUM),
    FP_DEG(355.43299958), DELTALONG(689050774.93988 / MILLENIUM),
    FP(0.0934006477),     FP(0.0009048438 / MILLENIUM * (1L << 29)),
    FP_DEG(336.06023395), FP_DEGSEC( 15980.45908 / MILLENIUM * (1L << 23)),
    FP_DEG(  1.84972648), FP_DEGSEC(  -293.31722 / MILLENIUM * (1L << 29)),
    FP_DEG( 49.55809321), FP_DEGSEC(-10620.90088 / MILLENIUM * (1L << 24)),
    1,7,1,6,
    { {   6345, FP_AU(   124e-7), FP_AU(  -621e-7) },
      {   7818, FP_AU(   621e-7), FP_AU(   532e-7) },
      {  15636, FP_AU(  -145e-7), FP_AU(  -694e-7) },
      {   7077, FP_AU(   208e-7), FP_AU(   -20e-7) },
      {   8184, FP_AU(    54e-7), FP_AU(   192e-7) },
      {  14163, FP_AU(   -57e-7), FP_AU(   -94e-7) },
      {   1107, FP_AU(    30e-7), FP_AU(   -71e-7) },
      {   4872, FP_AU(    15e-7), FP_AU(    73e-7) }},
    { {     10, FP_RAD(  2268e-7), FP_RAD(   854e-7) },
      {   6345, FP_RAD(  -979e-7), FP_RAD(  -205e-7) },
      {   7818, FP_RAD(   802e-7), FP_RAD(  -963e-7) },
      {   1107, FP_RAD(   602e-7), FP_RAD(  -240e-7) },
      {  15636, FP_RAD(  -668e-7), FP_RAD(   140e-7) },
      {   7077, FP_RAD(   -33e-7), FP_RAD(  -341e-7) },
      {   8184, FP_RAD(   345e-7), FP_RAD(   -97e-7) },
      {    532, FP_RAD(   201e-7), FP_RAD(  -232e-7) }},
  },
  /* JUPITER */
  { FP_AU(5.2026032092),  FP_AU(19132e-10 / MILLENIUM),
    FP_DEG( 34.35151874), DELTALONG(109256603.77991 / MILLENIUM),
    FP(0.0484979255),     FP(0.0016322542 / MILLENIUM * (1L << 28)),
    FP_DEG( 14.33120687), FP_DEGSEC(  7758.75163 / MILLENIUM * (1L << 24)),
    FP_DEG(  1.30326698), FP_DEGSEC(   -71.55890 / MILLENIUM * (1L << 30)),
    FP_DEG(100.46440702), FP_DEGSEC(  6362.03561 / MILLENIUM * (1L << 25)),
    2,6,0,5,
    { {   1760, FP_AU(-23437e-7), FP_AU(-14614e-7) },
      {   1454, FP_AU( -2634e-7), FP_AU(-19828e-7) },
      {   1167, FP_AU(  6601e-7), FP_AU( -5869e-7) },
      {    880, FP_AU(  6259e-7), FP_AU(  1881e-7) },
      {    279, FP_AU( -1507e-7), FP_AU( -4372e-7) },
      {   2640, FP_AU( -1821e-7), FP_AU( -2255e-7) },
      {     19, FP_AU(  2620e-7), FP_AU(   782e-7) },
      {   2047, FP_AU( -2115e-7), FP_AU(   930e-7) }},
    { {     19, FP_RAD(  7610e-7), FP_RAD(-56980e-7) },
      {   1760, FP_RAD( -4997e-7), FP_RAD(  8016e-7) },
      {   1454, FP_RAD( -7689e-7), FP_RAD(  1012e-7) },
      {    287, FP_RAD( -5841e-7), FP_RAD(  1448e-7) },
      {   1167, FP_RAD( -2617e-7), FP_RAD( -3024e-7) },
      {    880, FP_RAD(  1115e-7), FP_RAD( -3710e-7) },
      {    574, FP_RAD(  -748e-7), FP_RAD(   318e-7) },
      {   2640, FP_RAD(  -607e-7), FP_RAD(   503e-7) }},
  },
  /* SATURN */
  { FP_AU(9.5549091915),  FP_AU(-0.0000213896 / MILLENIUM),
    FP_DEG( 50.07744430), DELTALONG(43996098.55732 / MILLENIUM),
    FP(0.0555481426),     FP(-0.0034664062 / MILLENIUM * (1L << 27)),
    FP_DEG( 93.05723748), FP_DEGSEC( 20395.49439 / MILLENIUM * (1L << 23)),
    FP_DEG(  2.48887878), FP_DEGSEC(    91.85195 / MILLENIUM * (1L << 30)),
    FP_DEG(113.66550252), FP_DEGSEC( -9240.19942 / MILLENIUM * (1L << 24)),
    3,7,0,6,
    { {    574, FP_AU( 62911e-7), FP_AU(139737e-7) },
      {      0,FP_AU(-119919e-7), FP_AU(     0e-7) },
      {    880, FP_AU( 79336e-7), FP_AU( 24667e-7) },
      {    287, FP_AU( 17814e-7), FP_AU( 51123e-7) },
      {     19, FP_AU(-24241e-7), FP_AU( -5102e-7) },
      {   1760, FP_AU( 12068e-7), FP_AU(  7429e-7) },
      {   1167, FP_AU(  8306e-7), FP_AU( -4095e-7) },
      {    306, FP_AU( -4893e-7), FP_AU( -1976e-7) }},
    { {     19, FP_RAD(-18549e-7), FP_RAD(138606e-7) },
      {    574, FP_RAD( 30125e-7), FP_RAD(-13478e-7) },
      {    287, FP_RAD( 20012e-7), FP_RAD( -4964e-7) },
      {    306, FP_RAD(  -730e-7), FP_RAD(  1441e-7) },
      {   1760, FP_RAD(   824e-7), FP_RAD( -1319e-7) },
      {     12, FP_RAD(    23e-7), FP_RAD( -1482e-7) },
      {     31, FP_RAD(  1289e-7), FP_RAD(   427e-7) },
      {     38, FP_RAD(  -352e-7), FP_RAD(  1236e-7) }},
  },
  /* URANUS */
  { FP_AU(19.2184460618),  FP_AU(-3716e-10 / MILLENIUM),
    FP_DEG(314.05500511), DELTALONG(15424811.93933 / MILLENIUM),
    FP(0.0463812221),     FP(-0.0002729293 / MILLENIUM * (1L << 30)),
    FP_DEG(173.00529106), FP_DEGSEC(  3215.56238 / MILLENIUM * (1L << 26)),
    FP_DEG(  0.77319689), FP_DEGSEC(   -60.72723 / MILLENIUM * (1L << 30)),
    FP_DEG( 74.00595701), FP_DEGSEC(  2669.15033 / MILLENIUM * (1L << 26)),
    0,4,0,4,
    { {    204, FP_AU(389061e-7),FP_AU(-138081e-7) },
      {      0,FP_AU(-262125e-7), FP_AU(     0e-7) },
      {    177, FP_AU(-44088e-7), FP_AU( 37205e-7) },
      {   1265, FP_AU(  8387e-7), FP_AU(-49039e-7) },
      {      4, FP_AU(-22976e-7), FP_AU(-41901e-7) },
      {    385, FP_AU( -2093e-7), FP_AU(-33872e-7) },
      {    200, FP_AU(  -615e-7), FP_AU(-27037e-7) },
      {    208, FP_AU( -9720e-7), FP_AU(-12474e-7) }},
    { {      4,FP_RAD(-135245e-7), FP_RAD( 71234e-7) },
      {    204, FP_RAD(-14594e-7), FP_RAD(-41116e-7) },
      {    177, FP_RAD(  4197e-7), FP_RAD(  5334e-7) },
      {      8, FP_RAD( -4030e-7), FP_RAD( -4935e-7) },
      {     31, FP_RAD( -5630e-7), FP_RAD( -1848e-7) },
      {    200, FP_RAD( -2898e-7), FP_RAD(    66e-7) },
      {   1265, FP_RAD(  2540e-7), FP_RAD(   434e-7) },
      {    102, FP_RAD(  -306e-7), FP_RAD( -1748e-7) }},
  },
  /* NEPTUN */
  { FP_AU(30.1103868694),  FP_AU(-16635e-10 / MILLENIUM),
    FP_DEG(304.34866548), DELTALONG(7865503.20744 / MILLENIUM),
    FP(0.0094557470),     FP(0.0000603263 / MILLENIUM * (1L << 30)),
    FP_DEG( 48.12027554), FP_DEGSEC(  1050.71912 / MILLENIUM * (1L << 27)),
    FP_DEG(  1.76995259), FP_DEGSEC(     8.12333 / MILLENIUM * (1L << 30)),
    FP_DEG(131.78405702), FP_DEGSEC(  -221.94322 / MILLENIUM * (1L << 29)),
    0,3,0,1,
    { {      0,FP_AU(-412235e-7), FP_AU(     0e-7) },
      {    102,FP_AU(-157046e-7), FP_AU( 28492e-7) },
      {    106, FP_AU(-31430e-7), FP_AU(133236e-7) },
      {      4, FP_AU( 37817e-7), FP_AU( 69654e-7) },
      {     98, FP_AU( -9740e-7), FP_AU( 52322e-7) },
      {   1367, FP_AU(   -13e-7), FP_AU(-49577e-7) },
      {    487, FP_AU( -7449e-7), FP_AU(-26430e-7) },
      {    204, FP_AU(  9644e-7), FP_AU( -3593e-7) }},
    { {      4, FP_RAD( 89948e-7), FP_RAD(-47645e-7) },
      {    102, FP_RAD(  2103e-7), FP_RAD( 11647e-7) },
      {    106, FP_RAD(  8963e-7), FP_RAD(  2166e-7) },
      {      8, FP_RAD(  2695e-7), FP_RAD(  3194e-7) },
      {     98, FP_RAD(  3682e-7), FP_RAD(   679e-7) },
      {   1367, FP_RAD(  1648e-7), FP_RAD(     0e-7) },
      {    487, FP_RAD(   866e-7), FP_RAD(  -244e-7) },
      {    204, FP_RAD(  -154e-7), FP_RAD(  -419e-7) }},
  },
  /* PLUTO */
  { FP_AU(39.48168677  ),  FP_AU(-0.0076912 / MILLENIUM), /* a */
    FP_DEG(238.92881   ), DELTALONG(5227479       / MILLENIUM),   /* l */
    FP(0.24880766  ),     FP(0.0006465 / MILLENIUM * (1L << 30)), /* e */
    FP_DEG(224.06676   ), FP_DEGSEC( -1322.5     / MILLENIUM * (1L << 27)), /* p */
    FP_DEG( 17.14175   ), FP_DEGSEC(   110.7     / MILLENIUM * (1L << 30)), /* i */
    FP_DEG(110.30347   ), FP_DEGSEC(  -373.3     / MILLENIUM * (1L << 29)), /* o */
    0,3,0,1,
    { {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) },
      {      0,FP_AU(      0e-7), FP_AU(     0e-7) }},
    { {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) },
      {      0, FP_RAD(     0e-7), FP_RAD(     0e-7) }}
  }
};

static struct PlanetOrbitals planetOrbits[PLANET_DATA_COUNT];
static Int32 lastJday;

static EPHEMSECTION void EphemSetDay(Int32 jday) {
    Int16 planet, i;
    if (lastJday == jday)
	return;

    lastJday = jday;
    jday -= J2000;

    for (planet = 0; planet < PLANET_DATA_COUNT; planet++) {
	const struct PlanetConstants *pl = &planetConsts[planet];
	struct PlanetOrbitals *orb = &planetOrbits[planet];
	Int32 mu = fpmul(jday << 9, MU);
	Int32 sin, cos;

	orb->a = pl->a + fpmul(pl->da, jday);
	orb->l = pl->l + pl->dl_hi * jday + fpmul(pl->dl, jday);
	orb->e = pl->e + fpmul(pl->de, jday << pl->de_shift);
	orb->p = pl->p + fpmul(pl->dp, jday << pl->dp_shift);
	orb->i = pl->i + fpmul(pl->di, jday << pl->di_shift);
	orb->O = pl->O + fpmul(pl->dO, jday << pl->dO_shift);

	for (i = 0; i < 8; i++) {
	    fpsincos(pl->pa[i].p * mu, &sin, &cos);
	    orb->a += fpmul(pl->pa[i].C, cos) + fpmul(pl->pa[i].S, sin);
	    fpsincos(pl->pl[i].p * mu, &sin, &cos);
	    orb->l += fpmul(pl->pl[i].C, cos) + fpmul(pl->pl[i].S, sin);
	}
    }
}

/*
 * Solve Kepler's Equation.  All units in fixed points.
 */
static EPHEMSECTION Int32
Kepler(Int32 meanAnomaly, Int32 ecc)
{
    Int32 E, dE, d, sine, cose;

    /* Kepler's equation is
     *
     *  E = meanAnomaly + ecc * sin(E)
     *
     * In the above equation, E is in radians but we need the solution
     * in fixedpoint angles. Therefore,
     *
     *  E = meanAnomaly + ecc * 2 / PI * sin(E)
     */

    ecc = fpmul(ecc, FP_2_PI);
    E = meanAnomaly;
    do {
	fpsincos(E, &sine, &cose);
        d  = E - fpmul(ecc,sine) - meanAnomaly;
        dE = fpdiv(d, FP_ONE - fpmul(ecc,cose));
        E  = E - dE;
    } while (fpabs(dE) > 100);
    return E;
}

static EPHEMSECTION void
OrbitalToCartesian(Int32 axis, Int32 lng, Int32 ecc,
                   Int32 per, Int32 inc, Int32 asc,
                   Int32 *phdist, VectorType celestial)
{
    Int32 e, sine, cose, dist;
    Int32 hdist, sinlasc, xy;
    Int32 sinv, cosv, sinl, cosl, sini, cosi, sino, coso, sinp, cosp;

    e = Kepler(lng - per, ecc);
    fpsincos(e,&sine,&cose);
    dist = FP_ONE - fpmul(ecc, cose);
    cosv = fpdiv(cose - ecc, dist);
    sinv = fpmul(sine, fpdiv(fpsqrt(FP_ONE - fpsq(ecc)), dist));

    fpsincos(per, &sinp, &cosp);
    sinl = fpmul(sinv, cosp) + fpmul(cosv, sinp);
    cosl = fpmul(cosv, cosp) - fpmul(sinv, sinp);

    hdist = fpmul(axis, dist);
    if (phdist)
        *phdist = hdist;
    
    fpsincos(asc, &sino, &coso);
    fpsincos(inc, &sini, &cosi);

    sinlasc = fpmul(sinl, coso) - fpmul(cosl, sino);
    xy = fpmul(sinlasc, FP_ONE - cosi);
    celestial[0] = fpmul(cosl + fpmul(sino, xy), hdist);
    celestial[1] = fpmul(sinl - fpmul(coso, xy), hdist);
    celestial[2] = fpmul(fpmul(sinlasc, sini)  , hdist);
}

void
HelioPlanetPos(Int16 planet, Int32 jday, Int32 jtime,
               Int32 *pmeanlng, Int32 *phdist,
	       VectorType celestial)
{
    struct PlanetOrbitals *orb = &planetOrbits[planet];
    Int32 lng;

    if (jday != lastJday)
	EphemSetDay(jday);

    jtime += getTDBOffset(jday);
    lng = orb->l + fpmul(planetConsts[planet].dl_msec, jtime);
    if (pmeanlng)
	*pmeanlng = lng - orb->p;

    OrbitalToCartesian(orb->a, lng, orb->e, orb->p, orb->i, orb->O,
                       phdist, celestial);
}

