/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroLib.c 929 2009-01-21 23:08:53Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "AstroLib.h"
#include "constants.h"
#include "FixedPoint.h"
#include "AstroMatrix.h"
#include "AstroNutation.h"
#include "AstroMoon.h"
#include "AstroEphem.h"

#define FP_AU(x)  FP(x/40.0)
#define FP_RAD(x) FP((x)*2/3.14159265358979)

#define SMALL_DISTANCE_SQ 50

/* Correction since we measure in 40 AU but planet mag is based on AU */
#define GP16_MAGCORRECTION GP16(16.02060) /* 5 * log(40*40) */

/* Factor for transforming fpln_16 results in magnitude scale */
#define GP16_MAGFACTOR     GP16(34.74356) /* 5 / fpln_16(10) */

#define GREGORIAN_CUTOVER 2299161L

/*
 * Returns the Julian Day of noon of the given day.
 *
 */
Int32 ALIBSECTION
JDay(Int16 year, Int16 month, Int16 day)
{
    Int32 jday;
  
    if (month<3) {
        month += 12;
        year --;
    }
    
    jday = ((1461 * (Int32) year) >> 2)
        + ((unsigned) (month + 1) * 153L) / 5 + day + 1720995L;
    
    if (jday >= GREGORIAN_CUTOVER) {
        Int16 century = (unsigned) year / 100;
        jday -= (3 * century - 5) >> 2;
    }

    return jday;
}

ALIBSECTION Int32
RadialPos(Int32 radperday, Int16 shift, Int32 jday, Int32 jtime)
{
    Int32 pos, frac, tmp;
    
    frac = fpmul(jtime, 0x636BA876);
    jday += frac >> 27;
    frac &= (1L << 27) - 1;

    pos = radperday * (jday >> shift);
    tmp = ((frac << 3) >> shift);
    frac = ((jday << (30 - shift)) & (FP_ONE - 1)) + tmp; 
    pos += fpmul(radperday, frac);
    return pos;
}

/* Tropic year in days  */
#define TROPYEAR  (31556926.0 / 86400.0)  
/* Sidereal day in days */
#define SIDDAY    (1.0 + 1/TROPYEAR)
/* Sidereal day in angle per millisecond */
#define SIDDAYANG (SIDDAY * 4 * FP_ONE / DAY)

/* Get the GST offset for noon of the given day.  The GST offset
 * is the difference between greenwich sidereal time and UTC in
 * fixed point angle.
 */
Int32 ALIBSECTION
getGSTOffset(Int32 jday)
{
    Int32 D;

    D = (jday - 2451545);
    return 0xC7704C27
        + ((D>>7) * FP(512 / TROPYEAR))
        + fpmul((D&0x7f) << 23, FP(512 / TROPYEAR));
}

/*
 * Convert a Julian Time to GST
 *
 * Note that utc is in milliseconds, while gst is in fixed point angle.
 */
Int32 ALIBSECTION
UTCToGST(Int32 utc, Int32 gstOffset)
{
    return (fpmul(utc, FP(SIDDAYANG / 32)) << 5) + gstOffset;
}

/*
 * Convert GST to UTC.
 * @param gst       the GST to convert.  In fixed point angle.
 * @param gstOffset the GST offset for noon (UT) of the current julian day.
 * @param jtime     the current julian time (UT) in seconds since noon.  The
 *                  returned time is put in the "vicinity" of jtime.
 * @return The time in UTC in seconds since noon.
 */
Int32 ALIBSECTION
GSTToUTC(Int32 gst, Int32 gstOffset, Int32 jtime)
{
    Int32 jtimeOffset = fpmul(jtime, FP(SIDDAYANG / 32)) << 5;
    return fpmul(gst - gstOffset - jtimeOffset, FP(1 / SIDDAYANG)) + jtime;
}

/*
 * Convert a gst and longitude to LST.
 * all values are in fixed point angle
 */
Int32 ALIBSECTION
GSTToLST(Int32 gst, Int32 lng)
{
    return gst + lng;
}

/*
 * Convert a LST and longitude to GST
 * all values are in fixed point angle
 */
Int32 ALIBSECTION
LSTToGST(Int32 lst, Int32 lng)
{
    return lst - lng;
}

/*
 * Convert fixed point angle to degrees, minutes, seconds
 *
 * For simplicity the angle is considered as unsigned, i.e. degree is
 * in range 0..359, not -180..180.  If you want to display negative
 * angles with a "-" sign you should handle that yourself.
 */
void ALIBSECTION
AngToDMS(Int32 rad, Int16 *d, Int16 *m, Int16 *s)
{
    /* convert to unsigned and add half a second to rad so we
     * round to nearest
     */
    UInt32 urad = rad + 1657;

    /* Multiply urad with 360/512 = 45/64 = 3/4 * 15/16.
     * Then degree is in high 9 bits
     */
    urad -= (urad >> 2);
    urad -= (urad >> 4);
    *d   = urad >> 23;
    urad &= 0x007fffff;

    /* Multiply urad with 60/4 = 15 = 5 * 3 */
    urad += urad << 2;
    urad += urad << 1;
    *m   = urad >> 21;
    urad &= 0x001fffff;

    /* Multiply urad with 60/4 = 15 = 5 * 3 */
    urad += urad << 2;
    urad += urad << 1;
    *s   = urad >> 19;
}


/*
 * Convert fixed point angle to hour, minutes, seconds
 *
 * Note that all values are positive.
 */
void ALIBSECTION
AngToHMS(Int32 rad, Int16 *h, Int16 *m, Int16 *s)
{
    /* Add half a second to rad so we round to nearest and convert to
       unsigned */
    UInt32 urad = rad + 24855;

    /* Multiply urad with 24/32 = 3/4;  Then hour is in high 5 bits */
    urad -= (urad >> 2);
    *h    = urad >> 27;
    urad &= 0x07ffffff;

    /* Multiply rad with 60/4 = 15 = 5 * 3 */
    urad += urad << 2;
    urad += urad << 1;
    *m   = urad >> 25;
    urad &= 0x01ffffff;

    /* Multiply urad with 60/4 = 15 = 5 * 3 */
    urad += urad << 2;
    urad += urad << 1;
    *s   = urad >> 23;
}

/*
 * Convert DMS to fixed point angle
 *
 * Note that the assumption is that for negative values of DMS, all
 * d, m and s are negative.
 */
Int32 ALIBSECTION
DMSToAng(Int16 d, Int16 m, Int16 s)
{
    return d * 11930464L + m * 198841L + s * 3314L;
}

/*
 * Convert HMS to fixed point angle
 *
 * Note that the assumption is that for negative values of DMS, all
 * d, m and s are negative.
 */
Int32 ALIBSECTION
HMSToAng(Int16 d, Int16 m, Int16 s)
{
    return d * 178956970L + m * 2982616L + s * 49710L;
}

/*
 * Convert R.A./Dec to Alt/Az
 *
 * All in fixed point.
 */
void ALIBSECTION
RaDecToAltAz(Int32 lat, Int32 lst, VectorType radec, VectorType altaz)
{
    MatrixType rd2aa;
    MatRotate(rd2aa, FP_ONE - lat, FP_ONE + lst);
    MatVecMul(altaz, rd2aa, radec);
}


/*
 * Convert celestial co-ordinates (lat,long)
 * to right ascension and declination.
 *
 * All angles in radians.
 */
void ALIBSECTION
CelestialToRaDec(VectorType celestial,
                 Int32 jday,
		 VectorType radec)
{
    radec[0] = celestial[0];
    radec[1] = fpmul(celestial[1],COSEPS0) - fpmul(celestial[2],SINEPS0);
    radec[2] = fpmul(celestial[1],SINEPS0) + fpmul(celestial[2],COSEPS0);
}

/*
 * Generate matrix to convert right ascenstion and declination coordinates of
 * J2000 into Ra/Dec coordinate for given date.  This adopts for precession of
 * earth.
 */
#define DELTA_HIGH(x) ((Int32) (x))
#define DELTA_LOW(x)  FP((x) - (Int32)(x))
#define MEAN_VALUE(a, b, jday, jtime) (FP(a) \
	+ DELTA_HIGH((b)*FP_ONE/365250) * ((jday) - J2000) \
	+ fpmul(DELTA_LOW((b)*FP_ONE/365250), (jday) - J2000) \
	+ fpmul(FP((b)*FP_ONE/365250/DAY), jtime))

void ALIBSECTION
PrecessionMatrix(Int32 jday, Int32 jtime, MatrixType precession) {
    Int32 eps_a, psi_a, chi_a, t;
    Int32 dEps, dPsi;
    MatrixType tmp1, tmp2;
    static Int32 lastday, lasttime;
    static MatrixType lastPrecession; 

    /* Calculate time since J2000 in millenia (365250 days) as 
     * fixed point.
     */
    t = fpmul((jday - J2000)<<11, FP((1L<<19) / 365250.0))
        + (fpmul(jtime, FP((1L<<30)/(365250.0*24*3600*1000)*(1L<<15))) >> 15);

    if (jday != lastday || jtime != lasttime) {
	/* Numerical expressions for precession formulae
	 * and mean elements for the Moon and the planets
	 * J.L. Simon, et.al.
	 * Astronomy & Astrophysics 282, 663-683
	 *
	 * omega_a = 23d26'21.412" - ...
	 * psi_a   = 50385".0672 t - 107".2374 t^2 - ...
	 * chi_a   = 105".5794 t - 238".1379 t^2 - ...
	 * where t in thousands of julian years.
	 */
	psi_a = MEAN_VALUE(0, 50385.0672/90/3600, jday, jtime);
	chi_a = MEAN_VALUE(0,   105.5794/90/3600, jday, jtime);
	eps_a = MEAN_VALUE((23+(26+21.412/60)/60)/90, 468.0927/90/3600, 
			   jday, jtime);
	Nutation2000b(jday, jtime, &dPsi, &dEps);
	MatRotate(tmp1, -eps_a-dEps, -dPsi);
	MatRotate(tmp2,  eps_a,      chi_a);
	MatMatMul(precession, tmp1, tmp2);
	MatRotate(tmp2, -EPS0J2000, -psi_a);
	MatMatMul(tmp1, precession, tmp2);
	MatRotate(tmp2,  EPS0J2000,      0);
	MatMatMul(lastPrecession, tmp1, tmp2);
	lastday = jday;
	lasttime = jtime;
    }

    MemMove(precession, lastPrecession, sizeof(MatrixType));
}

#ifdef OLD_EPHEM
/*
 * Solve Kepler's Equation.  All units in fixed points.
 */
static Int32 ALIBSECTION
Kepler(Int32 MeanAnomaly, Int32 Eccentricity)
{
    Int32 E, dE, d;

    
    Eccentricity = fpmul(Eccentricity, FP_2_PI);
    E = MeanAnomaly;
    d = fpmul(Eccentricity, fpsin(E));

    while (fpabs(d) > 32) {
        dE = fpdiv(d, FP_ONE - fpmul(Eccentricity,fpcos(E)));
        E  = E - dE;
        d  = E - fpmul(Eccentricity,fpsin(E)) - MeanAnomaly;
    }
    return E;
}
#endif

/*
 * Calculate the rising and setting times for an object.
 *
 * Times in LST.  If the object does not rise/set at this
 * latitude, returns -1, and LSTr and LSTs are not set.
 *
 * ra and dec are in radians.
 */
static Int16 ALIBSECTION
RiseSet(VectorType radec,
        Int32 sinlat, Int32 coslat,
        Int32 px, Int32 *LSTt, Int32 *H)
{
    Int32 coslatsq = fpsq(coslat);
    Int32 sindecsq = fpsq(radec[2]);
    Int32 D = coslatsq - sindecsq;

    *LSTt = fpatan2(radec[1], radec[0]);
        
    if (D < 0) {
        if ((radec[2] < 0) ^ (sinlat < 0)) {
            *H = 0;
            return ASTROLIB_DONTRISESET;
        } else {
            *H = -2*FP_ONE;
            return ASTROLIB_CIRCUMPOLAR | ASTROLIB_DONTRISESET;
        }
    } else {
        Int32 y;
        Int32 sinpx   = fpsin(IOR_EST - px);
        Int32 sinpxsq = fpsq(sinpx);
        Int32 cosdec  = fpsqrt(FP_ONE - sindecsq);
        
        /* Time shift for parallax and refraction */
        y = fpatan2(fpmul(sinpx, cosdec), 
                    fpsqrt(D - fpmul(1 - sindecsq, sinpxsq)));
        
        /*
         * Hour angle
         */
        *H = fpatan2(fpsqrt(D), -fpmul(sinlat, radec[2]))
            + fpmul(y, cosdec);

        /* always/never arise due to parallax */
        if (*H < 0)
            return (y > 0 ? ASTROLIB_CIRCUMPOLAR : 0) | ASTROLIB_DONTRISESET;

        return 0;
    }
}

/* We measure distances in 40 AU (approx the radius of pluto's orbit) */

/* angular semidiameter @ 40 AU (in arc seconds)*/
static const Int16 angsemidau[] = { 
    GP16(  3.36),
    GP16(  8.41),        // Based on atmosphere, rather than 8.34 for the crust
    0, /* Earth */
    GP16(  4.68),
    GP16( 98.44),
    GP16( 82.73),
    GP16( 35.02),
    GP16( 33.50),
    GP16(  2.70),
    GP16(959.63)

};

/* Visual magnitude at distance 1 AU from earth and sun. */
static const Int16 vmau[] = {
    GP16(-0.42),
    GP16(-4.40),
    0, /* Earth */
    GP16(-1.52),
    GP16(-9.40),
    GP16(-8.88),
    GP16(-7.19),
    GP16(-6.87),
    GP16(-1.00)
};

#ifdef OLD_EPHEM
/*
 * Data for planetary equations.
 * see astrolib.h for definitions of MERCURY, MARS, etc..
 */

/* Average speed of planet in radians per day */
static const Int8 radshift[] = {
    5, 6, 7, 8, 11, 12, 13, 14, 15
};
static const Int32 radperday[] = {
    0x5D1F9611,
    0x48EA27BD,
    0x59B65454,
    0x5F6522CE,
    0x7900739e,
    0x6169d58b,
    0x4452a1f0,
    0x45af7499,
    0x5D12458E
};

/* Longitude at epoch relative to perihelion (Radians)*/
static const Int32 elong[] = {
    0xf43b4ffc,
    0xe170c0b0,
    0xfdd203c1,
    0xbc591dfc,
    0x366078e7,
    0x8a8b780f,
    0x45d0d11b,
    0xa6a44222,
    0xfe10c8d5
};

/* longitude at perihelion (radians) */
// XXX This has to stay in global section as it is accessed in Astro.c
static const Int32 sinplong[] = {
    0x3e6f25c1,
    0x2ffc19ab,
    0x3e7d961a,
    0xe5d75f6c,
    0x0fab0199,
    0x3feb9294,
    0x07ed63ed,
    0x2f918d40,
    0xd36f63da
};
// XXX This has to stay in global section as it is accessed in Astro.c
static const Int32 cosplong[] = {
    0x0e1201be,
    0xd5a69516,
    0xf22eac1a,
    0x3a68f290,
    0x3e0d714b,
    0xfcce1b71,
    0xc07e2b91,
    0x2ad0f64a,
    0xd210c9a6
};

/* Eccentricity of the orbit */
// XXX This has to stay in global section as it is accessed in Astro.c
static const Int32 ecc[] = {
    0x0d291750,
    0x006f0cfe,
    0x0111d367,
    0x05fa3337,
    0x031a543f,
    0x038ea39c,
    0x02f6ec5b,
    0x00938152,
    0x0fc2656b
};

/* precalculated values of sqrt((1 + ecc) / (1 - ecc)); */
static const Int32 sqrtecc[] = {
    0x4ed87307,
    0x406f6dff,
    0x4114271a,
    0x4648d897,
    0x432e8dba,
    0x43a96a35,
    0x4309581a,
    0x40942cd8,
    0x524b132c
};

/* Semi-major axis of the orbit */
static const Int32 sma[] = {
    FP_AU( 0.3870986),
    FP_AU( 0.7233316),
    FP_AU( 1.0000000),
    FP_AU( 1.5236883),
    FP_AU( 5.202561),
    FP_AU( 9.554747),
    FP_AU(19.21814),
    FP_AU(30.10957),
    FP_AU(39.3414),
};

/* inclination of the orbit */
static const Int32 cosinc[] = {
    0x3f85b7c7,
    0x3fe34103,
    0x40000000,
    0x3ff7766c,
    0x3ffbc26b,
    0x3ff08b10,
    0x3ffe8238,
    0x3ff82d50,
    0x3d282d56
};
static const Int32 sininc[] = {
    0x07cdfef2, 
    0x03ca1d9a, 
    0x00000000, 
    0x0210d952, 
    0x0174bdfe, 
    0x02c78306, 
    0x00dd0e59, 
    0x01fa3e79, 
    0x12dd08d9, 
};

/* longitude of the ascending node (sinus/cosinus) */
static const Int32 sinasc[] = {
    0x2fb84e84,
    0x3e4148fd,
    0x00000000,
    0x30a6d563,
    0x3ef53f98,
    0x3aa86ad5,
    0x3d7f8abb,
    0x2fce82de,
    0x3c15cab1,
};
static const Int32 cosasc[] = {
    0x2aa5c07d,
    0x0ed7ca14,
    0x40000000,
    0x2994d68b,
    0xf47f8df1,
    0xe666ef3d,
    0x11b81d93,
    0xd5732503,
    0xe9f5aa61
};

#define PLUTOEPOCH 2447161

void ALIBSECTION
HelioPlanetPos(Int16 planet, Int32 jday, Int32 jtime, Int32 *ms,
               Int32 *phdist, VectorType celestial)
{
    Int32 ma, e, sine2, cose2, d, sinv2, cosv2, sinv, cosv;
    Int32 sinl, cosl, sinlasc, xy, heldist;

    /*
     * Mean Anomaly
     */
    ma = elong[planet] + 
        RadialPos(radperday[planet], radshift[planet], 
                  jday - (planet == PLUTO ? PLUTOEPOCH : EPOCH),
                  jtime - HALFDAY);
    if (ms)
        *ms = ma;
    
    e = Kepler(ma, ecc[planet]);
    fpsincos(e/2, &sine2, &cose2);
    d = fpsqrt(fpdiv(FP_ONE - ecc[planet], 
                     FP_ONE + fpmul(ecc[planet], FP_ONE - 2 * fpsq(cose2))));
    
    /*
     * True Anomaly
     */
    cosv2 = fpmul(cose2, d);
    sinv2 = fpmul(sine2, fpmul(sqrtecc[planet], d));
    
    sinv = 2*fpmul(sinv2,cosv2);
    cosv = 2*fpmul(cosv2,cosv2) - FP_ONE;
    
    cosl = fpmul(cosv, cosplong[planet])
	- fpmul(sinv, sinplong[planet]);
    sinl = fpmul(sinv, cosplong[planet])
	+ fpmul(cosv, sinplong[planet]);
    
    // MEEUS AA 30.2: r = sma * (1 - ecc * cos e)
    heldist = fpmul(sma[planet], FP_ONE - fpmul(ecc[planet], fpcos(e))); 
    if (phdist)
        *phdist = heldist;


    sinlasc = fpmul(sinl, cosasc[planet]) - fpmul(cosl, sinasc[planet]);
    xy = fpmul(sinlasc, FP_ONE - cosinc[planet]);

    /*
     * Heliocentric coordinates of planet
     * x coordinate points in direction of asc.
     */
    celestial[0]  = fpmul(cosl + fpmul(sinasc[planet], xy), heldist);
    celestial[1]  = fpmul(sinl - fpmul(cosasc[planet], xy), heldist);
    celestial[2]  = fpmul(fpmul(sinlasc, sininc[planet])  , heldist);
}
#endif

void ALIBSECTION
SolarLong(Int32 jday, Int32 jtime, 
               Int32 *ms, Int32 *sinl, Int32 *cosl) {
    VectorType celestial;
    HelioPlanetPos(EARTH, jday, jtime, ms, NULL, celestial);

    *sinl = -celestial[1];
    *cosl = -celestial[0];
}

void ALIBSECTION
SolarPos(Int32 jday, Int32 jtime, VectorType radec)
{
    VectorType celestial;
    /*
     * Longitude of the earth relative to sun
     */
    HelioPlanetPos(EARTH, jday, jtime, NULL, NULL, celestial);


    /* We want coordinates of Sun, not of earth, so mirror coordinates */
    /* Fixme: HelioPlanetPos calculates position of earth-*moon*-barycenter */
    celestial[0] = -celestial[0];
    celestial[1] = -celestial[1];
    celestial[2] = -celestial[2];
    VecNormalize(celestial);
    
    /*
     * ra and dec.
     */
    CelestialToRaDec(celestial, jday, radec);
}

/*
 * Calculate the position of a planet for a given
 * julian date.
 */
void ALIBSECTION
PlanetPos(Int16 planet, Int32 jday, Int32 jtime, 
	  VectorType earth_celestial,
          VectorType radec,
          Int32 *pillf, Int16 *pmag, Int32 *pgeodist, 
          Int32 *psemid, Int32 *ppa)
{
    Int32 cosi;
    VectorType celestial;
    Int32 heldist, earthdistsq, geodistsq;
    Int32 geoheldist;
    Int32 mag, geodist, pa, pasq;

    HelioPlanetPos(planet, jday, jtime, NULL, &heldist, celestial);

    /*
     * Heliocentric coordinates of planet
     * x coordinate points in direction of asc.
     */
    celestial[0] -= earth_celestial[0];
    celestial[1] -= earth_celestial[1];
    celestial[2] -= earth_celestial[2];

    geodistsq = fpsq(celestial[0]) + fpsq(celestial[1]) + fpsq(celestial[2]);
    *pgeodist = geodist = fpsqrt(geodistsq);

    /* normalize celestial vector */
    celestial[0] = fpdiv(celestial[0], geodist);
    celestial[1] = fpdiv(celestial[1], geodist);
    celestial[2] = fpdiv(celestial[2], geodist);

    /* cosine rule for triangle sun, planet, earth:
     * MEEUS AA 41.2: cosi = (heldist^2 + geodist^2 - Er^2) / (2*heldist*geodist)
     */
    geoheldist = fpmul(geodist, heldist);
    earthdistsq = fpsq(earth_celestial[0])
	+ fpsq(earth_celestial[1])
	+ fpsq(earth_celestial[2]);
    cosi = fpdiv((Int32) (fpsq(heldist) + geodistsq - earthdistsq) >> 1,
                 geoheldist);

    /* Illuminated fraction. */
    // MEEUS AA 41.1: k = (1 + cosi) / 2
    *pillf = ((FP_ONE/2) + (cosi>>1));

    /* phase angle */
    pa = fpatan2(fpsqrt(FP_ONE - fpsq(cosi)), cosi);

    /* Scale semidiameter based on the distance from us. 
     * The 40 is for our AU factor, the 2 for "semi".
     * MEEUS AA 55.1:
     */
    *psemid = fpdiv(GP16TO32(angsemidau[planet])/20, geodist);

    /* Relative visual magnitude and semidiameter from Earth.
     * MEEUS AA 41.d: 
     * mag = vmau[planet] + 5 * log(geodist * heldist) + c[planet] * pa
     */
    mag = vmau[planet] + GP16_MAGCORRECTION + fpmul(GP16_MAGFACTOR, fpln_16(geoheldist));

    pasq = fpsq(pa);

    switch (planet) {
    case MERCURY:
        mag += fpmul(GP16(0.0380 * 90), pa) - fpmul(GP16(0.000273 * 90), pasq) + fpmul(GP16(0.000002 * 90), fpmul(pasq, pa));
        break;
    case VENUS:
        mag += fpmul(GP16(0.0009 * 90), pa) + fpmul(GP16(0.000239 * 90), pasq) - fpmul(GP16(0.00000065 * 90), fpmul(pasq, pa));
    case MARS:
        mag += fpmul(GP16(0.016 * 90), pa);
        break;
    case JUPITER:
        mag += fpmul(GP16(0.005 * 90), pa);
        break;
    case SATURN:
            // TODO: Add ring brightness based on plane of rings
        break;
    default:
        break;
    }
    *pmag = mag;
    *ppa = fpmul(celestial[1],earth_celestial[0]) 
	- fpmul(celestial[0],earth_celestial[1]) > 0
	? pa : -pa;

    CelestialToRaDec(celestial, jday, radec);
    return;
}

/*
 * Returns position (Ra/Dec) and parallax for a solar system object.
 * @param solarID  one of SUN, MOON or MERCURY..PLUTO.
 */
static void ALIBSECTION 
SolObjectPosPx(Int16 solarID, Int32 jday, Int32 jtime,
	       VectorType radec, Int32 *px) {
    VectorType earth_celestial;
    if (solarID == SUN) {
        SolarPos(jday, jtime, radec);
        *px = SOLAR_P - SOLAR_A / 2;
    } else if (solarID == MOON) {
        Int32 illf, dist, semid, pa;
        Int32 EC;
        Int16 mag;
        LunarPos(jday, jtime, &EC, radec,
                 &illf, &mag, &dist, &semid, &pa);
        *px = fpmul(MOON_P - MOON_A / 2, 
                    fpdiv(FP_ONE + fpmul(MOON_E, fpcos(EC)),
                          FP_ONE - fpsq(MOON_E)));
    } else {
        Int32 illf, dist, semid, pa;
        Int16 mag;
            
        HelioPlanetPos(EARTH, jday, jtime, NULL, NULL, earth_celestial);
        PlanetPos(solarID, jday, jtime, earth_celestial, radec, 
		  &illf, &mag, &dist, &semid, &pa);
                  
        /* Parallax */
        *px = fpdiv(FP(4.263452e-5), dist);
    }
}


/*
 * Returns the times of moon or planet rise and moon set in UT.
 * @param solarID  one of SUN, MOON or MERCURY..PLUTO.
 */
Int16 ALIBSECTION
SolObjectRiseSet(Int16 solarID, Int32 jday, Int32 jtime,
                 Int32 sinlat, Int32 coslat, Int32 lng,
                 Int32 *pUTr, Int32 *pUTs)
{
    VectorType radec;
    Int32 px;
    Int16 i, retry;
    Int16 flags;
    Int32 lstt, H;
    Int32 UTr, UTs, UTt;
    Int32 gstOffset;

    gstOffset = getGSTOffset(jday);

    /*
     * First Approximation of position.
     */
    SolObjectPosPx(solarID, jday, jtime, radec, &px);

    /*
     * Initial guess at rise and set times.
     */
    flags = RiseSet(radec, sinlat, coslat, px, &lstt, &H);

    /*
     * First approximation of rise and set.
     */
    UTt  = GSTToUTC(lstt - lng, gstOffset, jtime);
    H    = fpmul(H, FP(1/SIDDAYANG));
    UTr  = UTt - H;
    UTs  = UTt + H;
    retry = 100;

    while ((flags & ASTROLIB_DONTRISE) == 0 && retry--) {
        jtime = UTr;

        /*
         * Improved position at rise time.
         */
        SolObjectPosPx(solarID, jday, jtime, radec, &px);
        
        /*
         * Improved guess at rise time.
         * We ignore the set time.
         */
        i = RiseSet(radec, sinlat, coslat, px, &lstt, &H);

        if (i & ASTROLIB_DONTRISE) {
            flags |= i & (ASTROLIB_CIRCUMPOLAR | ASTROLIB_DONTRISE);
            break;
        }
        
        UTr = GSTToUTC(lstt - H - lng, gstOffset, jtime);
        if (fpabs(UTr - jtime) < 15000)
            break;
    }

    *pUTr = UTr;
    retry = 100;
    
    while ((flags & ASTROLIB_DONTSET) == 0 && retry--) {
        jtime = UTs;
        
        /*
         * Improved position at set time.
         */
        SolObjectPosPx(solarID, jday, jtime, radec, &px);
        
        /*
         * Improved guess at set time.
         * We ignore the rise time.
         */
        i = RiseSet(radec, sinlat, coslat, px, &lstt, &H);

        if (i & ASTROLIB_DONTSET) {
            flags |= i & (ASTROLIB_CIRCUMPOLAR | ASTROLIB_DONTRISE);
            break;
        }

        UTs = GSTToUTC(lstt + H - lng, gstOffset, jtime);
        if (fpabs(UTs - jtime) < 15000)
            break;
    }
    *pUTs = UTs;

    return flags;
}

/*
 * Returns the times of object rise set in UT.
 */
Int16 ALIBSECTION
ObjectRiseSet(Int32 jday, Int32 jtime,
	      VectorType radec,
	      Int32 sinlat, Int32 coslat, Int32 lng,
	      Int32 *pUTr, Int32 *pUTs)
{
    Int16 flags;
    Int32 lstt, H, UTt;
    Int32 gstOffset;

    gstOffset = getGSTOffset(jday);

    flags =  RiseSet(radec, sinlat, coslat, 0,
                     &lstt, &H);
    UTt  = GSTToUTC(lstt - lng, gstOffset, jtime);
    H    = fpmul(H, FP(1/SIDDAYANG));
    *pUTr  = UTt - H;
    *pUTs  = UTt + H;
    return flags;
}


UInt32 ALIBSECTION
SkyDistance(VectorType radec1, VectorType radec2)
{
    UInt32 distSq;
    Int32  delta[3];

    distSq = fpsq(radec1[0]-radec2[0])+
             fpsq(radec1[1]-radec2[1])+
             fpsq(radec1[2]-radec2[2]);

    if (FP_ONE <= distSq)
        return FP_ONE;

    if (SMALL_DISTANCE_SQ <= distSq) 
        return fpsqrt(distSq);

    delta[0] = radec1[0] - radec2[0];
    delta[1] = radec1[1] - radec2[1];
    delta[2] = radec1[2] - radec2[2];
    if (delta[0] < 0)
        delta[0] = -delta[0];
    if (delta[1] < 0)
        delta[1] = -delta[1];
    if (delta[2] < 0)
        delta[2] = -delta[2];

    return fpsqrt( fpsq(delta[0]<<8) + fpsq(delta[1]<<8) + fpsq(delta[2]<<8) ) >> 8;
}

#ifdef SELFTEST
void ReportTest(Char *text);

#define TEST(TEXT, COND) \
  if (!(COND)) { \
    ReportTest("Test Failed: \n" TEXT "\n" \
               #COND "\nin " __FILE__ "\n"); \
  }

void LibTest() {
    Int32 jday;
    Int32  Er, Eang;
    VectorType radec, celestial;
    Int32 ra, dec;
    Int32 ill, dist, semid, pa;
    Int32 sinlat, coslat;
    Int32 UTr, UTs;
    Int32 dPsi, dEps;
    Int16 mag;

    jday = JDay(1980, 1, 1);
    Nutation2000b(jday, -10L*60*60*1000, &dPsi, &dEps);
    TEST("Nutation Psi", fpabs(dPsi - FP_RAD(-0.000038028)) < FP_RAD(1E-6));
    TEST("Nutation Eps", fpabs(dEps - FP_RAD(-0.000042514)) < FP_RAD(1E-6));

    jday = JDay(2001, 1, 1);
    TEST("Jday", jday == 2451911L);
    
    HelioPlanetPos(EARTH, jday, 0,
                   NULL, &Er, celestial);
    Eang = fpatan2(celestial[1], celestial[0]);
    TEST("Eang", fpabs(Eang - FP_RAD(1.765176906)) < FP_RAD(1E-6));
    TEST("Erad", fpabs(Er - FP_AU(0.9833002263)) < FP_AU(2E-7));

    PlanetPos(MERCURY, jday, 0, celestial, radec,
              &ill, &mag, &dist, &semid, &pa);
    ra  = fpatan2(radec[1], radec[0]);
    dec = fpatan2(radec[2], fpsqrt(FP_ONE - fpsq(radec[2])));

/*  FIXME: Recompute position of Mercury taking TDB - UTC into account
    TEST("Ppos"  , 
         fpabs(ra - FP_RAD(-1.280773616)) + fpabs(dec - FP_RAD(-.428803467)) < FP(6E-8));
    TEST("pgeodist" , fpabs(dist - FP_AU(1.417871756)) < FP_AU(6E-7));
*/
    TEST("Pmag"  , fpabs(mag - GP16(-1.05)) < GP16(0.01));
    TEST("Pill"  , fpabs(ill - FP(.9927)) < FP(.0001));
    TEST("Psemid", fpabs(semid - GP(4.74)) < GP(0.01));

    fpsincos(FP( 41.25 / 90), &sinlat, &coslat);
    SolObjectRiseSet(MOON, jday, 0, sinlat, coslat, FP(-75.50 / 90),
                     &UTr, &UTs);

    TEST("MoonRise", fpabs(UTr - ((16-12)*60+34)*60000L) < 60000L);
    TEST("MoonSet",  fpabs(UTs - (( 4+12)*60+22)*60000L) < 60000L);
}
#endif

