/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroLib.h 1006 2010-10-02 02:54:19Z arpruss $
 * Copyright (C) 2000, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef AstroLib_h
#define AstroLib_h

#include "AstroMatrix.h"

#define ALIBSECTION __attribute__ ((section ("code2")))

#define ASTROLIB_DONTRISE    1
#define ASTROLIB_DONTSET     2
#define ASTROLIB_DONTRISESET 3  /* dont rise and dont set */
#define ASTROLIB_CIRCUMPOLAR 4

#define DAY 86400000
#define HALFDAY 43200000

/* Julian day J2000 */
#define J2000 2451545

/* These values are from 
 * Numerical expressions for precession formulae
 * and mean elements for the Moon and the planets
 * J.L. Simon, et.al.
 * Astronomy & Astrophysics 282, 663-683
 */
#define EPS0J2000   FP((23 + (26 + (21.412) / 60) / 60) / 90)
#define SINEPS0     FP(.3977769958)
#define COSEPS0     FP(.9174821315)

#define JDAYEPOCH 2416481       /* jday for Palm OS's epoch 1904-01-01
                                 * Do not forgot to subtract half a day.
                                 */


/*
 * Returns the Julian Day.
 *
 * This formula is valid for all dates from 1 to 9999
 */
Int32
JDay(Int16 year, Int16 month, Int16 day) ALIBSECTION;

Int32
RadialPos(Int32 radperday, Int16 shift, Int32 jday, Int32 jtime) ALIBSECTION;

Int32 getGSTOffset(Int32 jday) ALIBSECTION;

/*
 * Get the difference TDB - UT in milliseconds.
 */
static __inline__ Int32 getTDBOffset(Int32 jday)
{
    /* Very simplistic: We assume TDB = TT and difference of TAI and UTC (the
     * number of leap seconds) of 33 seconds.
     */
    return 33000 + 32184;
}

/*
 * Convert a Julian day to GST
 */
Int32 UTCToGST(Int32 utc, Int32 gstOffset) ALIBSECTION;

/*
 * Convert a gst and longitude to LST
 * longitude is in radians.
 */
Int32 GSTToLST(Int32 gst, Int32 lng) ALIBSECTION;

/*
 * Convert GST to UTC.
 * @param gst       the GST to convert.  In fixed point angle.
 * @param gstOffset the GST offset for noon (UT) of the current julian day.
 * @param jtime     the current julian time (UT) in seconds since noon.  The
 *                  returned time is put in the "vicinity" of jtime.
 * @return The time in UTC in seconds since noon.
 */
Int32 GSTToUTC(Int32 gst, Int32 gstOffset, Int32 jtime) ALIBSECTION;

/*
 * Convert a LST and longitude to GST.
 * longitude is in radians.
 */
Int32 LSTToGST(Int32 lst, Int32 lng) ALIBSECTION;


/*
 * Convert radians to degrees, minutes, seconds
 *
 * Note that for negative values of rad, all h, m and s will be negative.
 */
void AngToDMS(Int32 rad,
              Int16 *h, Int16 *m, Int16 *s) ALIBSECTION;

/*
 * Convert radians to hours, minutes, seconds
 */
void AngToHMS(Int32 rad,
              Int16 *h, Int16 *m, Int16 *s) ALIBSECTION;

/*
 * Convert DMS to decimal degrees. Note that for negative values, the
 * assumption is that if d, m and s are all negative.
 */
Int32 DMSToAng(Int16 d, Int16 m, Int16 s) ALIBSECTION;

/*
 * Convert HMS to fixed point angle
 *
 * Note that the assumption is that for negative values of DMS, all
 * d, m and s are negative.
 */
Int32 HMSToAng(Int16 h, Int16 m, Int16 s) ALIBSECTION;

/*
 * Convert R.A./Dec to Alt/Az
 *
 * ra = Right Ascension, dec = Declination
 * lst = observer's Local Sidereal Time, lat = observer's Latitude
 * alt = the final altitude, az = the final azimuth.
 *
 * All values in fixed point.  Note that the three coordinates
 * correspond to x, y, z in cartesian coordinates.
 */
void RaDecToAltAz(Int32 lat, Int32 lst, 
		  VectorType radec, VectorType altaz) ALIBSECTION;

/*
 * Convert celestial co-ordinates (lat,long)
 * to right ascension and declination.
 *
 * All values in fixed point.  Note that the three coordinates
 * correspond to x, y, z in cartesian coordinates.
 */
void CelestialToRaDec(VectorType celestial, 
		      Int32 jday, VectorType radec) ALIBSECTION;
/*
 * Return the longitude of the Sun for a given julian day.
 * If ms is !NULL then returns the mean anomoly, too.
 */
void SolarLong(Int32 jday, Int32 jtime,
	       Int32 *ms, Int32 *cosl, Int32 *sinl) ALIBSECTION;


void SolarPos(Int32 jday, Int32 jtime, VectorType radec) ALIBSECTION;

/*
 * Calculate the rising and setting times for the given solar system
 * object at the given (local) julian day and time.  All times are in
 * milliseconds UTC since local noon; negative before noon.
 *
 * The rise/set time are matching, i.e. UTr < UTs and (UTr+UTs)/2
 * denotes zenith transition time.  The chosen rise/set pair is the one,
 * whose transition time is nearest to the given time.
 * 
 * The rise/set time can be less or greater than -HALFDAY or HALFDAY
 * denoting the previous or next day.
 *
 * The return value is 0 if everthing is okay, or a bitmask of 
 * ASTROLIB_DONTRISE, ASTROLIB_DONTSET and ASTROLIB_CIRCUMPOLAR.
 * There are some cases where an object changes from circumpolar to
 * normal, in that case it can have a valid rise but an invalid set
 * time.  This is indicated by (ASTROLIB_CIRCUMPOLAR|ASTROLIB_DONTRISE). 
 *                                              
 * @param solarID One of MOON, SUN or MERCURY..PLUTO, but not EARTH :).
 * @param jday    The julian day.
 * @param jtime   The time in milliseconds since noon of jday (UTC).
 * @param sinlat  The observers latitude (sine).
 * @param coslat  The observers latitude (cosine).
 * @param lng     The observers longitude (fixed point angle).
 * @param UTr     pointer where rise time is stored.
 * @param UTs     pointer where set time is stored.
 * @return flags to indicate if rise/set make sense (see above).
 */
Int16
SolObjectRiseSet(Int16 solarID, Int32 jday, Int32 jtime,
                 Int32 sinlat, Int32 coslat, Int32 lng,
                 Int32 *UTr, Int32 *UTs) ALIBSECTION;

/*
 * Calculate the rising and setting times for the given Ra/Dec coords
 * at the given (local) julian day and time.  All times are in
 * milliseconds UTC since local noon; negative before noon.
 *
 * The rise/set time are matching, i.e. UTr < UTs and (UTr+UTs)/2
 * denotes zenith transition time.  The chosen rise/set pair is the one,
 * whose transition time is nearest to the given time.
 * 
 * The rise/set time can be less or greater than -HALFDAY or HALFDAY
 * denoting the previous or next day.
 *
 * The return value is 0 if everthing is okay, or a bitmask of 
 * ASTROLIB_DONTRISE, ASTROLIB_DONTSET and ASTROLIB_CIRCUMPOLAR.
 * There are some cases where an object changes from circumpolar to
 * normal, in that case it can have a valid rise but an invalid set
 * time.  This is indicated by (ASTROLIB_CIRCUMPOLAR|ASTROLIB_DONTRISE). 
 *                                              
 * @param jday    The julian day.
 * @param jtime   The time in milliseconds since noon of jday (UTC).
 * @param radec   The right ascension/declination vector.
 * @param sinlat  The observers latitude (sine).
 * @param coslat  The observers latitude (cosine).
 * @param lng     The observers longitude (fixed point angle).
 * @param UTr     pointer where rise time is stored.
 * @param UTs     pointer where set time is stored.
 * @return flags to indicate if rise/set make sense (see above).
 */
Int16
ObjectRiseSet(Int32 jday, Int32 jtime, VectorType radec,
	      Int32 sinlat, Int32 coslat, Int32 lng,
	      Int32 *pUTr, Int32 *pUTs) ALIBSECTION;

/*
 * Planet numbers
 */
#define MERCURY         0
#define VENUS           1
#define MOON            2
#define EARTH           2  /* for planet orbits */
#define MARS            3
#define JUPITER         4 
#define SATURN          5
#define URANUS          6
#define NEPTUNE         7
#define PLUTO           8
#define SUN             9
//#define JUPITER1        10
//#define JUPITER2        11
//#define JUPITER3        12
//#define JUPITER4        13

#define NUMPLANETS      10
//#define NUMJUPITERMOONS  4

/*
void 
HelioPlanetPos(Int16 planet, Int32 jday, Int32 jtime, 
	       Int32 *ms, Int32 *sinl, Int32 *cosl, Int32 *ecc) ALIBSECTION;
*/


/*
 * Calculate the position of a planet for a given
 * julian date. Returns the right ascension and
 * declination in radians. planet is the planet number
 * listed above.
 */
void
PlanetPos(Int16 planet, Int32 jday, Int32 jtime, 
	  VectorType earth_celestial,
          VectorType radec,
          Int32 *pillf, Int16 *pmag, Int32 *pgdist, 
          Int32 *psemid, Int32 *ppa) ALIBSECTION;

/*
 * Generate matrix to convert right ascenstion and declination coordinates of
 * J2000 into Ra/Dec coordinate for given date.  This adopts for precession of
 * earth.
 */
void
PrecessionMatrix(Int32 jday, Int32 jtime, MatrixType precession) ALIBSECTION;

UInt32 SkyDistance(VectorType radec1, VectorType radec2) ALIBSECTION;

typedef struct _AstroRecord
{
    Char description[32]; // max 32 chars
    short raH, raM, raS;
    short decH, decM, decS;
    Char magnitude[4]; // max 4 chars
    Char size[8]; // max 8 chars
} AstroRecord;

#define DMS_TO_ANG(d,m,s)   ((d) * 11930464L + (m) * 198841L + (s) * 3314L)


#ifdef SELFTEST
void LibTest(void);
#endif

#endif    // AstroLib_h
