/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroList.c 971 2010-04-12 22:38:03Z hoenicke $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#include <PalmOS.h>
#include <DateGlue.h>
#include <TxtGlue.h>
#include <WinGlue.h>
#include <FrmGlue.h>
#include <HostControl.h>
#include "PalmUtil.h"
#include "resource.h"
#include "Astro.h"
#include "AstroLib.h"
#include "AstroCat.h"
#include "AstroMoon.h"
#include "AstroSky.h"
#include "AstroQTree.h"
#include "AstroEphem.h"
#include "FixedPoint.h"
#include "AstroTime.h"
#include "AstroList.h"
#include "AstroBW.h"

#define MAX_LIST 32767
#define MAX_PATH   120

#define LIST_DBID 'list'

#define CATALOG_NAME_PREFIX "AstroInfo "
#define LIST_NAME_PREFIX "AstroInfo*"
#define LIST_NAME_PREFIX_LENGTH 10

#define LIST_PREF_ID 'Lp'

typedef enum {
    LIST_MORE_ADD_CATALOG = 0,
    LIST_MORE_ARRANGE_ALTAZ,
    LIST_MORE_ARRANGE_EQ,
    LIST_MORE_COPY_CLIPBOARD,
    LIST_MORE_COPY_CLIPBOARD_WITH_NOTES,
    LIST_MORE_COPY_NEW_LIST,
    LIST_MORE_SORT_ALT,
    LIST_MORE_SORT_CHECKED,
    LIST_MORE_SORT_CHECKED_TIME,
    LIST_MORE_SORT_MAG,
    LIST_MORE_SORT_NAME,
    LIST_MORE_SORT_RA,
    LIST_MORE_SORT_RISESET
} ListMoreActionType;

typedef enum {
     CHECKMARK_NO = 0,
     CHECKMARK_YES,
     CHECKMARK_QUERY
} ListCheckmarkType;

#define NUM_CHECKMARKS 3

typedef struct {
    Char        dbName[dmDBNameLength];
    Int16       selected;
} ListPrefType;

/* size must be even */
typedef struct {
    UInt32  catalogNumber;
    Char    objectName[FIELDLENGTH];
    Char    catalogName[FIELDLENGTH];
    Boolean checked;
    UInt8   reserved[3];
} BasicListItemType;

/* size must be even */
typedef struct {
    BasicListItemType basic;
    UInt32 localTime;
    Int32  jtime;
    Int32  jday;
    Char   location[FIELDLENGTH];  // currently holds city name
    Char   note[0];
} ExtendedListItemType;

typedef struct {
    Int32 phi;
    Int32 psi;
} Angles;

static UInt32* sortObjectList;
static DmOpenRef listRef = NULL;
static ListPrefType listPref;
static const Char listNamePrefix[] = LIST_NAME_PREFIX;
static Char defaultDBName[dmDBNameLength];
static UInt16 numItems = 0;
static Int32  phiSun = 0;

static UInt32 LISTSECTION
ListGetObject(BasicListItemType* item);


static void LISTSECTION
ListUpdateListName(Char* name, Boolean draw)
{
    static Char listName[FIELDLENGTH];
    ControlType* c = GetObjectPtr(tr_ListName);

    StrNCopy(listName, name, FIELDLENGTH);
    listName[FIELDLENGTH-1] = 0;

    CtlSetLabel(c, listName);
    if (draw)
        CtlDrawControl(c);
}



static Boolean LISTSECTION
ListGetRaDec(BasicListItemType* item, VectorType radec, Boolean j2000)
{
    VectorType radec2000;
    VectorType earth_celestial;
    MatrixType precession;
    Int32 illf, dist, semid, pa;
    Int16 mag;

    PrecessionMatrix(userPrefs.jday, userPrefs.jtime, precession);

    if (item->catalogName[0] == 0) {
        if (item->catalogNumber == SUN) {
	        SolarPos(userPrefs.jday, userPrefs.jtime, radec2000);
                /* TODO: semid and mag is dependent of distance sun->earth */
        } else if (item->catalogNumber == MOON) {
    	    Int32 EC, centerdist;
    	    Int32 sinlst, coslst;
    	    VectorType tmp, zenith;
            Int32 gst, lst, sinlat, coslat;

            gst = UTCToGST(userPrefs.jtime, currentGSTOffset);
            lst = GSTToLST(gst,userPrefs.Location[userPrefs.currentLoc].lng);

            fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

    	    LunarPos(userPrefs.jday, userPrefs.jtime, &EC, radec2000,
    		     &illf, &mag, &dist, &semid, &pa);
    	    /* adjust for topocentric moon.
    	     */
    	    fpsincos(lst, &sinlst, &coslst);
    	    centerdist =
    		fpdiv(FP(6378.14/384401),
    		      fpmul(EC, FP_ONE + fpmul(fpsq(sinlat),
    					       FP(6378.14/6356.755 - 1))));
    	    tmp[0] = fpmul(coslat, coslst);
    	    tmp[1] = fpmul(coslat, sinlst);
    	    tmp[2] = sinlat;
    	    MatTranspose(precession);
    	    MatVecMul(zenith, precession, tmp);
    	    MatTranspose(precession);
    	    radec2000[0] -= fpmul(zenith[0], centerdist);
    	    radec2000[1] -= fpmul(zenith[1], centerdist);
    	    radec2000[2] -= fpmul(zenith[2], centerdist);
        } else {
    	    HelioPlanetPos(EARTH, userPrefs.jday, userPrefs.jtime,
    			   NULL, NULL, earth_celestial);
            PlanetPos(item->catalogNumber, userPrefs.jday, userPrefs.jtime,
		        earth_celestial, radec2000, &illf, &mag, &dist, &semid, &pa);
        }
    } else {
        static AstroRecordType art;
        UInt16 i;
        UInt32 qtreeIndex;
        MemHandle handle;

        for (i = 0 ; i < numDBs; i++)
             if (catDBs[i]->isVisible &&
                      !StrCompare(catDBs[i]->name,item->catalogName))
                 break;

        if (i == numDBs)
            return false;

      	handle = CatalogGetHandle(catDBs[i], 0);
      	ErrFatalDisplayIf(handle == NULL,"Could not retrieve record!");
      	qtreeIndex = ((UInt16*)MemHandleLock(handle))[item->catalogNumber];
      	MemHandleUnlock(handle);
      	CatalogFreeHandle(catDBs[i], handle);

      	QTreeGetObject(catDBs[i], qtreeIndex, &art);
/*        UInt32 object;

        object = ListGetObject(item);

        if (object == NO_OBJECT)
            return false;

        GetCatalogObject(object - NUMPLANETS, &art, NULL); */

      	MemMove(radec2000, art.radec, sizeof(VectorType));
    }

    if (j2000)
        MemMove(radec, radec2000, sizeof(VectorType));
    else
        MatVecMul(radec, precession, radec2000);
    return true;
}



static Boolean LISTSECTION
ListGetMag(BasicListItemType* item, Int16* magP)
{
    VectorType radec2000;
    VectorType earth_celestial;
    Int32 illf, dist, semid, pa;

    if (item->catalogName[0] == 0) {
        if (item->catalogNumber == SUN) {
	        SolarPos(userPrefs.jday, userPrefs.jtime, radec2000);
            *magP = GP16(-26.7);
                /* TODO: semid and mag is dependent of distance sun->earth */
        } else if (item->catalogNumber == MOON) {
    	    Int32 EC;
            Int32 gst, lst, sinlat, coslat;

            gst = UTCToGST(userPrefs.jtime, currentGSTOffset);
            lst = GSTToLST(gst,userPrefs.Location[userPrefs.currentLoc].lng);

            fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

    	    LunarPos(userPrefs.jday, userPrefs.jtime, &EC, radec2000,
    		     &illf, magP, &dist, &semid, &pa);
        } else {
    	    HelioPlanetPos(EARTH, userPrefs.jday, userPrefs.jtime,
    			   NULL, NULL, earth_celestial);
            PlanetPos(item->catalogNumber, userPrefs.jday, userPrefs.jtime,
		        earth_celestial, radec2000, &illf, magP, &dist, &semid, &pa);
        }
    } else {
        static AstroRecordType art;
        UInt16 i;
        UInt32 qtreeIndex;
        MemHandle handle;

        for (i = 0 ; i < numDBs; i++)
             if (catDBs[i]->isVisible &&
                      !StrCompare(catDBs[i]->name,item->catalogName))
                 break;

        if (i == numDBs)
            return false;

      	handle = CatalogGetHandle(catDBs[i], 0);
      	ErrFatalDisplayIf(handle == NULL,"Could not retrieve record!");
      	qtreeIndex = ((UInt16*)MemHandleLock(handle))[item->catalogNumber];
      	MemHandleUnlock(handle);
      	CatalogFreeHandle(catDBs[i], handle);

      	QTreeGetObject(catDBs[i], qtreeIndex, &art);

      	*magP = art.magnitude;
    }
    
    return true;
}



static Boolean LISTSECTION
ListItemRiseSet(BasicListItemType* item, Int32* UTrP, Int32* UTsP, Int16* flagsP, Int16* upnowP)
{
    VectorType radec;

    if (!item->catalogName[0] || ListGetRaDec(item, radec, false)) {
        Int32  sinlat, coslat;

        fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

        if (!item->catalogName[0])
            *flagsP = SolObjectRiseSet(item->catalogNumber,
                                         userPrefs.jday, userPrefs.jtime,
                                         sinlat, coslat,
                                         userPrefs.Location[userPrefs.currentLoc].lng,
                                         UTrP, UTsP);
        else
            *flagsP = ObjectRiseSet(userPrefs.jday, userPrefs.jtime,
                                  radec,
                                  sinlat, coslat, userPrefs.Location[userPrefs.currentLoc].lng,
                                  UTrP, UTsP);

        /* Prepend a special symbol if object is up or circumpolar */
        *upnowP = UpNow(*flagsP, *UTrP, *UTsP, userPrefs.jtime);

        return true;
    }
    else {
        return false;
    }
}



static void LISTSECTION
ListDrawItem(Int16 itemNum, RectangleType* bounds, Char** itemsText)
{
    WinEraseRectangle(bounds, 0);

    if (listRef != NULL) {
        MemHandle h;

        h = DmQueryRecord(listRef, itemNum);

        if (h!=NULL) {
            FontID oldFont;
            BasicListItemType* item;
            Coord x;
            Coord checkWidth;
            Char text[2];

            item = MemHandleLock(h);

            x = bounds->topLeft.x;

            oldFont = FntSetFont( symbol11Font );
            checkWidth = FntCharWidth(1);

            text[1] = 0;

            switch(item->checked) {
                case CHECKMARK_YES:
                    text[0] = 1;
                    WinDrawChars(text, 1, x, bounds->topLeft.y);
                    break;
                case CHECKMARK_NO:
                    text[0] = 0;
                    WinDrawChars(text, 1, x, bounds->topLeft.y);
                    break;
                case CHECKMARK_QUERY:
                    FntSetFont(boldFont);
                    text[0] = '?';
                    WinDrawChars(text, 1, x + (checkWidth - FntCharWidth('?'))/2, bounds->topLeft.y);
                    break;
                default:
                    break;
            }

            FntSetFont(oldFont);

            x += checkWidth;

            x += FntCharWidth(' ');

            if (userPrefs.optFlags.moreAccurate
                && (userPrefs.optFlags.showAbove
                    || userPrefs.optFlags.showBelow
                    || userPrefs.optFlags.showNever)) {
                Int32 UTr, UTs;
                Int16 flags;
                Int16 upnow;

                if (ListItemRiseSet(item, &UTr, &UTs, &flags, &upnow)) {
                    FntSetFont(fontSmall);
                    if (upnow && userPrefs.optFlags.showAbove) {
                        WinDrawChars("\012", 1,
                                     x, bounds->topLeft.y + 2);
                    } else if ((flags & (ASTROLIB_DONTRISESET) && userPrefs.optFlags.showNever)) {
                        WinDrawChars("\026", 1,
                                     x, bounds->topLeft.y + 2);
                    } else if (!upnow && userPrefs.optFlags.showBelow) {
                        WinDrawChars("\013", 1,
                                     x, bounds->topLeft.y + 2);
                    }
                    FntSetFont(fontNormal);
                }
                x += 9;
            }

            WinGlueDrawTruncChars(item->objectName, StrLen(item->objectName), x, bounds->topLeft.y,
                bounds->extent.x-(x-bounds->topLeft.x));

            MemHandleUnlock(h);
        }
    }
}


static Boolean LISTSECTION
ListCreateDatabase(Char* name)
{
    LocalID id;

    if (errNone == DmCreateDatabase(0, name, APPID, LIST_DBID, false)) {
        id = DmFindDatabase(0, name);

        if (id != NULL) {
            UInt16 attr;
            Err    err;

            err = DmDatabaseInfo(0, id, NULL,
    			     &attr, NULL,      /* attr, vers */
    			     NULL, NULL, NULL,    /* time stamps */
    			     NULL, NULL, NULL,    /* modnum, app, sort*/
    			     NULL, NULL);         /* type, creator */

            if (err == errNone) {
                attr |= dmHdrAttrBackup;
                DmSetDatabaseInfo(0, id, NULL,
    			     &attr, NULL,      /* attr, vers */
    			     NULL, NULL, NULL,    /* time stamps */
    			     NULL, NULL, NULL,    /* modnum, app, sort*/
    			     NULL, NULL);
            }

            return true;
        }
    }

    return false;
}


static void LISTSECTION
AppendTime(Char* s, DateTimeType* dtP, UInt16 maxLen)
{
    Char temp[FIELDLENGTH];

    StrPrintF(temp, "%02d:%02d:%02d%s",
              userPrefs.optFlags.time24Hour ? dtP->hour : (dtP->hour + 11) % 12 + 1,
              dtP->minute, dtP->second,
              userPrefs.optFlags.time24Hour ? "" : (dtP->hour >= 12 ? "p" : "a"));

    StrNCat(s, temp, maxLen);
}


static void LISTSECTION
AppendDate(Char* s, DateTimeType* dtP, UInt16 maxLen)
{
    Char temp[FIELDLENGTH];
    Char temp2[FIELDLENGTH];

#ifdef NO_GLUE
    DateTemplateToAscii
#else
    DateGlueTemplateToAscii
#endif
                            (GetStringResource(s_DayMonthYearFormat, temp2),
                            dtP->month, dtP->day, dtP->year,
                            temp,
                            FIELDLENGTH);

    StrNCat(s, temp, maxLen);
}


static void LISTSECTION
ListUpdateItem(Boolean draw)
{
    Int16     i;
    Char      data[2*FIELDLENGTH];
    Char      note[2*FIELDLENGTH];
    FormType* formPtr = FrmGetActiveForm();
    UInt8     checkmark;

    checkmark = CHECKMARK_NO;

    data[0] = 0;
    note[0] = 0;

    i = LstGetSelection(FrmGetObjectPtrByID(formPtr, l_ObjectList));

    if (0<= i && listRef != NULL) {
        MemHandle h;

        h = DmQueryRecord(listRef, i);

        if (h!=NULL) {
            ExtendedListItemType* item;

            item = MemHandleLock(h);
            
            checkmark = item->basic.checked;

            if (item->basic.checked != CHECKMARK_NO) {
                DateTimeType dt;

                TimSecondsToDateTime(item->localTime, &dt);

                StrNCopy(data, item->location, 2*FIELDLENGTH-1);
                data[2*FIELDLENGTH-1] = 0;

                if (data[0])
                    StrNCat(data, ": ", 2*FIELDLENGTH);

                AppendDate(data, &dt, FIELDLENGTH*2);
                StrNCat(data, ", ", FIELDLENGTH*2);
                AppendTime(data, &dt, FIELDLENGTH*2);
            }

            if (sizeof(ExtendedListItemType) < MemPtrSize(item)) {
                Char* p;

                note[2*FIELDLENGTH-1] = 0;
                StrNCopy(note, item->note, 2*FIELDLENGTH);

                p = note;
                while(*p) {
                    if(*p == '\n')
                         *p = ' ';
                    p++;
                }
            }

            MemHandleUnlock(h);
        }
    }

    UpdateTextField(f_Checked, data);
    if(draw)
        FldDrawField(FrmGetObjectPtrByID(formPtr, f_Checked));
    UpdateTextField(f_Note, note);
    if(draw)
        FldDrawField(FrmGetObjectPtrByID(formPtr, f_Note));

    switch(checkmark) {
        case CHECKMARK_YES:
            Show(lbl_Query, false);
            Show(lbl_Checked, true);
            break;
        case CHECKMARK_QUERY:
            Show(lbl_Checked, false);
            Show(lbl_Query, true);
            break;
        default:
            Show(lbl_Checked, false);
            Show(lbl_Query, false);
            break;
    }
}


static void LISTSECTION
ListOpenDatabase(void)
{
    LocalID id;
    ListType* list;

    if (listRef != NULL) {
        DmCloseDatabase(listRef);
        listRef = NULL;
    }

    id = DmFindDatabase(0, listPref.dbName);

    if(id == NULL) {
         StrCopy(listPref.dbName, defaultDBName);
         listPref.selected = -1;
         ListCreateDatabase(listPref.dbName);
         id = DmFindDatabase(0, listPref.dbName);
         if(id == NULL) {
              listRef = NULL;
              numItems = 0;
              return;
         }
    }

    listRef = DmOpenDatabase(0, id, dmModeReadWrite);
    if (listRef == NULL) {
        numItems = 0;
    }
    else {
        numItems = DmNumRecords(listRef);
    }

    FrmGetActiveForm();
    list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList);

    LstSetListChoices(list, NULL, numItems);
    LstSetDrawFunction(list, ListDrawItem);
    LstSetSelection(list, listPref.selected);

    if(numItems) {
        if(listPref.selected<0) {
                LstMakeItemVisible(list, 0);
        }
        else {
            LstMakeItemVisible(list, listPref.selected);
        }
    }    

    ListUpdateListName(TrimName(listPref.dbName, listNamePrefix), false);

    ListUpdateItem(false);
}



static void LISTSECTION
ListDrawObjectList(void)
{
    LstDrawList(FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList));
    ListUpdateItem(true);
}


static void LISTSECTION
ListDrawAll(void)
{
    DrawDatePicker();
    DrawTimePicker();
    FrmDrawForm(FrmGetActiveForm());
    ListUpdateItem(true);
}



static void LISTSECTION
ListRename(void)
{
    Char dbName[dmDBNameLength];
    StrCopy(dbName, listNamePrefix);

    if (0==FrmCustomResponseAlert(a_RenameList, "", "", "", dbName+LIST_NAME_PREFIX_LENGTH,
               dmDBNameLength-LIST_NAME_PREFIX_LENGTH, NULL) &&
               dbName[LIST_NAME_PREFIX_LENGTH]) {
        LocalID id = DmFindDatabase(0, listPref.dbName);
        
        if (id == NULL ||
            errNone != DmSetDatabaseInfo(0, id, dbName,
    			     NULL, NULL,      /* attr, vers */
    			     NULL, NULL, NULL,    /* time stamps */
    			     NULL, NULL, NULL,    /* modnum, app, sort*/
    			     NULL, NULL) ) {
            FrmAlert(a_ErrRenameList);
            return;
        }

        StrCopy(listPref.dbName, dbName);
        ListUpdateListName(listPref.dbName+LIST_NAME_PREFIX_LENGTH, false);
    }
}



static void LISTSECTION
ListNew(Boolean copyOld)
{
    Char dbName[dmDBNameLength];
    StrCopy(dbName, listNamePrefix);

    if (0==FrmCustomResponseAlert(a_NewList, "", "", "", dbName+LIST_NAME_PREFIX_LENGTH,
               dmDBNameLength-LIST_NAME_PREFIX_LENGTH, NULL) &&
        dbName[LIST_NAME_PREFIX_LENGTH]) {

        if (!ListCreateDatabase(dbName)) {
            FrmAlert(a_ErrNewList);
            return;
        }

        if (copyOld) {
            LocalID id;

            id = DmFindDatabase(0, dbName);

            if (id != NULL) {
                DmOpenRef newRef;

                newRef = DmOpenDatabase(0, id, dmModeReadWrite);

                if (newRef != NULL) {
                     UInt16 i;

                     for (i = 0; i < numItems ; i++) {
                         MemHandle h;
                         
                         h = DmQueryRecord(listRef, i);
                         
                         if (h != NULL) {
                             MemHandle newH;
                             UInt16 pos;
                             
                             pos = i;

                             newH = DmNewRecord(newRef, &pos, MemHandleSize(h));

                             if (newH != NULL) {
                                 void* src = MemHandleLock(h);
                                 void* dest = MemHandleLock(newH);
                                 
                                 DmWrite(dest, 0, src, MemHandleSize(h));
                                 
                                 MemHandleUnlock(newH);
                                 DmReleaseRecord(newRef, pos, true);
                                 MemHandleUnlock(h);
                             }
                         }
                     }

                     DmCloseDatabase(newRef);
                }
            }
        }

        StrCopy(listPref.dbName, dbName);
        listPref.selected = -1;

        ListOpenDatabase();
        ListDrawAll();
    }
}



void LISTSECTION
ListLookupItem(LookupMethodType method)
{
    Int16 item;

    item = LstGetSelection(FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList));

    if (listRef == NULL || item<0) {
        LookupObject(GetCurrentObject(), NULL, method);
    }
    else {
        UInt32 object;
        MemHandle h;
        Char name[dmDBNameLength];

        object = NO_OBJECT;

        h = DmQueryRecord(listRef, item);

        if (h!=NULL) {
            BasicListItemType* item;

            item = MemHandleLock(h);

            if (item->catalogName[0]) {
                StrCopy(name, item->objectName);
            }
            else {
                object = item->catalogNumber;
            }

            MemHandleUnlock(h);
        }

        LookupObject(object, (object==NO_OBJECT) ? name : NULL, method);
    }
}


static void LISTSECTION
ListCopyClipboard(Boolean notes)
{
    UInt16 i;

    ClipboardAddItem(clipboardText, "", 0);

    for (i = 0 ; i < numItems ; i++) {
        MemHandle h;

        h = DmQueryRecord(listRef, i);

        if (h!=NULL) {
            BasicListItemType* item;
            Char*        s;

            item = MemHandleLock(h);

            s = "-*?" + (item->checked % NUM_CHECKMARKS);

            if (errNone != ClipboardAppendItem(clipboardText, s, 1) ||
                errNone != ClipboardAppendItem(clipboardText,
                               item->objectName, StrLen(item->objectName)) ||
                errNone != ClipboardAppendItem(clipboardText,
                               "\n", 1)) {
                MemHandleUnlock(h);
                FrmAlert(a_ClipboardTruncated);
                return;
            }

            if (notes && sizeof(ExtendedListItemType) < MemHandleSize(h)) {
                Char* note;

                note = (Char*)item + sizeof(ExtendedListItemType);
                if (errNone != ClipboardAppendItem(clipboardText, note, StrLen(note))||
                    errNone != ClipboardAppendItem(clipboardText, "\n", 1)) {
                    MemHandleUnlock(h);
                    FrmAlert(a_ClipboardTruncated);
                    return;
                }
            }
            MemHandleUnlock(h);
        }
    }
}


static void LISTSECTION
ListFormOpen(void)
{
    UInt16  size = 0;
    Char    defaultName[FIELDLENGTH];

    StrCopy(defaultDBName, listNamePrefix);
    GetStringResource(s_Default, defaultName);
    StrCat(defaultDBName, defaultName);

    MemSet(&listPref, sizeof(ListPrefType), 0);
    StrCopy(listPref.dbName, defaultDBName);
    listPref.selected = -1;

    if (noPreferenceFound != PrefGetAppPreferences(APPID, LIST_PREF_ID, NULL, &size, true)) {
        if(sizeof(ListPrefType)<size)
           size = sizeof(ListPrefType);
        PrefGetAppPreferences(APPID, LIST_PREF_ID, &listPref, &size, true);
    }

    ListOpenDatabase();
    ListDrawAll();
}



static void LISTSECTION
ListFormClose(void)
{
    if(listRef != NULL) {
        DmCloseDatabase(listRef);
        listRef = NULL;
    }

    PrefSetAppPreferences(APPID,
                          LIST_PREF_ID, 0,
                          &listPref,
                          sizeof(listPref),
                          true);
}


static void LISTSECTION
ListDeleteObject(Int16 index)
{
    if (errNone==DmRemoveRecord(listRef, index))  {
        ListType* list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList);

        numItems--;
        LstSetListChoices(list, NULL, numItems);
        listPref.selected = -1;
        LstSetSelection(list, -1);
        ListDrawObjectList();
    }
}


static void LISTSECTION
ListDelete(void)
{
    if (0 != FrmAlert(a_AreYouSure))
        return;

   if (!StrCompare(listPref.dbName, defaultDBName)) {
       ListType* list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList);

       while(numItems) {
           DmRemoveRecord(listRef, 0);
           numItems--;
       }

       listPref.selected = -1;
       LstSetListChoices(list, NULL, numItems);
       LstSetSelection(list, -1);

       ListDrawObjectList();
   }
   else {
       LocalID id;

       DmCloseDatabase(listRef);
       listRef = NULL;

       id = DmFindDatabase(0, listPref.dbName);

       if (id != NULL) {
           DmDeleteDatabase(0, id);
       }

       StrCopy(listPref.dbName, defaultDBName);
       listPref.selected = -1;
       ListOpenDatabase();
       FrmDrawForm(FrmGetActiveForm());
   }
}


static void LISTSECTION
ListAddObject(void)
{
    MemHandle h;
    UInt16    pos = numItems;
    BasicListItemType  itemIn;
    ListType*          list;
    UInt32             object;

    object = GetCurrentObject();

    if (object == NO_OBJECT) {
        FrmAlert(a_NoSelectedObject);
        return;
    }

    MemSet(&itemIn, sizeof(BasicListItemType), 0);

    if(object<NUMPLANETS) {
         itemIn.catalogNumber = object;
         SysStringByIndex(stbl_SolarObjects, object,
             itemIn.objectName, FIELDLENGTH);
    }
    else {
         UInt16 i;
         AstroRecordType art;

         object -= NUMPLANETS;

         for(i = 0; i < numDBs; i++) {
           	 if (!catDBs[i]->isVisible)
           	     continue;

             if(object < catDBs[i]->numRecords)
                  break;
             object -= catDBs[i]->numRecords;
         }

         if (i == numDBs) {
             return;
         }

         StrCopy(itemIn.catalogName, catDBs[i]->name);
         itemIn.catalogNumber = object;

         GetCatalogObject(GetCurrentObject()-NUMPLANETS, &art, NULL);
         StrCopy(itemIn.objectName, art.description);
    }
    
    if (MAX_LIST <= numItems) {
        FrmAlert(a_AddList);
        return;
    }

    h = DmNewRecord(listRef, &pos, sizeof(BasicListItemType));

    if (h != NULL) {
        BasicListItemType* itemOut;

        itemOut = MemHandleLock(h);
        DmWrite(itemOut, 0, &itemIn, sizeof(BasicListItemType));
        MemHandleUnlock(h);

        DmReleaseRecord(listRef, pos, true);

        list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList);
        numItems++;
        LstSetListChoices(list, NULL, numItems);
        LstSetSelection(list, pos);
        listPref.selected = pos;
        LstMakeItemVisible(list, pos);
        ListDrawObjectList();
    }
    else {
        FrmAlert(a_AddList);
    }
}


static void LISTSECTION
ListDrawCatalogName(Int16 itemNum, RectangleType* bounds, Char** itemsText)
{
    UInt16 count = 0;
    UInt16 i;

    for (i=0 ; i<numDBs; i++) {
        if (catDBs[i]->isVisible && !catDBs[i]->isStars) {
            if (count == itemNum) {
                Char* p = TrimName(catDBs[i]->name, dbPrefix);
                WinDrawChars(p, StrLen(p), bounds->topLeft.x, bounds->topLeft.y);
                break;
            }
            count++;
        }
    }
}


static void LISTSECTION
ListUpdating(Boolean state)
{
    if(state) {
        Char updating[FIELDLENGTH];

        GetStringResource(s_Updating, updating),
        ListUpdateListName(updating, true);
    }
    else {
        ListUpdateListName(TrimName(listPref.dbName, listNamePrefix), true);
    }
}



static void LISTSECTION
ListAddCatalog(void)
{
    UInt16 count;
    UInt16 i;
    ListType* list;
    Int16     sel;
    UInt32    offset;
    UInt16    j;
    BasicListItemType  itemIn;

    count  = 0;

    for (i=0; i<numDBs; i++) {
        if (catDBs[i]->isVisible && ! catDBs[i]->isStars)
            count++;
    }

    list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_Catalogs);
    LstSetListChoices(list, NULL, count);
    LstSetHeight(list, count);
    LstSetSelection(list, 0);
    LstSetDrawFunction(list, ListDrawCatalogName);

    sel = LstPopupList(list);

    if (sel<0)
        return;

    count = 0;
    offset = 0;

    for (i=0; i<numDBs; i++) {
        if (catDBs[i]->isVisible) {
            if (!catDBs[i]->isStars) {
                if (count == sel)
                    break;
                count++;
            }

            offset += catDBs[i]->numRecords;
        }
    }

    if (i==numDBs) {
        /* should never happen */
        return;
    }

    MemSet(&itemIn, sizeof(BasicListItemType), 0);
    StrCopy(itemIn.catalogName, catDBs[i]->name);

    ListUpdating(true);

    for (j = 0 ; j < catDBs[i]->numRecords ; j++, offset++) {
         AstroRecordType art;
         UInt16 pos;
         MemHandle h;

         itemIn.catalogNumber = j;
         GetCatalogObject(offset, &art, NULL);
         StrCopy(itemIn.objectName, art.description);

         pos = numItems;

         if (MAX_LIST <= pos) {
             FrmAlert(a_AddList);
             break;
         }

         h = DmNewRecord(listRef, &pos, sizeof(BasicListItemType));

         if (h != NULL) {
             BasicListItemType* itemOut;

             itemOut = MemHandleLock(h);
             DmWrite(itemOut, 0, &itemIn, sizeof(BasicListItemType));
             MemHandleUnlock(h);

             DmReleaseRecord(listRef, pos, true);
             numItems++;
         }
         else {
             FrmAlert(a_AddList);
             break;
         }
    }

    ListUpdating(false);
    LstSetListChoices(FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList),
        NULL, numItems);
    ListDrawObjectList();
}


static UInt32 LISTSECTION
ListGetObject(BasicListItemType* item)
{
   UInt32 object;

   if (item->catalogName[0] == 0) {
       object = item->catalogNumber;
   }
   else {
       UInt16 i;

       object = NUMPLANETS + item->catalogNumber;

       for (i = 0 ; i < numDBs; i++) {
           if (catDBs[i]->isVisible) {
               if(!StrCompare(catDBs[i]->name, item->catalogName)) {
                   break;
               }
               else {
                   object += catDBs[i]->numRecords;
               }
           }
       }

       if (i==numDBs) {
           return NO_OBJECT;
       }
   }

   return object;
}



static void LISTSECTION
ListGoObject(Int16 index)
{
   MemHandle h;
   BasicListItemType* item;
   UInt32             object;

   h = DmQueryRecord(listRef, index);

   if (h == NULL)
       return;

   item = MemHandleLock(h);

   object = ListGetObject(item);
   if (object == NO_OBJECT) {
       if (1==FrmCustomAlert(a_MissingCatalog, TrimName(item->catalogName, dbPrefix),
           "", "")) {
           EventType event;

           FrmGotoForm(f_Catalog);

           MemSet(&event, sizeof(EventType), 0);
           event.eType = catalogSelectEvent;
           EvtAddEventToQueue(&event);
       }
       MemHandleUnlock(h);
       return;
   }

   MemHandleUnlock(h);

   SetCurrentObject(object);
   SkyCenterObject(object);
   FrmGotoForm(f_ObjectDetail);
}


static void LISTSECTION
ListDrawListNames(Int16 itemNum, RectangleType* bounds, Char** itemsText)
{
    Char* p;
    Char s[FIELDLENGTH];
    FontID oldFont;
    
    if (itemsText[itemNum] == NULL) {
        GetStringResource(s_New, s);

        p = s;

        oldFont = FntSetFont(boldFont);
    }
    else {
        p = TrimName(itemsText[itemNum], listNamePrefix);

        oldFont = FntSetFont(stdFont);
    }

    WinDrawChars(p, StrLen(p), bounds->topLeft.x, bounds->topLeft.y);
    
    FntSetFont(oldFont);

    if (itemsText[itemNum] != NULL && itemsText[itemNum+1] == NULL) {
        /* Draw separator line for "New" */
        IndexedColorType oldColor = 0;
        if (CheckMinRomVersion(sysVersion35)) {
            oldColor = WinSetForeColor(UIColorGetTableEntryIndex(UIObjectFrame));
        }
        WinDrawLine(bounds->topLeft.x-2, bounds->topLeft.y+bounds->extent.y-1,
            bounds->topLeft.x+bounds->extent.x-1, bounds->topLeft.y+bounds->extent.y-1);
        if (CheckMinRomVersion(sysVersion35)) {
            WinSetForeColor(oldColor);
        }
    }
}



static Int32 LISTSECTION
ListGetPhi(BasicListItemType* item, Boolean j2000)
{
    VectorType radec;

    if (!ListGetRaDec(item, radec, j2000)) {
        return phiSun - 1;
    }

   	return fpatan2(radec[1], radec[0]);
}



static Int16 LISTSECTION
CompareRecordsAlt(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    Boolean have1;
    Boolean have2;
    VectorType radec1;
    VectorType radec2;
    Int32 alt1;
    Int32 alt2;
    MatrixType radec2altaz;
    Int32 lst;
    Int32 gst;

    have1 = ListGetRaDec(r1, radec1, false);
    have2 = ListGetRaDec(r2, radec2, false);

    if(!have1 && have2) {
        return 1;
    }
    else if(have1 && !have2) {
        return -1;
    }
    else if(!have1 && !have2) {
        return 0;
    }

    gst = UTCToGST(userPrefs.jtime, currentGSTOffset);
    lst = GSTToLST(gst,userPrefs.Location[userPrefs.currentLoc].lng);
    MatRotate(radec2altaz,
              FP_ONE-userPrefs.Location[userPrefs.currentLoc].lat,
              FP_ONE + lst);

    alt1 = VecScalarProd(radec2altaz[2], radec1);
    alt2 = VecScalarProd(radec2altaz[2], radec2);

    if (alt1>alt2)
        return -1;
    else if(alt1<alt2)
        return 1;
    else
        return 0;
}


static Int16 LISTSECTION
CompareRecordsMag(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    Boolean have1;
    Boolean have2;
    Int16 mag1;
    Int16 mag2;

    have1 = ListGetMag(r1, &mag1);
    have2 = ListGetMag(r2, &mag2);

    if(!have1 && have2) {
        return 1;
    }
    else if(have1 && !have2) {
        return -1;
    }
    else if(!have1 && !have2) {
        return 0;
    }

    return mag1-mag2;
}



/* This sort works as follows:
   1. Always up
   2. Currently up, sorted by setting time
   3. Currently down, sorted by rising time
   4. Never up
   5. Unavailable (catalog not loaded) */

static Int16 LISTSECTION
CompareRecordsRiseSet(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    Boolean have1;
    Boolean have2;
    Int32 UTr1;
    Int32 UTs1;
    Int32 UTr2;
    Int32 UTs2;
    Int16 flags1;
    Int16 flags2;
    Int16 upnow1;
    Int16 upnow2;

    have1 = ListItemRiseSet(r1, &UTr1, &UTs1, &flags1, &upnow1);
    have2 = ListItemRiseSet(r2, &UTr2, &UTs2, &flags2, &upnow2);

    if(!have1 && have2) {
        return 1;
    }
    else if(have1 && !have2) {
        return -1;
    }
    else if(!have1 && !have2) {
        return 0;
    }

    if(flags1 & ASTROLIB_CIRCUMPOLAR) {
        if(flags2 & ASTROLIB_CIRCUMPOLAR)
            return 0;
        else
            return -1;
    }

    if(flags2 & ASTROLIB_CIRCUMPOLAR)
        return 1;

    if(flags1 & ASTROLIB_DONTRISESET) {
         if(flags2 & ASTROLIB_DONTRISESET)
            return 0;
         else
             return 1;
    }

    if(flags2 & ASTROLIB_DONTRISESET)
        return -1;

    if (!upnow1 != !upnow2) {
        if(upnow1)
            return -1;
        else
            return 1;
    }

    if (upnow1) {
        if(UTs1 < UTs2)
            return -1;
        else if(UTs2 < UTs1)
            return 1;
        else
            return 0;
    }
    else {
        if(UTr1 < UTr2)
            return -1;
        else if(UTr2 < UTr1)
            return 1;
        else
            return 0;
    }
}


static Int16 LISTSECTION
CompareRecordsRA(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    UInt32 phi1;
    UInt32 phi2;

    phi1 = (UInt32)(ListGetPhi(r1, userPrefs.radec2000) - phiSun);
    phi2 = (UInt32)(ListGetPhi(r2, userPrefs.radec2000) - phiSun);

    if (phi1 < phi2)
        return -1;
    else if (phi2 < phi1)
        return 1;
    else
        return 0;
}


static Int16 LISTSECTION
CompareRecordsName(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    Char* n1;
    Char* n2;
    Int16 i;
    Int32 x1;
    Int32 x2;
    
    /* This currently only supports 8-bit character sets */

    n1 = ((BasicListItemType*)r1)->objectName;
    n2 = ((BasicListItemType*)r2)->objectName;

    i = 0;

    while(n1[i] == n2[i] && n1[i]) {
        i++;
    }

    if (!TxtGlueCharIsDigit(n1[i]) || !TxtGlueCharIsDigit(n2[i])) {
        return StrCaselessCompare(n1+i,n2+i);
    }

    while(0 <= i && TxtGlueCharIsDigit(n1[i])) {
        i--;
    }

    x1 = StrAToI(n1 + i + 1);
    x2 = StrAToI(n2 + i + 1);

    if (x1 < x2) {
        return -1;
    }
    else if (x2 < x1) {
        return 1;
    }
    else {
        return StrCaselessCompare(n1,n2);
    }
}


static Int16 LISTSECTION
CompareRecordsChecked(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    ListCheckmarkType c1;
    ListCheckmarkType c2;
    Int16 order[NUM_CHECKMARKS] = { 2, 0, 1 };

    c1 = ((BasicListItemType*)r1)->checked;
    c2 = ((BasicListItemType*)r2)->checked;

    return order[c1 % NUM_CHECKMARKS] - order[c2 % NUM_CHECKMARKS];
}


static Int16 LISTSECTION
CompareRecordsCheckedTime(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    ListCheckmarkType c1;
    ListCheckmarkType c2;

    c1 = ((BasicListItemType*)r1)->checked;
    c2 = ((BasicListItemType*)r2)->checked;

    if (c1 == CHECKMARK_NO)
        return (c2 == CHECKMARK_NO) ? 0 : 1;
    if (c2 == CHECKMARK_NO)
        return (c1 == CHECKMARK_NO) ? 0 : -1;
    if ( ((ExtendedListItemType*)r1)->localTime < 
         ((ExtendedListItemType*)r2)->localTime )
        return -1;
    else if ( ((ExtendedListItemType*)r2)->localTime <
         ((ExtendedListItemType*)r1)->localTime )
        return 1;
    else
        return 0;
}


static UInt16 LISTSECTION
ListGetSortOrder(UInt32 object)
{
    UInt16 i;

    for (i=0; i<numItems; i++)
        if (object == sortObjectList[i])
            return i;

    return 0;
}


static Int16 LISTSECTION
CompareRecordsByNumber(void* r1, void* r2, Int16 other,
   SortRecordInfoPtr r1s, SortRecordInfoPtr r2s, MemHandle appInfo)
{
    UInt16 order1;
    UInt16 order2;

    order1 = ListGetSortOrder(ListGetObject(r1));
    order2 = ListGetSortOrder(ListGetObject(r2));

    if(order1<order2)
        return -1;
    else if(order2<order1)
        return 1;
    else
        return 0;
}


static void LISTSECTION
ListSort(DmComparF f)
{
    ListType* list;

    if (numItems == 0 || listRef == NULL)
        return;

    if (f != NULL) {
        BasicListItemType sun;

        if (f == CompareRecordsRA) {
            sun.catalogNumber = SUN;
            sun.catalogName[0] = 0;

            phiSun = ListGetPhi(&sun, userPrefs.radec2000);
        }
    
        ListUpdating(true);
        DmInsertionSort(listRef, f, 0);
        ListUpdating(false);
    }

    listPref.selected = -1;

    list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_ObjectList);
    LstSetSelection(list, -1);
    LstMakeItemVisible(list, 0);

    ListDrawAll();
}



/* This is normalized by dividing by numItems, to avoid overflow */
static Int32 LISTSECTION
ListGetBestLength(Angles* angles, UInt16* pathOrder)
{
    Int32 pathLen;
    UInt16 i;
    UInt8* used;

    used = MemPtrNew((numItems+7)/8);
    if(used == NULL) {
        for(i=1; i<numItems; i++) {
            if (i != pathOrder[0])
                pathOrder[i] = i;
            else
                pathOrder[i] = 0;
        }
        return 0xFFFFFFFF;
    }

    MemSet(used, (numItems+7)/8, 0);

    pathLen = 0;

    used[pathOrder[0]>>3] |= (1<<(pathOrder[0]&7));

    for (i=1; i<numItems; i++) {
        UInt32  bestDistance = 0xFFFFFFFF;
        Int16  bestIndex = 0;
        UInt16 j;

        for (j=0; j<numItems; j++) {
            UInt32 distance;

            if(used[j>>3] & (1<<(j&7)))
                continue;

            /* This angle comparison seems to handle all the cases, because the
               angles are in the range -PI to PI, and because an Int32 wraps
               around when it gets more than PI. */
            distance = fpabs(angles[pathOrder[i-1]].phi-angles[j].phi)/2 +
                       fpabs(angles[pathOrder[i-1]].psi-angles[j].psi)/2;

            if (distance < bestDistance) {
                bestDistance = distance;
                bestIndex    = j;
            }
        }

        pathOrder[i] = bestIndex;
        used[bestIndex>>3] |= (1<<(bestIndex&7));
                               
        /* normalize by numItems to prevent Int32 overflow */
        pathLen += bestDistance / numItems;
    }

    MemPtrFree(used);

    return pathLen;
}


/* Arranges objects in a path by a greedy algorithm */
static void LISTSECTION
ListPath(Boolean doAltaz)
{
    UInt32 bestLength;
    UInt16 bestStart;
    UInt16 i;
    UInt16* pathOrder;
    Angles* angles;
    MatrixType radec2altaz;
    Int32 lst;
    Int32 gst;

    if (numItems == 0 || listRef == NULL)
        return;

    if (MAX_PATH < numItems) {
        FrmAlert(a_TooManyItems);
        return;
    }

    if (50 < numItems && 1 == FrmAlert(a_PathConfirm))
        return;

    pathOrder = MemPtrNew(sizeof(UInt16)*(UInt32)numItems);

    if (pathOrder == NULL) {
        FrmAlert(a_TooManyItems);
        return;
    }

    angles    = MemPtrNew(sizeof(Angles)*(UInt32)numItems);
    if (angles == NULL) {
        FrmAlert(a_TooManyItems);
        MemPtrFree(pathOrder);
        return;
    }

    ListUpdating(true);

    if(doAltaz) {
        gst = UTCToGST(userPrefs.jtime, currentGSTOffset);
        lst = GSTToLST(gst,userPrefs.Location[userPrefs.currentLoc].lng);
        MatRotate(radec2altaz,
                  FP_ONE-userPrefs.Location[userPrefs.currentLoc].lat,
                  FP_ONE + lst);
    }

    for (i=0; i<numItems; i++) {
        MemHandle h;
        BasicListItemType* item;
        VectorType radec;

        h = DmQueryRecord(listRef, i);

        if(h==NULL) {
            ListUpdating(false);
            MemPtrFree(pathOrder);
            MemPtrFree(angles);
            return;
        }

        item = MemHandleLock(h);

        if (!ListGetRaDec(item, radec, false)) {
            ListUpdating(false);
            FrmAlert(a_MissingCatalog);
            MemPtrFree(pathOrder);
            MemPtrFree(angles);
            return;
        }

        if (doAltaz) {
            VectorType altaz;

            MatVecMul(altaz, radec2altaz, radec);
        	angles[i].phi = fpatan2(altaz[0], altaz[1]);
        	angles[i].psi = fpatan2(altaz[2], fpsqrt(FP_ONE - fpsq(altaz[2])));
        }
        else {
        	angles[i].phi = fpatan2(radec[1], radec[0]);
        	angles[i].psi = fpatan2(radec[2], fpsqrt(fpsq(radec[0]) + fpsq(radec[1])));
        }

        MemHandleUnlock(h);
    }

    bestLength  = 0xFFFFFFFF;
    bestStart = 0;

    for (pathOrder[0]=0; pathOrder[0]<numItems; pathOrder[0]++) {
        Int32 pathLen;
        Char  msg[10];

        StrPrintF(msg, "%d%%", pathOrder[0]*100/numItems);
        ListUpdateListName(msg, true);

        pathLen = ListGetBestLength(angles, pathOrder);

        if (pathLen < bestLength) {
            bestStart  = pathOrder[0];
            bestLength = pathLen;
        }
    }

    pathOrder[0] = bestStart;
    ListGetBestLength(angles, pathOrder);

    MemPtrFree(angles);

    sortObjectList = MemPtrNew(sizeof(UInt32)*numItems);

    if(sortObjectList == NULL) {
        ListUpdating(false);
        MemPtrFree(pathOrder);
        return;
    }

    for (i=0; i<numItems; i++) {
        MemHandle h;
        BasicListItemType* item;

        h = DmQueryRecord(listRef, pathOrder[i]);

        item = MemHandleLock(h);

        sortObjectList[i] = ListGetObject(item);

        MemHandleUnlock(h);
    }

    MemPtrFree(pathOrder);

    ListSort(CompareRecordsByNumber);
    MemPtrFree(sortObjectList);
}


static Int16 LISTSECTION
CompareListNames(void* a, void* b, Int32 other)
{
    Char* s1 = *(Char**)a;
    Char* s2 = *(Char**)b;
    Int16 delta;

    delta = StrCaselessCompare(s1, s2);

    if (delta == 0)
        return 0;

    if (!StrCompare(s1, defaultDBName))
        return -1;

    if (!StrCompare(s2, defaultDBName))
        return 1;

    return delta;
}



static void LISTSECTION
ListOpen(void)
{
     UInt16 card;
     LocalID id;
     Boolean newSearch;
     DmSearchStateType state;
     UInt16 count;
     Char** names;
     UInt16 i;
     ListType* list;
     Int16  sel;
     UInt16 listIndex;
     FormPtr formPtr;
     RectangleType rect;
     Int16 curList;

     formPtr = FrmGetActiveForm();

     count = 0;
     newSearch = true;

     while(errNone == DmGetNextDatabaseByTypeCreator(newSearch, &state,
                                           LIST_DBID, APPID,
                                           false, &card, &id)) {
         count++;
         newSearch = false;
     }

     if(count==0)
         return;

     names = MemPtrNew(sizeof(Char*)*(count+1));

     if(names==NULL)
         return;

     newSearch = true;

     listIndex = FrmGetObjectIndex(formPtr, l_ListsList);
     FrmGetObjectBounds(formPtr, listIndex, &rect);

     rect.extent.x = 1;

     FntSetFont(fontNormal);
     
     curList = -1;

     for(i=0; i < count &&
              errNone == DmGetNextDatabaseByTypeCreator(newSearch, &state,
                                           LIST_DBID, APPID,
                                           false, &card, &id)
               ; i++) {
          Char name[dmDBNameLength];
          newSearch = false;
          if (errNone==DmDatabaseInfo(card, id, name,
    			     NULL, NULL,      /* attr, vers */
    			     NULL, NULL, NULL,    /* time stamps */
    			     NULL, NULL, NULL,    /* modnum, app, sort*/
    			     NULL, NULL))         /* type, creator */ {
              Coord width;
              Char* p;

              names[i] = MemPtrNew(StrLen(name)+1);

              if(names[i] == NULL)
                  break;

              StrCopy(names[i], name);

              p = TrimName(name, listNamePrefix);

              width = FntCharsWidth(p, StrLen(p));

              if (rect.extent.x < width)
                  rect.extent.x = width;
          }
     }

     rect.extent.x += 4;

     count = i;
     
     curList = 0;

     FrmSetObjectBounds(formPtr, listIndex, &rect);

     if (0<count)
	     SysInsertionSort(names, count, sizeof(names[0]), CompareListNames, 0);

     for (i = 0 ; i < count ; i++) {
        if (!StrCompare(names[i], listPref.dbName))
            curList = i;
     }
     
     names[count] = NULL; // "New"

     list = FrmGetObjectPtrByID(formPtr, l_ListsList);

     LstSetListChoices(list, names, count+1);
     LstSetHeight(list, count+1);
     LstSetSelection(list, curList);
     LstSetDrawFunction(list, ListDrawListNames);
     sel = LstPopupList(list);

     if (0 <= sel) {
         if (sel < count) {
             StrCopy(listPref.dbName, names[sel]);
             listPref.selected = -1;
             ListOpenDatabase();
             FrmDrawForm(FrmGetActiveForm());
         }
         else {
             ListNew(false);
         }
     }

     for(i=0; i<count; i++)
         MemPtrFree(names[i]);
     MemPtrFree(names);
}


static void LISTSECTION
ListToggleObject(Int16 index)
{
   MemHandle h;
   BasicListItemType* item;
   UInt8 newCheck;

   h = DmQueryRecord(listRef, index);

   if (h == NULL)
       return;

   item = MemHandleLock(h);

   newCheck = (item->checked + 1) % NUM_CHECKMARKS;

   if (newCheck == CHECKMARK_NO) {
       BasicListItemType newItem;

       MemMove(&newItem, item, sizeof(BasicListItemType));
       newItem.checked = newCheck;
       MemHandleUnlock(h);

       if (MemPtrSize(item)<=sizeof(ExtendedListItemType)) {
           /* no note appended, so can safely resize */
           h = DmResizeRecord(listRef, index, sizeof(BasicListItemType));
       }

       if (h != NULL) {
           item = MemHandleLock(h);
           DmWrite(item, 0, &newItem, sizeof(BasicListItemType));
           MemHandleUnlock(h);
       }
   }
   else {
       ExtendedListItemType* newItemP;
       Char*                loc;

       loc = userPrefs.Location[userPrefs.currentLoc].name;

       newItemP = MemPtrNew(sizeof(ExtendedListItemType));

       if (newItemP == NULL) {
           MemHandleUnlock(h);
           return;
       }

       MemMove(newItemP, item, sizeof(BasicListItemType));
       MemHandleUnlock(h);

       newItemP->basic.checked = newCheck;
       newItemP->localTime = TimGetSeconds();
       newItemP->jday      = userPrefs.jday;
       newItemP->jtime     = userPrefs.jtime;
       StrCopy(newItemP->location, loc);

       if (MemPtrSize(item)<sizeof(ExtendedListItemType)) {
           h = DmResizeRecord(listRef, index, sizeof(ExtendedListItemType));
       }

       if (h != NULL) {
           item = MemHandleLock(h);
           DmWrite(item, 0, newItemP, sizeof(ExtendedListItemType));
           MemHandleUnlock(h);
       }

       MemPtrFree(newItemP);
   }

   ListDrawObjectList();
}


static void LISTSECTION
ListNoteEdit(Int16 sel)
{
   MemHandle h;

   h = DmQueryRecord(listRef, sel);

   if (h != NULL) {
       if (MemHandleSize(h) <= sizeof(ExtendedListItemType)) {
           h = DmResizeRecord(listRef, sel, 1+sizeof(ExtendedListItemType));
           if (h != NULL) {
               ExtendedListItemType* p;
               Char                  zero;

               p = MemHandleLock(h);

               zero = 0;
               DmWrite(p, sizeof(ExtendedListItemType), &zero, 1);

               MemHandleUnlock(h);
           }
           else {
               return;
           }
       }
       
       listPref.selected = sel;
       FrmPopupForm(f_NoteEdit);
   }
}

/*
 * Events for the object list
 */
Boolean LISTSECTION
ListFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr formPtr = FrmGetActiveForm();
    static Coord lastPenUpX = -1;
    static Coord lastPenUpY = -1;

    if (DateChangeHandleEvent(event)
        || FormChangeHandleEvent(event))
        return true;

    switch (event->eType) {
        case keyDownEvent: {
            if (event->data.keyDown.chr == vchrRockerCenter)
                lastPenUpX = -1;
            // Use Page Up/Down buttons to scroll object list.

            if (event->data.keyDown.chr == pageUpChr ||
                event->data.keyDown.chr == pageDownChr ||
                event->data.keyDown.chr == vchrRockerDown ||
                event->data.keyDown.chr == vchrRockerUp)
            {
                ListType* list;
                Int16 numVis;

                if(InteractionMode())
                    break;

                list = FrmGetObjectPtrByID(formPtr, l_ObjectList);

                numVis = LstGetVisibleItems(list);

                if (0<numVis) {
                    Int16 delta;

                    if (1<numVis)
                        delta = numVis-1;
                    else
                        delta = numVis;

                    LstScrollList(list,
                        (event->data.keyDown.chr == pageUpChr ||
                            event->data.keyDown.chr == vchrRockerUp)?winUp:winDown,
                        delta);
                    handled = true;
                }
            }
            break;
        }

        case penUpEvent:
            lastPenUpX = event->screenX;
            lastPenUpY = event->screenY;
            break;

        case lstSelectEvent:
            if (event->data.lstSelect.listID == l_ObjectList) {
                listPref.selected = event->data.lstSelect.selection;

                if (0<=listPref.selected) {
                    RectangleType bounds;

                    FrmGetObjectBounds(formPtr,
                        FrmGetObjectIndex(formPtr, l_ObjectList), &bounds);

                    if (bounds.topLeft.x <= event->screenX &&
                        event->screenX < bounds.topLeft.x + 16/* &&
                        event->screenX == lastPenUpX && event->screenY == lastPenUpY*/) {
                        ListToggleObject(listPref.selected);
                    }
                }

                ListUpdateItem(true);
            }
            break;


        case frmUpdateEvent:
            if (event->data.frmUpdate.updateCode==updateListItem) {
                ListUpdateItem(true);
                handled = true;
            }
            else {
                ListDrawAll();
                handled = true;
            }
            break;
            
        case fldEnterEvent: {
            Int16 selection;

            selection = LstGetSelection(FrmGetObjectPtrByID(formPtr, l_ObjectList));

            if(event->data.fldEnter.fieldID == f_Note &&
                 listRef != NULL && 0 <= selection) {
                ListNoteEdit(selection);
                handled = true;
            }
            break;
        }
        
        case ctlSelectEvent: {
            Int16 selection;

            selection = LstGetSelection(FrmGetObjectPtrByID(formPtr, l_ObjectList));
            switch (event->data.ctlSelect.controlID) {
                case tr_ListName:
                    ListOpen();
                    handled = true;
                    break;
                case b_GoObject:
                    if (0 <= selection && listRef != NULL)
                        ListGoObject(selection);
                    handled = true;
                    break;
                case b_DeleteObject:
                    if (0 <= selection && listRef != NULL)
                        ListDeleteObject(selection);
                    handled = true;
                    break;
                case b_AddObject:
                    if (listRef != NULL)
                        ListAddObject();
                    handled = true;
                    break;
                case b_RenameList:
                    ListRename();
                    handled = true;
                    break;
                case b_DeleteList:
                    if(listRef != NULL)
                        ListDelete();
                    handled = true;
                    break;
                case b_MoreList:
                    if(listRef != NULL) {
                       switch(LstPopupList(FrmGetObjectPtrByID(formPtr,l_MoreList))) {
                           case LIST_MORE_ADD_CATALOG:
                               ListAddCatalog();
                               break;
                           case LIST_MORE_COPY_CLIPBOARD:
                               ListCopyClipboard(false);
                               break;
                           case LIST_MORE_COPY_CLIPBOARD_WITH_NOTES:
                               ListCopyClipboard(true);
                               break;
                           case LIST_MORE_COPY_NEW_LIST:
                               ListNew(true);
                               break;
                           case LIST_MORE_SORT_CHECKED:
                               ListSort(CompareRecordsChecked);
                               break;
                           case LIST_MORE_SORT_CHECKED_TIME:
                               ListSort(CompareRecordsCheckedTime);
                               break;
                           case LIST_MORE_SORT_RA:
                               ListSort(CompareRecordsRA);
                               break;
                           case LIST_MORE_SORT_RISESET:
                               ListSort(CompareRecordsRiseSet);
                               break;
                           case LIST_MORE_SORT_ALT:
                               ListSort(CompareRecordsAlt);
                               break;
                           case LIST_MORE_SORT_MAG:
                               ListSort(CompareRecordsMag);
                               break;
                           case LIST_MORE_SORT_NAME:
                               ListSort(CompareRecordsName);
                               break;
                           case LIST_MORE_ARRANGE_ALTAZ:
                               ListPath(true);
                               break;
                           case LIST_MORE_ARRANGE_EQ:
                               ListPath(false);
                               break;
                           default:
                               break;
                       }
                    }
                    handled = true;
                    break;
            }
            break;
        }

        case frmOpenEvent: {
            ListFormOpen();
            SetMenu(formPtr);
            handled = true;
            break;
	    }

        case listAddEvent:
            if (listRef != NULL)
                ListAddObject();
            handled = true;
            break;

        case frmCloseEvent:
            ListFormClose();
            handled = false;
            break;

        default:
            handled = false;
    }

    return handled;
}



static void LISTSECTION
NoteEditUpdateScrollBar( void )
{
    UInt16 pos;
    UInt16 tHeight;
    UInt16 fHeight;

    FldGetScrollValues(GetObjectPtr(f_Note), &pos, &tHeight, &fHeight);

    if ( tHeight <= fHeight ) {
        Show(scl_Note, false);
    }
    else {
        Show(scl_Note, true);
        SclSetScrollBar(GetObjectPtr(scl_Note), pos, 0, tHeight - fHeight, fHeight);
    }
}


static void LISTSECTION
NoteEditScrollField(Int16 newValue, Int16 oldValue)
{
    Int16 delta;

    delta = newValue - oldValue;

    if ( 0 < delta )
        FldScrollField( GetObjectPtr(f_Note), delta, winDown );
    else if ( delta < 0 )
        FldScrollField( GetObjectPtr(f_Note), -delta, winUp );
}



/*
 * Events for the object list
 */
Boolean LISTSECTION
NoteEditFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr formPtr = FrmGetActiveForm();

    switch (event->eType) {
        case frmCloseEvent:
            FldSetText(FrmGetObjectPtrByID(formPtr, f_Note), NULL,
                0, 0);
            break;

        case frmOpenEvent: {
            MemHandle h;
            BasicListItemType* item;

            if (listRef == NULL || listPref.selected < 0) {
                FrmReturnToForm(0);
                handled = true;
                break;
            }

            h = DmQueryRecord(listRef, listPref.selected);

            if (h==NULL) {
                FrmReturnToForm(0);
                handled = true;
                break;
            }

            item = MemHandleLock(h);
            UpdateTextField(f_Description, item->objectName);
            MemHandleUnlock(h);

            FldSetText(FrmGetObjectPtrByID(formPtr, f_Note), h,
                sizeof(ExtendedListItemType),
                MemHandleSize(h)-sizeof(ExtendedListItemType));
            NoteEditUpdateScrollBar();
            
            FrmSetFocus(formPtr, GetObjectIndex(f_Note));
            FrmGlueNavObjectTakeFocus(formPtr, f_Note);

            FrmDrawForm(formPtr);
            handled = true;
            break;
        }

        case sclRepeatEvent:
            if (event->data.sclRepeat.scrollBarID == scl_Note) {
                NoteEditScrollField(event->data.sclRepeat.newValue,
                    event->data.sclRepeat.value);
//                handled = true;
            }
            break;

        case fldChangedEvent:
            if (event->data.fldChanged.fieldID == f_Note) {
                NoteEditUpdateScrollBar();
                handled = true;
            }
            break;

        case ctlSelectEvent: {
            switch (event->data.ctlSelect.controlID) {
                case b_OK:
                    FldSetText(FrmGetObjectPtrByID(formPtr, f_Note), NULL,
                        0, 0);
                    FrmUpdateForm(f_List, updateListItem);
                    FrmReturnToForm(0);
                    handled = true;
                    break;

                case b_Delete: {
                    MemHandle h;

                    FldSetText(FrmGetObjectPtrByID(formPtr, f_Note), NULL,
                        0, 0);
                    h = DmQueryRecord(listRef, listPref.selected);

                    if (h!=NULL) {
                        BasicListItemType* item;
                        ListCheckmarkType  checked;

                        item = MemHandleLock(h);
                        checked = item->checked;
                        MemHandleUnlock(h);

                        if (checked == CHECKMARK_NO) {
                            DmResizeRecord(listRef, listPref.selected,
                                sizeof(BasicListItemType));
                        }
                        else {
                            DmResizeRecord(listRef, listPref.selected,
                                sizeof(ExtendedListItemType));
                        }
                    }
                    FrmUpdateForm(f_List, updateListItem);
                    FrmReturnToForm(0);
                    handled = true;
                    break;
                }
            }
        }

        case menuEvent:
            switch (event->data.menu.itemID) {
                case mi_InsertDate: {
                    Char date[FIELDLENGTH];
                    DateTimeType dt;

                    TimSecondsToDateTime(TimGetSeconds(), &dt);
                    date[0] = 0;
                    AppendDate(date, &dt, FIELDLENGTH);
                    FldInsert(GetObjectPtr(f_Note), date, StrLen(date));
                    handled = true;
                    break;
                }
                case mi_InsertTime: {
                    Char time[FIELDLENGTH];
                    DateTimeType dt;

                    TimSecondsToDateTime(TimGetSeconds(), &dt);
                    time[0] = 0;
                    AppendTime(time, &dt, FIELDLENGTH);
                    FldInsert(GetObjectPtr(f_Note), time, StrLen(time));
                    handled = true;
                    break;
                }
                case mi_InsertLocation: {
                    FldInsert(GetObjectPtr(f_Note), userPrefs.Location[userPrefs.currentLoc].name,
                        StrLen(userPrefs.Location[userPrefs.currentLoc].name));
                    handled = true;
                    break;
                }
                case mi_InsertJulianDate: {
                    UInt32 jday;
                    UInt32 jtime;
                    Char str[FIELDLENGTH];

                    GetJulianDate(&jday, &jtime, TimGetSeconds());
                    GetJulianDateString(str, jday, jtime);

                    FldInsert(GetObjectPtr(f_Note), str, StrLen(str));
                }
            }
        default:
            break;
    }

    return handled;
}
