/*
 * Astro Info - an astronomical visualizer for PalmOS devices.
 * 
 * $Id: AstroMatrix.h 613 2006-08-21 18:24:39Z hoenicke $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#ifndef AstroMatrix_h
#define AstroMatrix_h
#include "FixedPoint.h"

#define MATRIXSECTION __attribute__ ((section ("code2")))

/* The astronomers define many different polar and cartesian
 * coordinate systems, that we have to represent in astroinfo.  All
 * cartesian systems follow the right hand rule.
 *
 * Earth centric coordinates (origin in center of earth):
 *
 * RA/Dec (equatorial) coordinates: The Z axis in this system points to the
 * north pole, the X axis points to 0 degrees (the position of the sun at
 * vernal equinox), Y axis points to 90 degrees.  If we ignore precession
 * stars stay at a fixed position in this coordinate system.
 *
 *   X =  cos(ra)*cos(dec)
 *   Y =  sin(ra)*cos(dec)
 *   Z =  sin(dec)
 *
 * Alt/Az (celestial) coordinates: These are dependent of the position
 * of the observer. The Z axis points into the zenith, The X axis
 * points eastwards, the Y axis northwards.
 *   X =  sin(az)*cos(alt)
 *   Y =  cos(az)*cos(alt)
 *   Z =  sin(alt)
 *
 * Heliocentric coordinates:  The Z axis stands perpendicular to ecliptic
 * pointing north,  the X axis points to vernal equinox, the Y axis to
 * summer solstice.
 *
 * So for lat/long coordinates:
 *   X = cos(lng)*cos(lat)
 *   Y = sin(lng)*cos(lat)
 *   Z = sin(lat)
 */
typedef Int32 VectorType[3];

/* A matrix consists of three row vectors.
 * matrix[2][1] is element at row 2, column 1.
 */
typedef Int32 MatrixType[3][3];

#define VecScalarProd(v, w) \
 (fpmul((v)[0], (w)[0]) + fpmul((v)[1],(w)[1]) + fpmul((v)[2], (w)[2]))
#define VecFastScalarProd(v, w) \
 (fpmul_fast((v)[0], (w)[0]) \
  + fpmul_fast((v)[1],(w)[1]) \
  + fpmul_fast((v)[2], (w)[2]))

/**
 * Normalize v, such that ||v||=1.
 */
void VecNormalize (VectorType v) MATRIXSECTION;

/**
 * Multiply m * v and write result in dest.
 */
void MatVecMul (VectorType dest, MatrixType m, VectorType v) MATRIXSECTION;

/**
 * Create Rotation matrix that rotates a vector first by lng around
 * the z axis, and then by lat around the x axis.  This is the 
 * matrix:
 * (    coslng           sinlng           0   )
 * (-coslat*sinlng   coslat*coslng     sinlat )
 * ( sinlat*sinlng  -sinlat*coslng     coslat )
 */
void MatRotate(MatrixType dest, Int32 lat, Int32 lng) MATRIXSECTION;

/**
 * Transpose the matrix dest in place.  Note that all rotation matrices
 * are orthogonal, so that the transpose is the inverse.
 */
void MatTranspose(MatrixType dest) MATRIXSECTION;

/**
 * Multiply two matrices and write result in dest. The resulting matrix
 * first executes m2, then m1.
 */
void MatMatMul (MatrixType dest, MatrixType m1, MatrixType m2) MATRIXSECTION;

#endif
