/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroNutation.c 708 2008-01-12 15:20:02Z arpruss $
 * Copyright (C) 2000, Michael Heinz
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "AstroLib.h"
#include "AstroNutation.h"
#include "FixedPoint.h"

/*
 *  Nutation, IAU 2000B model.
 *
 *  1) The nutation components in longitude and obliquity are with
 *     respect to the equinox and ecliptic of date.  The obliquity at
 *     J2000 is assumed to be the Lieske et al. (1977) value of
 *     84381.448 arcsec.
 *
 *     The nutation model consists only of luni-solar terms, but includes
 *     also a fixed offset which compensates for certain long-period
 *     planetary terms.
 *
 *  2) This routine is an implementation of the IAU 2000B abridged
 *     nutation model formally adopted by the IAU General Assembly in
 *     2000.  The routine computes the MHB_2000_SHORT luni-solar nutation
 *     series (Luzum 2001), but without the associated corrections for
 *     the precession rate adjustments and the offset between the GCRS
 *     and J2000 mean poles.
 *
 *  4) The full IAU 2000A (MHB2000) nutation model contains nearly 1400
 *     terms.  The IAU 2000B model (McCarthy & Luzum 2001) contains only
 *     77 terms, plus additional simplifications, yet still delivers
 *     results of 1 mas accuracy at present epochs.  This combination of
 *     accuracy and size makes the IAU 2000B abridged nutation model
 *     suitable for most practical applications.
 *
 *     The routine delivers a pole accurate to 1 mas from 1900 to 2100
 *     (usually better than 1 mas, very occasionally just outside 1 mas).
 *     The full IAU 2000A model, which is implemented in the routine
 *     iau_NUT00A (qv), delivers considerably greater accuracy at
 *     current epochs;  however, to realize this improved accuracy,
 *     corrections for the essentially unpredictable free-core-nutation
 *     must also be included.
 *
 *  5) The present routine provides classical nutation.  The
 *     MHB_2000_SHORT algorithm, from which it is adapted, deals also
 *     with (i) the offsets between the GCRS and mean poles and (ii) the
 *     adjustments in longitude and obliquity due to the changed
 *     precession rates.  These additional functions, namely frame bias
 *     and precession adjustments, are supported by the SOFA routines
 *     iau_BI00 and iau_PR00.
 *
 *  6) The MHB_2000_SHORT algorithm also provides "total" nutations,
 *     comprising the arithmetic sum of the frame bias, precession
 *     adjustments, and nutation (luni-solar + planetary).  These total
 *     nutations can be used in combination with an existing IAU 1976
 *     precession implementation, such as iau_PMAT76, to deliver GCRS-to-
 *     true predictions of mas accuracy at current epochs.  However, for
 *     symmetry with the iau_NUT00A routine (qv for the reasons), the
 *     SOFA routines do not generate the "total nutations" directly.  They
 *     can of course easily be generated by calling iau_BI00, iau_PR00 and
 *     the present routine and adding the results.
 *
 *  References:
 *
 *     SOFA (Standards of Fundamental Astronomy) software library
 *     revision:  2005 August 24
 *
 */

/* angle seconds to fixed point */
#define AS2FP(x) ((x)/3600.0/90.0)

#define DELTA_HIGH(x) ((Int32) (x))
#define DELTA_LOW(x)  FP((x) - (Int32)(x))
#define MEAN_VALUE(a, b, jday, jtime) (FP(a) \
	+ DELTA_HIGH((b)*FP_ONE/365250) * ((jday) - J2000) \
	+ fpmul(DELTA_LOW((b)*FP_ONE/365250), (jday) - J2000) \
	+ fpmul(FP((b)*FP_ONE/365250/DAY/2), 2*jtime))

struct nutation_entry {
    Int16 l :3;
    Int16 lp:3;
    Int16 f2:2;
    Int16 d2:4;
    Int16 om:4;
    Int16 sp, stp;
    Int16 cp;
    Int16 ce, cte;
    Int16 se;
};

#define ENTRY(l,lp,F,D,om,sp,stp,cp,ce,cte,se) \
    { l, lp, (F/2), (D/2), om, \
      FP(sp/3600.0/90.0/1e6/2), \
      FP(stp/3600.0/90.0/1e6/365250*(1L<<24)), \
      FP(cp/3600.0/90.0/1e6), \
      FP(ce/3600.0/90.0/1e6), \
      FP(cte/3600.0/90.0/1e6/365250*(1L<<25)), \
      FP(se/3600.0/90.0/1e6) }

static const struct nutation_entry nutations[] = {
    /*     l   l'  F   D  om   sin p    tsin p cos p   cos e  tcos e  sin e*/
    ENTRY( 0,  0,  0,  0,  1,-17206416,-174666, 3339, 9205233,  9086,  1538),
    ENTRY( 0,  0,  2, -2,  2, -1317091,  -1675,-1370,  573034, -3015,  -459),
    ENTRY( 0,  0,  2,  0,  2,  -227641,   -234,  280,   97846,  -485,   137),
    ENTRY( 0,  0,  0,  0,  2,   207455,    207,  -70,  -89749,   470,   -29),
    ENTRY( 0,  1,  0,  0,  0,   147588,  -3633, 1182,    7387,  -184,  -192),
    ENTRY( 0,  1,  2, -2,  2,   -51682,   1226,  -52,   22439,  -677,   -17),
    ENTRY( 1,  0,  0,  0,  0,    71116,     73,  -87,    -675,     0,    36),
    ENTRY( 0,  0,  2,  0,  1,   -38730,   -367,   38,   20073,    18,    32),
    ENTRY( 1,  0,  2,  0,  2,   -30146,    -36,   82,   12903,   -63,    37),
    ENTRY( 0, -1,  2, -2,  2,    21583,   -494,   11,   -9593,   299,    13),
    ENTRY( 0,  0,  2, -2,  1,    12823,    137,   18,   -6898,    -9,     4),
    ENTRY(-1,  0,  2,  0,  2,    12346,     11,    2,   -5331,    32,     0),
    ENTRY( 1,  0,  0, -2,  0,   -15699,    -10,  -17,    -124,     0,    -8),
};


NUTATIONSECTION void 
Nutation2000b(Int32 jday, Int32 jtime, Int32 *deltaPsi, Int32 *deltaEps)
{
    Int32 el, elp, f2, d2, om;
    Int32 sinel, sinelp, sinf2, sind2, sinom[3];
    Int32 cosel, coselp, cosf2, cosd2, cosom[3];
    Int32 dp, de;
    const struct nutation_entry *nut;

    /*  -------------------
     *  LUNI-SOLAR NUTATION
     *  -------------------
     */

    jtime += getTDBOffset(jday);

    /* Mean anomaly of the Moon */
    el  = MEAN_VALUE( 485868.249036/90/3600, 17179159232.178/90/3600, jday, jtime);

    /* Mean anomaly of the Sun */
    elp = MEAN_VALUE(1287104.79305/90/3600 ,  1295965810.481/90/3600, jday, jtime);
    
    /* Mean argument of latitude of the Moon */
    f2  = MEAN_VALUE(335779.526232*2/90/3600, 17395272628.478*2/90/3600, jday, jtime);

    /*  Mean elongation of the Moon from the Sun. */
    d2  = MEAN_VALUE(1072260.70369*2/90/3600, 16029616012.090*2/90/3600, jday, jtime);

    /*  Mean longitude of the ascending node of the Moon. */
    om  = MEAN_VALUE(450160.398036/90/3600, -69628905.431/90/3600, jday, jtime);
    fpsincos(el, &sinel, &cosel);
    fpsincos(elp, &sinelp, &coselp);
    fpsincos(f2, &sinf2, &cosf2);
    fpsincos(d2, &sind2, &cosd2);
    fpsincos(om, &sinom[1], &cosom[1]);
    sinom[0] = 0; cosom[0] = FP_ONE;
    sinom[2] = 2*fpmul(sinom[1], cosom[1]);
    cosom[2] = fpsq(cosom[1]) - fpsq(sinom[1]);

    dp = 0;
    de = 0;
    for (nut = nutations; 
	 nut < nutations+sizeof(nutations)/sizeof(*nut); nut++) {
	Int32 sin, cos;
	Int16 j = nut->om;
	sin = sinom[j];
	cos = cosom[j];

	j = nut->l;
	if (j > 0) {
	    Int32 s = fpmul(sin, cosel) + fpmul(cos, sinel);
	    cos = fpmul(cos, cosel) - fpmul(sin, sinel);
	    sin = s;
	} else if (j < 0) {
	    Int32 s = fpmul(sin, cosel) - fpmul(cos, sinel);
	    cos = fpmul(cos, cosel) + fpmul(sin, sinel);
	    sin = s;
	}

	j = nut->lp;
	if (j > 0) {
	    Int32 s = fpmul(sin, coselp) + fpmul(cos, sinelp);
	    cos = fpmul(cos, coselp) - fpmul(sin, sinelp);
	    sin = s;
	} else if (j < 0) {
	    Int32 s = fpmul(sin, coselp) - fpmul(cos, sinelp);
	    cos = fpmul(cos, coselp) + fpmul(sin, sinelp);
	    sin = s;
	}

	j = nut->f2;
	if (j > 0) {
	    Int32 s = fpmul(sin, cosf2) + fpmul(cos, sinf2);
	    cos = fpmul(cos, cosf2) - fpmul(sin, sinf2);
	    sin = s;
	}

	j = nut->d2;
	if (j < 0) {
	    Int32 s = fpmul(sin, cosd2) - fpmul(cos, sind2);
	    cos = fpmul(cos, cosd2) + fpmul(sin, sind2);
	    sin = s;
	}

	dp += fpmul(sin, 2L*nut->sp + fpmul((Int32)nut->stp << 6, (jday-J2000))) +
	    fpmul(cos, nut->cp);
	de += fpmul(cos, nut->ce + fpmul((Int32)nut->cte << 5, (jday-J2000))) +
	    fpmul(sin, nut->se);
    }
    *deltaPsi = dp;
    *deltaEps = de;
}
