/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroOptions.c 1009 2010-10-03 01:49:11Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "resource.h"
#include "AstroOptions.h"
#include "AstroHorizon.h"
#include "Astro.h"
#include "AstroDST.h"
#include "FixedPoint.h"
#include "AstroTime.h"
#include "AstroLib.h"
#include "AstroSky.h"
#include "PalmUtil.h"
#include "AstroGlue.h"
#include "Grayscale.h"
#include "AstroResource.h"

#define MAX_CITY_LINE 80 // Max amount of data per city line
#define MAX_CITIES 1600

typedef struct {
    UInt16 compression;
    UInt16 unused;
    UInt32 textLength;
    UInt16 numRecords;
    UInt16 maxRecordSize;
    UInt32 currentPosition;
} DocHeaderType;

typedef Boolean (*CityAdderType)(Char* line1, Char* line2);

static void SetupGeneralOptions(EventPtr event) OPTIONSECTION;
static void UpdateGeneralOptions(EventPtr event) OPTIONSECTION;

static void SetupLocationOptions(void) OPTIONSECTION;
static void UpdateLocationOptions(EventPtr event) OPTIONSECTION;

static void SetupCatalogOptions(EventPtr event) OPTIONSECTION;
static void UpdateCatalogOptions(EventPtr event) OPTIONSECTION;

static void OptLookupCity(void) OPTIONSECTION;
static void SetupColorOptions(EventPtr event) OPTIONSECTION;
static void UpdateColorOptions(EventPtr event) OPTIONSECTION;
static Char *GetColorOrCatName(Int16 i) OPTIONSECTION;
static void OPTIONSECTION
CatNamesListColorsDrawData(Int16 itemNum, RectangleType *bounds, Char **itemsText);
static OPTIONSECTION UInt16 OptGetCities(CityAdderType CityAdder);

static void SetupSkyOptions(EventPtr event) OPTIONSECTION;
static void UpdateSkyOptions(EventPtr event) OPTIONSECTION;

static Char angleFormatStr[FIELDLENGTH];
static ListPtr         catNamePtr;
static Char**          cityNames = NULL;
static UInt16          numCities = 0;
static UInt16          currentCityNum;
static UInt16          currentCityCount;

static OPTIONSECTION void
DrawAngleFormatItem(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    Char str[64];

    SysStringByIndex(stbl_AngleFormats, (UInt16)itemNum, str, sizeof(str));
    WinDrawChars(str, StrLen(str), bounds->topLeft.x, bounds->topLeft.y);
}


static OPTIONSECTION void
OptSetPushButton(Int16 firstID, Int16 num, Int16 value) {
    Int16 i;
    for (i = 0; i < num; i++)
	SetControlValue(firstID+i, i == value);
    return;
}

static OPTIONSECTION Int16
OptGetPushButton(Int16 firstID, Int16 num) {
    Int16 i;
    for (i = 0; i < num; i++) {
	if (GetControlValue(firstID+i))
	    return i;
    }
    return 0;
}

static OPTIONSECTION Boolean
OptAddCity(Char* line1, Char* line2)
{
    Char* p;
    Char* q;

    if (MAX_CITIES <= numCities)
        return false;

    if (!StrCompare(line1, "North Pole") ||
        !StrCompare(line1, "South Pole"))
        return true;

    cityNames[numCities] = MemPtrNew(StrLen(line1)+1);

    if(cityNames[numCities] == NULL)
        return false;

    StrCopy(cityNames[numCities], line1);
    
    p = cityNames[numCities];
    q = p;

    while(*p) {
        /* This makes it easier to type names for quick finding */
        if(*p == ':')
            *q = ' ';
        else
            *q = *p;
        if (q == cityNames[numCities] || q[-1] != ' ' || *q != ' ')
            q++;
        p++;
    }
    *q = 0;

    numCities++;

    return true;
}


static OPTIONSECTION void
OptClearCities(void)
{
    UInt16 i;

    if (cityNames != NULL) {
        for (i=0; i < numCities; i++) {
            if(cityNames[i]!=NULL)
                MemPtrFree(cityNames[i]);
        }
        MemPtrFree(cityNames);
        cityNames = NULL;
    }

    numCities = 0;
}


static OPTIONSECTION Char*
OptGetNatural(Char* data, Int32* minP)
{
    if (*data < '0' || '9' < *data)
        return NULL;

    *minP = 0;

    while('0'<=*data && *data <= '9') {
        *minP = *minP * 10 + (*data - '0');
        data++;
    }
    
    return data;
}


static OPTIONSECTION Char*
OptGetMinutes(Char* data, Char unit, Int32* minP)
{
    Int32 sign;
    Int32 x1;
    Int32 x2;

    if (*data++ != '(') {
        return NULL;
    }
    
    if (*data != '-' && *data != '+' && !('0' <= *data && *data <= '9')) {
        while(*data && *data != ')')
            data++;
        if (!*data)
            return NULL;
        data++;
        while(*data == ' ')
            data++;
        if (*data != '(')
            return NULL;
        data++;
    }

    if (*data == '-') {
        sign = -1;
        data++;
    }
    else if (*data == '+') {
        sign = 1;
        data++;
    }
    else {
        sign = 1;
    }

    data = OptGetNatural(data, &x1);
    if (data == NULL)
        return NULL;

    if (*data++ != unit)
        return NULL;

    while(*data == ' ')
        data++;

    data = OptGetNatural(data, &x2);
    if (data == NULL)
        return NULL;

    if (*data++ != 'm' || *data++ != ')')
        return NULL;

    while(*data == ' ')
        data++;
        
    *minP = sign * (x1 * 60 + x2);

    return data;
}



static OPTIONSECTION void
OptSetCityData(Char* name, Char* data)
{
    Char* colon;
    Int32 latMin;
    Int32 longMin;
    Int32 tzMin;
    UInt16 i;
    UInt16 n;
    Int32 sign;

    /* Trim name */
    while (NULL != (colon = StrChr(name, ':'))) {
        name = colon+1;
    }

    data = OptGetMinutes(data, 'd', &latMin);

    if (data == NULL)
        return;

    data = OptGetMinutes(data, 'd', &longMin);

    if (data == NULL)
        return;

    data = OptGetMinutes(data, 'h', &tzMin);

    if (data == NULL)
        return;

    n = LstGetNumberOfItems(GetObjectPtr(l_dst));

    for (i=0; i<n; i++) {
        Char dstName[FIELDLENGTH];
        dstName[0] = 0;
        GetDSTName(i, dstName);
        if (!StrCompare(dstName, data))
            break;
    }

    if (i == n)
        i = 0;

    userPrefs.Location[userPrefs.currentLoc].dstRuleNr = i;
    SetDSTRule(userPrefs.Location[userPrefs.currentLoc].dstRuleNr,
               &userPrefs.Location[userPrefs.currentLoc].dstRule);

    StrNCopy(userPrefs.Location[userPrefs.currentLoc].name, name, FIELDLENGTH);
    userPrefs.Location[userPrefs.currentLoc].name[FIELDLENGTH-1] = 0;

    if (tzMin < 0) {
        sign = -1;
        tzMin = -tzMin;
    }
    else {
        sign = 1;
    }
    userPrefs.Location[userPrefs.currentLoc].tz =
        sign*tzMin*60000l;

    if (latMin < 0) {
        sign = -1;
        latMin = -latMin;
    }
    else {
        sign = 1;
    }
    userPrefs.Location[userPrefs.currentLoc].lat = sign * DMSToAng(latMin/60,latMin%60,0);
    if (longMin < 0) {
        sign = -1;
        longMin = -longMin;
    }
    else {
        sign = 1;
    }
    userPrefs.Location[userPrefs.currentLoc].lng = sign * DMSToAng(longMin/60,longMin%60,0);

    CloseDSTRules();
    SetupLocationOptions();
    FrmDrawForm(FrmGetActiveForm());
}


static OPTIONSECTION Boolean
OptSetCity(Char* line1, Char* line2)
{
    if (currentCityCount == currentCityNum) {
        OptSetCityData(line1, line2);
        return false;
    }

    currentCityCount++;
    return true;
}



static OPTIONSECTION void
OptLookupCity(void)
{
    OptClearCities();

    cityNames = MemPtrNew(sizeof(Char*) * MAX_CITIES);

    if (cityNames == NULL)
        return;

    numCities = 0;

    OptGetCities(OptAddCity);
    
    if (0 < numCities) {
        Int16 sel;
        ListType* list;

        list = GetObjectPtr(l_City);

        LstSetListChoices(list, cityNames, numCities);
        LstSetSelection(list, -1);
        LstMakeItemVisible(list, 0);

        sel = LstPopupList(list);

        if (0 <= sel) {
            currentCityNum  = sel;
            currentCityCount = 0;
            OptGetCities(OptSetCity);
        }
    }
    else {
        FrmAlert(a_NoLocations);
    }

    OptClearCities();
}

/*
 * Events for the option forms
 *
 * This is a single handler, with multiple paths to deal with the specific
 * fields on a form. Switching forms is handled by the menu handling code.
 */
Boolean
AllOptionsFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    Int8 changeloc = -1;
    FormPtr formPtr;
    UInt32  formID;

    formID  = FrmGetActiveFormID();
    formPtr = FrmGetActiveForm();

    if (FormChangeHandleEvent(event))
        return true;

    switch (event->eType) {

    case ctlSelectEvent:
        switch (event->data.ctlSelect.controlID) {
            case b_LookupCity: {
                OptLookupCity();
                handled = true;
                break;
            }
            case b_ResetColors: {
                UInt16 i;
                ColorModeType mode = GetColorMode();

                for (i=0; i<MAX_DBS; i++) {
                     if (catDBs[i] != NULL) {
                        switch (mode) {
                            case MODE_NIGHT:
                            case MODE_ONBLACK:
                                catDBs[i]->color[mode].r = 255;
                                catDBs[i]->color[mode].g = 255;
                                catDBs[i]->color[mode].b = 255;
                                break;
                            case MODE_ONWHITE:
                                catDBs[i]->color[mode].r = 0;
                                catDBs[i]->color[mode].g = 0;
                                catDBs[i]->color[mode].b = 0;
                                break;
                        }
                     }
                }

                for (i=0; i<NUM_COLOR_TYPES; i++) {
                    userPrefs.colors[mode][i] = defaultPrefs.colors[mode][i];
                }

                LstDrawList(catNamePtr);

                handled = true;
                break;
            }
            case p_NightMode:
            case p_onWhite:
            case p_onBlack: {
                userPrefs.prevColorMode = GetColorMode();
                userPrefs.optFlags.nightMode       = 
		    event->data.ctlSelect.controlID == p_NightMode;
                userPrefs.optFlags.blackBackground = 
		    event->data.ctlSelect.controlID != p_onWhite;
		if (colorSupport)
		    NightMode();
		FrmEraseForm(formPtr);
		FrmDrawForm(formPtr);
                handled = true;
                break;
            }
	case p_Opt_Gen:
	    if (formID != f_GeneralOptions)
		FrmGotoForm(f_GeneralOptions);
	    handled = true;
	    break;
	case p_Opt_Loc:
	    if (formID != f_LocationOptions)
		FrmGotoForm(f_LocationOptions);
	    handled = true;
	    break;
	case p_Opt_Cat:
	    if (formID != f_CatalogOptions)
		FrmGotoForm(f_CatalogOptions);
	    handled = true;
	    break;
	case p_Opt_Sky:
	    if (formID != f_SkyOptions)
		FrmGotoForm(f_SkyOptions);
	    handled = true;
	    break;
	case p_Opt_Color:
	    if (formID != f_ColorOptions)
		FrmGotoForm(f_ColorOptions);
	    handled = true;
	    break;

        case p_loc1:
            if (userPrefs.currentLoc != 0) {
                changeloc = 0;
            }
            break;

        case p_loc2:
            if (userPrefs.currentLoc != 1) {
                changeloc = 1;
            }
            break;

        case p_loc3:
            if (userPrefs.currentLoc != 2) {
                changeloc = 2;
            }
            break;

        case p_loc4:
            if (userPrefs.currentLoc != 3) {
                changeloc = 3;
            }
            break;

        case p_loc5:
            if (userPrefs.currentLoc != 4) {
                changeloc = 4;
            }
            break;

        default:
        }
        if (changeloc > -1) {
            // Save state then switch to different location
            UpdateLocationOptions(event);
            userPrefs.currentLoc = changeloc;
            SetupLocationOptions();

            handled = true;
            FrmDrawForm(formPtr);
        }
        break;

    case lstSelectEvent:
        if (event->data.lstSelect.listID == ls_ColorCatName) {
            Int16 i = event->data.lstSelect.selection;

            if ( i != noListSelection) {
                IndexedColorType  index;
                RGBColorType*     colorP;
                Char *name = GetColorOrCatName(i);

                if (i<NUM_COLOR_TYPES) {
                    colorP = &(userPrefs.colors[GetColorMode()][i]);
                }
                else {
                    colorP = &(catDBs[i-NUM_COLOR_TYPES]->color[GetColorMode()]);
                }

                index = WinRGBToIndex(colorP);

                if (UIPickColor(&index, NULL,
                             UIPickColorStartPalette, name, NULL)) {
                    WinIndexToRGB(index,colorP);
                }

                if ( userPrefs.optFlags.nightMode ) {
                    /* UIPickColor() obliterates UI colors */
                    NightMode();
                }

                FrmDrawForm(formPtr);
            }
            handled = true;
        }
        break;

    case frmOpenEvent:

        SetMenu(formPtr);

        // Different setup for each option form
        if (formID == f_GeneralOptions)
            SetupGeneralOptions(event);

        if (formID == f_LocationOptions)
            SetupLocationOptions();

        if (formID == f_CatalogOptions)
            SetupCatalogOptions(event);

        if (formID == f_SkyOptions)
            SetupSkyOptions(event);

        if (formID == f_ColorOptions)
            SetupColorOptions(event);

        FrmDrawForm(formPtr);

        // Nothing more to do
        handled = true;
        break;

    case frmCloseEvent:
        // Different preference saving code for each option form
        if (formID == f_GeneralOptions)
            UpdateGeneralOptions(event);

        if (formID == f_LocationOptions)
            UpdateLocationOptions(event);

        if (formID == f_CatalogOptions)
            UpdateCatalogOptions(event);

        if (formID == f_SkyOptions)
            UpdateSkyOptions(event);

        if (formID == f_ColorOptions)
            UpdateColorOptions(event);

        // Let PalmOS handle the rest of the tidyup
        handled = false;
        break;

    case popSelectEvent:
        switch (event->data.popSelect.listID) {
        
            case l_dst:
                GetDSTName(event->data.popSelect.selection, dstRuleName);
                SetControlLabel(event->data.popSelect.controlID, dstRuleName);
                handled = true;
                break;

            case l_AngleFormat:
                /* To trigger this event we must have had a valid list selection,
                 * so just grab string at same offset in our description list
                 * with no further validation required.
                 */
                SysStringByIndex(stbl_AngleFormats, event->data.popSelect.selection, angleFormatStr, sizeof(angleFormatStr));
                SetControlLabel(event->data.popSelect.controlID, angleFormatStr);
                handled = true;
                break;

            case l_TimeIncUnits:
                SetControlLabel(event->data.popSelect.controlID,
                            LstGetSelectionText(FrmGetObjectPtrByID(formPtr, event->data.popSelect.listID),
                                                event->data.popSelect.selection)
                                                );
                handled = true;
                break;
        }
        break;

    default:
        handled = false;
    }

    return handled;
}


/*****************************************************************************/

static OPTIONSECTION void 
SetupIncUnit(UInt16 fieldID, UInt16 unitID, UInt16 listID, 
	     UInt32 increment) {
    ListPtr listPtr = FrmGetObjectPtrByID(FrmGetActiveForm(), listID);
    Int16 unit = 0;
    if (increment == 0)
	increment = 1;
    if (!(increment % 60)) {
	unit++;
	increment /= 60;
	if (!(increment % 60)) {
	    unit++;
	    increment /= 60;
	    if (!(increment % 24)) {
		unit++;
		increment /= 24;
	    }
	}
    }
    SetFieldValue(fieldID, increment);
    LstSetSelection(listPtr, unit);
    SetControlLabel(unitID, LstGetSelectionText(listPtr, unit));
	
}

static OPTIONSECTION UInt32
UpdateIncUnit(UInt16 fieldID, UInt16 listID) {
    FormPtr formPtr = FrmGetActiveForm();
    UInt32 increment = GetNumField(formPtr, fieldID);
    Int16 unit = LstGetSelection(FrmGetObjectPtrByID(formPtr,listID));

    if (increment == 0)
        increment = 1; /* increment of 0 can hog the system */

    switch (unit) {
	case 3:
	    increment *= 24;
	    /*fall through*/
	case 2:
	    increment *= 60;
	    /*fall through*/
	case 1:
	    increment *= 60;
    }
    return increment;
}

static OPTIONSECTION void
SetupFormatOption(UInt16 listid, UInt16 triggerid, 
                  UInt16 listNum, Char *angleFormatStr)
{
    FormPtr  formPtr = FrmGetActiveForm();
    ListPtr templistPtr;

    templistPtr = (ListPtr) FrmGetObjectPtrByID(formPtr, listid);
    LstSetListChoices(templistPtr, NULL, 3);
    LstSetDrawFunction(templistPtr, DrawAngleFormatItem);
    LstSetSelection(templistPtr, listNum);
    SysStringByIndex(stbl_AngleFormats, listNum, angleFormatStr, FIELDLENGTH);
    SetControlLabel(triggerid, angleFormatStr);
}


OPTIONSECTION void SetupGeneralOptions(EventPtr event)
{
    RectangleType nativeRect = screenRect;
    Char    strColor[10];
    Char    str[50];

    SetControlValue(p_Opt_Gen, 1);

    // Choose time format for display
    SetControlValue(p_TIME24, userPrefs.optFlags.time24Hour);
    SetControlValue(p_TIME12, !userPrefs.optFlags.time24Hour);

    SetupIncUnit(f_TimeInc, tr_TimeIncUnits, l_TimeIncUnits,
		 userPrefs.timeInc);
    SetFieldValue(f_TimeInt, userPrefs.timeInt);
    
    // Choose action page up/down buttons perform
    SetControlValue(p_PB_DetRocker, 
                    userPrefs.pageBtnFunction == PAGEBTNROCKER);
    SetControlValue(p_PB_DateTime,
                    userPrefs.pageBtnFunction == PAGEBTNDATE);
    SetControlValue(p_PB_None,
                    userPrefs.pageBtnFunction == PAGEBTNOBJECT);
    SetControlValue(p_PB_Zoom,
                    userPrefs.pageBtnFunction == PAGEBTNZOOM);
    SetControlValue(c_RaDec2000,
                    userPrefs.radec2000);

    SetupFormatOption(l_AngleFormat, tr_AngleFormat,
                      (UInt16)userPrefs.angleFormat, angleFormatStr);

    if (winVersion >= 4) {
	WinPushDrawState();
	WinSetCoordinateSystem(kCoordinatesNative);
	WinScaleRectangle(&nativeRect);
	WinPopDrawState();
    }
    SysStringByIndex(stbl_Color, colorSupport ? 2 : grayscaleSupport ? 1 : 0,
		     strColor, sizeof(strColor));
    StrPrintF(str, "%dx%dx%ld (%s)",
              nativeRect.extent.x, nativeRect.extent.y,
              screenDepth, strColor);

    SetTextLabel(l_Resolution, str);
 }
 


OPTIONSECTION void UpdateGeneralOptions(EventPtr event)
{
    FormPtr formPtr = FrmGetActiveForm();
    userPrefs.optFlags.time24Hour = GetControlValue(p_TIME24);

    userPrefs.timeInc = UpdateIncUnit(f_TimeInc, l_TimeIncUnits);
    if (timePlayAdjust < 0)
        timePlayAdjust = -userPrefs.timeInc;
    else if (0 < timePlayAdjust)
        timePlayAdjust = userPrefs.timeInc;
          
    userPrefs.timeInt = GetNumField(formPtr, f_TimeInt);
    if (userPrefs.timeInt <= 0)
	userPrefs.timeInt = 1; /* Sanity check: 0 would hog the system */

    if (GetControlValue(p_PB_DetRocker))
        userPrefs.pageBtnFunction = PAGEBTNROCKER;
    else if (GetControlValue(p_PB_DateTime))
        userPrefs.pageBtnFunction = PAGEBTNDATE;
    else if (GetControlValue(p_PB_None))
        userPrefs.pageBtnFunction = PAGEBTNOBJECT;
    else
        userPrefs.pageBtnFunction = PAGEBTNZOOM;

    userPrefs.radec2000 = GetControlValue(c_RaDec2000);
    userPrefs.angleFormat = LstGetSelection(
	(ListPtr) FrmGetObjectPtrByID(formPtr, l_AngleFormat));

    CalculateDate();
}


/*****************************************************************************/


OPTIONSECTION void SetupLocationOptions(void)
 {
    FormPtr  formPtr = FrmGetActiveForm();
    ListPtr templistPtr;
    Int16 d, m, s;
    Int32 tz;

    SetControlValue(p_Opt_Loc, 1);

    SetControlValue(p_loc1, userPrefs.currentLoc == 0);
    SetControlValue(p_loc2, userPrefs.currentLoc == 1);
    SetControlValue(p_loc3, userPrefs.currentLoc == 2);
    SetControlValue(p_loc4, userPrefs.currentLoc == 3);
    SetControlValue(p_loc5, userPrefs.currentLoc == 4);

    UpdateTextField(f_City, userPrefs.Location[userPrefs.currentLoc].name);

    // local scope
    {
        Int16 entries = OpenDSTRules();

        if (userPrefs.Location[userPrefs.currentLoc].dstRuleNr >= entries)
            userPrefs.Location[userPrefs.currentLoc].dstRuleNr = 0;

        templistPtr = (ListPtr) FrmGetObjectPtrByID(formPtr, l_dst);

        LstSetListChoices(templistPtr, NULL, entries);
        LstSetDrawFunction(templistPtr, DSTListDrawData);
        LstSetTopItem(templistPtr, userPrefs.Location[userPrefs.currentLoc].dstRuleNr);
        LstSetSelection(templistPtr, userPrefs.Location[userPrefs.currentLoc].dstRuleNr);

        GetDSTName(userPrefs.Location[userPrefs.currentLoc].dstRuleNr, dstRuleName);
        SetControlLabel(pt_dst, dstRuleName);
    }

    /* Calculate absolute timezone in minutes; round to nearest */
    tz = ((userPrefs.Location[userPrefs.currentLoc].tz > 0
        ? userPrefs.Location[userPrefs.currentLoc].tz
            : -userPrefs.Location[userPrefs.currentLoc].tz) + 30000) / 60000;

    SetFieldValue(f_TZh, tz / 60);
    SetFieldValue(f_TZm, tz % 60);
    SetControlValue(p_TZw, (userPrefs.Location[userPrefs.currentLoc].tz <= 0));
    SetControlValue(p_TZe, (userPrefs.Location[userPrefs.currentLoc].tz > 0));
    
    AngToDMS(fpabs(userPrefs.Location[userPrefs.currentLoc].lat), &d, &m, &s);
    SetFieldValue(f_LATd, d);
    SetFieldValue(f_LATm, m);
    SetFieldValue(f_LATs, s);
    SetControlValue(p_LATs, (userPrefs.Location[userPrefs.currentLoc].lat < 0));
    SetControlValue(p_LATn, (userPrefs.Location[userPrefs.currentLoc].lat >= 0));
    
    AngToDMS(fpabs(userPrefs.Location[userPrefs.currentLoc].lng), &d, &m, &s);
    SetFieldValue(f_LNGd, d);
    SetFieldValue(f_LNGm, m);
    SetFieldValue(f_LNGs, s);
    SetControlValue(p_LNGe, (userPrefs.Location[userPrefs.currentLoc].lng > 0));
    SetControlValue(p_LNGw, (userPrefs.Location[userPrefs.currentLoc].lng <= 0));
}



OPTIONSECTION void UpdateLocationOptions(EventPtr event)
{
    FormPtr  formPtr = FrmGetActiveForm();
    UInt32   value;

    /* Note we don't update userPrefs.currentLoc here because the
     * location switching code is dynamic and is handled elsewhere.
     */

    StrCopy(userPrefs.Location[userPrefs.currentLoc].name,
            FldGetTextPtr(FrmGetObjectPtrByID(formPtr, f_City)));
    value = (GetNumField(formPtr,f_TZh) * 60
             + GetNumField(formPtr,f_TZm)) * 60000l;

    /* TZ is positive East of 0 UT */
    if (GetControlValue(p_TZw))
        value = -value;
    userPrefs.Location[userPrefs.currentLoc].tz = value;

    userPrefs.Location[userPrefs.currentLoc].dstRuleNr = LstGetSelection
        ((ListPtr) FrmGetObjectPtrByID(formPtr, l_dst));

    SetDSTRule(userPrefs.Location[userPrefs.currentLoc].dstRuleNr,
               &userPrefs.Location[userPrefs.currentLoc].dstRule);

    value = DMSToAng(GetNumField(formPtr,f_LATd), 
                     GetNumField(formPtr,f_LATm),
                     GetNumField(formPtr,f_LATs));
    if (GetControlValue(p_LATs))
        value = -value;
    userPrefs.Location[userPrefs.currentLoc].lat =  value;
    
    value = DMSToAng(GetNumField(formPtr,f_LNGd),
                     GetNumField(formPtr,f_LNGm),
                     GetNumField(formPtr,f_LNGs));
    if (GetControlValue(p_LNGw))
        value = -value;
    userPrefs.Location[userPrefs.currentLoc].lng = value;
    
    CloseDSTRules();
}


/*****************************************************************************/

OPTIONSECTION void SetupCatalogOptions(EventPtr event)
{
    SetControlValue(p_Opt_Cat, 1);
    SetControlValue(c_above, userPrefs.optFlags.showAbove);
    SetControlValue(c_below, userPrefs.optFlags.showBelow);
    SetControlValue(c_never, userPrefs.optFlags.showNever);
    SetControlValue(c_omitStars, userPrefs.optFlags.omitStars);
}



OPTIONSECTION void UpdateCatalogOptions(EventPtr event)
{
    userPrefs.optFlags.showAbove = GetControlValue(c_above);
    userPrefs.optFlags.showBelow = GetControlValue(c_below);
    userPrefs.optFlags.showNever = GetControlValue(c_never);
    userPrefs.optFlags.omitStars = GetControlValue(c_omitStars);
}


/*****************************************************************************/

OPTIONSECTION static Char *GetColorOrCatName(Int16 i) 
{
    static Char nameData[ 80 ];
    Char* name;

    switch(i) {
         case COLOR_PLANETS:
             SysCopyStringResource(nameData, s_Planets);
             break;
         case COLOR_SUN:
             SysCopyStringResource(nameData, s_Sun);
             break;
         case COLOR_MOON:
             SysCopyStringResource(nameData, s_Moon);
             break;
         case COLOR_LINES:
             SysCopyStringResource(nameData, s_Lines);
             break;
         case COLOR_EXTRAS:
             SysCopyStringResource(nameData, s_Extras);
             break;
         case COLOR_TEXT:
             SysCopyStringResource(nameData, s_Text);
             break;
         case COLOR_SHADOW:
             SysCopyStringResource(nameData, s_Shadow);
             break;
         case COLOR_CIRCLES:
             SysCopyStringResource(nameData, s_Circles);
             break;
         case COLOR_CONST_NAMES:
             SysCopyStringResource(nameData, s_ConstNames);
             break;
         default:
             StrCopy(nameData, catDBs[i-NUM_COLOR_TYPES]->name);
             break;
    }

    /* Skip AstroInfo prefix if present */
    if (!memcmp(nameData, "AstroInfo ", 10))
        name = nameData + 10;
    else
        name = nameData;

    return name;
}

/*
 * Display the list of catalogs of objects.
 */
static void OPTIONSECTION
CatNamesListColorsDrawData(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    RectangleType rect;
    RGBColorType *color;
    Char* name;

    name = GetColorOrCatName(itemNum);

    /* Show catalog name */
    FntSetFont(fontNormal);
    WinDrawChars(name, StrLen(name),
		 bounds->topLeft.x, bounds->topLeft.y);

    rect.topLeft.x = bounds->topLeft.x + 130;
    rect.extent.x  = 15 - 4;
    rect.topLeft.y = bounds->topLeft.y + 2;
    rect.extent.y  = bounds->extent.y - 4;

    if (userPrefs.optFlags.blackBackground !=
	userPrefs.optFlags.nightMode)
	WinDrawRectangleFrame( simpleFrame, &rect );
    else
	WinEraseRectangleFrame( simpleFrame, &rect );

    if (itemNum<NUM_COLOR_TYPES)
        color = & userPrefs.colors[GetColorMode()][itemNum];
    else
        color = & catDBs[itemNum-NUM_COLOR_TYPES]->color[GetColorMode()];

    WinPushDrawState();
    WinGlueSetBackColorRGB(color, NULL);
    WinFillRectangle( &rect, 0 );

    WinPopDrawState();
}


OPTIONSECTION void SetupColorOptions(EventPtr event)
{
    SetControlValue(p_Opt_Color, 1);

    if (CheckMinRomVersion(sysVersion35)) {
        FormType* formPtr;

        formPtr = FrmGetActiveForm();

        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, ls_ColorCatName));
        FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, b_ResetColors));

        catNamePtr = (ListPtr)FrmGetObjectPtrByID(formPtr, ls_ColorCatName);
        LstSetSelection(catNamePtr,-1);
    	LstSetListChoices(catNamePtr, (Char**)catDBs, numDBs+NUM_COLOR_TYPES);
    	LstSetTopItem(catNamePtr,0);
    	LstSetDrawFunction(catNamePtr, CatNamesListColorsDrawData);
    }

    if (colorSupport) {
	SetControlValue(p_NightMode , userPrefs.optFlags.nightMode);
    } else {
	FormPtr     formPtr = FrmGetActiveForm();
	ControlPtr  fld = (ControlPtr)FrmGetObjectPtrByID(formPtr, p_NightMode);
	CtlSetUsable(fld, false);
	userPrefs.optFlags.nightMode = false;
    }
    SetControlValue(p_onBlack, userPrefs.optFlags.blackBackground && 
        ! userPrefs.optFlags.nightMode);
    SetControlValue(p_onWhite, ! userPrefs.optFlags.blackBackground &&
        ! userPrefs.optFlags.nightMode);
}


OPTIONSECTION void SetupSkyOptions(EventPtr event)
{
    SetControlValue(p_Opt_Sky, 1);

    SetControlValue(p_radec, userPrefs.skyFlags.showRA);
    SetControlValue(p_altaz, !userPrefs.skyFlags.showRA);

    SetControlValue(p_ly, !userPrefs.parsec);
    SetControlValue(p_parsec, userPrefs.parsec);

    OptSetPushButton(p_Brightness, 8, userPrefs.skyFlags.brightness);

    SetControlValue(p_POSTEXT, userPrefs.skyFlags.displayPosText);
    SetControlValue(p_POSCOMP, userPrefs.skyFlags.displayPosGraph);
    SetControlValue(p_POSZOOM, userPrefs.skyFlags.displayZoom);

    OptSetPushButton(p_CompassSize, 3, userPrefs.skyFlags.sizePosCircles);

    if(CheckMinRomVersion(sysVersion35)) {
        FrmShowObject(FrmGetActiveForm(), 
            FrmGetObjectIndex(FrmGetActiveForm(), p_BelowHorizon+BELOW_HORIZON_FADE));
        OptSetPushButton(p_BelowHorizon, 3,
            userPrefs.optFlags.fadeBelowHorizon ? BELOW_HORIZON_FADE :
             ( userPrefs.skyFlags.showHorizon ? BELOW_HORIZON_DRAW : BELOW_HORIZON_OMIT ));
    }
    else {
        FrmHideObject(FrmGetActiveForm(), 
            FrmGetObjectIndex(FrmGetActiveForm(), p_BelowHorizon+BELOW_HORIZON_FADE));
        OptSetPushButton(p_BelowHorizon, 2,
            userPrefs.skyFlags.showHorizon ? BELOW_HORIZON_DRAW : BELOW_HORIZON_OMIT );
    }

    SetControlValue(c_const, userPrefs.skyFlags.showConstellation);
    SetControlValue(c_ConstNames, userPrefs.optFlags.showConstNames);
    SetControlValue(c_Timeout, ! userPrefs.skyFlags.noDisplayTimeout);
    SetControlValue(c_FastScroll, ! userPrefs.slowScroll);
    SetControlValue(c_ShowAllNonStars, userPrefs.skyFlags.showAllNonStars);

    OptSetPushButton(p_DetRows, 7, userPrefs.showDetailRows);
    OptSetPushButton(p_DetSpacing, 3, userPrefs.spacingDetailRows);
}

OPTIONSECTION void
UpdateColorOptions(EventPtr event)
{
    userPrefs.optFlags.blackBackground = GetControlValue(p_onBlack) ||
        userPrefs.optFlags.nightMode;
}


OPTIONSECTION void
UpdateSkyOptions(EventPtr event)
{
    userPrefs.skyFlags.showRA = GetControlValue(p_radec);
    userPrefs.parsec = GetControlValue(p_parsec);

    userPrefs.skyFlags.brightness = OptGetPushButton(p_Brightness, 8);
    
    /* It's not gray then its none. 
     * Unimplemented--if you don't like bitmaps don't install them :)
    userPrefs.optFlags.showImage = GetControlValue(p_OBJECTIMAGE);
    */
    
    userPrefs.skyFlags.displayPosText = GetControlValue(p_POSTEXT);
    userPrefs.skyFlags.displayPosGraph = GetControlValue(p_POSCOMP);
    userPrefs.skyFlags.displayZoom = GetControlValue(p_POSZOOM);

    userPrefs.skyFlags.sizePosCircles = OptGetPushButton(p_CompassSize, 3);

    userPrefs.skyFlags.showHorizon       = GetControlValue(p_BelowHorizon+BELOW_HORIZON_DRAW);
    userPrefs.optFlags.fadeBelowHorizon  = GetControlValue(p_BelowHorizon+BELOW_HORIZON_FADE);
    userPrefs.skyFlags.showConstellation = GetControlValue(c_const);
    userPrefs.optFlags.showConstNames    = GetControlValue(c_ConstNames);
    userPrefs.skyFlags.noDisplayTimeout  = ! GetControlValue(c_Timeout);
    userPrefs.skyFlags.showAllNonStars   = GetControlValue(c_ShowAllNonStars);

    userPrefs.slowScroll                 = ! GetControlValue(c_FastScroll);

    userPrefs.showDetailRows = OptGetPushButton(p_DetRows, 7);
    userPrefs.spacingDetailRows = OptGetPushButton(p_DetSpacing, 3);
}



static OPTIONSECTION UInt16
UncompressDOCRecord(UInt16 compression, MemHandle h, Char* out, UInt16 bufSize)
{
    UInt16 i;
    UInt16 inSize;
    UInt16 j;
    UInt8* in;

    if(h==NULL)
         return 0;

    inSize  = MemHandleSize(h);

    in = MemHandleLock(h);

    if (compression<2) {
        MemMove(out, in, inSize);
        j = inSize;
    }
    else {
        i       = 0;
        j       = 0;

        while(i<inSize && j<bufSize) {
            UInt16 c;

            c = in[i++];

            if(1<=c && c<=8) {
                /* literal c bytes */
                while(c--) {
                    out[j++] = in[i++];
                }
            }
            else if(c<=0x7F) {
                out[j++] = c;
            }
            else if(0xC0<=c) {
                /* quick way of encoding space */
                out[j++] = ' ';
                out[j++] = c & 0x7F;
            }
            else {
                /* reverse lookup */
                UInt16 delta;
                UInt16 count;

                c = (c << 8) | in[i++];

                delta = (c & 0x3FFF) >> 3;
                count = (c & 0x0007) + 3;

                do {
                    out[j] = out[j-delta];
                    j++;
                } while(--count);
            }
        }
    }

    MemHandleUnlock(h);

    return j;
}


static OPTIONSECTION UInt16
OptGetCities(CityAdderType CityAdder)
{
    LocalID id;
    UInt16  count;
    DmOpenRef ref;
    Char*     cityLines[2];
    MemHandle h;

    count = 0;

    id = DmFindDatabase(0, "Lat/Long/TZ List");

    if (id==NULL)
        return 0;

    ref = DmOpenDatabase(0, id, dmModeReadOnly);

    if (ref==NULL)
        return 0;

    cityLines[0] = MemPtrNew(MAX_CITY_LINE+1);
    if (cityLines[0] == NULL) {
        return 0;
    }
    cityLines[1] = MemPtrNew(MAX_CITY_LINE);
    if (cityLines[1] == NULL) {
        MemPtrFree(cityLines[0]);
        return 0;
    }

    if (NULL != (h=DmQueryRecord(ref, 0))) {
        DocHeaderType* head;
        UInt16         numRecords;
        UInt16         compression;
        UInt16         rec;
        UInt16         line;
        UInt16         offsetInLine;
        UInt16         needToSkip = 4;
        UInt16         bufSize;
        Char*          outBuf;

        head = MemHandleLock(h);

        numRecords = head->numRecords;
        compression = head->compression;
        bufSize     = head->maxRecordSize;

        MemHandleUnlock(h);

        outBuf = MemPtrNew(bufSize);

        if (outBuf != NULL) {
            line = 0;
            offsetInLine = 0;
    
            for (rec=0 ; rec<head->numRecords; rec++) {
                 UInt16 size;
                 UInt16 i;

                 size = UncompressDOCRecord(compression, 
                            DmQueryRecord(ref, rec+1), outBuf, bufSize);

                 if (size == 0)
                     break;

                 for (i=0; i < size; i++) {
                     if(needToSkip) {
                         if(outBuf[i] == '\n')
                             needToSkip--;
                         continue;
                     }
                     if(outBuf[i] == '\n') {
                         if(line==1) {
                             cityLines[1][offsetInLine] = 0;
                             line++;
                             count++;
                             if (!CityAdder(cityLines[0], cityLines[1])) {
                                 goto DONE;
                             }
                         }
                         else if(line==0) {
                             cityLines[0][offsetInLine] = 0;
                             line++;
                             offsetInLine = 0;
                         }
                         else if(line==2) {
                             line=0;
                             offsetInLine = 0;
                         }
                     }
                     else {
                         if (line<2 && offsetInLine < MAX_CITY_LINE) {
                             cityLines[line][offsetInLine++] = outBuf[i];
                         }
                     }
                 }
            }
    
            if(line==1) {
                cityLines[1][offsetInLine] = 0;
                CityAdder(cityLines[0], cityLines[1]);
            }

        DONE:
            MemPtrFree(outBuf);
        }
    }

    MemPtrFree(cityLines[0]);
    MemPtrFree(cityLines[1]);
    DmCloseDatabase(ref);

    return count;
}

/*****************************************************************************/

// This is for moon phase calendar. Disabled for now.
#if 0
static Boolean
PhaseOptionsFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();

    switch (event->eType) {
    case ctlSelectEvent:
         if (event->data.ctlSelect.controlID == b_OK) {

                userPrefs.optFlags.blackBackground =
                        GetControlValue(p_OBJECTBGBL);

                FrmUpdateForm(userPrefs.activeFormId, updateDateChanged);
                FrmReturnToForm(0);
                handled = true;
         }
        break;

    case frmOpenEvent:
        SetControlValue(p_OBJECTBGWH, ! userPrefs.optFlags.blackBackground);
        SetControlValue(p_OBJECTBGBL, userPrefs.optFlags.blackBackground);
         
        FrmDrawForm(formPtr);

        handled = true;
        break;
        
    default:
        handled = false;
    }

    return handled;
}
#endif
