/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: AstroPhaseCal.c 944 2009-04-24 00:54:43Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <TblGlue.h>
#include <BmpGlue.h>
#include "AstroResource.h"
#include "resource.h"
#include "AstroPhaseCal.h"
#include "Astro.h"
#include "AstroDetails.h"
#include "AstroSky.h"
#include "AstroTime.h"
#include "AstroTable.h"
#include "AstroGlue.h"
#include "Grayscale.h"
#include "AstroLib.h"
#include "AstroMoon.h"
#include "FixedPoint.h"
#include "Bitmap.h"


static Boolean PCSECTION
PhaseCalendarSelect(Int16 row, Int16 column)
{
    Int32   jday;
    Int16   day, offset, mag;
    Int32   EC, illf, dist, semid, pa;
    Int32   sinlat, coslat;
    DateType date;
    Int32 UTr;
    Int32 UTs;
    Int32 LTr;
    Int32 LTs;
    Char  rise[FIELDLENGTH];
    Char  set[FIELDLENGTH];
    Char  transit[FIELDLENGTH];
    Char  flstr[3];
    Int32 dstoffset;
    Int16 inMonth;
    Int16 i;
    VectorType radec2000;
    
    TblUnhighlightSelection(GetObjectPtr(tbl_PhaseCal));

    if (0==row)
        return false;

    DateDaysToDate(userPrefs.jday - JDAYEPOCH, &date);

    offset = DayOfWeek(date.month, 1, date.year+1904);
    inMonth = DaysInMonth(date.month, date.year+1904);
    for (i=0; i < inMonth; i++) {
        if (row == (((i + offset) / 7) % 5) + 1 &&
            column == ((i + offset) % 7) )
           break;
    }

    if (i == inMonth)
        return false;

    day = i+1;

    fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

    jday = JDay(date.year+1904, date.month, day);

    LunarPos(jday, userPrefs.jtime, &EC, radec2000, &illf, &mag, &dist, &semid, &pa);

    SolObjectRiseSet(MOON, jday, userPrefs.jtime,
                                 sinlat, coslat,
                                 userPrefs.Location[userPrefs.currentLoc].lng,
                                 &UTr, &UTs);

    dstoffset = getDSTOffset(jday, userPrefs.jtime,
                             userPrefs.Location[userPrefs.currentLoc].tz,
                             &userPrefs.Location[userPrefs.currentLoc].dstRule);

    LTr = UTr + userPrefs.Location[userPrefs.currentLoc].tz + dstoffset;
    LTs = UTs + userPrefs.Location[userPrefs.currentLoc].tz + dstoffset;

    FormatTime(LTr, rise, FIELDLENGTH, flstr, true);
    StrCat(rise, flstr);
    FormatTime(LTs, set, FIELDLENGTH, flstr, true);
    StrCat(set, flstr);
    FormatTime((LTr+LTs)/2,  transit, FIELDLENGTH, flstr, true);
    StrCat(transit, flstr);

    FrmCustomAlert(a_MoonRiseSet+GetMoonPhaseNumber(pa), rise, set, transit);

    return true;
}


/*
 * Events for the Phase Calendar form
 */
Boolean PCSECTION
PhaseCalendarFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();

    if (DateChangeHandleEvent(event)
        || FormChangeHandleEvent(event))
        return true;

    switch (event->eType) {

    case frmOpenEvent:
        SetMenu(formPtr);

        DrawTimePicker();
        DrawDatePicker();

        InitPhaseCalendarTable();
        UpdatePhaseCalendarTable();
        FrmDrawForm(formPtr);

        handled = true;
        break;

    case tblSelectEvent:
        if (event->data.tblSelect.tableID == tbl_PhaseCal) {
            handled = PhaseCalendarSelect(event->data.tblSelect.row,
                event->data.tblSelect.column);
            break;
        }
        handled = false;
        break;

    case frmUpdateEvent:
        switch (event->data.frmUpdate.updateCode) {

            // Second flag to each function determines whether to
            // perform a redraw of top half of screen.  This may be
            // meaningless for some objects and is simply ignored.
        case updateDateChanged:
        case updateDetailChanged:
		    UpdatePhaseCalendarTable();
    		FrmDrawForm(formPtr);
            handled = true;
            break;
        }
        break;

    case frmCloseEvent:
        AstroFreeTable((TablePtr)
                       FrmGetObjectPtrByID(formPtr, tbl_PhaseCal));
        handled = false;
        break;

    default:
        handled = false;
    }

    return handled;
}



void PCSECTION
PhaseTableDrawItem(void *tblP,
                   Int16 row, Int16 column, RectanglePtr bounds)
{
    FontID curFont;
    TablePtr thetable = (TablePtr) tblP;
    MemHandle handle = (MemHandle) TblGlueGetItemPtr(thetable, row, column);
    Char *text = NULL;
    VectorType radec;
    Int32   jday, EC, illf, dist, semid, pa;
    Int16   opts, len, day, offset, mag;
    Int16     space;
    Coord   y, size, width, height;
    DateType date;
    Err     err;
    Int16   transparentColor;
    Coord   adjSize;
    BitmapPtr bmp1, bmp2, tmpBmp2;
    Coord x;
    Char str[7];

    if (handle == NULL)
        return;

    /* PalmOS 3.0 needs this */
    if (!CheckMinRomVersion(sysVersion30))
        WinEraseRectangle(bounds, 0);

    DateDaysToDate(userPrefs.jday - JDAYEPOCH, &date);
    text = (Char*) MemHandleLock(handle);
    len = StrLen(text);
    
    if (len == 0)
        return;
    
    /* Grab the day of the month from the the displayed number! */
    day = StrAToI(text);

    opts = TblGetItemInt(thetable, row, column);
    curFont = FntSetFont(TblGetItemFont(thetable, row, column));

    /* Work out the correct horizontal offset based on the alignment */
    if ((opts & 3) == leftAlign) {
        space = 0;
    } else {
        space = bounds->extent.x - FntCharsWidth(text, len);
        if ((opts & 3) == centerAlign)
            space /= 2;
    }
    WinSetUnderlineMode(noUnderline);

    WinDrawChars(text, len, bounds->topLeft.x + space, bounds->topLeft.y);
    y = FntLineHeight();
    FntSetFont(curFont);
    MemHandleUnlock(handle);

    /* If we're in the headings or an empty cell then return */
    if (row == 0 || day == 0)
        return;

    offset = DayOfWeek(date.month, 1, date.year+1904);

    /* Phase calendar does not work under 3.0-3.3 due to missing bitmap
     * support.
     */
    if (!CheckMinRomVersion(sysVersion35))
	return;

    jday = JDay(date.year+1904, date.month, day);
    LunarPos(jday, userPrefs.jtime, &EC, radec,
	     &illf, &mag, &dist, &semid, &pa);

    size = bounds->extent.y - y;
    if (size > bounds->extent.x)
        size = bounds->extent.x;
    if (size < 2)
        return;

    if (screenDepth < 16)
        transparentColor = (1 << screenDepth)-1;
    else
        transparentColor = 0;

    bmp1 = SkyMapCreatePlanetBitmap(MOON, pa,
                                    transparentColor, transparentColor);
    if (bmp1 == NULL)
        /* User did not install planet bitmaps? */
        return;

    adjSize = size * screenDensity / 72;
    tmpBmp2 = BmpCreate(adjSize, adjSize, screenDepth, NULL, &err);

    ErrFatalDisplayIf(tmpBmp2 == NULL, "Could not create temporary bitmap");

    if (72<screenDensity) {
        bmp2 = (BitmapPtr) BmpCreateBitmapV3(tmpBmp2, screenDensity,
                                           BmpGetBits(tmpBmp2), NULL);
        ErrFatalDisplayIf(bmp2 == NULL, "Could not create temporary bitmap");
    }
    else {
        bmp2 = tmpBmp2;
    }

    /* Draw bitmap.  Rotate it if user lives in the southern hemisphere. */
    BmpGlueGetDimensions(bmp1, &width, &height, NULL);
    BmpScaleRotateBitmap(bmp2, adjSize/2, adjSize/2, adjSize, adjSize,
                         bmp1, 0, 0, width, height,
                         0, (userPrefs.Location[userPrefs.currentLoc].lat < 0
                             ? - FP_ONE : FP_ONE), 1,
                         transparentColor,
                         (transparentColor==0) ? winMask : winPaint);
    x = bounds->topLeft.x + (bounds->extent.x - size) / 2;
    y = bounds->topLeft.y + y;
    WinDrawBitmap(bmp2, x, y);
    BmpDelete(bmp2);
    if (tmpBmp2 != bmp2) {
        BmpDelete(tmpBmp2);
    }
    BmpDelete(bmp1);

    if ( 8 <= screenDepth ) {
        RGBColorType c;
        
        WinPushDrawState();

        StrPrintF(str,"%u%%",(UInt16)((FPTOGP(illf)*100+GP_ONE/2)/GP_ONE));

        curFont = FntSetFont(fontSmall);

        x = x + size/2 - FntCharsWidth(str,StrLen(str))/2;
        y = y + size/2 - FntCharHeight()/2;

        c.r = 51;
        c.g = 204;
        c.b = 255;
        WinSetTextColorRGB(&c, NULL);
        WinSetDrawMode(winOverlay);
        WinPaintChars(str,StrLen(str),x,y);

        WinPopDrawState();
    }
}

