/*
 * Astro QTree - an astronomical visualizer for PalmOS devices.
 *
 * $Id: AstroQTree.c 994 2010-06-24 22:45:56Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <PceNativeCall.h>
#include "Astro.h"
#include "AstroCat.h"
#include "AstroLib.h"
#include "AstroQTree.h"
#include "AstroSky.h"
#include "AstroMatrix.h"
#include "FixedPoint.h"
#include "AstroPnoUtil.h"
#include "Grayscale.h"
#include "resource.h"

#include "palmprof.h"

#undef DEBUG_QTREE

static Int32  xMin;
static Int32  xMax;
static Int32  yMin;
static Int32  yMax;

#define INVALID_CATNUM 0xFFFF

#define LOAD_SMALL_RADEC(r,p) { MemMove(r, ((QTreeObjectSmall*)(p))->radecData, 3); \
                MemMove((r)+1, ((QTreeObjectSmall*)(p))->radecData+3, 3); \
                MemMove((r)+2, ((QTreeObjectSmall*)(p))->radecData+6, 3); }

static QTREESECTION void
QTreeIterate (Int16 recidx,
              SkyDrawContext *ctx, DBRecord *db, DrawObjectFunc drawObject,
              Int32 minDec, Int32 maxDec, Int32 minRa, Int32 maxRa
#ifdef DEBUG_QTREE
              ,Int32 minD, Int32 maxD, Int32 minR, Int32 maxR
#endif
	      ) {
    MemHandle handle;
    QTreeHeader *header;
    UInt8 *ptr;
    Int16 i, rec1, rec2;
    Int32 minRa1, maxRa2;
    UInt16        magHigh;
    
    magHigh = (UInt16)db->magHigh << 8;

    handle = CatalogGetHandle(db, recidx);
    if(handle == NULL)
        return;
    header = MemHandleLock(handle);
    ptr = (UInt8*) (header+1);

    {
        Int32* rotate0;
        Int32* rotate1;
        Int32* rotate2;
        Int32* zenith;
        UInt16 qTreeObjectSize;
        VectorType fullRadec;
        Int32* radec = fullRadec;

        qTreeObjectSize = db->qTreeObjectSize;

        rotate0 = ctx->rotate[0];
        rotate1 = ctx->rotate[1];
        rotate2 = ctx->rotate[2];
        zenith  = ctx->zenith;

        if (qTreeObjectSize <= SIZE_QTreeObjectMedium) {
            MemSet(fullRadec, sizeof(VectorType), 0);
        }

        for (i = 0; i < header->numObj; i++, ptr += qTreeObjectSize) {
            Int32 tx, ty;
            UInt16 colorOffset;
            Int16  mag;

            mag = MAGNITUDE(qTreeObjectSize, ptr, magHigh);

            if (mag > ctx->magnitude)
                break;

            if (qTreeObjectSize <= SIZE_QTreeObjectMedium) {
                LOAD_SMALL_RADEC(radec, ptr);
            }
            else {
                radec = ((QTreeObject*)ptr)->radec;
            }    

            if (VecFastScalarProd(rotate1, radec) <= 0 ) {
                /* other side of globe */
                continue;
            }

            colorOffset = ctx->colorOffset;
            if (ctx->horizon != BELOW_HORIZON_DRAW && VecFastScalarProd(zenith, radec) <= 0) {
                if (ctx->horizon == BELOW_HORIZON_OMIT)
        	        continue;
       	        colorOffset = ctx->colorOffset + ctx->fadeOffset;
    	    }

            /* Be more accurate for smaller zoom levels */
            if (userPrefs.zoomValue < 18) {
                tx = VecScalarProd(rotate0, radec);
                if (tx < xMin || xMax < tx) {
                    continue;
                }
                ty = VecScalarProd(rotate2, radec);
            } else {
                tx = VecFastScalarProd(rotate0, radec);
                if (tx < xMin || xMax < tx) {
                    continue;
                }
                ty = VecFastScalarProd(rotate2, radec);
            }

            if (ty < yMin || yMax < ty )
                continue;

#ifdef TEST_NEW_PROJECTION
            /* broken after recent optimizations that removed tz */
            if (fpabs(tx) > fpmul(tz, SKY_WIDTH/2+1)
                || fpabs(ty) > fpmul(tz, SKY_HEIGHT/2+1))
                continue;

            tx = fpdiv(tx, tz);
            ty = fpdiv(ty, tz);
#endif

            drawObject(ctx, 
                mag, tx >> userPrefs.zoomValue,
                ty >> userPrefs.zoomValue, colorOffset);
        }
    }
    
    minRa1 = minRa;
    maxRa2 = maxRa;

    if (header->limMag <= ctx->magnitude) {
        Int32   coord = header->coord;

        rec1 = header->lowIdx;
        rec2 = header->highIdx;

        if (header->raSplit) {
            if (minRa > maxRa) {
                if (coord < minRa)
                    minRa1 = -2*FP_ONE;
                if (coord > maxRa)
                    maxRa2 = FP_ONE + (FP_ONE-1);
            } else {
                if (coord < minRa)
                    rec1 = 0;
                if (coord > maxRa)
                    rec2 = 0;
            }
        } else {
            if (coord < minDec)
                rec1 = 0;
            if (coord > maxDec)
                rec2 = 0;
        }
    } else {
        rec1 = rec2 = 0;
    }

    MemHandleUnlock(handle);
    CatalogFreeHandle(db, handle);
    if (rec1)
        QTreeIterate (rec1, ctx, db, drawObject,
                      minDec, maxDec, minRa1, maxRa
#ifdef DEBUG_QTREE
                      ,minD, header->raSplit ? maxD : coord,
                      minR, header->raSplit ? coord : maxR
#endif
                      );
    if (rec2)
        QTreeIterate (rec2, ctx, db, drawObject,
                      minDec, maxDec, minRa, maxRa2
#ifdef DEBUG_QTREE
                      ,header->raSplit ? minD : coord, maxD,
                      header->raSplit ? coord : minR, maxR
#endif
                      );
#ifdef DEBUG_QTREE
    if (!rec1 && !rec2) {
        VectorType pos, newpos;
        Int32 sd,cd,sr1,cr1, sr2,cr2;
        Int16 xul, xur, xll, xlr, yul, yur, yll, ylr;
        fpsincos(minR, &sr1, &cr1);
        fpsincos(maxR, &sr2, &cr2);
        fpsincos(minD, &sd, &cd);

        pos[0] = fpmul(cr1, cd);
        pos[1] = fpmul(sr1, cd);
        pos[2] = sd;
        MatVecMul(newpos, ctx->rotate, pos);
        xll = newpos[0] >> (3 + userPrefs.zoomValue);
        yll = newpos[2] >> (3 + userPrefs.zoomValue);

        pos[0] = fpmul(cr2, cd);
        pos[1] = fpmul(sr2, cd);
        MatVecMul(newpos, ctx->rotate, pos);
        xlr = newpos[0] >> (3 + userPrefs.zoomValue);
        ylr = newpos[2] >> (3 + userPrefs.zoomValue);

        fpsincos(maxD, &sd, &cd);

        pos[0] = fpmul(cr1, cd);
        pos[1] = fpmul(sr1, cd);
        pos[2] = sd;
        MatVecMul(newpos, ctx->rotate, pos);
        xul = newpos[0] >> (3 + userPrefs.zoomValue);
        yul = newpos[2] >> (3 + userPrefs.zoomValue);

        pos[0] = fpmul(cr2, cd);
        pos[1] = fpmul(sr2, cd);
        MatVecMul(newpos, ctx->rotate, pos);
        xur = newpos[0] >> (3 + userPrefs.zoomValue);
        yur = newpos[2] >> (3 + userPrefs.zoomValue);

        WinEraseLine(SKY_WIDTH/2 + xll, SKY_HEIGHT/2 + yll,
                     SKY_WIDTH/2 + xlr, SKY_HEIGHT/2 + ylr);
        WinEraseLine(SKY_WIDTH/2 + xul, SKY_HEIGHT/2 + yul,
                     SKY_WIDTH/2 + xur, SKY_HEIGHT/2 + yur);
        WinEraseLine(SKY_WIDTH/2 + xll, SKY_HEIGHT/2 + yll,
                     SKY_WIDTH/2 + xul, SKY_HEIGHT/2 + yul);
        WinEraseLine(SKY_WIDTH/2 + xlr, SKY_HEIGHT/2 + ylr,
                     SKY_WIDTH/2 + xur, SKY_HEIGHT/2 + yur);
    }
#endif
}

void QTREESECTION
QTreeDrawSph (SkyDrawContext *ctx, DrawObjectFunc drawObject)
{
    UInt16 i;
    UInt16 nonStars;
    Int32 radius, scaler;
    Int32 minRa, maxRa;
    Int32 minDec, maxDec;
    Int32 raCnt, decCnt;
    UInt32 proc;
    Int16  magLimit;

#ifdef DEBUG_QTREE
    Err error;
    WinHandle win, oldwin;

    WinEraseLine(7*SKY_WIDTH/16, 7*SKY_HEIGHT/16,
                 9*SKY_WIDTH/16, 7*SKY_HEIGHT/16);
    WinEraseLine(7*SKY_WIDTH/16, 7*SKY_HEIGHT/16,
                 7*SKY_WIDTH/16, 9*SKY_HEIGHT/16);
    WinEraseLine(7*SKY_WIDTH/16, 9*SKY_HEIGHT/16,
                 9*SKY_WIDTH/16, 9*SKY_HEIGHT/16);
    WinEraseLine(9*SKY_WIDTH/16, 7*SKY_HEIGHT/16,
                 9*SKY_WIDTH/16, 9*SKY_HEIGHT/16);
#endif

    raCnt  = fpatan2(ctx->rotate[1][1], ctx->rotate[1][0]);
    decCnt = fpatan2(ctx->rotate[1][2], 
		     fpsqrt(FP_ONE - fpsq(ctx->rotate[1][2])));

    radius = fpmul(SKY_WIDTH + SKY_HEIGHT, FP_2_PI/2);
    if (radius >= 1L << (30 - userPrefs.zoomValue))
        radius = FP_ONE-1;
    else
        radius <<= userPrefs.zoomValue;
    
    minDec = decCnt - radius;
    maxDec = decCnt + radius;
    if (minDec < -FP_ONE)
        minDec = -FP_ONE;
    if (maxDec > FP_ONE)
        maxDec = FP_ONE;

    scaler = fpcos(fpabs(minDec) < fpabs(maxDec) ? maxDec : minDec);
    if (radius > scaler) {
        minRa  = -2*FP_ONE;
        maxRa  = FP_ONE + (FP_ONE-1);
    } else {
        radius = fpdiv(radius, scaler);
        minRa = raCnt - radius;
        maxRa = raCnt + radius;
    }

#ifdef DEBUG_QTREE
    {
        VectorType pos, newpos;
        Int32 sd,cd,sr1,cr1, sr2,cr2;
        Int16 xul, xur, xll, xlr, yul, yur, yll, ylr;
        fpsincos(minRa, &sr1, &cr1);
        fpsincos(maxRa, &sr2, &cr2);
        fpsincos(minDec, &sd, &cd);

        pos[0] = fpmul(cr1, cd);
        pos[1] = fpmul(sr1, cd);
        pos[2] = sd;
        MatVecMul(newpos, ctx->rotate, pos);
        xll = newpos[0] >> (3 + userPrefs.zoomValue);
        yll = newpos[2] >> (3 + userPrefs.zoomValue);

        pos[0] = fpmul(cr2, cd);
        pos[1] = fpmul(sr2, cd);
        MatVecMul(newpos, ctx->rotate, pos);
        xlr = newpos[0] >> (3 + userPrefs.zoomValue);
        ylr = newpos[2] >> (3 + userPrefs.zoomValue);

        fpsincos(maxDec, &sd, &cd);

        pos[0] = fpmul(cr1, cd);
        pos[1] = fpmul(sr1, cd);
        pos[2] = sd;
        MatVecMul(newpos, ctx->rotate, pos);
        xul = newpos[0] >> (3 + userPrefs.zoomValue);
        yul = newpos[2] >> (3 + userPrefs.zoomValue);

        pos[0] = fpmul(cr2, cd);
        pos[1] = fpmul(sr2, cd);
        MatVecMul(newpos, ctx->rotate, pos);
        xur = newpos[0] >> (3 + userPrefs.zoomValue);
        yur = newpos[2] >> (3 + userPrefs.zoomValue);

        WinEraseLine(SKY_WIDTH/2 + xll, SKY_HEIGHT/2 + yll,
                     SKY_WIDTH/2 + xlr, SKY_HEIGHT/2 + ylr);
        WinEraseLine(SKY_WIDTH/2 + xul, SKY_HEIGHT/2 + yul,
                     SKY_WIDTH/2 + xur, SKY_HEIGHT/2 + yur);
        WinEraseLine(SKY_WIDTH/2 + xll, SKY_HEIGHT/2 + yll,
                     SKY_WIDTH/2 + xul, SKY_HEIGHT/2 + yul);
        WinEraseLine(SKY_WIDTH/2 + xlr, SKY_HEIGHT/2 + ylr,
                     SKY_WIDTH/2 + xur, SKY_HEIGHT/2 + yur);
    }
#endif

    if (errNone == FtrGet(sysFileCSystem, sysFtrNumProcessorID, &proc) &&
             sysFtrNumProcessorIsARM(proc)
        && ctx->bits != NULL && (screenDepth == 16 || screenDepth == 8)) {

        ctx->qTreePno = MemPtrNew(sizeof(QTreePnoType));

        if ( ctx->qTreePno != NULL ) {

            ctx->qTreePno->handle = DmGetResource(RESOURCE_TYPE_PNO,
                RESOURCE_NUM_QTREE_PNO);

            if (ctx->qTreePno->handle == NULL) {
                MemPtrFree(ctx->qTreePno);
                ctx->qTreePno=NULL;
            }
            else {
                ctx->qTreePno->function = MemHandleLock(ctx->qTreePno->handle);
            }
        }
    }
    else {
        ctx->qTreePno = NULL;
    }

    xMin    = (Int32)(-(SKY_WIDTH/2+1)) << userPrefs.zoomValue;
    xMax    = (Int32)((SKY_WIDTH/2+1)) << userPrefs.zoomValue;
    yMin    = (Int32)(-(SKY_HEIGHT/2+1)) << userPrefs.zoomValue;
    yMax    = (Int32)((SKY_HEIGHT/2+1)) << userPrefs.zoomValue;

    magLimit = ctx->magnitude;

    for (nonStars = 0; nonStars <= 1; nonStars++) {
        for (i = 0; i < numDBs; i++) {

            if (catDBs[i]->isVisible && nonStars == !catDBs[i]->isStars) {
                ctx->magnitude = magLimit;
                if (nonStars && userPrefs.skyFlags.showAllNonStars)
                    ctx->magnitude = 32767;

                ctx->colorOffset = NUM_COLOR_TYPES+i;


                if (catDBs[i]->minMag > ctx->magnitude)
                    continue;

                if (ctx->qTreePno != NULL) {
                    UInt32 colorArray;
                    QTreePnoDataType* d;
                    
                    d = &(ctx->qTreePno->data);
                    
                    MemMove(d->rotate,ctx->rotate,sizeof(MatrixType));
                    MemMove(d->zenith,ctx->zenith,sizeof(VectorType));
                    d->minDec    = minDec;
                    d->maxDec    = maxDec;
                    d->minRa     = minRa;
                    d->maxRa     = maxRa;
                    d->maxMag    = ctx->magnitude;
                    d->skyHeight = SKY_HEIGHT;
                    d->skyWidth  = SKY_WIDTH;
                    d->zoomValue = userPrefs.zoomValue;
                    d->horizon   = ctx->horizon;
                    d->bits      = ctx->bits;
                    d->depth     = screenDepth;
                    d->rowBytes      = ctx->rowBytes;
                    d->maxBrightnessForColorList =
                        MAX_BRIGHTNESS;
                    d->maxBrightnessForDrawingCode =
                        ctx->maxBrightnessForDrawingCode;

                    colorArray = (UInt32)(ctx->colors[ctx->colorOffset]);
                    d->colors = (UInt16*)SWAP32(colorArray);
                    colorArray = (UInt32)(ctx->colors[ctx->colorOffset+ctx->fadeOffset]);
                    d->fadedColors = (UInt16*)SWAP32(colorArray);
                    d->colorNum = NUM_COLOR_TYPES+i;
                    d->qTreeObjectSize = catDBs[i]->qTreeObjectSize;
                    d->magHigh         = catDBs[i]->magHigh;
                    d->catDB           = catDBs[i];
                    d->rootIdx         = catDBs[i]->qtreeRootBlock;

                    PceNativeCall(ctx->qTreePno->function, d);
                } else {
                    QTreeIterate(catDBs[i]->qtreeRootBlock,
                                 ctx, catDBs[i], drawObject, minDec, maxDec, minRa, maxRa
#ifdef DEBUG_QTREE
                                 ,-FP_ONE, FP_ONE,
                                 -2*FP_ONE, FP_ONE + (FP_ONE-1)
#endif
                                 );
                }
            }
        }

    }

    if (ctx->qTreePno != NULL) {
        MemHandleUnlock(ctx->qTreePno->handle);
        DmReleaseResource(ctx->qTreePno->handle);
        MemPtrFree(ctx->qTreePno);
        ctx->qTreePno = NULL;
    }
}

static void QTREESECTION
QTreeFindRecursive (DBRecord *db, Int16 recidx,
                    Int32 ra, Int32 dec,
                    VectorType radec, Int16 magnitude,
                    FinderType* finder)
{
    MemHandle handle;
    QTreeHeader *header;
    UInt8 *ptr;
    Int16 i, rec1, rec2;
    UInt16 qTreeObjectSize;
    VectorType fullRadec;
    Int32* objectRadec = fullRadec;
    UInt16 magHigh;
    Int32 click;
    Int32 split;

    handle = CatalogGetHandle(db, recidx);
    header = MemHandleLock(handle);
    ptr = (UInt8*) (header+1);
    magHigh = (UInt16)db->magHigh << 8;

    qTreeObjectSize = db->qTreeObjectSize;

    for (i = 0; i < header->numObj;
                    i++,
                    ptr += qTreeObjectSize) {
        UInt32 dist;

        if (MAGNITUDE(qTreeObjectSize,ptr,magHigh) > magnitude)
            break;
            
        if (qTreeObjectSize <= SIZE_QTreeObjectMedium) {
            LOAD_SMALL_RADEC(objectRadec, ptr);
        }
        else {
            objectRadec = ((QTreeObject*)ptr)->radec;
        }

        dist = SkyDistance(objectRadec, radec);

        if (dist < finder->ignoreDistance)
            SkyMapUpdateFinder(finder, dist, INVALID_OBJECT_INDEX, 
                db->catNum, recidx, ptr - (UInt8*)(header+1));
    }

    if (header->limMag <= magnitude) {
        click = (header->raSplit ? ra: dec);
        split = header->coord;

        if (click < split) {
            rec1 = header->lowIdx;
            rec2 = header->highIdx;
        } else {
            rec1 = header->highIdx;
            rec2 = header->lowIdx;
        }
    } else {
        rec1 = rec2 = 0;
        click = split = 0;
    }
        
    MemHandleUnlock(handle);
    CatalogFreeHandle(db, handle);
    if (rec1)
        QTreeFindRecursive (db, rec1, ra, dec,
                            radec, magnitude, finder);
    
    if (rec2) {
        /* Check, whether we should look into the other node of the
         * tree.  This depends on the distance between split line
         * and click point and the best distance found so far.
         */

        /* Calculate distance of coordinates to split line */
        UInt32 distSq = fpsq(click-split);
        /* Adjust for declination, not exact, but good enough */
        if (header->raSplit)
            distSq = fpmul(distSq, FP_ONE - fpsq(radec[2]));
        
        /* Exact distance is fpsq(2*fpsin(fpsqrt(dist)/2)).
         * We approximate it with "dist" in radians.
         * To convert dist into radians multiply with PI/2.
         * As this would leave the fixed point range we multiply
         * bestDist with 2/PI instead.
         */
        if (fpmul(fpsq(finder->ignoreDistance), FP_2_PI) > distSq)
            QTreeFindRecursive (db, rec2, ra, dec,
                                radec, magnitude, finder);
    }
}

void QTREESECTION
QTreeFindObject (VectorType radec, Int16 magnitude,
                 FinderType* finder)
{
    Int32 ra  = fpatan2(radec[1], radec[0]);
    Int32 dec = fpatan2(radec[2], fpsqrt(FP_ONE - fpsq(radec[2])));
    Int16 i;

    for (i = 0; i < numDBs; i++) {
        if (catDBs[i]->isVisible) {
        Int16 searchMagnitude;

        if (userPrefs.skyFlags.showAllNonStars && !catDBs[i]->isStars)
            searchMagnitude = 32767;
        else
            searchMagnitude = magnitude;

        if(!catDBs[i]->isStars || !userPrefs.optFlags.omitStars)
    	    QTreeFindRecursive(catDBs[i],
    			       catDBs[i]->qtreeRootBlock,
    			       ra, dec,
    			       radec, searchMagnitude, finder);
        }
    }

    for (i=0; i < finder->numObjects ; i++) {
        if (finder->found[i].objectIndex == INVALID_OBJECT_INDEX) {
            UInt32 itemToFind;
            UInt32 offset;
            UInt16    j;
            UInt16    catNum;
            DBRecord *db;

            catNum = finder->found[i].catNum;

            offset = NUMPLANETS;
            for (j = 0; j < catNum; j++) {
                if (catDBs[j]->isVisible) {
              	    offset += catDBs[j]->numRecords;
                }
            }

            db = catDBs[catNum];
            itemToFind = finder->found[i].offsetInNode / db->qTreeObjectSize +
                ((UInt32) finder->found[i].recidx) * db->maxQTreeNodeSize;

            finder->found[i].objectIndex =
                offset + CatalogGetIndexFromQTree(db, itemToFind);
        }
    }
}


void QTREESECTION
QTreeGetObject(DBRecord *db, UInt32 index, AstroRecordType *art)
{
    MemHandle handle;
    QTreeHeader *header;
    QTreeObject *ptr;

    handle = CatalogGetHandle(db, index / db->maxQTreeNodeSize);
    header = MemHandleLock(handle);
    ptr = (QTreeObject*) ( (UInt8*)(header+1) +
                                db->qTreeObjectSize * ( index % db->maxQTreeNodeSize ) );
    if (db->qTreeObjectSize <= SIZE_QTreeObjectMedium) {
        MemSet(art->radec, sizeof(VectorType), 0);
        LOAD_SMALL_RADEC(art->radec, ptr);
    }
    else {
        art->radec[0] = ptr->radec[0];
        art->radec[1] = ptr->radec[1];
        art->radec[2] = ptr->radec[2];
    }

    art->magnitude = MAGNITUDE(db->qTreeObjectSize, ptr, (UInt16)db->magHigh << 8);

    if ( SIZE_QTreeObject_up_to_sx <= db->qTreeObjectSize )
        art->width     = ptr->sx;
    else
        art->width     = 0;

    if ( SIZE_QTreeObject_up_to_sy <= db->qTreeObjectSize )
        art->minoraxis = ptr->sy;
    else
        art->minoraxis = art->width;
        
    art->distance = -1;

    if ( SIZE_QTreeObject_up_to_extraData <= db->qTreeObjectSize ) {
        if (db->distance) {
            float d;

            MemMove(&d, &(ptr->extraData.distance), sizeof(float));
            art->distance = d;
            art->objclass = 0xFFFF;
        }
        else {
            art->objclass = ptr->extraData.typeIndex.type;
        }
    }

    MemHandleUnlock(handle);
    CatalogFreeHandle(db, handle);
}
