#ifndef ARM
#define ARM
#endif

#include <Standalone.h>
#include <PalmOS.h>
#include <PceNativeCall.h>
#include <ByteOrderUtils.h>
#include "resource.h"
#include "AstroQTreePno.h"
#include "AstroQTree.h"
#include "FixedPoint.h"
#include "stubs.h"

STANDALONE_CODE_RESOURCE_ID(RESOURCE_NUM_QTREE_PNO);

#define Read68KUnaligned16(addr)  \
	(((((unsigned char *)(addr))[0]) << 8) | \
	 ((((unsigned char *)(addr))[1])))
#define Read68KUnaligned24(addr)  \
	(((((unsigned char *)(addr))[0]) << 24) | \
	 ((((unsigned char *)(addr))[1]) << 16) | \
	 ((((unsigned char *)(addr))[2]) <<  8))


/* ARM hack to get rid of .GOT (global offset table).
 * Problem is that arm-gcc automatically creates a got for constant strings,
 * but PalmOS cannot handle it. This works around this.
 *
 * If a function needs a constant string "str", add the macro in front of
 * the function as 
 *   CONST_STR_DEF(str_val, "str"); 
 * and then replace "str" with str_val() in the function.
 */
#define CONST_STR_DEF(name, str) \
    static const char *name() { \
      const char *x;                             \
      __asm__("adrl\t%0, .Lsz" #name : "=r" (x)); \
      return x; } \
    __asm__("\n.Lsz" #name ":\n\t.ASCIZ \"" str "\"");



static void DrawObjects(QTreePnoDataType* d, int recidx, int numObj, UInt8* ptr);
static void QTreeIterate(DBRecord *catDB, Int16 recidx,
                         QTreePnoDataType* d, 
                         Int32 minDec, Int32 maxDec, Int32 minRa, Int32 maxRa);

UInt32 NativeFunction
    (
    const void*      emulStateP,
    void*            userData68KP,
    Call68KFuncType* call68KFuncP
    )
{
    int i;
    QTreePnoDataType* d = (QTreePnoDataType*) userData68KP;
    d->maxMag       = EndianSwap16(d->maxMag);
    d->skyHeight    = EndianSwap16(d->skyHeight);
    d->skyWidth     = EndianSwap16(d->skyWidth);
    d->rowBytes     = EndianSwap16(d->rowBytes);
    d->bits         = (UInt8*) EndianSwap32(d->bits);

    for (i=0; i<3 ;i++) {
        UInt32 j;
        
        d->zenith[i] = EndianSwap32(d->zenith[i]);
        
        for(j=0; j<3; j++) {
            d->rotate[i][j] = EndianSwap32(d->rotate[i][j]);
        }
    }

    d->xMin    = (Int32)(-(d->skyWidth/2+1)) << d->zoomValue;
    d->xMax    = (Int32)((d->skyWidth/2+1)) << d->zoomValue;
    d->yMin    = (Int32)(-(d->skyHeight/2+1)) << d->zoomValue;
    d->yMax    = (Int32)((d->skyHeight/2+1)) << d->zoomValue;

    QTreeIterate((DBRecord*)Read68KUnaligned32(&d->catDB),
                 EndianSwap16(d->rootIdx), d, 
                 EndianSwap32(d->minDec), EndianSwap32(d->maxDec),
                 EndianSwap32(d->minRa), EndianSwap32(d->maxRa));
    return 0;
}

static MemHandle CatalogGetHandle(DBRecord* db, UInt16 index)
{
   if (db->vfs) {
       MemHandle h;
       
       if (errNone != VFSFileDBGetRecord(Read68KUnaligned32(&db->dbInfo.vfs.fileRef), index, &h, NULL, NULL))
           return NULL;

       return h;
   }
   else {
       return DmQueryRecord((DmOpenRef)Read68KUnaligned32(&db->dbInfo.ram.db), index);
   }
}

static void
CatalogFreeHandle(DBRecord* db, MemHandle h)
{
    if (h != NULL && db->vfs)
        MemHandleFree(h);
}

static void
QTreeIterate (DBRecord *catDB, Int16 recidx, QTreePnoDataType* d, 
              Int32 minDec, Int32 maxDec, Int32 minRa, Int32 maxRa) {

    MemHandle handle;
    QTreeHeader *header;
    Int16 rec1, rec2;
    Int32 minRa1, maxRa2;

    handle = CatalogGetHandle(catDB, recidx);
    if(handle == NULL)
        return;
    header = MemHandleLock(handle);
    
    DrawObjects(d, recidx, header->numObj, (UInt8*) (header+1));

    minRa1 = minRa;
    maxRa2 = maxRa;
    
    if (Read68KUnaligned16(&header->limMag) <= d->maxMag) {
        Int32   coord = Read68KUnaligned32(&header->coord);
        rec1 = Read68KUnaligned16(&header->lowIdx);
        rec2 = Read68KUnaligned16(&header->highIdx);
        if (header->raSplit) {
            if (minRa > maxRa) {
                if (coord < minRa)
                    minRa1 = -2*FP_ONE;
                if (coord > maxRa)
                    maxRa2 = FP_ONE + (FP_ONE-1);
            } else {
                if (coord < minRa)
                    rec1 = 0;
                if (coord > maxRa)
                    rec2 = 0;
            }
        } else {
            if (coord < minDec)
                rec1 = 0;
            if (coord > maxDec)
                rec2 = 0;
        }
    } else {
        rec1 = rec2 = 0;
    }

    MemHandleUnlock(handle);
    CatalogFreeHandle(catDB, handle);

    if (rec1)
        QTreeIterate (catDB, rec1, d, minDec, maxDec, minRa1, maxRa);
    if (rec2)
        QTreeIterate (catDB, rec2, d, minDec, maxDec, minRa, maxRa2);
}


static void BmpBitsDrawHLine(void* bits, UInt8 depth, Int32 x,
                             Int32 xEnd, UInt16 color68K)
{
    if (depth == 16) {
        for (; x < xEnd; x++ )
             ((UInt16*)bits)[x] = color68K;
    }
    else {
        UInt8 color = color68K >> 8;
        for (; x < xEnd; x++)
             ((UInt8*)bits)[x] = color;
    }
}



static void DrawObjectBmp(QTreePnoDataType* d, Int16 mag, Int32 tx, Int32 ty, Boolean fade)
{
    Int32  y, brightness;
    UInt16 color;
    UInt16* colors;
    void *bits;

    if (fade)
        colors = d->fadedColors;
    else
        colors = d->colors;

    tx = d->skyWidth/2  + tx;
    ty = d->skyHeight/2 - ty;

    /* (3*mag) >> 10 is approximately (3/4)*magnitude since mag is GP16 */
    /* userPrefs.zoomValue varies between 1 and 24, with 24 being the lowest
       magnification */
    mag = 25 - (Int16) d->zoomValue - ((3*mag) >> 10);

    if (mag <= 0) {
    	if (tx >= 0 && tx < d->skyWidth && ty >= 0 && ty < d->skyHeight) {
    	    brightness = d->maxBrightnessForColorList + mag;
    	    if (brightness <= 0)
    		    brightness = 1;

    	    /* Draw single pixel */
    	    if (d->depth==16)
                ((UInt16*)((UInt8*)d->bits + ty * d->rowBytes))[tx] =
                     colors[brightness];
            else
                ((UInt8*)d->bits + ty * d->rowBytes)[tx] =
                     colors[brightness] >> 8;
    	}
    	return;
    }

    if (mag > d->maxBrightnessForDrawingCode)
        mag = d->maxBrightnessForDrawingCode;

    /* mag adjustment */
    if (mag == 4)
        mag = 5;

    tx -= mag/2;
    ty -= mag/2;
    bits  = (UInt8*)d->bits + ty * d->rowBytes;
    color = colors[d->maxBrightnessForColorList];
    for (y = 0; y <= mag; y++) {
        if (ty >= 0 && ty < d->skyHeight) {
            Int32 xoff = tx;
            Int32 xend = tx+mag+1;
            if (mag >= 2 && (y == 0 || y == mag)) {
                xoff++;
                xend--;
            }
            if (xoff < 0)
                xoff = 0;
            if (xend >= d->skyWidth)
                xend = d->skyWidth;
            if (xoff < xend)
                BmpBitsDrawHLine(bits, d->depth, xoff, xend, color);
        }
        ty++;
        bits += d->rowBytes;
    }
}

#ifdef DEBUG_PNO        
CONST_STR_DEF(formatStr, "%d,%d: %x %lx %lx %lx");
#endif

static void DrawObjects(QTreePnoDataType* d, int recIdx, int numObj, UInt8* ptr)
{
    int  i;
    UInt32  qTreeObjectSize;
    Boolean fade;
    VectorType radec;
    UInt16  magHigh;
    
    magHigh = ((UInt16)d->magHigh) << 8;

    qTreeObjectSize = d->qTreeObjectSize;

    for (i = numObj; 0 < i ; i--, ptr += qTreeObjectSize) {

        Int32 tx;
        Int32 ty;
        Int16 mag;
        
        mag = MAGNITUDE(qTreeObjectSize, ptr, magHigh);

        if ( mag > d->maxMag )
            return;
        
        if (qTreeObjectSize <= SIZE_QTreeObjectMedium) {
            radec[0] = Read68KUnaligned24(((QTreeObjectSmall*)ptr)->radecData);
            radec[1] = Read68KUnaligned24(((QTreeObjectSmall*)ptr)->radecData+3);
            radec[2] = Read68KUnaligned24(((QTreeObjectSmall*)ptr)->radecData+6);
        } else {
            radec[0] = Read68KUnaligned32(&((QTreeObject*)ptr)->radec[0]);
            radec[1] = Read68KUnaligned32(&((QTreeObject*)ptr)->radec[1]);
            radec[2] = Read68KUnaligned32(&((QTreeObject*)ptr)->radec[2]);
        }

        if (VecFastScalarProd(d->rotate[1], radec) <= 0) {
            /* other side of globe */
            continue;
        }
        
        fade = false;
        
        if (d->horizon != BELOW_HORIZON_DRAW
            && VecFastScalarProd(d->zenith, radec) <= 0) {
            if (d->horizon == BELOW_HORIZON_OMIT)
                continue;
            fade = true;
        }
        
        if (d->zoomValue < 18) {
            tx = VecScalarProd(d->rotate[0], radec);
            if (tx < d->xMin || d->xMax < tx)
                continue;
            
            ty = VecScalarProd(d->rotate[2], radec);
        }
        else {
            tx = VecFastScalarProd(d->rotate[0], radec);
            if (tx < d->xMin || d->xMax < tx)
                continue;
            
            ty = VecFastScalarProd(d->rotate[2], radec);
        }
        
        if (ty < d->yMin || d->yMax < ty )
             continue;
        
        tx >>= d->zoomValue;
        ty >>= d->zoomValue;

#ifdef DEBUG_PNO        
        if (recIdx==4299) {
            char msg[400];
            StrPrintF(msg, formatStr(), recIdx, numObj - i,
                      Read68KUnaligned16(ptr+9),
                      Read68KUnaligned24(ptr+0),
                      Read68KUnaligned24(ptr+3),
                      Read68KUnaligned24(ptr+6));
            WinDrawChars(msg, StrLen(msg), 0, 50);
            FrmAlert(a_Warning);
        }
#endif
        DrawObjectBmp(d, mag, tx, ty, fade);
    }
}

STUB_StrPrintF
STUB_FrmAlert
STUB_WinDrawChars
STUB_VFSFileDBGetRecord
STUB_MemHandleLock
STUB_MemHandleUnlock
STUB_MemHandleFree
STUB_DmQueryRecord
