/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroSS.c 963 2010-01-07 21:40:21Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "Astro.h"
#include "AstroSS.h"
#include "AstroSky.h"
#include "AstroTime.h"
#include "AstroTable.h"
#include "resource.h"
#include "AstroLib.h"
#include "AstroDetails.h"
#include "FixedPoint.h"
#include "Grayscale.h"

// Palm Reporter support
#include <HostControl.h>

/*
 * Events for the entry Solar System form.
 */
Boolean SSSECTION
SolarSystemFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr formPtr = FrmGetActiveForm();

    if (DateChangeHandleEvent(event)
        || FormChangeHandleEvent(event))
        return true;

    switch (event->eType) {
        case tblSelectEvent:
	    SetCurrentObject(event->data.tblSelect.row);
	    SkyCenterObject(event->data.tblSelect.row);
	    FrmGotoForm(f_ObjectDetail);
            handled = true;
            break;

        case frmOpenEvent:
            DrawTimePicker();
            DrawDatePicker();

            InitSolarTable(formPtr);
            ShowSolarSystem(formPtr);

            SetMenu(formPtr);
            FrmDrawForm(formPtr);

            handled = true;
            break;

        case frmCloseEvent:
            AstroFreeTable((TablePtr) 
                           FrmGetObjectPtrByID(formPtr, tbl_Details));
            handled = false;
            break;

        case frmUpdateEvent:
            switch (event->data.frmUpdate.updateCode) {
            case updateDateChanged:
            case updateDateChangedAutomatic:
                if(CheckMinRomVersion(sysVersion35))
                    WinScreenLock(winLockCopy);
                ShowSolarSystem(formPtr);
                if(CheckMinRomVersion(sysVersion35))
                    WinScreenUnlock();
                break;
            }
            FrmDrawForm(formPtr);
            handled = true;
            break;
            
        default:
            handled = false;
    }

    return handled;
}


/*
 * This can be used to display rise, trans and set information on any
 * form with two or three time fields.  (Transit time is optional and
 * can be excluded by passing transid as -1)
 */
static void SSSECTION
ShowRiseSet(TablePtr table, Int8 flags, Int32 LTr, Int32 LTs,
            Boolean showndmarker, 
	    Int8 riserow, Int8 risecol, 
	    Int8 transrow, Int8 transcol, 
	    Int8 setrow, Int8 setcol)
{
    Char str[FIELDLENGTH];
    Char flstr[3];
    if (flags & ASTROLIB_DONTRISE) {
        GetStringResource(flags & ASTROLIB_CIRCUMPOLAR ? s_CirPl: s_NvrUp, 
                          str);
	UpdateTblItemText(table, riserow, risecol, str);
    } else {
        // Force 24h format
        FormatTime(LTr,  str, FIELDLENGTH, flstr, true);

        // We only use a separate column for next day in Solar System form
        if (showndmarker) {
            UpdateTblItemText(table, riserow, risecol + 1, flstr);
        } else {
            StrCat(str, flstr);
        }
	UpdateTblItemText(table, riserow, risecol, str);
    }

    TblSetItemFont(table, riserow, risecol, ((FrmGetActiveFormID() == f_SolarSystem) ? 0 : fontSmall));

    // We don't always show the transit time
    if (transrow >= 0) {
        if ((flags & (ASTROLIB_DONTRISE | ASTROLIB_DONTSET))) {
            GetStringResource(flags & ASTROLIB_CIRCUMPOLAR ? s_CirPl: s_NvrUp,
                              str);
	    UpdateTblItemText(table, transrow, transcol, str);
                             
        } else {
            // Force 24h format
            FormatTime((LTs + LTr) / 2,  str, FIELDLENGTH, flstr, true);

			// We only use a separate column for next day in Solar System form
	    if (showndmarker) {
		UpdateTblItemText(table, transrow, transcol + 1, flstr);
	    } else {
		StrCat(str, flstr);
	    }
	    UpdateTblItemText(table, transrow, transcol, str);
        }
        TblSetItemFont(table, transrow, transcol, ((FrmGetActiveFormID() == f_SolarSystem) ? 0 : fontSmall));
    }
    
    if (flags & ASTROLIB_DONTSET) {
        GetStringResource(flags & ASTROLIB_CIRCUMPOLAR ? s_CirPl: s_NvrUp, 
                          str);
	UpdateTblItemText(table, setrow, setcol, str);
    } else {
        // Force 24h format
        FormatTime(LTs,  str, FIELDLENGTH, flstr, true);

        // We only use a separate column for next day in Solar System form
        if (showndmarker) {
	    UpdateTblItemText(table, setrow, setcol + 1, flstr);
        } else {
            StrCat(str, flstr);
        }
	UpdateTblItemText(table, setrow, setcol, str);
    }
    TblSetItemFont(table, setrow, setcol, ((FrmGetActiveFormID() == f_SolarSystem) ? 0 : fontSmall));
}

void SSSECTION
ShowSolarSystem(FormPtr formPtr)
{
    Int16         planet, flags;
    Int32       sinlat, coslat;
    Int32       UTr, UTs;
    Int32       tz;
    Int16			font = fontNormal;
    Char        str[FIELDLENGTH];

    TablePtr    tablePlanets = (TablePtr)FrmGetObjectPtrByID(formPtr, tbl_Details);

    tz = currentTzDST;

    fpsincos(userPrefs.Location[userPrefs.currentLoc].lat, &sinlat, &coslat);

    for (planet = SUN; planet >= MERCURY; planet--)
    {
        SysStringByIndex(stbl_SolarObjects, planet, str, FIELDLENGTH);

        TblSetItemFont(tablePlanets, planet, 0, fontNormal);

	flags = SolObjectRiseSet(planet, userPrefs.jday, userPrefs.jtime, 
				 sinlat, coslat, userPrefs.Location[userPrefs.currentLoc].lng, &UTr, &UTs);
	
	// Bold planet name if above the horizon at current time
	font = UpNow(flags, UTr + tz, UTs + tz, 
		     userPrefs.jtime + currentTzDST) ? boldFont : stdFont;
	TblSetItemFont(tablePlanets, planet, 0, font);
	
	ShowRiseSet(tablePlanets, flags, UTr + tz, UTs + tz, 
			true, planet, 1, planet, 3, planet, 5);
        
        UpdateTblItemText(tablePlanets, planet, 0, str);
        TblSetItemInt(tablePlanets, planet, 0, rightAlign);
    }
}
