/*
 * Astro Sky - an astronomical visualizer for PalmOS devices.
 *
 * $Id: AstroSky.c 994 2010-06-24 22:45:56Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <BmpGlue.h>
#include "Bitmap.h"
#include "AstroGlue.h"
#include "constants.h"
#include "resource.h"
#include "Astro.h"
#include "AstroBW.h"
#include "AstroLib.h"
#include "AstroMoon.h"
#include "AstroEphem.h"
#include "AstroCat.h"
#include "AstroSky.h"
#include "AstroDetails.h"
#include "AstroTable.h"
#include "Grayscale.h"
#include "AstroMatrix.h"
#include "FixedPoint.h"
#include "AstroTime.h"
#include "AstroRocker.h"
#include "AstroQTree.h"
#include "AstroAlign.h"
#include "AstroFetchImage.h"
#include "Bitmap.h"
#include "PalmUtil.h"

#include "palmprof.h"
#include <HostControl.h>

#define FADE_NUM     45
#define FADE_DEN    100 // Fading is FADE_NUM / FADE_DEN

#define MAX_ZOOM     (screenDensity >= kDensityDouble ? 23 : 24)

#define MAX_DSS_RANGE DMS_TO_ANG(2,30,0)

#define FIND_PIXEL_PRECISION 2ul

#define NUM_NON_CUSTOM_CIRCLES 12
static Int32 circles[NUM_NON_CUSTOM_CIRCLES] = {
     DMS_TO_ANG(10,0,0),
     DMS_TO_ANG(5,0,0),
     DMS_TO_ANG(4,0,0),
     DMS_TO_ANG(3,0,0),

     DMS_TO_ANG(2,0,0),
     DMS_TO_ANG(1,30,0),
     DMS_TO_ANG(1,0,0),
     DMS_TO_ANG(0,45,0),

     DMS_TO_ANG(0,30,0),
     DMS_TO_ANG(0,15,0),
     DMS_TO_ANG(0,7,30),
     DMS_TO_ANG(0,5,0)
};

static struct {
    UInt16 d;
    UInt16 m;
    UInt16 s;
    UInt16 name;
} customCircleFields[NUM_CUSTOM_CIRCLES] = {
     { f_custom1d, f_custom1m, f_custom1s, f_custom1name },
     { f_custom2d, f_custom2m, f_custom2s, f_custom2name },
     { f_custom3d, f_custom3m, f_custom3s, f_custom3name },
     { f_custom4d, f_custom4m, f_custom4s, f_custom4name },
     { f_custom5d, f_custom5m, f_custom5s, f_custom5name },
     { f_custom6d, f_custom6m, f_custom6s, f_custom6name },
     { f_custom7d, f_custom7m, f_custom7s, f_custom7name },
};

static const UInt16 circleIDs[] = {
     c_10deg,
     c_5deg,
     c_4deg,
     c_3deg,

     c_2deg,
     c_1_5deg,
     c_1deg,
     c_45min,

     c_30min,
     c_15min,
     c_7_5min,
     c_5min,

     c_custom1,
     c_custom2,
     c_custom3,
     c_custom4,
     c_custom5,
     c_custom6,
     c_custom7,
};

#define SKY_FULLSCREEN_X (lowresSkyRect.topLeft.x + lowresSkyRect.extent.x - 8)
#define SKY_FULLSCREEN_Y (lowresSkyRect.topLeft.y + lowresSkyRect.extent.y - 8)

#define SKY_ORIENTATION_X (lowresSkyRect.topLeft.x + lowresSkyRect.extent.x - 16)
#define SKY_ORIENTATION_Y (lowresSkyRect.topLeft.y + lowresSkyRect.extent.y - 8)

#define SKY_POS_TIMEOUT (2 * ticksPerSec)
#define SKY_CNTX (SKY_X + SKY_WIDTH / 2)
#define SKY_CNTY (SKY_Y + SKY_HEIGHT / 2)

#define IS_FULLSCREEN (FrmGetActiveFormID() == f_FullScreen)

/* This sets the number of ms to wait between auto-updates.  This avoids the
   device becoming non-responsive due too heavy redraw load. */
#define MINIMUM_WAIT_MS  200

static PlanetInfoType planetInfo[NUMPLANETS];
static VectorType ecliptic;
static BitmapPtr bitmap = NULL;
static BitmapPtr tmpBitmap = NULL;
static WinHandle skywin;
static UInt16    numLinePositions = 0;
static DmOpenRef lineDB = NULL;
static MemHandle lineHandle = NULL;
static VectorType *linePositions;
static Boolean*  lineNewLine;
static UInt32    lastDateChangedRedrawTicks = 0;
static Boolean   needFullDraw = false;
static Boolean   fastModePenDown = false;

typedef struct {
    VectorType center;
    Int32      r;
} ConstArea;

ConstArea* constAreas;
Char*      constNames;
UInt16     numConstAreas = 0;
static MemHandle constAreasHandle = NULL;

RectangleType lowresSkyRect;
RectangleType skyRect;

static Boolean displayPos = false;
static UInt32 displayPosTimeout = 0;

static void SkyMapDescribe(FormPtr formPtr, UInt32 objectNr) SKYSECTION;
static Boolean InSkyMapWindow(void) SKYSECTION;

static RockerType DetailRocker, ZoomRocker;
static DmOpenRef planetBmpDB;

static UInt32 currentLst, currentLat;
static Int32 currentAlt, currentAz;

static Coord maxBrightness;

static MatrixType precession;
static VectorType zenith;


SKYSECTION ColorModeType GetColorMode(void)
{
    if (userPrefs.optFlags.nightMode)
        return MODE_NIGHT;
    else if (userPrefs.optFlags.blackBackground)
        return MODE_ONBLACK;
    else
        return MODE_ONWHITE;
}

/* Generic Cohen-Sutherland line clipping function. */
/* Based on code by Sam Hocevar under the WTFPL:
   http://sam.zoy.org/wtfpl/COPYING */

static SKYSECTION Int32 MulDiv(UInt32 a, Int32 b, UInt32 c)
{
    UInt32 tmp;
    UInt32 sign = 0;
    if (b < 0) {
	b = -b; sign = -1;
    }
    umulw(b, a, tmp);
    udivw(tmp, c, tmp);
    return (tmp ^ sign) - sign;
}

static SKYSECTION Boolean ClipPoint(Int32* x1P, Int32* y1P, Int32 x2, Int32 y2)
{
    Int32 x1 = *x1P;
    Int32 y1 = *y1P;

    if (x1 < 0) {
	if (x2 < 0)
	    return false;
	y1 = y2 - MulDiv(x2 - 0, y2 - y1, x2 - x1);
	x1 = 0;
    } else if (x1 > SKY_WIDTH-1) {
	if (x2 > SKY_WIDTH-1) 
	    return false;
	y1 = y2 - MulDiv((SKY_WIDTH-1) - x2, y2 - y1, x1 - x2);
	x1 = SKY_WIDTH-1;
    }

    if (y1 < 0) {
	if (y2 < 0)
	    return false;
	x1 = x2 - MulDiv(y2 - 0, x2 - x1,  y2 - y1);
	y1 = 0;
    } else if (y1 > SKY_HEIGHT-1) {
	if (y2 > SKY_HEIGHT-1) 
	    return false;
	x1 = x2 - MulDiv((SKY_HEIGHT-1) - y2, x2 - x1,  y1 - y2);
	y1 = SKY_HEIGHT-1;
    }
    if (x1 < 0 || x1 > SKY_WIDTH-1)
	return false;
    *x1P = x1;
    *y1P = y1;
    return true;
}


static void SKYSECTION
FreeLines(void)
{
    if (lineHandle != NULL) {
        DmReleaseRecord(lineDB,0,false);
        MemHandleUnlock(lineHandle);
        lineHandle = NULL;
        numLinePositions = 0;
    }
    if (constAreasHandle != NULL) {
        DmReleaseRecord(lineDB,1,false);
        MemHandleUnlock(constAreasHandle);
        constAreasHandle = NULL;
        numConstAreas = 0;
    }
    if (lineDB != NULL) {
        DmCloseDatabase(lineDB);
        lineDB=NULL;
    }
}


static void SKYSECTION
LoadLines(void)
{
    lineDB = DmOpenDatabaseByTypeCreator(LINES_DBID, APPID, dmModeReadOnly);

    if (lineDB == NULL)
        return;

    lineHandle = DmGetRecord(lineDB, 0);

    if (lineHandle != NULL) {
        UInt8* data;

        data = MemHandleLock(lineHandle);
        numLinePositions = MemHandleSize(lineHandle) / (sizeof(VectorType)+sizeof(Boolean));
        lineNewLine   = (Boolean*)( data + numLinePositions * sizeof(VectorType));
        linePositions = (VectorType*)data;
    }

    constAreasHandle = DmGetRecord(lineDB, 1);

    if (constAreasHandle != NULL) {
        UInt8* data;

        data = MemHandleLock(constAreasHandle);
        if(*(UInt16*)data != 0) {
            /* old version */
            MemHandleUnlock(constAreasHandle);
            DmReleaseRecord(lineDB, 1, false);
            constAreasHandle = NULL;
        }
        else {
            numConstAreas = *(UInt16*)(data + 2);
            constAreas    = (void*)(data + 4);
            constNames    = (void*)(data + 4 + numConstAreas*sizeof(ConstArea));
        }
    }
}



void SKYSECTION
SkyMapFreeCatalog() {
    Int16 i;

    FreeLines();

    for(i = 0; i < NUMPLANETS; i++) {
        if (planetInfo[i].bmpPlanet) {
            BmpDelete(planetInfo[i].bmpPlanet);
            planetInfo[i].bmpPlanet = NULL;
        }
    }
    if (planetBmpDB)
        DmCloseDatabase(planetBmpDB);
}



void SKYSECTION
SkyMapLoadCatalog() {
    HostTraceOutputTL(appErrorClass, "SkyMapLoadCatalog");

    LoadLines();
    planetBmpDB = DmOpenDatabaseByTypeCreator(PLANETS_DBID,
                                              APPID, dmModeReadOnly);
}



void SKYSECTION
SkyMapInit(FormPtr formPtr)
{
    HostTraceOutputTL(appErrorClass, "SkyMapInit");

    FrmDrawForm(formPtr);
    FntSetFont(fontNormal);

    /* Completely full screen sky map? */
    if (IS_FULLSCREEN) {
        /* create space for description box */
        descRect.extent.x  = screenRect.extent.x - 32;
        descRect.extent.y  = FntLineHeight();
        descRect.topLeft.x = 16;
        descRect.topLeft.y = screenRect.extent.y - descRect.extent.y;

        detailRect.topLeft.x = 0;
        detailRect.topLeft.y = descRect.topLeft.y;
        detailRect.extent.x  = screenRect.extent.x;
        detailRect.extent.y  = 0;

        skyRect.topLeft.x = skyRect.topLeft.y = 0;
        skyRect.extent.x = screenRect.extent.x;
        skyRect.extent.y = screenRect.extent.y; //detailRect.topLeft.y;
    } else {
        FntSetFont(fontNormal);
        /* create space for description box */
        descRect.extent.x  = screenRect.extent.x - 32;
        descRect.extent.y  = FntCharHeight();
        descRect.topLeft.x = 16; // Hardcoded as we need to always have room for rocker buttons
        descRect.topLeft.y = screenRect.extent.y - descRect.extent.y;

        FntSetFont(fontSmall);
        /* Create space for details table. */
        detailRect.extent.x  = screenRect.extent.x;
        detailRect.extent.y  = userPrefs.showDetailRows * (FntLineHeight() + userPrefs.spacingDetailRows);
        detailRect.topLeft.x = 0;
        detailRect.topLeft.y = descRect.topLeft.y - detailRect.extent.y - 1;

        /* Fit sky map with other display objects */
        skyRect.topLeft.x = 0;
        skyRect.topLeft.y = GetTitleAreaHeight();
        skyRect.extent.x = screenRect.extent.x;
        skyRect.extent.y = detailRect.topLeft.y - skyRect.topLeft.y;

        /* Initialise the detail rocker with an arbitrarily high range.
         * This is reset to suit the object in MarkSkyObject().
         */
        InitRocker(&DetailRocker, 0, userPrefs.detailPage, 99);

        // We want the down button to increment.
        SetRockerButtons(&DetailRocker, r_DetUp, r_DetDown);

        userPrefs.detailPage = DetailRocker.Value;

        /* Zoom rocker is straightforward. */
        InitRocker(&ZoomRocker, 1, userPrefs.zoomValue, MAX_ZOOM);

        SetRockerButtons(&ZoomRocker, r_ZoomIn, r_ZoomOut);
        ZoomRocker.zoom = 1;

        DrawDatePicker();
        DrawTimePicker();
    }
    lowresSkyRect = skyRect;
    maxBrightness = 2;

    if (CheckMinRomVersion(sysVersion35)) {
        Err error;

        if (winVersion >= 4) {
            WinPushDrawState();
            WinSetCoordinateSystem(kCoordinatesNative);
            WinScaleRectangle(&skyRect);
            maxBrightness = WinScaleCoord(maxBrightness, false);
            tmpBitmap = BmpCreate(SKY_WIDTH, SKY_HEIGHT,
                                  screenDepth, NULL, &error);
            ErrFatalDisplayIf(error!=errNone, "Could not create bitmap!");
            bitmap = (BitmapPtr) BmpCreateBitmapV3(tmpBitmap, screenDensity,
                                                   BmpGetBits(tmpBitmap),
                                                   NULL);
            ErrFatalDisplayIf(bitmap==NULL, "Could not create bitmap!");
            WinPopDrawState();
        } else {
            bitmap = BmpCreate(SKY_WIDTH, SKY_HEIGHT, screenDepth,
                               NULL, &error);
            ErrFatalDisplayIf(error!=errNone, "Could not create bitmap!");
        }
        skywin = WinCreateBitmapWindow(bitmap, &error);
    }

    SetMenu(formPtr);

    SkyMapSetPosTime(userPrefs.jday, userPrefs.jtime,
                     userPrefs.Location[userPrefs.currentLoc].lat,
                     userPrefs.Location[userPrefs.currentLoc].lng);
}


static SKYSECTION void
SkyStoreViewpoint(void)
{
    if (userPrefs.skyFlags.showRA) {
	userPrefs.dec = currentAlt;
	userPrefs.ra  = FP_ONE - currentAz;
    } else {
	MatrixType rd2aa;
    MatrixType rotaa;
	VectorType radec;
	/*
	 * Build radec/altAz conversion matrix
	 */
	MatRotate(rd2aa, FP_ONE-currentLat, FP_ONE + currentLst);
	MatTranspose(rd2aa);
        MatRotate(rotaa, currentAlt, -currentAz);
	MatVecMul(radec, rd2aa, rotaa[1]);
	userPrefs.ra  = fpatan2(radec[1], radec[0]);
	userPrefs.dec = fpatan2(radec[2], fpsqrt(FP_ONE - fpsq(radec[2])));
    }
}

static SKYSECTION void SkyMapCenterRaDec(VectorType radec)
{
    userPrefs.ra = fpatan2(radec[1], radec[0]);
    userPrefs.dec = fpatan2(radec[2], fpsqrt(FP_ONE - fpsq(radec[2])));

    if (userPrefs.skyFlags.showRA) {
	currentAlt = userPrefs.dec;
	currentAz = FP_ONE - userPrefs.ra;
    } else {
	VectorType altaz;

        RaDecToAltAz(currentLat, currentLst, radec, altaz);
	currentAz = fpatan2(altaz[0], altaz[1]);
        currentAlt = fpatan2(altaz[2], fpsqrt(FP_ONE - fpsq(altaz[2])));
    }
}

/* Transform planets to the given day and time.
 * lst contains local siderean time as angle value,
 * lat contains latitude as angle value.
 */
void SKYSECTION
SkyMapSetPosTime(Int32 jday, Int32 jtime,
                 Int32 lat, Int32 lng)
{
    Int32 gst, lst, dist, illf, semid;
    Int32 sinlat, coslat, sinlst, coslst;
    Int32 sundist;
    VectorType tmp, earth_celestial;
    UInt32 i;

    HostTraceOutputTL(appErrorClass, "SkyMapSetPosTime");
        
    gst = UTCToGST(jtime, getGSTOffset(jday));
    lst = GSTToLST(gst, lng);
    fpsincos(lat, &sinlat, &coslat);
    fpsincos(lst, &sinlst, &coslst);

    /* Build zenith vector in ra/dec coordinates.
     * This is needed for cutting against horizon and for parallax computation
     */
    tmp[0] = fpmul(coslat, coslst);
    tmp[1] = fpmul(coslat, sinlst);
    tmp[2] = sinlat;
    PrecessionMatrix(jday, jtime, precession);
    MatTranspose(precession);
    MatVecMul(zenith, precession, tmp);
    MatTranspose(precession);

    /* The following code calculates the vector that stands
     * perpendicular to the ecliptic pointing north.
     */
    ecliptic[0] = 0;
    ecliptic[1] = SINEPS0;
    ecliptic[2] = COSEPS0;

    /* Compute position of earth */
    HelioPlanetPos(EARTH, jday, jtime, NULL, &sundist, earth_celestial);

    for(i = 0; i < NUMPLANETS; i++) {
        if (i == SUN) {
            SolarPos(jday, jtime, planetInfo[i].pos);
            /* TODO: semid is dependent of distance sun->earth */
            planetInfo[i].semid = fpdiv(SOLAR_A / 2 / 40, sundist);
            planetInfo[i].pa = 0;
            planetInfo[i].magnitude = GP16(-26.7);
            planetInfo[i].colorType = COLOR_SUN;
        } else if (i == MOON) {
            Int32 EC, centerdist, adjust;
            planetInfo[i].colorType = COLOR_MOON;
            LunarPos(jday, jtime, &EC, planetInfo[i].pos,
                     &illf, &planetInfo[i].magnitude,
                     &dist, &semid, &planetInfo[i].pa);
	    /* adjust for topocentric moon.
	     */
	    centerdist = 
		fpdiv(FP(6378.14/384401),
		      fpmul(EC, FP_ONE + fpmul(fpsq(sinlat), 
					       FP(6378.14/6356.755 - 1))));
	    planetInfo[i].pos[0] -= fpmul(zenith[0], centerdist);
	    planetInfo[i].pos[1] -= fpmul(zenith[1], centerdist);
	    planetInfo[i].pos[2] -= fpmul(zenith[2], centerdist);
	    adjust = fpdiv(FP_ONE, fpsqrt(fpsq(planetInfo[i].pos[0])
				     +fpsq(planetInfo[i].pos[1])
				     +fpsq(planetInfo[i].pos[2])));
	    planetInfo[i].pos[0] = fpmul(adjust, planetInfo[i].pos[0]);
	    planetInfo[i].pos[1] = fpmul(adjust, planetInfo[i].pos[1]);
	    planetInfo[i].pos[2] = fpmul(adjust, planetInfo[i].pos[2]);
		
            planetInfo[i].semid = fpmul(semid, FP((1L<<14)/90.0/3600.0));
        } else {
            Int32 semid;
            planetInfo[i].colorType = COLOR_PLANETS;
            /* Transform semid in GP seconds to FP angle value */
            PlanetPos(i, jday, jtime, earth_celestial,
		      planetInfo[i].pos,
                      &illf, &planetInfo[i].magnitude, 
                      &dist, &semid, &planetInfo[i].pa);
            planetInfo[i].semid = fpmul(semid, FP((1L<<14)/90.0/3600.0));
        }
        
        if (planetInfo[i].bmpPlanet) {
            BmpDelete(planetInfo[i].bmpPlanet);
            planetInfo[i].bmpPlanet = NULL;
        }
    }

    currentLat = lat;
    currentLst = lst;
    if (userPrefs.skyFlags.showRA) {
	currentAlt = userPrefs.dec;
	currentAz = FP_ONE - userPrefs.ra;
    } else {
	UInt32 cosdec;
	VectorType altaz;
	fpsincos(userPrefs.ra, &tmp[1], &tmp[0]);
	fpsincos(userPrefs.dec, &tmp[2], &cosdec);
	tmp[0] = fpmul(tmp[0], cosdec);
	tmp[1] = fpmul(tmp[1], cosdec);
	
        RaDecToAltAz(currentLat, currentLst, tmp, altaz);
	currentAz = fpatan2(altaz[0], altaz[1]);
        currentAlt = fpatan2(altaz[2], fpsqrt(FP_ONE - fpsq(altaz[2])));
    }
}



SKYSECTION void
SkyMapSetCoord(Boolean radec)
{
    userPrefs.skyFlags.showRA = radec;
}

void SKYSECTION
SkyMapFree(FormPtr formPtr)
{
    HostTraceOutputTL(appErrorClass, "SkyMapFree");

    if (CheckMinRomVersion(sysVersion35) && bitmap) {
        BmpDelete(bitmap);
        if (tmpBitmap)
            BmpDelete(tmpBitmap);
        WinDeleteWindow(skywin, false);

        /* Need to clear this otherwise we release the memory but not
         * the pointer. */
        bitmap = NULL;
    }
}

static SKYSECTION Int16
SkyMapLimitingMag (Boolean fastMode)
{
    /* Zooming in corresponds to factor 4 (2*width and 2*height)
     * This translates to approximately 1.5 magnitudes
     */
    Int16 mag0;
    Int16 mag;
    UInt32 proc;

    mag0 = userPrefs.skyFlags.brightness * GP16(0.6)
        + (24 - userPrefs.zoomValue) * GP16(1.3) + GP16(1.8);
    mag = mag0 - ( fastMode ? GP16(1.5) : 0 );
    
    if (mag < GP16(8) && fastMode && 
        errNone == FtrGet(sysFileCSystem, sysFtrNumProcessorID, &proc) &&
                 sysFtrNumProcessorIsARM(proc)) {
        if (GP16(8) < mag0)
            mag = GP16(8);
        else
            mag = mag0;
    }

    return mag;
}


/*
 * Create a bitmap for the planet.  This bitmap will have the screen
 * depth and the same resolution as the bitmap in the database.  It
 * will already be "phased", but not rotated.  The bitmap is stored
 * into the planet info structure.
 *
 * planet          : Planet number
 * pa              : Phase angle
 * transparentColor: The color of transparent pixels
 * darkColor       : The color of shadow pixels
 *
 * Returns bitmap, or 0 if error occured.  You should delete it with
 * BmpDelete if you're finished with it.
 */
SKYSECTION BitmapPtr
SkyMapCreatePlanetBitmap(Int16 planet, Int32 pa,
                         Int16 transparentColor, Int16 darkColor)
{
    MemHandle handle;
    BitmapPtr bmpSrc, bmpDest;
    RectangleType bounds;
    Err error;
    WinHandle win, oldwin;
    Int32 horiz1, horiz2;
    Int16 x0, y0, x1, x2, y, radius;
    Int32 gpradius;
    Int32 pixelsize;

    bounds.topLeft.x = bounds.topLeft.y = 0;

    handle = DmGetResource('Tbmp', PLANETS_BMPID + planet);
    if (!handle)
        return 0;
    HostTraceOutputTL(appErrorClass, "DrawObjectImage Got planet image");
    bmpSrc = MemHandleLock(handle);
    BmpGlueGetDimensions(bmpSrc, &bounds.extent.x, &bounds.extent.y, NULL);

    bmpDest = BmpCreate(bounds.extent.x, bounds.extent.y, screenDepth, NULL, 
                        &error);

    if (error) {
        MemHandleUnlock(handle);
        return 0;
    }

    win = WinCreateBitmapWindow(bmpDest, &error);

    if (error) {
        BmpDelete(bmpDest);
        MemHandleUnlock(handle);
        return 0;
    }
    oldwin = WinSetDrawWindow(win);

    if (screenDepth >= 16) {
        RGBColorType prevrgb;
        RGBColorType rgb = { 0, (transparentColor >> 8) & 0xf8, 
                             (transparentColor >> 3) & 0xfc,
                             (transparentColor << 3) & 0xf8 };
        WinGlueSetForeColorRGB(&rgb, &prevrgb);
    } else
        WinSetForeColor(transparentColor);

    WinDrawRectangle(&bounds, 0);
    WinDrawBitmap(bmpSrc, 0, 0);

    x0 = bounds.extent.x/2;
    y0 = radius = bounds.extent.y/2;
    gpradius = radius * GP_ONE;
    pixelsize = (FP_ONE / radius);

    if (pa < 0) {
        /*waxing*/
        horiz1 = -gpradius;
        horiz2 = fpmul(gpradius, -fpcos(pa));
    } else {
        /*waning*/
        horiz1 = fpmul(gpradius, fpcos(pa));
        horiz2 = gpradius;
    }

    if (screenDepth >= 16) {
        RGBColorType prevrgb;
        RGBColorType rgb = { 0, (darkColor >> 8) & 0xf8, 
                             (darkColor >> 3) & 0xfc,
                             (darkColor << 3) & 0xf8 };
        WinGlueSetForeColorRGB(&rgb, &prevrgb);
    } else
        WinSetForeColor(darkColor);
    for (y = 0; y < radius; y++) {
        Int32 circx = fpsqrt(FP_ONE - fpsq(y * pixelsize + pixelsize/2));
        x1 = GP_ROUND(fpmul(horiz1, circx));
        x2 = GP_ROUND(fpmul(horiz2, circx));
    
        WinDrawLine(x0+x1, y0-1-y, x0+x2, y0-1-y);
        WinDrawLine(x0+x1, y0+y,   x0+x2, y0+y);
    }

    WinSetDrawWindow(oldwin);
    WinDeleteWindow(win, false);
    MemHandleUnlock(handle);
    return bmpDest;
}


static Err SKYSECTION
DrawObjectImage(SkyDrawContext *ctx, PlanetInfoType *planet,
		Int16 tx, Int16 ty)
{
    Coord width, height;
    BitmapPtr bmp;
    Int16 diam;
    Int32 ex, ey, escale;

    HostTraceOutputTL(appErrorClass, "DrawObjectImage");

    if (!planet->bmpPlanet)
        planet->bmpPlanet = SkyMapCreatePlanetBitmap
            (planet - planetInfo, planet->pa,
             ctx->transparentColor, ctx->colors[COLOR_SHADOW][MAX_BRIGHTNESS]);
    if (!planet->bmpPlanet)
        return sysErrNoFreeResource;

    bmp = planet->bmpPlanet;

    // How big to draw the object in pixels
    diam = fpmul(planet->semid, FP_PI_2)
	>> (userPrefs.zoomValue-1);

    /*
     *  Calculate the angle of the ecliptic vector.
     */
    ex = VecScalarProd(ctx->rotate[0], ecliptic);
    ey = VecScalarProd(ctx->rotate[2], ecliptic);
    escale = fpsqrt(fpsq(ex)+fpsq(ey));

    BmpGlueGetDimensions(bmp, &width, &height, NULL);
    BmpScaleRotateBitmap(ctx->dest, SKY_WIDTH/2 + tx, SKY_HEIGHT/2 - ty,
                         /* may not be circular */
                         (UInt32) diam * width / height, diam,
                         bmp, 0, 0,
                         width, height,
                         fpdiv(ex,escale), fpdiv(ey, escale),
                         XFLIP * YFLIP,
                         ctx->transparentColor, winPaint);
    return 0;
}

/* We use a static draw order.  Sometime we should draw them ordered by
 * distance.  For VENUS, MERCURY and SUN this ordering may be wrong.
 */
static const unsigned short drawOrder[] = {
    PLUTO, NEPTUNE, URANUS, SATURN, JUPITER, MARS, SUN, MERCURY, VENUS, MOON
};

static SKYSECTION void
PlanetsDrawSph (SkyDrawContext *ctx, DrawObjectFunc drawObject)
{
    UInt16 i;

    /*
     * Get the cat strings
     */
    for(i = 0; i < NUMPLANETS; i++) {
        Int32 tx, ty;
        Int32 radius;
        UInt16 colorOffset;
	    PlanetInfoType *planet = &planetInfo[drawOrder[i]];

        if ( planet->magnitude > ctx->magnitude && ! userPrefs.skyFlags.showAllNonStars )
            continue;

        colorOffset = 0;

        if (VecFastScalarProd(ctx->rotate[1], planet->pos) <= 0) {
            /* other side of globe */
            continue;
        }

        if ( ctx->horizon != BELOW_HORIZON_DRAW &&
             VecFastScalarProd(ctx->zenith, planet->pos) < 0) {
            if (ctx->horizon == BELOW_HORIZON_OMIT)
                continue;

            colorOffset = ctx->fadeOffset;
        }

        radius = fpmul(planet->semid, FP_PI_2) >> userPrefs.zoomValue;

        /* Be more accurate for smaller zoom levels */
        if (userPrefs.zoomValue < 18) {
            tx = VecScalarProd(ctx->rotate[0], planet->pos);
            ty = VecScalarProd(ctx->rotate[2], planet->pos);
        } else {
            tx = VecFastScalarProd(ctx->rotate[0], planet->pos);
            ty = VecFastScalarProd(ctx->rotate[2], planet->pos);
        }

        tx >>= (userPrefs.zoomValue);
        ty >>= (userPrefs.zoomValue);

        if (fpabs(tx) > SKY_WIDTH/2+radius
            || fpabs(ty) > SKY_HEIGHT/2+radius)
            continue;

        if (radius < 2
            || !CheckMinRomVersion(sysVersion35)
            || !ctx->dest
            || DrawObjectImage(ctx, planet, tx, ty) != 0) {
            drawObject(ctx, planet->magnitude, tx, ty, planet->colorType + colorOffset);
        }
    }
}



static SKYSECTION void
ConstellationNamesDrawSph(SkyDrawContext* ctx)
{
    void (*fastWinDrawChars)(Char*, UInt16, Coord, Coord);
    RGBColorType textColor;
    RGBColorType fadedTextColor;
    RGBColorType oldTextColor;
    WinDrawOperation oldMode = winPaint;
    UInt16 i;
    Char* p;

    FntSetFont(stdFont);

    if (CheckMinRomVersion(sysVersion35)) {
        RGBColorType back;

        GetBackColor(&back);
        textColor = (userPrefs.colors[GetColorMode()][COLOR_CONST_NAMES]);

        WinGlueSetTextColorRGB(NULL, &oldTextColor);

        fadedTextColor.r = ((UInt16)back.r*(FADE_DEN-FADE_NUM) + (UInt16)textColor.r*FADE_NUM)/FADE_DEN;
        fadedTextColor.g = ((UInt16)back.g*(FADE_DEN-FADE_NUM) + (UInt16)textColor.g*FADE_NUM)/FADE_DEN;
        fadedTextColor.b = ((UInt16)back.b*(FADE_DEN-FADE_NUM) + (UInt16)textColor.b*FADE_NUM)/FADE_DEN;

	    oldMode = WinSetDrawMode(winOverlay);
	    fastWinDrawChars = SysGetTrapAddress(sysTrapWinPaintChars);
    } else {
        Boolean inverted;

	    inverted = userPrefs.optFlags.blackBackground
	      ^ userPrefs.optFlags.nightMode;
        fastWinDrawChars = SysGetTrapAddress((inverted ? sysTrapWinEraseChars
    					 : sysTrapWinDrawChars));
    }

    p = constNames;

    for(i=0, p=constNames; i<numConstAreas; i++, p+=StrLen(p)+1) {
        Boolean faded;
        Int32 tx, ty;
        Coord w;
        Int32 tr;

        if (VecFastScalarProd(ctx->rotate[1], constAreas[i].center) <= 0) {
            /* other side */
            continue;
        }

        faded = false;

        if ( ctx->horizon != BELOW_HORIZON_DRAW &&
             VecFastScalarProd(ctx->zenith, constAreas[i].center) < 0 ) {

            if (ctx->horizon == BELOW_HORIZON_OMIT) {
                continue;
            }
            else {
                faded = true;
            }
        }

        if (userPrefs.zoomValue < 18) {
            tx = VecScalarProd(ctx->rotate[0], constAreas[i].center);
            ty = VecScalarProd(ctx->rotate[2], constAreas[i].center);
        } else {
            tx = VecFastScalarProd(ctx->rotate[0], constAreas[i].center);
            ty = VecFastScalarProd(ctx->rotate[2], constAreas[i].center);
        }

        tx >>= (userPrefs.zoomValue);
        ty >>= (userPrefs.zoomValue);
        tx = SKY_WIDTH/2 + tx;
	    ty = SKY_HEIGHT/2 - ty;

	    if (tx<0 || ty<0 || SKY_WIDTH <=tx || SKY_HEIGHT <=ty)
	        continue;

        w = FntCharsWidth(p, StrLen(p));

        tr = constAreas[i].r >> userPrefs.zoomValue;
        if (tr > SKY_WIDTH * 3 / 2 || tr < 2 * w / 3)
            continue;

        if (CheckMinRomVersion(sysVersion35)) {
            WinGlueSetTextColorRGB(faded?&fadedTextColor:&textColor, NULL);
        }

        fastWinDrawChars(p, StrLen(p), tx-w/2, ty-FntLineHeight()/2);
    }

    if (CheckMinRomVersion(sysVersion35)) {
	    WinSetDrawMode(oldMode);
        WinGlueSetTextColorRGB(&oldTextColor, NULL);
    }
}


static SKYSECTION void
LinesDrawSph (SkyDrawContext *ctx)
{
    UInt16 i;
    Int32 txLast = -1;
    Int32 tyLast = -1;
    Int8  clipbitsLast = 15;
    Boolean inverted;
    Boolean forceNewLine;
    void (*fastWinDrawLine) (Coord, Coord, Coord, Coord);

    Int32 radius;
    Int32 limitZ;
    UInt16 colorOffset;
    UInt16 nextColorOffset;
    RGBColorType normalColor;
    RGBColorType fadedColor;

    /* No need to show constellations on such a high zoom level.
     * and more important: it will overflow MulDiv
     */
    if (userPrefs.zoomValue < 15)
	return;

    radius = fpmul(SKY_WIDTH + SKY_HEIGHT, FP_2_PI/2);
    if (radius >= 3L << (28 - userPrefs.zoomValue))
        radius = FP_ONE-1;
    else
        radius = (radius << userPrefs.zoomValue) + FP(0.25);
    limitZ = fpcos(radius) - (4l << 18);

    if (CheckMinRomVersion(sysVersion35)) {
        RGBColorType back;
        WinPushDrawState();
        normalColor = (userPrefs.colors[GetColorMode()][COLOR_LINES]);
        GetBackColor(&back);
        fadedColor.r = ((UInt16)back.r*(FADE_DEN-FADE_NUM) + (UInt16)normalColor.r*FADE_NUM)/FADE_DEN;
        fadedColor.g = ((UInt16)back.g*(FADE_DEN-FADE_NUM) + (UInt16)normalColor.g*FADE_NUM)/FADE_DEN;
        fadedColor.b = ((UInt16)back.b*(FADE_DEN-FADE_NUM) + (UInt16)normalColor.b*FADE_NUM)/FADE_DEN;
	    WinGlueSetForeColorRGB(&normalColor,NULL);
	inverted = false;
    } else {
	inverted = userPrefs.optFlags.blackBackground
	    ^ userPrefs.optFlags.nightMode;
    }
    fastWinDrawLine = SysGetTrapAddress((inverted ? sysTrapWinEraseLine
					 : sysTrapWinDrawLine));
    forceNewLine = true;
    colorOffset     = 0;
    nextColorOffset = 0;

    for(i = 0; i < numLinePositions; i++) {
        Int32 tx, ty;
	    Int8  clipbits = 0;

        if (VecFastScalarProd(ctx->rotate[1], linePositions[i]) <= limitZ) {
            /* other side of globe */
            forceNewLine = true;
            nextColorOffset = 0;
            continue;
        }

        if ( ctx->horizon != BELOW_HORIZON_DRAW &&
             VecFastScalarProd(ctx->zenith, linePositions[i]) < 0 ) {
            if (ctx->horizon == BELOW_HORIZON_OMIT) {
                forceNewLine = true;
                nextColorOffset = 0;
                continue;
            }
            else {
                colorOffset = ctx->fadeOffset;
                nextColorOffset = ctx->fadeOffset;
            }
        }
        else {
            colorOffset     = nextColorOffset;
            nextColorOffset = 0;
        }

        /* Be more accurate for smaller zoom levels */
        if (userPrefs.zoomValue < 18) {
            tx = VecScalarProd(ctx->rotate[0], linePositions[i]);
            ty = VecScalarProd(ctx->rotate[2], linePositions[i]);
        } else {
            tx = VecFastScalarProd(ctx->rotate[0], linePositions[i]);
            ty = VecFastScalarProd(ctx->rotate[2], linePositions[i]);
        }

        tx >>= (userPrefs.zoomValue);
        ty >>= (userPrefs.zoomValue);
	tx = SKY_WIDTH/2 + tx;
	ty = SKY_HEIGHT/2 - ty;
	if (tx < 0)
	    clipbits |= 1;
	else if (tx >= SKY_WIDTH)
	    clipbits |= 2;

	if (ty < 0)
	    clipbits |= 4;
	else if (ty >= SKY_HEIGHT)
	    clipbits |= 8;

        if( ! forceNewLine && ! lineNewLine[i]
	    && (clipbits & clipbitsLast) == 0 ) {
	    Int32 lx1 = txLast;
	    Int32 ly1 = tyLast;
	    Int32 lx2 = tx;
	    Int32 ly2 = ty;
	    if ((clipbitsLast == 0
		 || ClipPoint(&lx1, &ly1, lx2, ly2))
		&& (clipbits == 0
		    || ClipPoint(&lx2, &ly2, lx1, ly1))) {
             if (colorOffset) {
	             WinGlueSetForeColorRGB(&fadedColor,NULL);
                 fastWinDrawLine(lx1,ly1,lx2,ly2);
	             WinGlueSetForeColorRGB(&normalColor,NULL);
             }
             else {
                 fastWinDrawLine(lx1,ly1,lx2,ly2);
             }
        }
        }

        txLast = tx;
        tyLast = ty;
	clipbitsLast = clipbits;

        forceNewLine = false;
    }
    if (CheckMinRomVersion(sysVersion35))
        WinPopDrawState();
}


static const UInt32 steps[] = {1, 2, 4, 8, 15, 30,
                               60, 120, 300, 600, 1200,
                               3600, 7200, 18000, 36000, 72000 };

static void SKYSECTION
SkyMapDrawCircle(SkyDrawContext *ctx, Int32 diameterAngle)
{
    Int32 sinRadius;
    Int32 cosRadius;
    Int32 pixelRadius;
    Coord x;
    Coord y;
    Coord halfWidth;
    Coord halfHeight;
    Int16 diff;
    Coord cx;
    Coord cy;
    Boolean inverted;

    fpsincos(diameterAngle / 2, &sinRadius, &cosRadius);

    pixelRadius = sinRadius >> userPrefs.zoomValue;

    halfWidth  = SKY_WIDTH  / 2;
    halfHeight = SKY_HEIGHT / 2;

    if ( pixelRadius > halfWidth && pixelRadius > halfHeight )
        return;

    if (CheckMinRomVersion(sysVersion35)) {
        WinPushDrawState();
	    WinGlueSetForeColorRGB
	    (&(userPrefs.colors[GetColorMode()][COLOR_CIRCLES]),NULL);
	inverted = false;
    } else {
	inverted = userPrefs.optFlags.blackBackground
	    ^ userPrefs.optFlags.nightMode;
    }

    y = 0;
	x = pixelRadius;
	diff = 4*x-1;

	if (ctx->bits) {
	    void *bits;
	    UInt32 yBytes, xBytes;
	    UInt16 color = ctx->colors[COLOR_CIRCLES][MAX_BRIGHTNESS];
	    yBytes = 0;
	    xBytes = ctx->rowBytes * (Int32)x;

	    cx = halfWidth;
	    cy = halfHeight;
	    
	    bits = (void*) ctx->bits + ((Int32)cy * ctx->rowBytes);

	    while (y <= x) {
        if(x<halfWidth && y<halfHeight) {
            ctx->fastBmpBitsDrawPixel(bits + yBytes, screenDepth,
					  cx + x, color);
            ctx->fastBmpBitsDrawPixel(bits + yBytes, screenDepth,
					  cx - x, color);
            ctx->fastBmpBitsDrawPixel(bits - yBytes, screenDepth,
					  cx + x, color);
            ctx->fastBmpBitsDrawPixel(bits - yBytes, screenDepth,
					  cx - x, color);
        }
        if(y<halfWidth && x<halfHeight) {
            ctx->fastBmpBitsDrawPixel(bits + xBytes, screenDepth,
					  cx + y, color);
            ctx->fastBmpBitsDrawPixel(bits + xBytes, screenDepth,
					  cx - y, color);
            ctx->fastBmpBitsDrawPixel(bits - xBytes, screenDepth,
					  cx + y, color);
            ctx->fastBmpBitsDrawPixel(bits - xBytes, screenDepth,
					  cx - y, color);
        }
		y++;
		yBytes += ctx->rowBytes;
		diff -= 8*y+4;
		if (diff < 0) {
		    x--;
		    xBytes -= ctx->rowBytes;
		    diff += 8*x;
		}
	    }
	} else {
        RectangleType oldClip;
        RectangleType newClip;

        void (*fastWinDrawLine) (Coord, Coord, Coord, Coord) =
        SysGetTrapAddress((inverted ? sysTrapWinEraseLine
                               : sysTrapWinDrawLine));
	    void (*fastWinDrawPixel) (Coord, Coord, Coord, Coord) =
		CheckMinRomVersion(sysVersion35)
		? SysGetTrapAddress(inverted ? sysTrapWinErasePixel
				    : sysTrapWinDrawPixel)
		: fastWinDrawLine;

        cx = SKY_CNTX;
        cy = SKY_CNTY;

		WinGetClip(&oldClip);

		newClip.topLeft.x = SKY_X;
		newClip.topLeft.y = SKY_Y;
		newClip.extent.x  = SKY_WIDTH;
		newClip.extent.y  = SKY_HEIGHT;

		WinSetClip(&newClip);

	    while (y <= x) {
		fastWinDrawPixel(cx + x, cy + y,
				 cx + x, cy + y);
		fastWinDrawPixel(cx - x, cy + y,
				 cx - x, cy + y);
		fastWinDrawPixel(cx + x, cy - y,
				 cx + x, cy - y);
		fastWinDrawPixel(cx - x, cy - y,
				 cx - x, cy - y);
		fastWinDrawPixel(cx + y, cy + x,
				 cx + y, cy + x);
		fastWinDrawPixel(cx - y, cy + x,
				 cx - y, cy + x);
		fastWinDrawPixel(cx + y, cy - x,
				 cx + y, cy - x);
		fastWinDrawPixel(cx - y, cy - x,
				 cx - y, cy - x);

		y++;
		diff -= 8*y+4;
		if (diff < 0) {
		    x--;
		    diff += 8*x;
		}
	    }

	    WinSetClip(&oldClip);
	}

    if (CheckMinRomVersion(sysVersion35))
        WinPopDrawState();

}


static void SKYSECTION
SkyMapDrawPos(SkyDrawContext *ctx, FormPtr formPtr)
{
    Boolean inverted;
    FontID oldFont;
    void (*fastWinDrawLine) (Coord, Coord, Coord, Coord);
    Coord xstart, ystart;

    HostTraceOutputTL(appErrorClass, "SkyMapDrawPos Start");

    oldFont = FntSetFont(fontSmall);

    if (CheckMinRomVersion(sysVersion35)) {
        WinPushDrawState();
	WinGlueSetForeColorRGB
	    (&(userPrefs.colors[GetColorMode()][COLOR_EXTRAS]),NULL);
	inverted = false;
    } else {
	inverted = userPrefs.optFlags.blackBackground
	    ^ userPrefs.optFlags.nightMode;
    }

    fastWinDrawLine =
        SysGetTrapAddress((inverted ? sysTrapWinEraseLine
                               : sysTrapWinDrawLine));

    if (bitmap) {
        xstart = ystart = 0;
    } else {
        xstart = SKY_X;
        ystart = SKY_Y;
    }
    if (userPrefs.skyFlags.displayPosText) {
        Char str1[FIELDLENGTH], str2[FIELDLENGTH];

        HostTraceOutputTL(appErrorClass, "SkyMapDrawPos displayPosText");

        if (userPrefs.skyFlags.showRA) {
            /* Ra/Dec */
            /* In ra/dec mode ra is encoded 90 degree - az */
	        SysStringByIndex(stbl_AngleFormatsHMS_sprintf,
			     userPrefs.angleFormat, str2, FIELDLENGTH);
            FormatAngle(str1, FP_ONE - currentAz, str2, false);
            StrPrintF(str2, "%s%s",
                      GetStringResource(s_Ra, str2), str1);


        } else {
            /* Az/Alt */
	    SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			     userPrefs.angleFormat, str2, FIELDLENGTH);
            FormatAngle(str1, currentAz, str2, false);
            StrPrintF(str2, "%s%s",
                      GetStringResource(s_Az, str2), str1);
        }
        ctx->fastWinDrawChars(str2, StrLen(str2),
                         xstart + 1, ystart + 1);

	SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
			 userPrefs.angleFormat, str2, FIELDLENGTH);
        FormatAngle(str1, currentAlt, str2, true);
        StrPrintF(str2, "%s%s",
                  GetStringResource(userPrefs.skyFlags.showRA ? s_Dec : s_Alt,
                                    str2), str1);

        ctx->fastWinDrawChars(str2, StrLen(str2),
                         xstart + 1, ystart + 1 + FntLineHeight());
     }

    if (userPrefs.skyFlags.displayPosGraph) {
        Coord x, y;
        Int16 diff;
        Int32 sin1, cos1, sin2, cos2;
        Coord skyPosRadius, charWidth, charHeight;
        Coord skyPosCirc1x, skyPosCirc1y, skyPosCirc2x, skyPosCirc2y;

        HostTraceOutputTL(appErrorClass, "SkyMapDrawPos displayPosGraph");

        skyPosRadius = 9 + 3 * userPrefs.skyFlags.sizePosCircles;
        charWidth = 5;
        charHeight = 8;
        if (winVersion >= 4) {
            skyPosRadius = WinScaleCoord(skyPosRadius, false);
            charWidth = WinScaleCoord(charWidth, false);
            charHeight = WinScaleCoord(charHeight, false);
        }

        skyPosCirc2x = xstart + SKY_WIDTH  - skyPosRadius - charWidth/2;
        skyPosCirc2y = ystart + skyPosRadius;
        skyPosCirc1x = skyPosCirc2x - skyPosRadius - charWidth/2;
        skyPosCirc1y = skyPosCirc2y;
            
	if (ctx->bits) {
	    void *bits = (void*) ctx->bits + (skyPosCirc1y * ctx->rowBytes);
	    UInt16 yBytes, xBytes;
	    UInt16 color = ctx->colors[COLOR_EXTRAS][MAX_BRIGHTNESS];
	    y = 0;
	    x = skyPosRadius;
	    yBytes = 0;
	    xBytes = ctx->rowBytes * x;
	    diff = 4*x-1;
	    while (y <= x) {
		ctx->fastBmpBitsDrawPixel(bits + yBytes, screenDepth,
					  skyPosCirc1x + x, color);
		ctx->fastBmpBitsDrawPixel(bits + yBytes, screenDepth,
					  skyPosCirc1x - x, color);
		ctx->fastBmpBitsDrawPixel(bits - yBytes, screenDepth,
					  skyPosCirc1x + x, color);
		ctx->fastBmpBitsDrawPixel(bits - yBytes, screenDepth,
					  skyPosCirc1x - x, color);
		ctx->fastBmpBitsDrawPixel(bits + xBytes, screenDepth,
					  skyPosCirc1x + y, color);
		ctx->fastBmpBitsDrawPixel(bits + xBytes, screenDepth,
					  skyPosCirc1x - y, color);
		ctx->fastBmpBitsDrawPixel(bits - xBytes, screenDepth,
					  skyPosCirc1x + y, color);
		ctx->fastBmpBitsDrawPixel(bits - xBytes, screenDepth,
					  skyPosCirc1x - y, color);
		ctx->fastBmpBitsDrawPixel(bits + yBytes, screenDepth,
					  skyPosCirc2x + x, color);
		ctx->fastBmpBitsDrawPixel(bits - yBytes, screenDepth,
					  skyPosCirc2x + x, color);
		ctx->fastBmpBitsDrawPixel(bits + xBytes, screenDepth,
					  skyPosCirc2x + y, color);
		ctx->fastBmpBitsDrawPixel(bits - xBytes, screenDepth,
					  skyPosCirc2x + y, color);
		y++;
		yBytes += ctx->rowBytes;
		diff -= 8*y+4;
		if (diff < 0) {
		    x--;
		    xBytes -= ctx->rowBytes;
		    diff += 8*x;
		}
	    }
	} else {
	    void (*fastWinDrawPixel) (Coord, Coord, Coord, Coord) =
		CheckMinRomVersion(sysVersion35)
		? SysGetTrapAddress(inverted ? sysTrapWinErasePixel
				    : sysTrapWinDrawPixel)
		: fastWinDrawLine;
	    y = 0; 
	    x = skyPosRadius;
	    diff = 4*x-1;
	    while (y <= x) {
		fastWinDrawPixel(skyPosCirc1x + x, skyPosCirc1y + y,
				 skyPosCirc1x + x, skyPosCirc1y + y);
		fastWinDrawPixel(skyPosCirc1x - x, skyPosCirc1y + y,
				 skyPosCirc1x - x, skyPosCirc1y + y);
		fastWinDrawPixel(skyPosCirc1x + x, skyPosCirc1y - y,
				 skyPosCirc1x + x, skyPosCirc1y - y);
		fastWinDrawPixel(skyPosCirc1x - x, skyPosCirc1y - y,
				 skyPosCirc1x - x, skyPosCirc1y - y);
		fastWinDrawPixel(skyPosCirc1x + y, skyPosCirc1y + x,
				 skyPosCirc1x + y, skyPosCirc1y + x);
		fastWinDrawPixel(skyPosCirc1x - y, skyPosCirc1y + x,
				 skyPosCirc1x - y, skyPosCirc1y + x);
		fastWinDrawPixel(skyPosCirc1x + y, skyPosCirc1y - x,
				 skyPosCirc1x + y, skyPosCirc1y - x);
		fastWinDrawPixel(skyPosCirc1x - y, skyPosCirc1y - x,
				 skyPosCirc1x - y, skyPosCirc1y - x);

		fastWinDrawPixel(skyPosCirc2x + x, skyPosCirc2y + y,
				 skyPosCirc2x + x, skyPosCirc2y + y);
		fastWinDrawPixel(skyPosCirc2x + x, skyPosCirc2y - y,
				 skyPosCirc2x + x, skyPosCirc2y - y);
		fastWinDrawPixel(skyPosCirc2x + y, skyPosCirc2y + x,
				 skyPosCirc2x + y, skyPosCirc2y + x);
		fastWinDrawPixel(skyPosCirc2x + y, skyPosCirc2y - x,
				 skyPosCirc2x + y, skyPosCirc2y - x);
		y++;
		diff -= 8*y+4;
		if (diff < 0) {
		    x--;
		    diff += 8*x;
		}
	    }
	}

        if (userPrefs.skyFlags.showRA) {
            MemHandle handle;
            Char* strPtr;

            ctx->fastWinDrawChars("12", 2,
                             skyPosCirc1x - charWidth,
                             skyPosCirc1y + skyPosRadius + 1);
            ctx->fastWinDrawChars("18", 2, 
                             skyPosCirc1x - skyPosRadius - 2*charWidth, 
                             skyPosCirc1y - charHeight/2);

            handle = DmGetResource('tSTL', stbl_CompassPoints);
            strPtr = MemHandleLock(handle);
            strPtr += StrLen(strPtr)+3;

            strPtr += StrLen(strPtr)+1;
            strPtr += StrLen(strPtr)+1;
            ctx->fastWinDrawChars(strPtr, StrLen(strPtr),
                             skyPosCirc2x - charWidth/2,
                             skyPosCirc2y + skyPosRadius + 1);
            MemHandleUnlock(handle);

            fpsincos(FP_ONE - currentAz, &sin1, &cos1);
        } else {
            Char* strPtr;
            MemHandle handle = DmGetResource('tSTL', stbl_CompassPoints);
            strPtr = MemHandleLock(handle);
            strPtr += StrLen(strPtr)+3;
            // Top

            strPtr += StrLen(strPtr)+1;
            // Right

            strPtr += StrLen(strPtr)+1;
            // Bottom

            ctx->fastWinDrawChars(strPtr, StrLen(strPtr),
                             skyPosCirc1x - charWidth/2,
                             skyPosCirc1y + skyPosRadius + 1);

            strPtr += StrLen(strPtr)+1;
            // Left
            ctx->fastWinDrawChars(strPtr, StrLen(strPtr),
                             skyPosCirc1x - skyPosRadius - charWidth,
                             skyPosCirc1y - charHeight/2);
            MemHandleUnlock(handle);

            fpsincos(currentAz, &sin1, &cos1);
        }
	fpsincos(currentAlt, &sin2, &cos2);

        fastWinDrawLine(skyPosCirc1x, skyPosCirc1y,
                        skyPosCirc1x + fpmul(skyPosRadius-1, sin1),
                        skyPosCirc1y - fpmul(skyPosRadius-1, cos1));
        fastWinDrawLine(skyPosCirc2x, skyPosCirc2y,
                        skyPosCirc2x + fpmul(skyPosRadius-1, cos2),
                        skyPosCirc2y - fpmul(skyPosRadius-1, sin2));
    }

    /* Optionally draw a horizontal line with current FOV like:
     *
     *              32'
     *           |------|
     */
    if (userPrefs.skyFlags.displayZoom) {
        Char text[4];
        Char *units = "\"'\260*";
        UInt32 step;
        Int16 len;
        Coord x, y, xtext, width;
        Int16 i;

        x = xstart + 1;
        y = ystart + SKY_HEIGHT - 3;
        if (IS_FULLSCREEN) {
            y -= FntLineHeight() + 3;
        }
        width = (SKY_WIDTH / 4);

        for (i = sizeof(steps)/sizeof(steps[0]) - 1; i >= 0; i--) {

            if ((steps[i] * FP(3.14159 / 180 / 60 / 60)) >> userPrefs.zoomValue < width) {
                step = steps[i];
                len = (Int16) ((step * FP(3.14159 / 180 / 60 / 60))>> userPrefs.zoomValue);
                HostTraceOutputTL(appErrorClass, "SkyMapDrawPos displayZoom %lu %hu %hu %hd", steps[i], userPrefs.zoomValue, width, len);

                while (step >= 60) {
                    step /= 60;
                    units++;
                }

                StrPrintF(text, "%d%c", (Int16) step, *units);

                xtext = x + (len - FntCharsWidth(text, StrLen(text))) / 2;
                if (x < 0)
                    x = 0;

                ctx->fastWinDrawChars(text, StrLen(text),
                                 xtext, y - FntLineHeight());
                fastWinDrawLine(x, y - 1, x, y + 1);
                fastWinDrawLine(x, y, xstart + len, y);
                fastWinDrawLine(x + len, y - 1, x + len, y + 1);
                break;
            }
        }
    }
    if (CheckMinRomVersion(sysVersion35)) {
	WinPopDrawState();
    }
    FntSetFont(oldFont);
}


static const Int16 scaleMag[5] = {0, 1, 2, 3, 5};

static SKYSECTION void
DrawObjectBmp(SkyDrawContext *ctx, Int16 mag, Int16 tx, Int16 ty, UInt16 colorNum) {
    Int16  y, brightness;
    UInt16 color;
    Char *bits;

    tx = SKY_WIDTH/2  + tx;
    ty = SKY_HEIGHT/2 - ty;

    /* (3*mag) >> 10 is approximately (3/4)*magnitude since mag is GP16 */
    /* userPrefs.zoomValue varies between 1 and 24, with 24 being the lowest
       magnification */
    mag = 25 - userPrefs.zoomValue - ((3*mag) >> 10);

    if (mag <= 0) {
	if (tx >= 0 && tx < SKY_WIDTH
	    && ty >= 0 && ty < SKY_HEIGHT) {

	    brightness = MAX_BRIGHTNESS + mag;
	    if (brightness <= 0)
		brightness = 1;

	    /* Draw single pixel */
	    ctx->fastBmpBitsDrawPixel((UInt16*)(ctx->bits + (Int32)ty * ctx->rowBytes),
			     screenDepth, tx, ctx->colors[colorNum][brightness]);
	}
	return;
    }

    if (mag > maxBrightness)
        mag = maxBrightness;

    mag = scaleMag[mag];

    tx -= mag/2;
    ty -= mag/2;
    bits = ctx->bits + ((Int32)ty * ctx->rowBytes);
    color = ctx->colors[colorNum][MAX_BRIGHTNESS];
    for (y = 0; y <= mag; y++) {
        if (ty >= 0 && ty < SKY_HEIGHT) {
            Int16 xoff = tx;
            Int16 xend = tx+mag+1;
            if (mag >= 2 && (y == 0 || y == mag)) {
                xoff++;
                xend--;
            }
            if (xoff < 0)
                xoff = 0;
            if (xend > SKY_WIDTH)
                xend = SKY_WIDTH;
            if (xoff < xend)
                ctx->fastBmpBitsDrawHLine((UInt16*) bits, screenDepth,
                                 xoff, xend, color);
        }
        ty++;
        bits += ctx->rowBytes;
    }
}



static SKYSECTION void
DrawObject20(SkyDrawContext *ctx, Int16 mag, Int16 tx, Int16 ty, UInt16 catNum) {
    Coord x, y;
    RectangleType rect;

    x = SKY_CNTX + tx;
    y = SKY_CNTY - ty;

    mag = 25 - userPrefs.zoomValue - ((3*mag) >> 10);
    if (mag > maxBrightness)
        mag = maxBrightness;
    if (mag <= 0)
        mag = 0;
    if (mag < 2) {
        rect.topLeft.x = x - mag/2;
        rect.topLeft.y = y - mag/2;
        rect.extent.x = mag+1;
        rect.extent.y = mag+1;
        ctx->fastWinPaintRectangle(&rect, 0);
    } else {
        rect.topLeft.x = x - mag/2;
        rect.topLeft.y = y - mag/2+1;
        rect.extent.x = mag+1;
        rect.extent.y = mag-1;
        ctx->fastWinPaintRectangle(&rect, 0);
        rect.topLeft.x = x - mag/2+1;
        rect.topLeft.y = y - mag/2;
        rect.extent.x = mag-1;
        rect.extent.y = mag+1;
        ctx->fastWinPaintRectangle(&rect, 0);
    }
}


void SKYSECTION
SkyGetPlanetInfo(PlanetInfoType* p, UInt32 planetNumber)
{
    *p = planetInfo[planetNumber];
}


void SKYSECTION
SkyGetRaDec2AltAz(MatrixType rotate)
{
    MatrixType rd2aa;

    MatRotate(rd2aa, FP_ONE-currentLat, FP_ONE + currentLst);
    MatMatMul(rotate, rd2aa, precession);
}


static void SKYSECTION
SkyBuildRotateMatrix(MatrixType rotate)
{
    MatrixType tmp1, tmp2;
    if (userPrefs.skyFlags.showRA) {
        MatRotate(tmp1, currentAlt, -currentAz);
    } else {
	MatrixType rd2aa;

	/*
	 * Build radec/altAz conversion matrix
	 */
        MatRotate(rd2aa, FP_ONE-currentLat, FP_ONE + currentLst);
        MatRotate(tmp2, currentAlt, -currentAz);
        MatMatMul(tmp1, tmp2, rd2aa);
    }
    MatMatMul(rotate, tmp1, precession);
    if (XFLIP == -1) {
        rotate[0][0] = -rotate[0][0];
        rotate[0][1] = -rotate[0][1];
        rotate[0][2] = -rotate[0][2];
    }
    if (YFLIP == -1) {
        rotate[2][0] = -rotate[2][0];
        rotate[2][1] = -rotate[2][1];
        rotate[2][2] = -rotate[2][2];
    }
}


static SKYSECTION void
SkyMapGetObjectName(Char* str, UInt32 markedObject)
{
    AstroRecordType art;

	if (markedObject < NUMPLANETS) {
	    if (markedObject == SUN) {
		GetStringResource(s_OurOwnStar, str);
	    } else if (markedObject == MOON) {
		GetMoonPhaseDescription(planetInfo[markedObject].pa, str);
	    } else {
		SysStringByIndex(stbl_SolarObjects, markedObject,
				 str, FIELDLENGTH);
	    }
	} else {
	    GetCatalogObject(markedObject - NUMPLANETS, &art, NULL);
	    StrNCopy(str, art.description, FIELDLENGTH);
	}
 }


static SKYSECTION void
MarkSkyObject(UInt32 markedObject, Int16 erase)
{
    RectangleType oldClip, rect;
    MatrixType rotate;
    Int32 tx, ty;
    Int32 unscaledx;
    Int32 unscaledy;
    Int16 showdiam, limdiam;
    if (markedObject == NO_OBJECT)
	    return;

    SkyBuildRotateMatrix(rotate);

    /* Size box to encompass object */
    if (markedObject < NUMPLANETS) {
        unscaledx = VecScalarProd(rotate[0], planetInfo[markedObject].pos);
        unscaledy = VecScalarProd(rotate[2], planetInfo[markedObject].pos);

        showdiam = (fpmul(planetInfo[markedObject].semid, FP_PI_2)
		    >> (userPrefs.zoomValue-1)) + 1;

    } else {
        AstroRecordType art;
        if (!GetCatalogObject(markedObject - NUMPLANETS, &art, NULL))
            return;
        unscaledx = VecScalarProd(rotate[0], art.radec);
        unscaledy = VecScalarProd(rotate[2], art.radec);
        showdiam = (((UInt32)art.width * FP(3.14159 / (1L<<21)))
                    >> userPrefs.zoomValue) + 1;
    }

    limdiam = 5;
    if (winVersion >= 4)
        limdiam = WinScaleCoord(limdiam - 2, false) + 2;
    if (showdiam < limdiam)
        showdiam = limdiam;

    HostTraceOutputTL(appErrorClass, "MarkSkyObject showdiam %hd", showdiam);
    tx = ( unscaledx >> userPrefs.zoomValue );
    ty = ( unscaledy >> userPrefs.zoomValue );

    WinGetClip(&oldClip);
    WinSetClip(&skyRect);

    if (IS_FULLSCREEN) {
        RectangleType rect;
	Coord textWidth;
	Char str[64];
	Int16 len;
	SkyMapGetObjectName(str, markedObject);

	FntSetFont(fontNormal);
	len = StrLen(str);
        rect = descRect;
        if (winVersion >= 4)
            WinScaleRectangle(&rect);
        textWidth = FntCharsWidth(str, len);
	rect.topLeft.x += (rect.extent.x - textWidth) / 2;
        rect.extent.x = textWidth;
	if (erase) {
	    if (userPrefs.optFlags.blackBackground !=
		userPrefs.optFlags.nightMode)
		WinDrawRectangle(&rect, 0);
	    else
		WinEraseRectangle(&rect, 0);
	} else {
	    if (userPrefs.optFlags.blackBackground !=
		userPrefs.optFlags.nightMode)
		/* Draw description inverted onto skymap */
		WinDrawInvertedChars(str, len, rect.topLeft.x, rect.topLeft.y);
	    else
		WinDrawChars(str, len, rect.topLeft.x, rect.topLeft.y);
	}
    }

    if (fpabs(tx) > SKY_WIDTH/2+showdiam/2
        || fpabs(ty) > SKY_HEIGHT/2+showdiam/2) {
        Int32 x = unscaledx;
        Int32 y = unscaledy;
        Int32 r1, r2;

        if (fpmul(SKY_WIDTH, fpabs(y)) < fpmul(SKY_HEIGHT, fpabs(x))) {
            r2 = fpdiv(SKY_WIDTH / 2, fpabs(x));
        } else {
            r2 = fpdiv(SKY_HEIGHT / 2, fpabs(y));
        }
        r1 = r2 - r2/16;

        WinInvertLine(SKY_CNTX+fpmul(x, r1),
                      SKY_CNTY-fpmul(y, r1),
                      SKY_CNTX+fpmul(x, r2),
                      SKY_CNTY-fpmul(y, r2));
    }
    else {
        rect.topLeft.x = SKY_CNTX + tx - showdiam / 2;
        rect.topLeft.y = SKY_CNTY - ty - showdiam / 2;
        rect.extent.x = showdiam + 1;
        rect.extent.y = 1;
        WinInvertRectangle(&rect, 0);
        rect.topLeft.y += showdiam;
        WinInvertRectangle(&rect, 0);
        rect.topLeft.y -= showdiam - 1;
        rect.extent.x = 1;
        rect.extent.y = showdiam - 1;
        WinInvertRectangle(&rect, 0);
        rect.topLeft.x += showdiam;
        WinInvertRectangle(&rect, 0);
    }


    WinSetClip(&oldClip);
}


SKYSECTION void GetBackColor(RGBColorType* back)
{
    if (GetColorMode() == MODE_ONWHITE) {
        back->r = 255;
        back->g = 255;
        back->b = 255;
    }
    else {
        back->r = 0;
        back->b = 0;
        back->g = 0;
    }
}

void SKYSECTION
SkyMapDraw(FormPtr formPtr, Boolean fast)
{
    RectangleType oldClip;
    Boolean scrLocked = false;
    SkyDrawContext ctx;
    DrawObjectFunc drawObject;
    FontID oldFont;
    PointType p;
    WinHandle oldwin = 0;
    Char c;
    RGBColorType back;
    Int16 i;
    
    HostTraceOutputTL(appErrorClass, "SkyMapDraw");
    
    needFullDraw = fast;

    if (userPrefs.optFlags.fadeBelowHorizon && CheckMinRomVersion(sysVersion35)) {
        ctx.horizon = BELOW_HORIZON_FADE;
        ctx.fadeOffset = NUM_COLOR_TYPES+numDBs;
    }
    else if (userPrefs.skyFlags.showHorizon) {
        ctx.horizon = BELOW_HORIZON_DRAW;
        ctx.fadeOffset = 0;
    }
    else {
        ctx.horizon = BELOW_HORIZON_OMIT;
        ctx.fadeOffset = 0;
    }

    ctx.colors = MemPtrNew((ctx.fadeOffset+NUM_COLOR_TYPES+numDBs)*sizeof(ObjectColorsType));

    if (ctx.colors == NULL)
        return;

    switch(screenDepth) {
        case 16:
            ctx.fastBmpBitsDrawPixel = BmpBitsDrawPixel16;
            ctx.fastBmpBitsDrawHLine = BmpBitsDrawHLine16;
            break;
        case 8:
            ctx.fastBmpBitsDrawPixel = BmpBitsDrawPixel8;
            ctx.fastBmpBitsDrawHLine = BmpBitsDrawHLine8;
            break;
        default:
            ctx.fastBmpBitsDrawPixel = BmpBitsDrawPixel;
            ctx.fastBmpBitsDrawHLine = BmpBitsDrawHLine;
            break;
    }
    
    ctx.maxBrightnessForDrawingCode = maxBrightness;
    ctx.magnitude = SkyMapLimitingMag(fast);

    ctx.inverted = userPrefs.optFlags.blackBackground;

    ctx.fastWinPaintRectangle = (void (*) (RectangleType *, UInt16))
        SysGetTrapAddress(ctx.inverted ^ userPrefs.optFlags.nightMode
			  ? sysTrapWinEraseRectangle
                          : sysTrapWinDrawRectangle);
    ctx.fastWinDrawChars = (void (*) (const Char *, Int16, Coord, Coord))
        SysGetTrapAddress((ctx.inverted ^ userPrefs.optFlags.nightMode)
                          ? sysTrapWinDrawInvertedChars
                          : sysTrapWinDrawChars);

    if (CheckMinRomVersion(sysVersion35)) {
        scrLocked = (WinScreenLock(winLockCopy) != NULL);
        WinPushDrawState ();
        WinGlueSetCoordinateSystem(kCoordinatesNative);
    }

    WinGetClip(&oldClip);
    WinSetClip(&skyRect);

    if (bitmap) {
        Int16 j;

        GetBackColor(&back);

        for (i=0 ; i<NUM_COLOR_TYPES + numDBs ; i++) {
            RGBColorType in;

            if ( i < NUM_COLOR_TYPES ) {
                in = userPrefs.colors[GetColorMode()][i];
            }
            else {
                in = catDBs[i-NUM_COLOR_TYPES]->color[GetColorMode()];
            }

            for (j = 0; j <= MAX_BRIGHTNESS; j++) {
                RGBColorType rgb;
                Int16 v;

                if ( ! userPrefs.skyFlags.showAllNonStars ||
                     ( NUM_COLOR_TYPES <= i && catDBs[i-NUM_COLOR_TYPES]->isStars ) ) {
                    v = ( j == MAX_BRIGHTNESS ) ? ( MAX_BRIGHTNESS + 1 ) : j;
                }
                else {
                    v = MAX_BRIGHTNESS + 1;
                }

                rgb.r = ( in.r * v + back.r * ( MAX_BRIGHTNESS + 1 - v ) ) / (MAX_BRIGHTNESS+1);
                rgb.g = ( in.g * v + back.g * ( MAX_BRIGHTNESS + 1 - v ) ) / (MAX_BRIGHTNESS+1);
                rgb.b = ( in.b * v + back.b * ( MAX_BRIGHTNESS + 1 - v ) ) / (MAX_BRIGHTNESS+1);

                if (screenDepth < 16) {
                    ctx.colors[i][j] = WinRGBToIndex(&rgb);
                } else {
                    ctx.colors[i][j] = ((rgb.r>>3) << 11)
                        | ((rgb.g>>2)<<5) | (rgb.b>>3);
                }

                if (ctx.fadeOffset) {
                    rgb.r = ((UInt16)rgb.r*FADE_NUM + back.r*(FADE_DEN-FADE_NUM))/FADE_DEN;
                    rgb.g = ((UInt16)rgb.g*FADE_NUM + back.g*(FADE_DEN-FADE_NUM))/FADE_DEN;
                    rgb.b = ((UInt16)rgb.b*FADE_NUM + back.b*(FADE_DEN-FADE_NUM))/FADE_DEN;

                    if (screenDepth < 16) {
                        ctx.colors[i+ctx.fadeOffset][j] = WinRGBToIndex(&rgb);
                    } else {
                        ctx.colors[i+ctx.fadeOffset][j] = ((rgb.r>>3) << 11)
                            | ((rgb.g>>2)<<5) | (rgb.b>>3);
                    }
                }
            }
        }

        if (screenDepth < 16)
            ctx.transparentColor = (1 << screenDepth)-1;
        else
            ctx.transparentColor = 0;
        ctx.bits = (Char*) BmpGetBits(bitmap);
        ctx.dest = bitmap;
        BmpGlueGetDimensions(bitmap, NULL, NULL, &ctx.rowBytes);
        drawObject = DrawObjectBmp;
        MemSet(ctx.bits, (UInt32)ctx.rowBytes * SKY_HEIGHT,
               (ctx.inverted ^ (screenDepth >= 16)) ? 0xff : 0);
        oldwin = WinSetDrawWindow(skywin);
    } else {
        ctx.bits = NULL;
        ctx.dest = NULL;
        drawObject = DrawObject20;

        if (ctx.inverted)
            WinDrawRectangle(&skyRect, 0);
        else
            WinEraseRectangle(&skyRect, 0);

    }

    MemMove(ctx.zenith, zenith, sizeof(VectorType));

    SkyBuildRotateMatrix(ctx.rotate);

    for (i=0; i<sizeof(circleIDs)/sizeof(*circleIDs);i++) {
         if(userPrefs.circles & (1l<<i)) {
              if (i<NUM_NON_CUSTOM_CIRCLES)
                  SkyMapDrawCircle(&ctx,circles[i]);
              else
                  SkyMapDrawCircle(&ctx,CUSTOM_CIRCLES(i-NUM_NON_CUSTOM_CIRCLES));
         }
    }

    if (userPrefs.skyFlags.showConstellation) {
        LinesDrawSph(&ctx);
    }
    
    if (userPrefs.optFlags.showConstNames) {
        ConstellationNamesDrawSph(&ctx);
    }

#ifdef DEBUG_PNO        
    if (bitmap) {
        if (scrLocked)
            WinScreenUnlock();
        WinSetDrawWindow(oldwin);
    }
#endif
    QTreeDrawSph(&ctx, drawObject);
#ifdef DEBUG_PNO        
    if (bitmap) {
        oldwin = WinSetDrawWindow(skywin);
        scrLocked = (WinScreenLock(winLockCopy) != NULL);
    }
#endif
    PlanetsDrawSph(&ctx, drawObject);

    if (displayPos)
        SkyMapDrawPos(&ctx, formPtr);

    oldFont = FntSetFont(fontSmall);
    c = IS_FULLSCREEN ? 21 : 20;
    p.x = skyRect.extent.x - FntCharWidth(c);
    p.y = skyRect.extent.y - FntCharHeight();
    if (!bitmap) {
        p.x += SKY_X;
        p.y += SKY_Y;
    }
    ctx.fastWinDrawChars(&c, 1, p.x, p.y);
    c = 23 + userPrefs.optFlags.orientation;
    if (winVersion >= 4) {
        p.x -= WinScaleCoord(2, false);
    }
    else {
        p.x -= 2;
    }
    p.x -= FntCharWidth(c);
    ctx.fastWinDrawChars(&c, 1, p.x, p.y);

    FntSetFont(oldFont);

    if (bitmap) {
        WinSetDrawWindow(oldwin);
        WinDrawBitmap(bitmap, SKY_X, SKY_Y);
    }

    /* Mark current object */
    MarkSkyObject(GetCurrentObject(), 0);

    WinSetClip(&oldClip);

    if (CheckMinRomVersion(sysVersion35)) {
        WinPopDrawState ();
        if (scrLocked)
            WinScreenUnlock();
    }
    
    MemPtrFree(ctx.colors);
}

static void SKYSECTION
SkyShift(FormPtr formPtr, Int16 x, Int16 y, Boolean fast)
{
    Int32 yZoom;

    HostTraceOutputTL(appErrorClass, "SkyShift");

    if (winVersion >= 4) {
        WinPushDrawState();
        WinSetCoordinateSystem(kCoordinatesNative);
        x = WinScaleCoord(x, false);
        y = WinScaleCoord(y, false);
        WinPopDrawState();
    }

    currentAz -= (Int32) x << userPrefs.zoomValue;
    yZoom = (Int32) y << userPrefs.zoomValue;

    currentAlt += yZoom;
    if (currentAlt > FP_ONE - (1L << userPrefs.zoomValue))
        currentAlt = FP_ONE - (1L << userPrefs.zoomValue);
    else if (currentAlt < - FP_ONE + (1L << userPrefs.zoomValue))
        currentAlt = -FP_ONE + (1L << userPrefs.zoomValue);

    SkyMapDraw(formPtr, fast);
}



static void SKYSECTION
SkyMapDrawWithPos(FormPtr formPtr)
{
    HostTraceOutputTL(appErrorClass, "SkyMapDrawWithPos");

    /* set timeout for display position */
    displayPos = true;

    displayPosTimeout = TimGetTicks() + SKY_POS_TIMEOUT;

    if (! userPrefs.skyFlags.noDisplayTimeout) {
        EvtGlueSetNullEventTick(displayPosTimeout);
    }

    SkyMapDraw(formPtr, false);
}



static SKYSECTION Boolean
SkyMapGetLocation(Coord x, Coord y, VectorType radec)
{
    MatrixType rotate;
    VectorType st;
    Int32 tmp;

    if (winVersion >= 4) {
        WinPushDrawState();
        WinSetCoordinateSystem(kCoordinatesNative);
        x = WinScaleCoord(x, false);
        y = WinScaleCoord(y, false);
        WinPopDrawState();
    }

    st[0] = (Int32) ((x - SKY_CNTX)) << userPrefs.zoomValue;
    st[2] = (Int32) ((SKY_CNTY - y)) << userPrefs.zoomValue;
    tmp = FP_ONE - fpsq(st[0]) - fpsq(st[2]);

    if (tmp < 0) {
        /* in case of roundoff errors */
        tmp = 0;
    }

    st[1] = fpsqrt(tmp);
    SkyBuildRotateMatrix(rotate);
    MatTranspose(rotate);
    MatVecMul(radec, rotate, st);

    return true;
}



static SKYSECTION Int32
SkyMapDistance(Coord x1, Coord y1, Coord x2, Coord y2)
{
    VectorType radec1;
    VectorType radec2;
    Int32      scalarProduct;
    Int32      angle;

    if ( ! SkyMapGetLocation(x1, y1, radec1) ||
         ! SkyMapGetLocation(x2, y2, radec2) )
        return -1;

    scalarProduct = VecScalarProd(radec1,radec2);

    if ( scalarProduct < -FP_ONE || FP_ONE < scalarProduct )
        return -1;

    angle = fpatan2(fpsqrt(FP_ONE - fpsq(scalarProduct)), scalarProduct);

    return angle;
}


static SKYSECTION Int32
SkyMapDistanceNative(Coord x1, Coord y1, Coord x2, Coord y2)
{
    if (winVersion >= 4) {
        WinPushDrawState();
        WinSetCoordinateSystem(kCoordinatesNative);
        x1 = WinUnscaleCoord(x1, true);
        y1 = WinUnscaleCoord(y1, true);
        x2 = WinUnscaleCoord(x2, true);
        y2 = WinUnscaleCoord(y2, true);
        WinPopDrawState();
    }

    return SkyMapDistance(x1, y1, x2, y2);
}


static SKYSECTION void
SkyMapDSSLookup(void)
{
    VectorType radec;
    VectorType st;
    MatrixType rotate;

    dssXSize = 2*SkyMapDistanceNative(SKY_CNTX,SKY_CNTY,SKY_CNTX+SKY_WIDTH/2-1,SKY_CNTY);
    dssYSize = 2*SkyMapDistanceNative(SKY_CNTX,SKY_CNTY,SKY_CNTX,SKY_CNTY+SKY_HEIGHT/2-1);

    if(dssXSize < 0 || MAX_DSS_RANGE < dssXSize ||
       dssYSize < 0 || MAX_DSS_RANGE < dssYSize) {
        if (1==FrmAlert(a_TooBig))
            return;

        dssXSize = MAX_DSS_RANGE;
        dssYSize = MAX_DSS_RANGE;
    }

    SkyBuildRotateMatrix(rotate);

    if (rotate[0][2] == 0)
        dssRotate = 0;
    else
        dssRotate = -fpatan2(rotate[0][2], rotate[2][2]);

    dssXFlip = XFLIP < 0;
    if (YFLIP<0) {
        dssRotate += -2*FP_ONE;
        dssXFlip = !dssXFlip;
    }

    st[0] = 0;
    st[1] = FP_ONE;
    st[2] = 0;

    MatTranspose(rotate);
    MatVecMul(radec, rotate, st);

	dssRA  = fpatan2(radec[1], radec[0]);
	dssDec = fpatan2(radec[2], fpsqrt(fpsq(radec[0]) + fpsq(radec[1])));

    FrmPopupForm(f_DSSImage);
}



SKYSECTION void
SkyMapUpdateFinder(FinderType* finder, UInt32 dist,
    UInt32 objectIndex, UInt16 catNum, UInt16 recidx, UInt16 offsetInNode)
{
    Int16 insertAt;
    Int16 i;
    
    if (finder->ignoreDistance <= dist)
        return;
        
    insertAt = MAX_FIND_ITEMS;

    for (i=finder->numObjects-1; 0 <= i ; i--) {
        if (finder->found[i].dist<=dist) {
            break;
        }
    }

    insertAt = i+1;

    if (MAX_FIND_ITEMS <= insertAt)
        return;

    if (dist < finder->precision) {
        Int16 n;

        if (finder->numObjects == MAX_FIND_ITEMS)
            finder->numObjects--;

        n = finder->numObjects - insertAt;

        if ( 0 < n )
            MemMove(finder->found + insertAt + 1, finder->found + insertAt,
                sizeof(FoundObjectInfo)*n);

        finder->numObjects++;

        if (insertAt+1 < finder->numObjects &&
            finder->precision <= finder->found[insertAt+1].dist) {
            /* Trim off less precise options */
            finder->numObjects = insertAt+1;
        }

        if (finder->numObjects == MAX_FIND_ITEMS) {
            finder->ignoreDistance = finder->found[MAX_FIND_ITEMS-1].dist;
        }
        else {
            finder->ignoreDistance = finder->precision;
        }
    }
    else {
        insertAt = 0;
        finder->numObjects = 1;
        finder->ignoreDistance = dist;
    }

    finder->found[insertAt].dist         = dist;
    finder->found[insertAt].objectIndex  = objectIndex;
    finder->found[insertAt].catNum       = catNum;
    finder->found[insertAt].recidx       = recidx;
    finder->found[insertAt].offsetInNode = offsetInNode;
}


static void DBSECTION
SkyMapDrawFoundObjectName(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    FinderType* finder = (void*)itemsText;
    UInt32 markedObject;
    Char   str[FIELDLENGTH];

    markedObject = finder->found[itemNum].objectIndex;

    if (markedObject == INVALID_OBJECT_INDEX)
        return;

    SkyMapGetObjectName(str, markedObject);
    WinDrawChars(str, StrLen(str), bounds->topLeft.x, bounds->topLeft.y);
}


/* Find object at the given coordinates */
static SKYSECTION UInt32
SkyMapFindObject(Coord x, Coord y)
{
    UInt16 i;
    UInt32 objectNr;
    Boolean cutHoriz;
    VectorType radec;
    FinderType* finder;
    Int16 magnitude = SkyMapLimitingMag(false);

    objectNr = NO_OBJECT;
    cutHoriz = !(userPrefs.skyFlags.showRA || userPrefs.skyFlags.showHorizon);

    if (!SkyMapGetLocation(x, y, radec))
        return objectNr;

    finder = MemPtrNew(sizeof(FinderType));

    if (finder == NULL)
        return objectNr;

    finder->numObjects     = 0;
    finder->precision      = ((FIND_PIXEL_PRECISION*screenDensity/72) << userPrefs.zoomValue)+1;
    finder->ignoreDistance = FP_ONE;

    for(i=0; i < NUMPLANETS; i++) {
        UInt32 dist;
        if (planetInfo[i].magnitude > magnitude && !userPrefs.skyFlags.showAllNonStars)
            continue;

        dist = SkyDistance(planetInfo[i].pos, radec);
        if (dist < finder->ignoreDistance) {
             SkyMapUpdateFinder(finder, dist, i, 0,0,0);
        }
    }

    QTreeFindObject(radec, magnitude, finder);

    if (1 < finder->numObjects) {
        Int16 index;

        ListType* l = FrmGetObjectPtrByID(FrmGetActiveForm(), l_FoundObjects);

        LstSetListChoices(l, (void*)finder, finder->numObjects);
        LstSetHeight(l, finder->numObjects);
        LstSetDrawFunction(l, SkyMapDrawFoundObjectName);
        LstSetSelection(l, 0);

        index = LstPopupList(l);

        LstSetListChoices(l, NULL, 0);

        if (index < 0) {
             objectNr = userPrefs.currentObj;
        }
        else if (finder->found[index].objectIndex != INVALID_OBJECT_INDEX) {
             objectNr = finder->found[index].objectIndex;
             if(objectNr == userPrefs.currentObj) {
                 /* It will be confusing if we toggle under these circumstances */
                 objectNr = NO_OBJECT;
             }
        }
    }
    else if (1 == finder->numObjects &&
             finder->found[0].objectIndex != INVALID_OBJECT_INDEX) {
        objectNr = finder->found[0].objectIndex;
    }

    MemPtrFree(finder);

    return objectNr;
}


/* Describe the current object. */
static SKYSECTION void
SkyMapDescribe(FormPtr formPtr, UInt32 objectNr)
{
    Int16 numDetailRows;

    HostTraceOutputTL(appErrorClass, "SkyMapDescribe: %hu", objectNr);

    if (objectNr < NUMPLANETS + GetTotalRecords()) {
        if (objectNr >= NUMPLANETS) {
            numDetailRows = MAXOBJDETAILROWS;
        } else {
            numDetailRows = MAXSSDETAILROWS;
        }

        if (AlignValid()) {
            numDetailRows++;
        }

        if (!IS_FULLSCREEN) {
            /* Display rocker max value is based on max rows to
             * display less the number to be displayed at any one
             * time. That way the max value for the rocker has the
             * last row to display at the bottom of the form.
             *
             * It also handles the special condition of zero detail
             * rows and full screen.
             */
            InitRocker(&DetailRocker,
                       0,
                       userPrefs.detailPage,
                       userPrefs.showDetailRows > numDetailRows
		       || userPrefs.showDetailRows == 0
                       ? 0 : numDetailRows - userPrefs.showDetailRows);
            SetRockerButtons(&DetailRocker, r_DetUp, r_DetDown);
            userPrefs.detailPage = DetailRocker.Value;
            HostTraceOutputTL(appErrorClass, "SkyMapDescribe %hd %hd", 
                              numDetailRows, userPrefs.showDetailRows);
        }
        ShowObjectDetail(objectNr);

    } else {
        if (!IS_FULLSCREEN) {
            InitRocker(&DetailRocker, 0, 0, 0);
            SetRockerButtons(&DetailRocker, r_DetUp, r_DetDown);
            ClearObjectDetail();
        }
    }
}


/* Describe the current object. */
static SKYSECTION void
SkyMapSelectObject(FormPtr formPtr, UInt32 newObject)
{
    /* Select no object if the marked object was selected */
    if (newObject == userPrefs.currentObj) {
	    newObject = NO_OBJECT;
    }
    
    if (winVersion >= 4) {
        WinPushDrawState ();
        WinSetCoordinateSystem(kCoordinatesNative);
    }
    /* Remove mark from old object */
    if (userPrefs.currentObj != NO_OBJECT) {
        MarkSkyObject(GetCurrentObject(), 1);
    }
    
    /* Mark and describe new object */
    userPrefs.currentObj = newObject;
    if (newObject != NO_OBJECT) {
        MarkSkyObject(newObject, 0);
    }
    if (winVersion >= 4)
        WinPopDrawState ();
    SkyMapDescribe(formPtr, newObject);
}

/* This only works for delta = +-1 */
static void SKYSECTION
ChangeObject(FormPtr formPtr, Int8 delta)
{
    UInt32 currObj = GetCurrentObject();

    if (currObj < NUMPLANETS) {
        currObj += delta;
        if (((Int32)currObj) > SUN)
            currObj = MERCURY;
        if (((Int32)currObj) < MERCURY)
            currObj = SUN;

    } else {
        if (filterString[0] && NUMPLANETS <= currObj) {
            UInt32 filteredPos;
            filteredPos = BWGetFilteredPosition(currObj-NUMPLANETS);
            if (delta < 0 && filteredPos == 0)
                filteredPos = totalRecords - 1;
            else if (0 < delta && totalRecords <= filteredPos + 1)
                filteredPos = 0;
            else
                filteredPos += delta;
            currObj = BWGetFilteredObject(filteredPos)+NUMPLANETS;
        }
        else {
            currObj += delta;
            if (currObj < NUMPLANETS)
                currObj = NUMPLANETS + GetTotalRecords() - 1;
            else if (currObj == NUMPLANETS + GetTotalRecords())
                currObj = NUMPLANETS;
        }
    }
    userPrefs.currentObj = currObj;
    SkyCenterObject(currObj);
    SkyMapDescribe(formPtr, GetCurrentObject());
    SkyMapDraw(formPtr, false);
}


static void SKYSECTION
SkyMapMeasure(FormPtr formPtr, Coord firstX, Coord firstY)
{
    Coord x;
    Coord y;
    Coord prevX;
    Coord prevY;
    Boolean penDown;
    Char str2[FIELDLENGTH];
    void (*fastWinDrawLine) (Coord, Coord, Coord, Coord);
    FontID oldFont;
    Boolean inverted;
    void (*fastWinDrawChars) (const Char *, Int16, Coord, Coord);
    RectangleType src;
    WinHandle saved;
    Err err;

    saved = WinCreateOffscreenWindow( lowresSkyRect.extent.x,
                lowresSkyRect.extent.y, nativeFormat, &err );
    WinCopyRectangle(WinGetActiveWindow(), saved, &lowresSkyRect, 0, 0, winPaint );

    src = lowresSkyRect;
    src.topLeft.x = 0;
    src.topLeft.y = 0;

    if ( saved == NULL )
        return;

    SysStringByIndex(stbl_AngleFormatsDMS_sprintf,
        userPrefs.angleFormat, str2, FIELDLENGTH);

    oldFont = FntSetFont(fontSmall);

    x = firstX;
    y = firstY;

    if (CheckMinRomVersion(sysVersion35)) {
        WinPushDrawState();
	    WinGlueSetForeColorRGB
	    (&(userPrefs.colors[GetColorMode()][COLOR_EXTRAS]),NULL);
	    inverted = false;
    } else {
	inverted = userPrefs.optFlags.blackBackground
	    ^ userPrefs.optFlags.nightMode;
    }

    fastWinDrawLine =
        SysGetTrapAddress((inverted ? sysTrapWinEraseLine
                               : sysTrapWinDrawLine));
    fastWinDrawChars = (void (*) (const Char *, Int16, Coord, Coord))
        SysGetTrapAddress((inverted ^ userPrefs.optFlags.nightMode)
                          ? sysTrapWinDrawInvertedChars
                          : sysTrapWinDrawChars);

    prevX = -1000;
    prevY = -1000;
    penDown = true;

    do {
         Char str1[FIELDLENGTH];

         if (prevX != x || prevY != y) {
             RectangleType textBox;
             Int32         distance;

             if ( RctPtInRectangle( x, y, &lowresSkyRect ) &&
                  0 <= ( distance = SkyMapDistance( firstX, firstY, x, y ) ) ) {
                 FormatAngle(str1, distance, str2, false);
    
                 textBox.extent.x = FntCharsWidth(str1, StrLen(str1));
                 textBox.extent.y = FntCharHeight();
                 textBox.topLeft.x = x - textBox.extent.x;
                 textBox.topLeft.y = y - textBox.extent.y - 2;
    
                 if ( RctPtInRectangle( firstX, firstY, &textBox ) ) {
                     textBox.topLeft.x = x;
                 }
    
                 if ( lowresSkyRect.topLeft.x + lowresSkyRect.extent.x <
                          textBox.topLeft.x + textBox.extent.x )
                     textBox.topLeft.x = lowresSkyRect.topLeft.x
                                             + lowresSkyRect.extent.x
                                             - textBox.extent.x;

                 if ( textBox.topLeft.x < lowresSkyRect.topLeft.x )
                     textBox.topLeft.x = lowresSkyRect.topLeft.x;
    
                 if ( textBox.topLeft.y < lowresSkyRect.topLeft.y )
                     textBox.topLeft.y = lowresSkyRect.topLeft.y;
    
                 WinCopyRectangle(saved, WinGetActiveWindow(), &src,
                     lowresSkyRect.topLeft.x, lowresSkyRect.topLeft.y, winPaint);
    
                 fastWinDrawLine( firstX, firstY, x, y );

                 fastWinDrawChars(str1, StrLen(str1), textBox.topLeft.x,
                     textBox.topLeft.y);
             }

             prevX = x;
             prevY = y;
         }

         EvtGetPen(&x, &y, &penDown);
         WinDisplayToWindowPt(&x, &y);
    } while(penDown);

    WinCopyRectangle(saved, WinGetActiveWindow(), &src,
        lowresSkyRect.topLeft.x, lowresSkyRect.topLeft.y, winPaint);

    if (CheckMinRomVersion(sysVersion35)) {
	    WinPopDrawState();
    }
    FntSetFont(oldFont);

    WinDeleteWindow(saved, false);
}


static void SKYSECTION
SkyMapGoToRADec(void)
{
    FormType* formPtr = FrmInitForm(f_GetRADec);
    UInt16 button;
    Int32  ra;
    Int32  dec;
    Int32  sinra;
    Int32  cosra;
    Int32  sindec;
    Int32  cosdec;
    VectorType radec;
    VectorType radec2000;
    ControlType* decPlus;
    ControlType* decMinus;

    decPlus = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, p_DecPlus));
    decMinus = FrmGetObjectPtr(formPtr, FrmGetObjectIndex(formPtr, p_DecMinus));

    CtlSetValue(decPlus, true);
    CtlSetValue(decMinus, false);

    button = FrmDoDialog(formPtr);

    if (button==1) {
        FrmDeleteForm(formPtr);
        return;
    }

    dec = DMSToAng(GetNumField(formPtr,f_Decd),
                 GetNumField(formPtr,f_Decm),
                 GetNumField(formPtr,f_Decs));

    if (CtlGetValue(decMinus))
        dec = -dec;

    ra  = HMSToAng(GetNumField(formPtr,f_RAd),
                 GetNumField(formPtr,f_RAm),
                 GetNumField(formPtr,f_RAs));

    FrmDeleteForm(formPtr);

    fpsincos(dec, &sindec, &cosdec);
    fpsincos(ra, &sinra, &cosra);

    if (userPrefs.radec2000) {
        radec2000[0] = fpmul(cosra, cosdec);
        radec2000[1] = fpmul(sinra, cosdec);
        radec2000[2] = sindec;

        MatVecMul(radec, precession, radec2000);
    }
    else {
        radec[0] = fpmul(cosra, cosdec);
        radec[1] = fpmul(sinra, cosdec);
        radec[2] = sindec;
    }

    SkyMapCenterRaDec(radec);
    SkyMapDrawWithPos(FrmGetActiveForm());
}


static void DBSECTION
SkyMapDrawConstellationName(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    Char* p = constNames;
    while(itemNum--) {
        p += 1 + StrLen(p);
    }

    WinDrawChars(p, StrLen(p), bounds->topLeft.x, bounds->topLeft.y);
}


static void SKYSECTION
SkyMapGoToConstellation(Char* name)
{
    Int16      choice;
    Int16      z;

    ListType* l;

    if (numConstAreas == 0) {
        FrmAlert(a_NoConstellations);
        return;
    }

    choice = -1;

    if (name != NULL) {
        Char*      p;

        p = constNames;

        for (choice = 0; choice<numConstAreas; choice++) {
            if (!StrCaselessCompare(name, p))
                break;
            p += StrLen(p) + 1;
        }

        if (choice == numConstAreas)
            choice = -1;
    }

    if (choice < 0) {
        Coord width;
        Int16 i;
        UInt16 index;
        Char* p;
        RectangleType bounds;

        p = constNames;

        FntSetFont(stdFont);
        width = 0;
        for (i=0; i<numConstAreas; i++) {
            Coord w;

            w = FntCharsWidth(p, StrLen(p));

            if (width<w)
                width = w;

            p += StrLen(p) + 1;
        }

        l = FrmGetObjectPtrByID(FrmGetActiveForm(), l_Constellations);
        index = FrmGetObjectIndex(FrmGetActiveForm(), l_Constellations);

        FrmGetObjectBounds(FrmGetActiveForm(), index, &bounds);
        bounds.extent.x = 10 + width;
        FrmSetObjectBounds(FrmGetActiveForm(), index, &bounds);

        LstSetListChoices(l, (void*)constNames, numConstAreas);
        LstSetHeight(l, numConstAreas);
        LstSetDrawFunction(l, SkyMapDrawConstellationName);
        LstSetSelection(l, userPrefs.lastConstellation);
        LstMakeItemVisible(l, userPrefs.lastConstellation);

        choice = LstPopupList(l);
    }

    if (choice<0)
        return;

    userPrefs.lastConstellation = choice;

    for ( z = 1 ;
          z <= MAX_ZOOM;
          z++ ) {
         if ( (constAreas[choice].r >> z) < SKY_WIDTH / 2 &&
              (constAreas[choice].r >> z) < SKY_HEIGHT / 2 ) {
             break;
         }
    }

    userPrefs.zoomValue = z;
    SetRocker(&ZoomRocker, z);
    ZoomRocker.zoom = 1;

    SkyMapCenterRaDec(constAreas[choice].center);

    SkyMapDrawWithPos(FrmGetActiveForm());
}


static void SKYSECTION SetAngleFields(FormType* formPtr, UInt16 degID,
    UInt16 minID, UInt16 secID, Int32 angle)
{
    UInt16 deg;
    UInt16 min;
    UInt16 sec;

    AngToDMS(angle, &deg, &min, &sec);
    SetNumField(formPtr, degID, deg);
    SetNumField(formPtr, minID, min);
    SetNumField(formPtr, secID, sec);
}


void SKYSECTION
SkyMapChooseCircles(void)
{
    FrmPopupForm(f_ChooseCircles);
}


static Boolean SKYSECTION
SkyMapGadgetHandler(FormPtr formPtr, EventPtr event)
{
    static Int16 firstX, firstY, lastX, lastY;
    Boolean handled = false;
    switch (event->eType) {
    case penDownEvent:
        if (userPrefs.skyFlags.measureMode &&
            ! (event->screenX >= SKY_ORIENTATION_X &&
               event->screenY >= SKY_FULLSCREEN_Y)
             ) {
            SkyMapMeasure(formPtr, event->screenX, event->screenY);
            fastModePenDown = false;
        }
        else {
            lastX = firstX = event->screenX;
            lastY = firstY = event->screenY;
            fastModePenDown = ! userPrefs.slowScroll;
        }
        handled = true;
        break;

    case penMoveEvent:
        if (abs(firstX - event->screenX) > 5
            || abs(firstY - event->screenY) > 5)
            firstX = -1000;

        if (!userPrefs.skyFlags.measureMode) {
            SkyShift(formPtr, 
                     (event->screenX - lastX)*XFLIP, 
                     (event->screenY - lastY)*YFLIP,
                ! userPrefs.slowScroll);
            fastModePenDown = ! userPrefs.slowScroll;
        }
        else {
            fastModePenDown = false;
        }

        lastX = event->screenX;
        lastY = event->screenY;

        handled = userPrefs.slowScroll;
        break;

    case penUpEvent:
        if(needFullDraw) {
            SkyShift(formPtr, 
                     (event->screenX - lastX)*XFLIP,
                     (event->screenY - lastY)*YFLIP,
                     false);
        }
        fastModePenDown = false;
        if (abs(firstX - lastX) <= 5
            && abs(firstY - lastY) <= 5 ) {
            if (lastX >= SKY_FULLSCREEN_X
                && lastY >= SKY_FULLSCREEN_Y) {

                FrmGotoForm(IS_FULLSCREEN ? f_ObjectDetail : f_FullScreen);
            } else if (lastX >= SKY_ORIENTATION_X
                && lastY >= SKY_ORIENTATION_Y) {
                ListType* list;
                Int16     newOrientation;
                OrientationType oldOrientation;

                list = FrmGetObjectPtrByID(FrmGetActiveForm(), l_Orientation);
                oldOrientation = userPrefs.optFlags.orientation;
                LstSetSelection(list, userPrefs.optFlags.orientation);
                newOrientation = LstPopupList(list);
                if (0 <= newOrientation) {
                    userPrefs.optFlags.orientation = newOrientation;
                    if (userPrefs.optFlags.orientation != oldOrientation) {
                        FrmUpdateForm(FrmGetActiveFormID(), frmRedrawUpdateCode);
                    }
                }
            }
            else if(!userPrefs.skyFlags.measureMode) {
                UInt32 newObject;

                newObject = SkyMapFindObject(firstX, firstY);
                if ( NO_OBJECT != newObject )
                    SkyMapSelectObject(formPtr, newObject);
            }
        }
        handled = true;
        break;
        
    default:
    }
    return handled;
}

void SkyCenterObject (UInt32 objectNr)
{
    VectorType radec;

    SkyMapSetPosTime (userPrefs.jday, userPrefs.jtime, 
                      userPrefs.Location[userPrefs.currentLoc].lat,
                      userPrefs.Location[userPrefs.currentLoc].lng);
    /* Size box to encompass object */
    if (objectNr < NUMPLANETS) {
	MatVecMul(radec, precession, planetInfo[objectNr].pos);
    } else {
	AstroRecordType art;
	if (!GetCatalogObject(objectNr - NUMPLANETS, &art, NULL))
	    return;
	MatVecMul(radec, precession, art.radec);
    }

    SkyMapCenterRaDec(radec);
}


SKYSECTION static Boolean
InSkyMapWindow(void)
{
    UInt16 formID;

    formID = FrmGetActiveFormID();

    return (formID == f_ObjectDetail || formID == f_FullScreen) &&
           FrmGetWindowHandle(FrmGetActiveForm()) == WinGetDrawWindow();
}


SKYSECTION static Boolean
SkyMapZoomOut(void)
{
    userPrefs.zoomValue = MAX_ZOOM;
    SkyMapCenterRaDec(zenith);
    SkyMapDrawWithPos(FrmGetActiveForm());
    
    return true;
}


/*
 * Events for the sky view form
 *
 * You should not call functions directly that are also handled by form messages
 * as the order may not be as expected such as redrawing a form before it is setup.
 *
 * Instead send a form message to be handled by the event loop.
 */
Boolean SKYSECTION
SkyFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    static Boolean grabbedPen = false;
    static Boolean delayedTimeUpdate = false;
    FormPtr formPtr = FrmGetActiveForm();

    HostTraceOutputTL(appErrorClass, "SkyMapHandleEvent %d", event->eType);

    /* reset displayPos if timeout happened */
    if (event->eType == nilEvent && displayPos && ! userPrefs.skyFlags.noDisplayTimeout) {
        UInt32 ticks = TimGetTicks();
        if (ticks >= displayPosTimeout) {
            displayPos = false;
            if (userPrefs.skyFlags.displayPosText
                || userPrefs.skyFlags.displayPosGraph
                || userPrefs.skyFlags.displayZoom)
                FrmUpdateForm(FrmGetActiveFormID(), updatePosTimeout);
        } else {
            EvtGlueSetNullEventTick(displayPosTimeout);
        }
    }

    if (IS_FULLSCREEN) {
        /* Any activity on the zoom rocker? */
        if (event->eType == keyDownEvent) {
            if (event->data.keyDown.chr == pageUpChr
                && userPrefs.zoomValue > 1) {
                userPrefs.zoomValue--;
                SkyMapDrawWithPos(formPtr);
                return true;
            }
            else if (event->data.keyDown.chr == pageDownChr
                     && userPrefs.zoomValue < MAX_ZOOM) {
                userPrefs.zoomValue++;
                SkyMapDrawWithPos(formPtr);
                return true;
            }
            else if ((event->data.keyDown.modifiers & commandKeyMask) &&
                     event->data.keyDown.chr == vchrRockerCenter) {
                FrmGotoForm(f_ObjectDetail);
                return true;
            }
            else if (event->data.keyDown.chr == vchrRockerRight) {
                ChangeObject(formPtr, 1);
                return true;
            }
            else if (event->data.keyDown.chr == vchrRockerCenter) {
                FrmGotoForm(f_FullScreen);
                return true;
            }
        }
    } else {
        if (event->eType == keyDownEvent &&
             (event->data.keyDown.modifiers & commandKeyMask)) {
            if ( event->data.keyDown.chr == vchrRockerLeft ||
                 (userPrefs.pageBtnFunction ==  PAGEBTNOBJECT &&
                  event->data.keyDown.chr == pageUpChr) ) {
                ChangeObject(formPtr, -1);
                return true;
            }
            else if (event->data.keyDown.chr == vchrRockerRight  ||
                 (userPrefs.pageBtnFunction == PAGEBTNOBJECT &&
                  event->data.keyDown.chr == pageDownChr)) {
                ChangeObject(formPtr, 1);
                return true;
            }
            else if (event->data.keyDown.chr == vchrRockerCenter) {
                FrmGotoForm(f_FullScreen);
                return true;
            }
        }

        /* Any activity on the detail rocker? */
        if (RockerChangeHandleEvent(&DetailRocker, event)) {
            userPrefs.detailPage = DetailRocker.Value;
            FrmUpdateForm(FrmGetActiveFormID(), updateDetailsScrolled);
            return true;
        }

        /* Any activity on the zoom rocker? */
        if (RockerChangeHandleEvent(&ZoomRocker, event)) {
            userPrefs.zoomValue = ZoomRocker.Value;
            SkyMapDrawWithPos(formPtr);
            return true;
        }
    }

    if (DateChangeHandleEvent(event)
        || FormChangeHandleEvent(event))
        return true;

    HostTraceOutputTL(appErrorClass, "Switch %d", event->eType);
    switch (event->eType) {

    case zoomOutEvent:
        SkyMapZoomOut();
        handled = true;
        break;

    case chooseCirclesEvent:
        SkyMapChooseCircles();
        handled = true;
        break;

    case gotoRaDecEvent:
        SkyMapGoToRADec();
        handled = true;
        break;

    case dssLookupEvent:
        SkyMapDSSLookup();
        handled = true;
        break;

    case gotoConstellationEvent:
        SkyMapGoToConstellation(NULL);
        handled = true;
        break;

    case findConstellationEvent:
        SkyMapGoToConstellation(astroInfoArgument+1);
        handled = true;
        break;

    case gotoAlignEvent:
        SkyStoreViewpoint();
        SkyMapSetPosTime (userPrefs.jday, userPrefs.jtime,
                          userPrefs.Location[userPrefs.currentLoc].lat,
                          userPrefs.Location[userPrefs.currentLoc].lng);
        Align(GetCurrentObject());
        handled = true;
        break;

    case ctlRepeatEvent:
        switch (event->data.ctlRepeat.controlID) {

        case r_ObjPrev:
            ChangeObject(formPtr, -1);
            CtlHandleEvent(event->data.ctlRepeat.pControl, event);
            handled = true;
            break;

        case r_ObjNext:
            ChangeObject(formPtr, +1);
            CtlHandleEvent(event->data.ctlRepeat.pControl, event);
            handled = true;
            break;

        default:
        }
        break;

    case penDownEvent: {
        if (RctPtInRectangle (event->screenX, event->screenY,
                              &lowresSkyRect)) {
            grabbedPen = true;
        }
    }
        /* fall through */
    case penMoveEvent:
    case penUpEvent:
        if (grabbedPen) {
            displayPos = true;
            displayPosTimeout = 0xffffffff;
            handled = SkyMapGadgetHandler(formPtr, event);
            if (event->eType == penUpEvent) {
                grabbedPen = false;

                /* set timeout for display position */
                displayPosTimeout = TimGetTicks() + SKY_POS_TIMEOUT;
                if (! userPrefs.skyFlags.noDisplayTimeout)
                    EvtGlueSetNullEventTick(displayPosTimeout);
            }
        }
        break;

    case winDisplayChangedEvent:
        if (!InSkyMapWindow())
            break;
        SkyMapFree(formPtr);
        /* fall through... */
    case frmOpenEvent:
        SkyMapInit(formPtr);
        SkyMapDescribe(formPtr, GetCurrentObject());

        /* display position for the next few seconds */
        displayPosTimeout = TimGetTicks() + SKY_POS_TIMEOUT;
        if (!userPrefs.skyFlags.noDisplayTimeout)
            EvtGlueSetNullEventTick(displayPosTimeout);
        displayPos = true;

        SkyMapDraw(formPtr, false);
//#define BENCHMARK
#ifdef BENCHMARK
        {
          UInt16 i;
          UInt32 t;
          Char   s[55];
          t = TimGetTicks();
          for ( i = 0 ; i < 100 ; i++ ) {
               SkyMapDraw(formPtr, false);
          }
          StrPrintF(s,"<%ld>",TimGetTicks()-t);
          WinResetClip();
          WinDrawChars(s,StrLen(s),0,0);
          SysTaskDelay(100);
        }
#endif

        handled = true;
        break;

    case frmUpdateEvent:
        switch (event->data.frmUpdate.updateCode) {
        case updateDateChangedAutomatic:
            if (TimGetTicks() < lastDateChangedRedrawTicks +
                    (UInt32)MINIMUM_WAIT_MS * SysTicksPerSecond() / 1000) {
                /* try again in a second */
                EvtGlueSetNullEventTick( SysTicksPerSecond() );
                handled = true;
                break;
            }
            /* fall through */
        case updateDateChanged:
            if (!InSkyMapWindow()) {
                handled = true;
                break;
            }

            SkyStoreViewpoint();
            SkyMapSetPosTime (userPrefs.jday, userPrefs.jtime,
                              userPrefs.Location[userPrefs.currentLoc].lat,
                              userPrefs.Location[userPrefs.currentLoc].lng);
            SkyMapDescribe(formPtr, GetCurrentObject());
            SkyMapDraw(formPtr, fastModePenDown);

            lastDateChangedRedrawTicks = TimGetTicks();

            handled = true;
            break;

        case updateDetailChanged:
            SkyMapDescribe(formPtr, GetCurrentObject());
            handled = true;
            break;

        case updateZoomChanged:
            SkyMapDrawWithPos(formPtr);
            handled = true;
            break;

        case updatePosTimeout:
            if (!InSkyMapWindow()) {
                delayedTimeUpdate = true;
                handled = true;
                break;
            }
            delayedTimeUpdate = false;
            SkyMapDraw(formPtr, fastModePenDown);
            handled = true;
            break;

        case updateDetailsScrolled:
            ShowObjectDetail(GetCurrentObject());
            handled = true;
            break;

        default:
            FrmDrawForm(formPtr);
            SkyStoreViewpoint();
            SkyMapSetPosTime (userPrefs.jday, userPrefs.jtime,
                              userPrefs.Location[userPrefs.currentLoc].lat,
                              userPrefs.Location[userPrefs.currentLoc].lng);
            SkyMapDescribe(formPtr, GetCurrentObject());
            SkyMapDraw(formPtr, false);
            handled = true;
            break;
        }
        break;

    case winEnterEvent:
        if (delayedTimeUpdate) {
            FrmUpdateForm(FrmGetActiveFormID(), updatePosTimeout);
        }
        break;

    case frmCloseEvent:
	    SkyStoreViewpoint();
        SkyMapFree(formPtr);

        handled = false;
        break;
    default:
    }

    HostTraceOutputTL(appErrorClass, "SkyMapHandleEvent done");
    return handled;
}



void SKYSECTION
UpdatePhaseCalendarTable()
{
    FormPtr  formPtr = FrmGetActiveForm();
    TablePtr thetable = (TablePtr) FrmGetObjectPtrByID(formPtr, tbl_PhaseCal);
    Int16 i, offset;
    Char   str[FIELDLENGTH];
    DateType date;
    DateDaysToDate(userPrefs.jday - JDAYEPOCH, &date);

    offset = DayOfWeek(date.month, 1, date.year+1904);

    // Preserve the day headings
    ClearTable(tbl_PhaseCal, 1, 0);

    // Draw days
    for (i=0; i < DaysInMonth(date.month, date.year+1904); i++) {
        UpdateTblItemText(thetable, (((i + offset) / 7) % 5) + 1, ((i + offset) % 7), StrIToA(str, i + 1));
    }
}



static void SKYSECTION LoadCustomCircleNames( void )
{
    UInt16 size;
    UInt16 i;
    static Char empty[NUM_CUSTOM_CIRCLES] = { 0,0,0, 0,0,0, 0 };
    Char*  data;
    UInt16 offset;
    Boolean needToFree;

    size = 0;
    data = NULL;
    needToFree = false;

    if (noPreferenceFound != PrefGetAppPreferences(APPID, 'Ci', NULL, &size, true)
        && 0 < size) {

        data = MemPtrNew(size);

        if (data != NULL) {
            MemSet(data, size, 0);
            PrefGetAppPreferences(APPID, 'Ci',data,&size,true);
            needToFree = true;
        }
    }

    if (data == NULL) {
        size = NUM_CUSTOM_CIRCLES;
        data = empty;
    }

    offset = 0;

    for (i=0; i < NUM_CUSTOM_CIRCLES; i++) {
         if (size <= offset || data[offset] == 0) {
             Char name[FIELDLENGTH];
             GetStringResource(s_Custom, name);
             UpdateTextField(customCircleFields[i].name, name);
         }
         else {
             UpdateTextField(customCircleFields[i].name, data+offset);
         }
         offset += StrLen(data+offset) + 1;
    }

    if (needToFree)
        MemPtrFree(data);
}


static void SKYSECTION SaveCustomCircleNames( void )
{
    UInt16 length;
    UInt16 offset;
    UInt16 i;
    FormType* form;
    Char*     data;

    length = 0;
    form = FrmGetActiveForm();

    for (i=0; i < NUM_CUSTOM_CIRCLES; i++) {
         FieldType* f;
         Char*      name;

         f = (FieldPtr) FrmGetObjectPtrByID(form,
                 customCircleFields[i].name);

         name = FldGetTextPtr(f);
         if ( name == NULL )
             length++;
         else
             length += 1 + StrLen(name);
    }

    data = MemPtrNew(length);

    if (data==NULL)
        return;

    offset = 0;

    for (i=0; i < NUM_CUSTOM_CIRCLES; i++) {
         FieldType* f;
         Char*      name;

         f = (FieldPtr) FrmGetObjectPtrByID(form,
                 customCircleFields[i].name);

         name = FldGetTextPtr(f);
         if ( name == NULL ) {
             data[offset++] = 0;
         }
         else {
             StrCopy(data+offset, name);
             offset += 1 + StrLen(name);
         }
    }

    PrefSetAppPreferences(APPID, 'Ci', 1, data,
            offset, true);
            
    MemPtrFree(data);
}



Boolean SKYSECTION
ChooseCirclesFormHandleEvent(EventPtr event)
{
     Boolean handled = false;
     FormType* formPtr = FrmGetActiveForm();
     UInt16 i;

     switch(event->eType) {
         case frmOpenEvent:
            /* reset displayPos if timeout happened */
             for (i=0; i<NUM_CUSTOM_CIRCLES; i++) {
                Int32 c = CUSTOM_CIRCLES(i);

                if (c != 0) {
                    SetAngleFields(formPtr, customCircleFields[i].d,
                        customCircleFields[i].m,customCircleFields[i].s,
                        c);
                }
             }

             for (i=0; i<sizeof(circleIDs)/sizeof(*circleIDs); i++)
                  CtlSetValue(FrmGetObjectPtr(formPtr,
                                 FrmGetObjectIndex(formPtr, circleIDs[i])),
                                 (userPrefs.circles>>i)&1);

             LoadCustomCircleNames();

             FrmDrawForm(formPtr);
             handled = true;
             break;
          case ctlSelectEvent:
              switch (event->data.ctlSelect.controlID) {
                  case b_OK:
                      for (i=0; i<NUM_CUSTOM_CIRCLES; i++) {
                           Int32 value = DMSToAng(
                                   GetNumField(formPtr,customCircleFields[i].d),
                                   GetNumField(formPtr,customCircleFields[i].m),
                                   GetNumField(formPtr,customCircleFields[i].s));

                           if(i<FIRST_CUSTOM_CIRCLES)
                               userPrefs.customCircles[i] = value;
                           else
                               userPrefs.moreCustomCircles[i-FIRST_CUSTOM_CIRCLES] = value;
                      }

                      userPrefs.circles = 0;
                      for (i=0; i<sizeof(circleIDs)/sizeof(*circleIDs); i++)
                           if (CtlGetValue(FrmGetObjectPtr(formPtr,
                                           FrmGetObjectIndex(formPtr, circleIDs[i]))))
                                userPrefs.circles |= 1l<<i;
                      SaveCustomCircleNames();
                      FrmUpdateForm(userPrefs.activeFormId, frmRedrawUpdateCode);
                      /* fall through */
                   case b_Cancel:
                      FrmReturnToForm(0);
                      handled = true;
                      break;
              }
              break;
          default:
              break;
     }

     return handled;
}
