/*
 * Astro Sky - an astronomical visualizer for PalmOS devices.
 *
 * $Id: AstroSky.h 968 2010-04-11 13:31:28Z hoenicke $
 * Copyright (C) 2001, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#ifndef AstroSky_h
#define AstroSky_h

#include "AstroMatrix.h"
#include "AstroQTreePno.h"
#include "AstroHorizon.h"

#define SKYSECTION __attribute__ ((section ("code5")))

#define PLANETS_DBID  'PlIm'
#define PLANETS_BMPID 500

#define ASTRO_SKY_PROJ_SPH   0
#define ASTRO_SKY_PROJ_CYL   1
#define ASTRO_SKY_PROJ_CON   2

#define SKY_WIDTH (skyRect.extent.x)
#define SKY_HEIGHT (skyRect.extent.y)
#define SKY_X (skyRect.topLeft.x)
#define SKY_Y (skyRect.topLeft.y)

#define MAX_BRIGHTNESS 3

typedef UInt16 ObjectColorsType[MAX_BRIGHTNESS+1];

typedef struct {
    void    (*fastBmpBitsDrawHLine)(UInt16*, Int16, Coord, Coord, UInt16);
    void    (*fastBmpBitsDrawPixel)(UInt16*, Int16, Coord, UInt16);
    void    (*fastWinPaintRectangle) (RectangleType *, UInt16);
    void    (*fastWinDrawChars) (const Char *, Int16, Coord, Coord);
    ObjectColorsType* colors;
    MatrixType rotate;
    VectorType zenith;
    BitmapPtr dest;
    Char   *bits;
    Coord   maxBrightnessForDrawingCode;
    UInt16  rowBytes;
    UInt16  transparentColor;
    UInt16  colorOffset;
    UInt16  fadeOffset;
    Int16   magnitude;
    Boolean inverted;
    UInt8   horizon;
    Boolean nonStarMode;
    QTreePnoType* qTreePno;
} SkyDrawContext;

typedef struct {
    Int32    ra;
    Int32    dec;
} LineDataType;

typedef struct {
    VectorType pos;
    Int16 magnitude;
    Int32 pa;
    Int32 semid;
    BitmapPtr bmpPlanet;
    ColorTypeType colorType;
} PlanetInfoType;

#define INVALID_OBJECT_INDEX 0xFFFFFFFFul
#define INVALID_DISTANCE     0xFFFFFFFFul
#define MAX_FIND_ITEMS       12

typedef struct {
    UInt32 dist;
    UInt32 objectIndex;
    UInt16 catNum;
    UInt16 recidx;
    UInt16 offsetInNode;
} FoundObjectInfo;

typedef struct {
    UInt32 precision;
    FoundObjectInfo found[MAX_FIND_ITEMS];
    UInt32 ignoreDistance;
    UInt16 numObjects;
} FinderType;

typedef void (*DrawObjectFunc) (SkyDrawContext *ctx,
                                Int16 mag, Int16 x, Int16 y, UInt16 colorNum);


/* Read in the object catalog and set variables */
void SkyMapInit (FormPtr formPtr) SKYSECTION;

/* Set position and time */
void SkyMapSetPosTime(Int32 jday, Int32 jtime,
                      Int32 lat, Int32 lng) SKYSECTION;

void SkyMapSetCoord(Boolean radec) SKYSECTION;

void SkyGetPlanetInfo(PlanetInfoType* p, UInt32 planetNumber) SKYSECTION;

void
SkyGetRaDec2AltAz(MatrixType rotate) SKYSECTION;

/* Free the star catalog */
void SkyMapFree (FormPtr formPtr) SKYSECTION;

/* Redraw the sky. */
void SkyMapDraw (FormPtr formPtr, Boolean fast) SKYSECTION;

/* Center the object */
void SkyCenterObject(UInt32 objectNr) SKYSECTION;

/* Handle a event on the Sky Map form */
Boolean SkyFormHandleEvent(EventPtr event) SKYSECTION;

/* Load in current catalog and convert coordinates to cartesian. */
void SkyMapLoadCatalog(void) SKYSECTION;
void SkyMapFreeCatalog(void) SKYSECTION;

SKYSECTION void
SkyMapUpdateFinder(FinderType* finder, UInt32 dist,
    UInt32 objectIndex, UInt16 catNum, UInt16 recidx, UInt16 offsetInNode);

/*
 * Create a bitmap for the planet.  This bitmap will have the screen
 * depth and the same resolution as the bitmap in the database.  It
 * will already be "phased", but not rotated.
 *
 * planet          : Planet number
 * pa              : Phase angle
 * transparentColor: The color of transparent pixels
 * darkColor       : The color of shadow pixels
 *
 * Returns bitmap, or 0 if error occured.  You should delete it with
 * BmpDelete if you're finished with it.
 */
extern BitmapPtr SkyMapCreatePlanetBitmap
(Int16 planet, Int32 pa, Int16 transparentColor, Int16 darkColor) SKYSECTION;

void UpdatePhaseCalendarTable(void) SKYSECTION;
ColorModeType GetColorMode(void) SKYSECTION;
void GetBackColor(RGBColorType* back) SKYSECTION;
void SkyMapChooseCircles(void) SKYSECTION;
Boolean ChooseCirclesFormHandleEvent(EventPtr event) SKYSECTION;

extern Char filterString[FIELDLENGTH];
extern UInt32 totalRecords;
#endif
