/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroTable.c 945 2009-04-26 22:05:19Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <TblGlue.h>
#include "resource.h"
#include "AstroTable.h"
#include "Astro.h"
#include "AstroGlue.h"
#include "AstroLib.h"
#include "PalmUtil.h"
#include "AstroPhaseCal.h"
#include "Grayscale.h"

#include <HostControl.h>

void ATABLESECTION
AstroFreeTable(TableType *thetable)
{
    Int8 numColumns = TblGlueGetNumberOfColumns(thetable);
    Int8 numRows = TblGetNumberOfRows(thetable);
    Int8 c, r;
     
    HostTraceOutputTL(appErrorClass, "AstroFreeTable");

    for (c = 0; c < numColumns; c++) {
        for (r = 0; r < numRows; r++) {
            MemHandle ptr = (MemHandle) TblGlueGetItemPtr(thetable, r, c);
            if (ptr) {
                MemHandleFree(ptr);
                TblSetItemPtr(thetable, r, c, NULL);
            }
        }
    }
}



void ATABLESECTION
AstroTableDrawItem(void *tblP, 
                   Int16 row, Int16 column, RectanglePtr bounds) 
{
    FontID curFont;
    TablePtr thetable = (TablePtr) tblP;
    MemHandle handle = (MemHandle) TblGlueGetItemPtr(thetable, row, column);
    Char *text = NULL;
    Int16  opts, len;
    Int16  space;

    /* PalmOS 3.0 needs this */
    if (!CheckMinRomVersion(sysVersion30))
		    WinEraseRectangle(bounds, 0);

    if (handle == NULL)
        return;

    text = (Char*) MemHandleLock(handle);
    len = StrLen(text);
    opts = TblGetItemInt(thetable, row, column);
    curFont = FntSetFont(TblGetItemFont(thetable, row, column));

    // Work out the correct horizontal offset based on the alignment
    if ((opts & 3) == leftAlign) {
        space = 0;
    } else {
        space = bounds->extent.x - FntCharsWidth(text, len);
        if ((opts & 3) == centerAlign)
            space /= 2;
    }
    WinSetUnderlineMode(noUnderline);

    WinDrawChars(text, len, bounds->topLeft.x + space, bounds->topLeft.y);
    FntSetFont(curFont);
    MemHandleUnlock(handle);
}



/*
 * Set the field text.  It allocates a copy on the heap, so you can give
 * a local variable as text.
 */
void ATABLESECTION
UpdateTblItemText(TablePtr thetable, Int8 row, Int8 col, Char *text)
{
    MemHandle handle;
    Int16 size;

    if (text == NULL)
        return;

    size = StrLen(text) + 1;
        
    handle = (MemHandle)TblGlueGetItemPtr(thetable, row, col);
        
    if (handle != NULL) {
        if (MemHandleSize(handle) < size) {
            MemHandleFree(handle);
            handle = MemHandleNew(size);
        }
    } else
        handle = MemHandleNew(size);
            
    StrCopy (MemHandleLock(handle), text);
    TblSetItemPtr(thetable, row, col, handle);
    MemHandleUnlock(handle);
}



void ATABLESECTION
ClearTable(UInt16 objid, UInt8 startingrow, UInt8 startingcol)
{
    FormPtr  formPtr = FrmGetActiveForm();
    TablePtr thetable = (TablePtr) FrmGetObjectPtrByID(formPtr, objid);

    Int16 row, col;
    Int16 rows;
    Int16 cols;
    
    cols = TblGlueGetNumberOfColumns(thetable);

    if (tbl_PhaseCal == objid)
        rows = TblGetNumberOfRows(thetable);
    else
        rows = userPrefs.showDetailRows;

    for (row = startingrow; row < rows; row++) {
        for (col = startingcol; col < cols; col++) {
            MemHandle handle = (MemHandle) TblGlueGetItemPtr(thetable, row, col);
            if (handle)
                MemHandleFree(handle);
            TblSetItemPtr(thetable, row, col, NULL);

			// XXX testing
			//UpdateTblItemText(thetable, row, col, "");
        }
    }
}



void ATABLESECTION
InitPhaseCalendarTable()
{
    FormPtr  formPtr = FrmGetActiveForm();
    TablePtr thetable = (TablePtr) FrmGetObjectPtrByID(formPtr, tbl_PhaseCal);

    Int16 row, col;
    Char   str[FIELDLENGTH];

    for (row = 0; row < TblGetNumberOfRows(thetable); row++) {

        if (row > 0)
            TblSetRowHeight(thetable, row, 26);

        TblSetRowUsable(thetable, row, true);
        TblSetRowSelectable(thetable, row, row > 0);

        for (col = 0; col < TblGlueGetNumberOfColumns(thetable); col++) {
            TblSetItemPtr(thetable, row, col, NULL);
            TblSetItemStyle(thetable, row, col, customTableItem);

            if (row == 0) {
                TblSetItemInt(thetable, row, col, centerAlign);
                TblSetItemFont(thetable, row, col, fontSmall);
            } else {
                TblSetItemInt(thetable, row, col, centerAlign);
                TblSetItemFont(thetable, row, col, 130);
            }
        }
    }

    for (col = 0; col < TblGlueGetNumberOfColumns(thetable); col++) {
        TblSetCustomDrawProcedure(thetable, col, PhaseTableDrawItem);
        TblSetColumnUsable(thetable, col, true);
        TblSetColumnSpacing(thetable, col, 0);

        SysStringByIndex(stbl_ShortWeekdays, col, str, FIELDLENGTH);
        UpdateTblItemText(thetable, 0, col, str);
    }        
    /* XXX Bug in PilRC?? the usable flag isn't set automatically */
    FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, tbl_PhaseCal));
}



void ATABLESECTION
InitSolarTable(FormPtr formPtr)
{
    TablePtr thetable = (TablePtr)FrmGetObjectPtrByID(formPtr, tbl_Details);
    Int16 row, col;

    for (col = 0; col < TblGlueGetNumberOfColumns(thetable); col++) {
        TblSetCustomDrawProcedure(thetable, col, AstroTableDrawItem);
        TblSetColumnUsable(thetable, col, true);
        TblSetColumnSpacing(thetable, col, 0);
    }
    
    for (row = MERCURY; row <= SUN; row++) {
        TblSetRowUsable(thetable, row, true);
        for (col = 0; col < TblGlueGetNumberOfColumns(thetable); col++) {
            TblSetItemPtr(thetable, row, col, NULL);
            TblSetItemInt(thetable, row, col,
                          col == 2 || col == 4 || col == 6 ? leftAlign :
                          col == 0 ? centerAlign : rightAlign);
            TblSetItemFont(thetable, row, col,
                           col == 2 || col == 4 || col == 6 ? fontSmall : fontNormal);
            TblSetItemStyle(thetable, row, col, customTableItem);
        }
    }

    /* XXX Bug in PilRC?? the usable flag isn't set automatically */
    FrmShowObject(formPtr, FrmGetObjectIndex(formPtr, tbl_Details));
}

