/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: AstroTime.c 963 2010-01-07 21:40:21Z arpruss $
 * Copyright (C) 2002, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <DateGlue.h>
#include "resource.h"
#include "Astro.h"
#include "AstroTime.h"
#include "AstroLib.h"
#include "PalmUtil.h"
#include "FixedPoint.h"
#include "AstroDST.h"
#include "AstroGlue.h"
#include <HostControl.h>



static Char datestr[FIELDLENGTH];
static Char datefmt[FIELDLENGTH];
static Char timestr[FIELDLENGTH];
Int32 currentTzDST;
Int32 currentGSTOffset;

Int32 timePlayAdjust = 0;
static UInt32 timeNextTrigger = 0;

void ATIMESECTION
CalculateDate()
{
    Int32 dstoffset;
    dstoffset = getDSTOffset(userPrefs.jday, userPrefs.jtime,
                             userPrefs.Location[userPrefs.currentLoc].tz, 
                             &userPrefs.Location[userPrefs.currentLoc].dstRule);
    currentTzDST = userPrefs.Location[userPrefs.currentLoc].tz + dstoffset;
    while (userPrefs.jtime + currentTzDST >= HALFDAY) {
	userPrefs.jtime -= DAY;
	userPrefs.jday++;
    }
    while (userPrefs.jtime + currentTzDST < -HALFDAY) {
	userPrefs.jtime += DAY;
	userPrefs.jday--;
    }
    currentGSTOffset = getGSTOffset(userPrefs.jday);
}

/*
 * Change the base time (used for various calculations) by 'minutes'.
 */
static Boolean ATIMESECTION
ChangeTime(Int32 seconds)
{
    Int32 days = seconds / 86400;
    userPrefs.jtime += (seconds % 86400) * 1000L;
    userPrefs.jday  += days;
    CalculateDate();
    return true;
}


 
/*
 * Display the time selector and update the time with the user's
 * selection.
 *
 * Return boolean indicating whether time has actually changed to reduce unnecessary recalc
 */
static Boolean ATIMESECTION
PickTime()
{
    Char str[FIELDLENGTH];
    Int16 day;
    Int16 hours, minutes;

    /* Provide a default of the current setting. */
    minutes = (Int16)((userPrefs.jtime + currentTzDST) / 60000L) + 720;
    day = minutes / 1440;
    minutes %= 1440;
    if (minutes < 0) {
        minutes += 1440;
        day--;
    }
    hours = minutes / 60;
    minutes = minutes % 60;
    
    /* PalmOS 3.1+ has a nice time picker. */
    if (CheckMinRomVersion(sysVersion31)) {
        

        if (!SelectOneTime(&hours, &minutes, 
                           GetStringResource(s_SelectTime, str)))
            return false;
    } else {
        /* Pre PalmOS 3.1 it's a bit clumsier. */

        TimeType    s,e;
        
        e.hours = s.hours = hours;
        e.minutes = s.minutes = minutes;

        if (!SelectTimeV33(&s, &e, false, 
                           GetStringResource(s_SetStartTimeOnly, str), 0)
            || *((Int16*)&s) == -1)
            return false;
    }
    userPrefs.jtime = (((day * 24) + hours) * 60 + minutes) * 60000L
        - currentTzDST - HALFDAY;
    userPrefs.optFlags.isNow = false;

    CalculateDate();
    return true;
}



/*
 * Display the date selector and update the current date with the
 * user's selection. dP is a pointer to the selector resource
 * to refresh.
 *
 * Return boolean indicating whether date has actually changed to reduce unnecessary recalc
 */
 
/* TODO: I am not sure this works perfectly across DST changes */ 
static Boolean ATIMESECTION
PickDate()
{
    Int16    adjtimedays;
    Char str[FIELDLENGTH];
    DateType date;
    Int16    day, month, year;

    adjtimedays = (userPrefs.jtime + DAY + HALFDAY + currentTzDST ) / DAY - 1;
    DateDaysToDate(userPrefs.jday + adjtimedays - JDAYEPOCH, &date);
    day   = date.day;
    month = date.month;
    year  = date.year+1904;
    if (!SelectDay(selectDayByDay, &month, &day, &year,
                   GetStringResource(s_SelectDate, str)))
        return false;

    userPrefs.jday = JDay(year, month, day) - adjtimedays;
    userPrefs.optFlags.isNow = false;
    CalculateDate();
    return true;
}



/*
 * Events for the sidereal time form
 */
Boolean ATIMESECTION
TimebasesFormHandleEvent(EventPtr event)
{
    Boolean handled = false;
    FormPtr  formPtr = FrmGetActiveForm();
    
    if (DateChangeHandleEvent(event)
        || FormChangeHandleEvent(event))
        return true;

    switch (event->eType) {
        case frmOpenEvent:
            DrawTimePicker();
            DrawDatePicker();
            ShowTimebases(formPtr);

            SetMenu(formPtr);
            FrmDrawForm(formPtr);

            handled = true;
            break;

        case frmUpdateEvent:
            switch (event->data.frmUpdate.updateCode) {
            
            case updateDateChanged:
                ShowTimebases(formPtr);
                break;
            }
            handled = false;
            break;

        default:
            handled = false;
    }

    return handled;
}



Boolean
DateChangeHandleEvent(EventPtr event)
{
    static Boolean active = false;
    Boolean changed = false;
    Boolean handledEvent = false;
    Int32 oldLocalDay, oldLocalTime;
    UInt16 changeCode;

    oldLocalTime = (userPrefs.jtime + currentTzDST);
    oldLocalDay = userPrefs.jday + (oldLocalTime + DAY + HALFDAY) / DAY - 1;
    
    changeCode = updateDateChanged;

    switch (event->eType) {
        case ctlSelectEvent:
            switch (event->data.ctlSelect.controlID) {
            
                case s_Date:
                    changed = PickDate();
                    if (changed) {
                        userPrefs.optFlags.timeStopped = true;
                        userPrefs.optFlags.isNow = false;
                    }
                    handledEvent = true;
                    break;
                    
                case s_Time:
                    changed = PickTime();
                    if (changed) {
                        userPrefs.optFlags.timeStopped = true;
                        userPrefs.optFlags.isNow = false;
                    }
                    handledEvent = true;
                    break;

                case r_DateDownStep:
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = true;
                    changed = ChangeTime(-userPrefs.timeInc);
                    handledEvent = false; /* allow to repeat */
                    break;
                    
                case r_DateDownPlay:
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = false;
                    timeNextTrigger = 0;
                    timePlayAdjust = -userPrefs.timeInc;
                    handledEvent = true;
                    break;

                /* If Stop button is pressed then stop increment 
                 * and/or realtime 
                 */
                case r_Stop:
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = true;
                    handledEvent = true;
                    break;
                
                /* If Now button is pressed then set to realtime */
                case r_Now:
                    userPrefs.optFlags.isNow = true;
                    userPrefs.optFlags.timeStopped = false;
                    changed = SetNow();
                    timeNextTrigger = 0;
                    handledEvent = true;
                    break;

                case r_DateUpPlay:
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = false;
                    timePlayAdjust = userPrefs.timeInc;
                    timeNextTrigger = 0;
                    handledEvent = true;
                    break;

                case r_DateUpStep:
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = true;
                    changed = ChangeTime(userPrefs.timeInc);
                    handledEvent = false; /* allow to repeat */
                    break;

                default:
            }
            break;
    
        case keyDownEvent:
            if (userPrefs.pageBtnFunction == PAGEBTNDATE) {
                if (event->data.keyDown.chr == pageUpChr) {
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = true;
                    changed = ChangeTime(userPrefs.timeInc);
                    handledEvent = true;
                } else if (event->data.keyDown.chr == pageDownChr) {
                    userPrefs.optFlags.isNow = false;
                    userPrefs.optFlags.timeStopped = true;
                    changed = ChangeTime(-userPrefs.timeInc);
                    handledEvent = true;
                }
            }
            break;

	case winExitEvent:
	    active = false;
	    break;

	case winEnterEvent: {
	    FormPtr form = FrmGetFirstForm();
	    if (event->data.winEnter.enterWindow == (WinHandle) form) {
		UInt16 formID = FrmGetFormId(form);
		switch (formID) {
		    case f_SolarSystem:
		    case f_PhaseCalendar:
		    case f_Catalog:
		    case f_ObjectDetail:
		    case f_FullScreen:
		    case f_Timebases:
		    case f_List:
			active = true;
			break;
		}
	    }
	    break;
	}


        default:
    }        

    if (active) {

	if (!userPrefs.optFlags.timeStopped) {
	    /* If we have pressed play and the next interval has passed
	     * then increment time
	     */
	    Boolean forcedChange;
	    UInt32 now = TimGetSeconds();
	    Boolean setNullEventTick = (event->eType == nilEvent) ||
            (timeNextTrigger == 0);
            
            forcedChange = changed;

	    if (timeNextTrigger < now + userPrefs.timeInt) {
		    if (userPrefs.optFlags.isNow)
		        changed = SetNow();
            else
		        changed = ChangeTime(timePlayAdjust);

            if(changed && !forcedChange)
                changeCode = updateDateChangedAutomatic;

            timeNextTrigger += userPrefs.timeInt;

		    if (timeNextTrigger < now + userPrefs.timeInt)
		        timeNextTrigger = now + userPrefs.timeInt;
        }

	    if (setNullEventTick)
	        EvtGlueSetNullEventTick(TimGetTicks() +
				    ticksPerSec * (timeNextTrigger - now));
	}

	if (changed) {
	    UInt16 formID;
	    Int32 newLocalTime = userPrefs.jtime + currentTzDST;
	    Int32 newLocalDay = userPrefs.jday
		+ (newLocalTime + DAY + HALFDAY) / DAY - 1;

	    /* These checks avoid redraw if nothing visible changes */
	    /* Don't trigger redraw for phase calendar
	     * if day does not change
	     */
	    formID = FrmGetActiveFormID();
	    if (formID != f_FullScreen) {
		if ((newLocalTime - oldLocalTime) % DAY != 0)
		    DrawTimePicker();
		if (oldLocalDay != newLocalDay)
		    DrawDatePicker();
	    }

	    if (oldLocalDay != newLocalDay
		|| (oldLocalTime != newLocalTime
		    && formID != f_PhaseCalendar)) {
		FrmUpdateForm(formID, changeCode);
            }
	    
	}
    }
	
    return handledEvent;
}



void ATIMESECTION
DrawDatePicker()
{
    DateType date;
    DateDaysToDate(userPrefs.jday - JDAYEPOCH + 
                   (userPrefs.jtime + currentTzDST + DAY + HALFDAY) / DAY - 1,
                   &date);
#ifdef NO_GLUE
    DateTemplateToAscii
#else
    DateGlueTemplateToAscii
#endif
                            (GetStringResource(s_DayMonthFormat, datefmt),
                            date.month, date.day, date.year+1904,
                            datestr,
                            FIELDLENGTH);

    SetControlLabel(s_Date, datestr);
}



void ATIMESECTION
DrawTimePicker()
{
    Int32 time = userPrefs.jtime + currentTzDST + HALFDAY;
    Int16 hour, minute, second;

    while (time < 0)
        time += DAY;
    while (time >= DAY)
        time -= DAY;
    time /= 1000L;
    second = time % 60;
    minute = time / 60;
    hour   = minute / 60;
    minute = minute % 60;

    StrPrintF(timestr, "%02d:%02d:%02d%s",
              userPrefs.optFlags.time24Hour ? hour : (hour + 11) % 12 + 1,
              minute, second,
              userPrefs.optFlags.time24Hour ? "" : (hour >= 12 ? "p" : "a"));

    SetControlLabel(s_Time, timestr);
}



/*
 * Time is measured in seconds since midday. Positive forward, negative back.
 *
 */
Char * ATIMESECTION
FormatTime(Int32 time, Char *s, UInt16 size, Char *flags, Boolean force24H)
{
    Int32       hr, min;
    Int16         day;
    Char        str[10];

    time += HALFDAY;
    day = 0;
    while (time < 0) {
        time += DAY;
        day--;
    }
    while (time >= DAY) {
        time -= DAY;
        day++;
    }
        
    min = time / 60000;
    hr  = (min / 60) % 24; 
    min = min % 60;
    
    TimeToAscii(hr, min, 
                userPrefs.optFlags.time24Hour || force24H
                ? tfColon24h : sysPrefs.timeFormat, str);
    StrCopy(s, "");
    if (hr < 10) {
        StrCopy(s, "0");
    }
    StrCat(s, str);
    
    /* Indicate next, previous, current day. */
    if (flags) {
        if (userPrefs.Location[userPrefs.currentLoc].tz != currentTzDST)
            *flags++ = '';

        if (day < 0)
            *flags++ = '-';
        else if (day > 0)
            *flags++ = '+';
        
        *flags = 0;
    }
    return s;
}




/*
 * Display the current Greenwich and Local Sidereal times
 */
void ATIMESECTION
ShowTimebases(FormPtr formPtr)
{
    Int32 gst, lst;
    Char str[FIELDLENGTH];

    gst = UTCToGST(userPrefs.jtime, currentGSTOffset);
    lst = GSTToLST(gst, userPrefs.Location[userPrefs.currentLoc].lng);
    
    GetJulianDateString(str, userPrefs.jday, userPrefs.jtime);
    UpdateTextField(f_Jday, str);

    FormatTime(fpmul(gst + -2*FP_ONE, DAY/4), str, FIELDLENGTH, NULL, true);
    UpdateTextField(f_GST, str);
    FormatTime(fpmul(lst + -2*FP_ONE, DAY/4), str, FIELDLENGTH, NULL, true);
    UpdateTextField(f_LST, str);
}



/* Determines whether an object is above the horizon based on rise/set time. */

Int16 ATIMESECTION
UpNow(Int16 flags, Int32 rise, Int32 set, Int32 now)
{
    switch (flags & ASTROLIB_DONTRISESET) {
    case 0:
        return (rise <= now) && (now <= set);

    case ASTROLIB_DONTRISE:
        return now <= set;

    case ASTROLIB_DONTSET:
        return rise <= now;

    case ASTROLIB_DONTRISESET:
        return flags & ASTROLIB_CIRCUMPOLAR;
    }
    // Only get here if something is VERY wrong!
    abort();
}


void ATIMESECTION
GetJulianDate(UInt32* jdayP, UInt32* jtimeP, UInt32 seconds)
{
    Int32  dstoffset;

    *jdayP = seconds / 86400 + JDAYEPOCH;
    *jtimeP = (seconds % 86400) * 1000
        - userPrefs.Location[userPrefs.currentLoc].tz - HALFDAY;

    dstoffset = getDSTOffset(userPrefs.jday, userPrefs.jtime,
                             userPrefs.Location[userPrefs.currentLoc].tz,
                             &userPrefs.Location[userPrefs.currentLoc].dstRule);

    *jtimeP -= dstoffset;
}


void ATIMESECTION
GetJulianDateString(Char* str, UInt32 jday, UInt32 jtime)
{
    Int32  frac, deltaday;

    frac     = jtime / (Int32) (DAY/10000L);
    deltaday = frac / 10000L;
    frac     %= 10000L;
    if (frac < 0) {
        frac += 10000L;
        deltaday--;
    }
    StrPrintF(str, "%ld.%04ld", jday + deltaday, frac);
}


/*
 * Set the base date and TOD (used for various calculations) to 'now'.
 */
Boolean ATIMESECTION
SetNow()
{
    UInt32 seconds;
    Int32  dstoffset;

    /*
     * Check whether we need to change something.
     */

    seconds = TimGetSeconds();
    if ((Int32)(seconds - (UInt32)(userPrefs.jday - JDAYEPOCH) * 86400UL)
        * 1000L == userPrefs.jtime + currentTzDST)
        return false;

    /*
     * Set the date and time to now.
     */
    GetJulianDate(&userPrefs.jday, &userPrefs.jtime, seconds);
    dstoffset = getDSTOffset(userPrefs.jday, userPrefs.jtime,
                             userPrefs.Location[userPrefs.currentLoc].tz,
                             &userPrefs.Location[userPrefs.currentLoc].dstRule);
    currentTzDST = userPrefs.Location[userPrefs.currentLoc].tz + dstoffset;
    currentGSTOffset = getGSTOffset(userPrefs.jday);
    return true;
}

