/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: Bitmap.c 967 2010-04-07 07:44:22Z hoenicke $
 * Copyright (C) 2001, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <BmpGlue.h>
#include "Bitmap.h"
#include "FixedPoint.h"
#include "HostControl.h"
#include "Grayscale.h"


static const Int8 log2depth[17] = { -1, 0, 1, -1, 2, -1, -1, -1, 3,
				    -1, -1, -1, -1, -1, -1, -1, 4 };

static BITMAPSECTION void
BmpTransformLine4(UInt8 *ptr, UInt8 shift, Int16 lwidth, 
                  UInt8 *sptr, UInt16 sxl, UInt16 syl, 
                  Int32 dsptr, UInt16 dxxl, UInt16 dyxl,
                  UInt16 srowBytes, UInt8 transparentColor)
{
    UInt8 pixel;
    if (!lwidth)
	return;

    pixel = *ptr;
    switch (shift) {
    case  0:
	goto shift_0;
    case  4:
	goto shift_1;
    }

    for (;;) {
        /* Invariant:
         * ptr              points to current pixel in dest.
         * sptr, sxl, syl   point to the corresponding pixel in src.
         * Note that this is the inner loop; it should be fast.
         */
        UInt8 spix;
        
        pixel = *ptr;
    shift_0:
        spix = ((*sptr) << ((sxl >> 13) & 0x4)) & 0xf0;
        if (spix != transparentColor) {
            pixel ^= (pixel ^ spix) & 0xf0;
        }

        sptr += dsptr;
        syl += dyxl;
        if (syl < dyxl)
            sptr += srowBytes;
        sxl += dxxl;
        if (sxl < dxxl)
            sptr++;
	if (!--lwidth)
	    break;

    shift_1:
        spix = ((*sptr) << ((sxl >> 13) & 0x4)) & 0xf0;
        if (spix != transparentColor) {
            pixel ^= (pixel ^ (spix>>4)) & 0xf;
        }

        sptr += dsptr;
        syl += dyxl;
        if (syl < dyxl)
            sptr += srowBytes;
        sxl += dxxl;
        if (sxl < dxxl)
            sptr++;
	if (!--lwidth)
	    break;

        *ptr++ = pixel;
    }
    *ptr++ = pixel;
}

static BITMAPSECTION void
BmpTransformLine8(UInt8 *ptr, Int16 lwidth, 
                  UInt8 *sptr, UInt16 sxl, UInt16 syl, 
                  Int32 dsptr, UInt16 dxxl, UInt16 dyxl,
                  UInt16 srowBytes, UInt8 transparentColor)
{
    while (lwidth-- > 0) {
        UInt8 spix;
        
        spix = *sptr;
        if (spix != transparentColor)
            *ptr = spix;

        sptr += dsptr;
        syl += dyxl;
        if (syl < dyxl)
            sptr += srowBytes;
        sxl += dxxl;
        if (sxl < dxxl)
            sptr++;

        ptr++;
    }
}

static BITMAPSECTION void
BmpTransformLine16(UInt16 *ptr, Int16 lwidth, 
                   UInt16 *sptr, UInt16 sxl, UInt16 syl, 
                   Int32 dsptr, UInt16 dxxl, UInt16 dyxl,
                   UInt16 srowBytes, UInt16 transparentColor)
{
    while (lwidth-- > 0) {
        UInt16 spix;
        
        spix = *sptr;
        if (spix != transparentColor)
            *ptr = spix;

        sptr = (void*) sptr + dsptr;
        syl += dyxl;
        if (syl < dyxl)
            sptr = (void*) sptr + srowBytes;
        sxl += dxxl;
        if (sxl < dxxl)
            sptr++;

        ptr++;
    }
}

BITMAPSECTION void 
BmpScaleRotateBitmap
(BitmapPtr dest, Coord cx, Coord cy, Coord dwidth, Coord dheight,
 BitmapPtr src,  Coord sx, Coord sy, Coord swidth, Coord sheight,
 Int32 sina, Int32 cosa, Int16 xFlip,
 Int16 transparentColor, WinDrawOperation drawMode)
{
    Coord width, height, lastsy;
    UInt16 rowBytes, srowBytes;
    UInt16 *dptr, *sptr;
    Int32 sgx, sgy, dxx, dyx, dxy, dyy;
    Int32 sx0, sx1, sy0, sy1;
    Int32  dsptr;
    UInt16 dyxl, dxxl;
    UInt16 pixelMask;
    Int16 depthShift,  ppwShift, ppwMask;
    Int16 depth;

    /* Check depth */
    depth = BmpGlueGetBitDepth(src);
    ErrFatalDisplayIf(depth != BmpGlueGetBitDepth(dest), 
		      "depth of bitmap mismatch");
    ErrFatalDisplayIf(depth > 16, "depth is too big");
    depthShift = log2depth[depth];
    ErrFatalDisplayIf(depthShift == -1, "depth is not a power of two");

    /* We we later shift all pixel data into the high bits.
     * pixelMask masks the pixel data.
     * ppwShift is the amount we shift the x offsets to the right to get 
     * the word offset.
     * ppwMask masks the lower part of the x offset.
     * transparentColor is shifted to the pixel data position.
     */
    pixelMask = (0xffff << (16 - depth));
    ppwShift  = 4 - depthShift;
    ppwMask   = (0xf << depthShift) & 0xf;
    transparentColor <<= (16 - depth);

    /* The following pixel coordinates are stored as GP fixed point numbers.
     * The x coordinate is in words (16 bits) not pixel.
     * The y coordinate is in rows as expected.
     */

    /* Calculate delta values
     * dij: change in i coordinate of source image if we change
     *      j coordinate of dest image by one.
     */
    dxx = ((((Int32) swidth) << 16) / dwidth);
    dyy = (((Int32) sheight) << 16) / dheight;
    dxy = (fpmul(sina, dxx) >> ppwShift);
    dxx = (fpmul(cosa, dxx) >> ppwShift) * xFlip;
    dyx = fpmul(-sina, dyy) * xFlip;
    dyy = fpmul(cosa, dyy);

    /* convert source center coords to GP */
    sgx = ((Int32) (2*sx + swidth)) << (15 - ppwShift);
    sgy = ((Int32) (2*sy + sheight)) << 15;

    /* convert source border coords to GP */
    sx0 = (UInt32) sx << (16 - ppwShift);
    sx1 = (UInt32) (sx + swidth) << (16 - ppwShift);
    sy0 = (UInt32) sy << 16;
    sy1 = (UInt32) (sy + sheight) << 16;

    /* Get destination bitmap dimension, we will need to clip to it.
     * Also get the rowBytes and data for both bitmaps.
     */
    BmpGlueGetDimensions(dest, &width, &height, &rowBytes);
    BmpGlueGetDimensions(src, NULL, NULL, &srowBytes);
    dptr = BmpGetBits(dest);
    sptr = BmpGetBits(src);
    lastsy = 0;

    /* Put the non-fraction part of dyx and dxx as constant offset dsptr */
    dsptr  = (dyx >> 16) * (Int32) srowBytes + (Coord) (dxx >> 16)*2;
    dyxl   = dyx & 0xffff;
    dxxl   = dxx & 0xffff;
    
    /* Now find out the lines of destination that matter.
     */
    {
        Coord rheight;
        Coord rh2;
        /* Get real height of destination rectangle and top line */
        rheight = fpabs(fpmul(cosa, dheight)) + fpabs(fpmul(sina, dwidth));
        rh2 = (rheight+1) / 2;
        
        if (cy > rh2) {
            /* go to minimum y coordinate of dest bitmap that matters 
             * and adjust center and height
             */
            dptr = (void*) dptr + (Int32) rowBytes * (cy - rh2);
            height -= (cy - rh2);
            cy   = rh2;
        }
        /* Move in the source image to the top row of dest */
        sgx -= dxy * cy + dxx * cx;
        sgy -= dyy * cy + dyx * cx;

        /* adjust height of dest */
        if (cy + rh2 < height)
            height = cy + rh2;
    }

    while (height-- > 0) {
        /* Invariant:
         * dptr points to the first pixel of current line in dest
	 * sptr points always to the start of source bitmap
         * sgx, sgy    point to the pixel in src corresponding to dptr.
         */
        UInt16 *ptr, *lsptr;
        UInt16 pixel;
        Int16 shift;
        Int32 slx = sgx;
        Int32 sly = sgy;
        UInt16 sxl, syl;
        Coord xl, lwidth;

        UInt32 q;
        xl = 0;
        lwidth = width;
        if (dxx > 0) {
            if (slx < sx0) {
                q = (UInt32) (sx0 - slx + (dxx-1)) / (UInt32) dxx;
                if (q >= lwidth)
                    goto next_scanline;
                xl += q;
                lwidth -= q;
                slx += q * dxx;
                sly += q * dyx;
            }
            if (slx > sx1)
                goto next_scanline;

            q = (UInt32) (sx1 - slx + (dxx-1)) / (UInt32) dxx;
            if (lwidth > q)
                lwidth = q;
        } else if (dxx < 0) {
            if (slx > sx1) {
                q = (UInt32) (slx - sx1 + (-dxx-1)) / (UInt32) -dxx;
                if (q >= lwidth)
                    goto next_scanline;
                xl += q;
                lwidth -= q;
                slx += q * dxx;
                sly += q * dyx;
            }
            if (slx < sx0)
                goto next_scanline;
            
            q = (UInt32) (slx - sx0) / (UInt32) -dxx + 1;
            if (lwidth > q)
                lwidth = q;
        }
        if (dyx > 0) {
            if (sly < sy0) {
                q = (UInt32) (sy0 - sly + (dyx-1)) / (UInt32) dyx;
                if (q >= lwidth)
                    goto next_scanline;
                xl += q;
                lwidth -= q;
                slx += q * dxx;
                sly += q * dyx;
            }
            if (sly > sy1)
                goto next_scanline;
            
            q = (UInt32) (sy1 - sly + (dyx-1)) / (UInt32) dyx;
            if (lwidth > q)
                lwidth = q;
        } else if (dyx < 0) {
            if (sly > sy1) {
                q = (UInt32) (sly - sy1 + (-dyx-1)) / (UInt32) -dyx;
                if (q >= lwidth)
                    goto next_scanline;
                xl += q;
                lwidth -= q;
                slx += q * dxx;
                sly += q * dyx;
            }
            if (sly < sy0)
                goto next_scanline;
            
            q = (UInt32) (sly - sy0) / (UInt32) -dyx + 1;
            if (lwidth > q)
                lwidth = q;
        }

        ptr = dptr + (xl >> ppwShift);
        shift = (xl << depthShift) & ppwMask;

	/* lsptr points to the word in source bitmap that contains
	 * the pixel slx, sly.  The numbers sxl, syl contain the
	 * fractional part of pixel position.
	 */

        lsptr = (void*)sptr + (sly >> 16) * (Int32) srowBytes
	    + (Coord) (slx >> 16) * 2;
        sxl = slx & 0xffff;
        syl = sly & 0xffff;

	/* First we use some optimized versions of the most common 
	 * bitmap depths and operations.
	 */
        if (drawMode == winPaint) {
	    switch (depth) {
	    case 4:
		BmpTransformLine4((UInt8*)ptr + (shift >> 3), shift & 4,
                                  lwidth,
                                  (UInt8*)lsptr + (sxl >> 15),
                                  sxl << 1, syl,
                                  dsptr + (dxxl >> 15), 
                                  dxxl << 1, dyxl, srowBytes,
                                  transparentColor >> 8);
		goto next_scanline;
	    case 8:
		/* convert to byte based coordinates */
		BmpTransformLine8((UInt8*)ptr + (shift >> 3), lwidth, 
                                  (UInt8*)lsptr + (sxl >> 15), 
                                  sxl << 1, syl,
                                  dsptr + (dxxl >> 15), 
                                  dxxl << 1, dyxl, srowBytes, 
                                  transparentColor >> 8);
		goto next_scanline;
	    case 16:
		BmpTransformLine16(ptr, lwidth, lsptr, sxl, syl,
				   dsptr, dxxl, dyxl, srowBytes, 
				   transparentColor);
		goto next_scanline;
	    }
	}

	/* The non-optimized allpurpose code */
	pixel = *ptr;
	while (lwidth-- > 0) {
	    /* Invariant:
	     * ptr, shift point to current pixel in dest.
	     * slx, sly   point to the corresponding pixel in src.
	     * Note that this is the inner loop; it should be fast.
	     */
	    UInt16 spix;
	    
	    if (shift == 16) {
		*ptr++ = pixel;
		shift = 0;
		pixel = *ptr;
	    }
	    
	    spix = ((*lsptr) << ((sxl >> 12) & ppwMask)) & pixelMask;
	    if (spix != transparentColor) {
		spix >>= shift;
		switch (drawMode) {
                case winPaint:
		    pixel ^= (pixel ^ spix) & (pixelMask >> shift);
		    break;
                case winErase:
		    pixel &= spix;
		    break;
		case winMask:
		    pixel &= ~spix;
		    break;
		case winInvert:
		    pixel ^= spix;
		    break;
		case winOverlay:
		    pixel |= spix;
		    break;
		case winPaintInverse:
		    pixel ^= (pixel ^ ~spix) & (pixelMask >> shift);
		    break;
		default:
		    ErrFatalDisplayIf(1, "unsupported draw mode.");
		}
	    }
	    
	    shift += 1 << depthShift;
	    
	    lsptr = (void*) lsptr + dsptr;
	    syl += dyxl;
	    if (syl < dyxl)
		lsptr = (void*) lsptr + srowBytes;
	    sxl += dxxl;
	    if (sxl < dxxl)
		lsptr++;
	}
	*ptr  = pixel;
    next_scanline:
	dptr = (void*) dptr + rowBytes;
        sgx  += dxy;
        sgy  += dyy;
    }
}

BITMAPSECTION void
BmpBitsDrawHLine16(UInt16 *bits, Int16 depth,
                  Coord x_start, Coord x_end, UInt16 color)
{
    Int16 count;
    count = x_end - x_start;
    bits += x_start;
    while (count--) {
        *bits++ = color;
    }
}


BITMAPSECTION void
BmpBitsDrawHLine8(UInt16 *bits0, Int16 depth,
                  Coord x_start, Coord x_end, UInt16 color)
{
    Int16 count;
    UInt8* bits = ( UInt8* )bits0;
    count = x_end - x_start;
    bits += x_start;
    while (count--) {
        *bits++ = color;
    }
}


BITMAPSECTION void
BmpBitsDrawHLine (UInt16 *bits, Int16 depth,
                  Coord x_start, Coord x_end, UInt16 color)
{
    Int16 depthBits = log2depth[depth];
    Int16 offset;
    Int16 i;
    UInt16 mask;

    /* extend color to one word */
    for (i = depth; i < 16; i += i) {
        color |= (color << i);
    }

    /* adjust start/end so that they are in bits, instead of pixels */
    x_start <<= depthBits;
    x_end <<= depthBits;

    /* point bits to first pixel and adjust x_start, x_end */
    offset = x_start >> 4;
    bits += offset;
    x_start &= 0xf;
    x_end -= offset << 4;

    /* first word */
    mask = 0xffff >> x_start;
    if (x_end < 16)
	goto finish;

    *bits ^= (*bits ^ color) & mask;
    bits++;
    x_end -= 16;

    /* middle bytes */
    while (x_end >= 16) {
        *bits++ = color;
        x_end -= 16;
    }

    mask = 0xffff;
 finish:
    /* last byte */
    if (x_end > 0) {
        mask &= 0xffff << (16 - x_end);
        *bits ^= (*bits ^ color) & mask;
    }
}


BITMAPSECTION void
BmpBitsDrawPixel16(UInt16 *bits, Int16 depth, Coord x_start, UInt16 color)
{
    bits[x_start] = color;
}


BITMAPSECTION void
BmpBitsDrawPixel8(UInt16 *bits, Int16 depth, Coord x_start, UInt16 color)
{
    ((UInt8*)bits)[x_start] = color;
}


BITMAPSECTION void
BmpDrawHLine (BitmapPtr dest, 
              Coord x_start, Coord x_end, Coord y, Int16 color)
{
    Int16 depth;
    UInt16* bits;
    Coord width, height;
    UInt16 rowBytes;

    depth = BmpGlueGetBitDepth(dest);
    bits  = BmpGetBits(dest);
    BmpGlueGetDimensions(dest, &width, &height, &rowBytes);

    if (y < 0 || y >= height)
        return;

    if (x_start < 0) 
        x_start = 0;
    if (x_end >= width)
        x_end = width;

    if (x_start < x_end) {
        bits = (void*)bits + rowBytes * y;
        BmpBitsDrawHLine(bits, depth, x_start, x_end, color);
    }
}


BITMAPSECTION void
BmpBitsDrawPixel (UInt16 *bits, Int16 depth,
                  Coord x, UInt16 color)
{
    Int16 depthBits = log2depth[depth];
    Int16 shift;
    UInt16 mask;

    /* adjust x so that it is in bits, instead of pixels */
    x <<= depthBits;

    bits += x >> 4;
    shift = (16 - depth - (x & 0xf));
    mask = ((1 << depth) - 1) << shift;
    color = color << shift;
    *bits ^= (*bits ^ color) & mask;
}
