/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: Bitmap.h 850 2008-12-05 00:56:53Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
#ifndef Bitmap_h
#define Bitmap_h

#define BITMAPSECTION __attribute__ ((section ("code3")))

/** 
 * The function to rotate or scale bitmaps.  The destination bitmap
 * must have the same depth as the source bitmap.  It copies a rotated
 * part of the source bitmap to the destination bitmap.
 *
 *  src bitmap      dest bitmap
 *
 * .----------.   .---,---,---.
 * | .-----.  |   |  /    /   |
 * | |     |  |   |  ~-_ /    |
 * | |_____|  |   |     ~     |
 * `----------'   `-----------'
 *
 * As you can see the rectangle can go beyond the bounds of the
 * destination image; it gets clipped then.  It can be scaled in x and
 * y independendly.  And finally different drawing modes may be used
 * (see documentation of WinDrawOperation in PalmOS reference manual).
 *
 * For each bitmap you specify the location and size of the rectangle
 * you want to work on.  For the destination rectangle you give the
 * center.  The content of source rectangle is rotated clockwise
 * around the center.  sina, cosa are a FP fixed point values.  as
 * used at other places in AstroInfo.
 *
 * @param dest             Pointer to destination bitmap.
 * @param cx, cy           center position of destination rectangle.
 * @param width, height    size of destination rectangle.
 * @param src              Pointer to sourc bitmap.
 * @param sx, sy           position of source rectangle.
 * @param swidth, sheight  size of source rectangle.
 * @param sina, cosa       sin and cos of rotation angle (clockwise).
 * @param transparentColor The color denoting transparent pixels.
 *                         -1 to disable.
 * @param drawMode The draw mode (All but WinSwap are supported).  */
extern void
BmpScaleRotateBitmap
(BitmapPtr dest, Coord cx, Coord cy, Coord width, Coord height,
 BitmapPtr src,  Coord sx, Coord sy, Coord switdh, Coord sheight,
 Int32 sina, Int32 cosa, Int16 xFlip,
 Int16 transparentColor, WinDrawOperation drawMode) BITMAPSECTION;

/**
 * Draw a horizontal line in the bitmap from (x_start, 0) inclusive to
 * (x_end, 0) exclusive.  It takes a pointer into the right row of the
 * bitmap.
 *
 * @param bits             Pointer to destination bitmap, must be in
 *                         uncompressed standard format.
 * @param depth            Depth of destination bitmap.
 * @param x_start          starting x coordinate (inclusive).
 * @param x_end            ending x coordinate (exclusive).
 * @param color            Color in which to draw.
 */
extern void
BmpBitsDrawHLine (UInt16 *bits, Int16 depth,
                  Coord x_start, Coord x_end, UInt16 color) BITMAPSECTION;

/* the following only works for depth = 8 */
extern void
BmpBitsDrawHLine8(UInt16 *bits, Int16 depth,
                  Coord x_start, Coord x_end, UInt16 color) BITMAPSECTION;
/* the following only works for depth = 16 */
extern void
BmpBitsDrawHLine16(UInt16 *bits, Int16 depth,
                  Coord x_start, Coord x_end, UInt16 color) BITMAPSECTION;


/**
 * Draw a horizontal line in the bitmap from (x_start, y) inclusive to
 * (x_end, y) exclusive.  If x_end <= x_start then don't draw anything.
 *
 * @param dest             Pointer to destination bitmap, must be in
 *                         uncompressed standard format.
 * @param x_start          starting x coordinate (inclusive).
 * @param x_end            ending x coordinate (exclusive).
 * @param y                y coordinate.
 * @param color            Color in which to draw.
 */
extern void
BmpDrawHLine (BitmapPtr dest,
              Coord x_start, Coord x_end, Coord y, Int16 color) BITMAPSECTION;


/**
 * Draw a horizontal line in the bitmap from (x_start, 0) inclusive to
 * (x_end, 0) exclusive.  It takes a pointer into the right row of the
 * bitmap.
 *
 * @param bits             Pointer to destination bitmap, must be in
 *                         uncompressed standard format.
 * @param depth            Depth of destination bitmap.
 * @param x                x coordinate.
 * @param color            Color in which to draw.
 */
extern void
BmpBitsDrawPixel (UInt16 *bits, Int16 depth,
                  Coord x, UInt16 color) BITMAPSECTION;


/* The following works only for depth = 8 */
extern void
BmpBitsDrawPixel8(UInt16 *bits, Int16 depth,
                  Coord x, UInt16 color) BITMAPSECTION;

/* The following works only for depth = 16 */
extern void
BmpBitsDrawPixel16(UInt16 *bits, Int16 depth,
                  Coord x, UInt16 color) BITMAPSECTION;


#endif
