/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 * 
 * $Id: CityTime.c 613 2006-08-21 18:24:39Z hoenicke $
 * Copyright (C) 2001, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "Astro.h"
#include "AstroDST.h"
#include "AstroLib.h"
#include "CityTime.h"

static DmOpenRef cityDB = NULL;

/* This special rule has always DST on:
 *
 * Start month is 0, end month is 15, so all days are
 * automatically between them.
 */
static const DSTRule alwaysDST = {
    /* startTime, startMon */ 0, 0,
    /* endTime,   endMon   */ 0, 15,
    /* isUTC */               0,
    /* startDOW, startDay */  7, 0, 
    /* endDay, endDOW */      0, 7, 
    /* savings */             60
};

static CITYSECTION void
CityConvertDST(dstType *cityDST, DSTRule *astroDST)
{
    switch (cityDST->dstManual) {
    case 0:
        /* Manual DST with DST off. */
        astroDST->savings = 0;
        break;
    case 1:
        /* Manual DST with DST on. */
        MemMove(astroDST, &alwaysDST, sizeof(alwaysDST));
        break;
    default:
        astroDST->savings = 60;
        astroDST->isUTC = 0;

        /* Start time is fixed, see CityTime DST.txt */
        astroDST->startTime = 2*60;
        astroDST->startMon = cityDST->start.month;
        if (cityDST->start.weekNo == 5) {
            astroDST->startMon++;
            astroDST->startDay = -6;
        } else
            astroDST->startDay = cityDST->start.weekNo * 7 - 6;
        astroDST->startDOW = cityDST->start.weekDay;

        astroDST->endTime = 1*60;
        astroDST->endMon = cityDST->stop.month;
        if (cityDST->stop.weekNo == 5) {
            astroDST->endMon++;
            astroDST->endDay = -6;
        } else
            astroDST->endDay = cityDST->stop.weekNo * 7 - 6;
        astroDST->endDOW = cityDST->stop.weekDay;
    }
}

static CITYSECTION UInt32
CityCoord(degMinType coord) {
    if (coord.deg < 0 && coord.min > 0)
        coord.min = -coord.min;
    return DMSToAng(coord.deg, coord.min, 0);
}

UInt32
CityOpen() {
    cityDB = DmOpenDatabaseByTypeCreator(CityTimeDBType, CityTimeAppType,
                                         dmModeReadOnly);
    if (!cityDB)
        return 0;

    return DmNumRecords(cityDB);
}

void
CityClose() {
    if (cityDB) {
        Err err = DmCloseDatabase(cityDB);
        ErrFatalDisplayIf(err != errNone,
                          "Could not close city time database!");
        cityDB = NULL;
    }
}

void
CitySetCity(Int16 recNr) {
    MemHandle handle = DmQueryRecord(cityDB, recNr);
    CityRecordTypePtr cityRec = (CityRecordTypePtr) MemHandleLock(handle);
    userPrefs.Location[userPrefs.currentLoc].tz = cityRec->timeFromGmt * 900000L;
    userPrefs.Location[userPrefs.currentLoc].lat = CityCoord(cityRec->locn.lat);
    userPrefs.Location[userPrefs.currentLoc].lng = CityCoord(cityRec->locn.lon);
    CityConvertDST(&cityRec->dst, &userPrefs.Location[userPrefs.currentLoc].dstRule);
    MemHandleUnlock(handle);
}

void
CityListDrawData(Int16 itemNum, RectangleType *bounds, Char **itemsText)
{
    MemHandle handle;
    CityRecordTypePtr cityRec;

    handle = DmQueryRecord(cityDB, itemNum - 1);
    cityRec = (CityRecordTypePtr)MemHandleLock(handle);
    WinDrawChars(cityRec->cityName, StrLen(cityRec->cityName), 
                 bounds->topLeft.x, bounds->topLeft.y);
    MemHandleUnlock(handle);
}

UInt16
CityGetPrefs()
{
    CityTime30PreferenceType cityPrefs;
    UInt16 cityPrefsSize = sizeof(cityPrefs);
    if ((PrefGetAppPreferences(CityTimeAppType, CityTimePrefID,
                               &cityPrefs, &cityPrefsSize,
                               true) == noPreferenceFound))
        return -1;

    return cityPrefs.savedCity[0];
}
