/*
 * Astro Sky - an astronomical visualizer for PalmOS devices.
 * 
 * $Id: Grayscale.c 827 2008-11-10 13:31:09Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include "PalmUtil.h"
#include "Grayscale.h"
#include "Astro.h"

Boolean grayscaleSupport;
Boolean colorSupport;
Int32   screenDepth;
Int16   screenDensity = 72;
UInt32  winVersion;

static Int32 oldDepth;
static Boolean oldColor;

/* Enable Grayscale or color mode, if Palm supports it. 
 */
void EnableGraphicsMode ()
{
    UInt32    val;
    UInt32    depthMask;
    Int32     depth;
    Int16       i;
    WinHandle oldDrawWindow;
    Err error = 0;

#ifdef HANDERA_SUPPORT
    UInt32    version;
#endif

#if WE_WOULD_SUPPORT_20
    /* First check if this PalmOS is recent enough for grayscale/color. 
     *
     * WinScreenMode maps to ScrDisplayMode function versions of
     * PalmOS pre 3.5.  Features are equivalent and name change was
     * simply for consistency within the API. Obviously older devices
     * will report many features as not present.
     */
    if (!CheckMinRomVersion(sysVersion30)) { 
        oldDepth = depth = screenDepth = 1;
        oldColor = colorSupport = false;
        grayscaleSupport = false;
        return;
    }
#endif

    // Windows manager version >= 4 can support hires
    error = FtrGet(sysFtrCreator, sysFtrNumWinVersion, &winVersion);
    if (winVersion >= 4) {
        WinScreenGetAttribute(winScreenDensity, &val);
        screenDensity = val;
    }    

#ifdef HANDERA_SUPPORT
    /* Handera support */
    if (FtrGet(TRGSysFtrID, TRGVgaFtrNum, &version) == 0)
        if (sysGetROMVerMajor(version) >= 1)
        {
            SetPDAType(pdaHandera);

            VgaSetScreenMode(screenMode1To1, rotateModeNone);
            
            fontSmall = stdFont;
            fontNormal = VgaBaseToVgaFont(stdFont);
            fontBold = VgaBaseToVgaFont(boldFont);
        }
#endif

    depth = 1;       

    /* When this function is called we normally don't have a draw window.
     * We work on the display window. 
     */
    oldDrawWindow = WinSetDrawWindow(WinGetDisplayWindow());

    /* Query old depth */
    WinScreenMode(winScreenModeGet, NULL, NULL, &oldDepth, &oldColor);

    /* Query supported depth and colormode */
    WinScreenMode(winScreenModeGetSupportedDepths,
                  NULL, NULL, &depthMask, NULL);

    /* We only support 16, 8, 4 and 1 bit */
    depthMask &= (1<<15) | (1 << 7) | (1 << 3) | (1 << 0);
    WinScreenMode(winScreenModeGetSupportsColor,
                  NULL, NULL, NULL, &colorSupport);

    /* Switch to maximum supported depth
     */
    depth = oldDepth;
    for (i = oldDepth + 1; i < 32; i++) {
        if ((depthMask & (1UL << (i-1))))
            depth = i;
    }

    /*
     * Force into colour mode if available in case we came here from
     * somewhere weird.
     */
    if (oldDepth != depth || oldColor != colorSupport)
        WinScreenMode(winScreenModeSet, NULL, NULL, &depth, &colorSupport);

    screenDepth = depth;
    grayscaleSupport = depth > 1;
        
    /* Restore old draw window (which is normally NULL) */
    WinSetDrawWindow(oldDrawWindow);
}

void DisableGraphicsMode() {
    Err err;


    if (oldDepth != screenDepth || oldColor != colorSupport) {

        err = WinScreenMode(winScreenModeSet, NULL, NULL, &oldDepth, &oldColor);
        ErrFatalDisplayIf(err!=errNone, "Could not restore depth!");
    }

    grayscaleSupport = false;
}


void GetScreenDimensions(FormPtr formPtr, RectangleType *screenRect)
{

    FrmGetFormBounds(formPtr, screenRect);
    screenRect->extent.x -= 2;
    screenRect->extent.y -= 2;
}



Int16 GetTitleAreaHeight()
{
    return 15;
}
