/*
 * Astro Info - a an astronomical calculator/almanac for PalmOS devices.
 *
 * $Id: PalmUtil.c 987 2010-05-24 19:07:08Z arpruss $
 * Copyright (C) 2001, Astro Info SourceForge Project
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <PalmOS.h>
#include <UIColor.h>
#include <FrmGlue.h>
#include "PalmUtil.h"
#include "Astro.h"
#include "resource.h"
#include "Grayscale.h"

#include <HostControl.h>
void _UI_BLACK(UInt16 libRef, UInt8 flag) SYS_TRAP(0xA821);
static void SetDisplayBorder(Boolean value);
static void SaveUIColors(void);
static void RestoreUIColors(void);

static RGBColorType savedUIColors[UILastColorTableEntry];
static Boolean inNightMode = false;
UInt32 romVersion;
UInt16 ticksPerSec;
RockerType currentRocker;

void* FrmGetObjectPtrByID(FormPtr formPtr, UInt16 objid)
{
    UInt16  objIndex = FrmGetObjectIndex(formPtr, objid);

    if (objIndex == frmInvalidObjectId) {
        Char str[100];
        StrPrintF(str, "Missing form element! %d", objid);
        ErrDisplay(str);
        return NULL;
    }

    return (FieldPtr) FrmGetObjectPtr(formPtr, objIndex);
}



/* Retrieve the current value integer value of a field */
Int16 GetNumField(FormPtr formPtr, UInt16 id)
{
    Int16 value;
    Char* p;

    FieldPtr fld = (FieldPtr) FrmGetObjectPtrByID(formPtr, id);
    p = FldGetTextPtr(fld);
    if (p==NULL)
        return 0;
    value = StrAToI(p);
    return value;
}

/* Set the current value integer value of a field */
void SetNumField(FormPtr formPtr, UInt16 id, Int16 value)
{
    Char s[maxStrIToALen+1];
    FieldPtr fld = (FieldPtr) FrmGetObjectPtrByID(formPtr, id);

    StrIToA(s, value);

    FldDelete(fld, 0, 0x7FFF);
    FldInsert(fld, s, StrLen(s));
    
    /* Kludge: Flush out fldChanged event */
    if (EvtEventAvail()) {
        EventType e;
        
        EvtGetEvent(&e, 0);
    }
}

void
GotoApplication(UInt32 ApplId)
{
    Err err;
    UInt16 cardNo;
    LocalID dbID;
    DmSearchStateType dss;

    /*
     * Bounce the application
     */
    err = DmGetNextDatabaseByTypeCreator(true,
                                         &dss,
                                         sysFileTApplication,
                                         ApplId,
                                         true,
                                         &cardNo,
                                         &dbID);
    ErrFatalDisplayIf(err != errNone,"Failed to launch application!");
    
    SysUIAppSwitch(cardNo,dbID,
                   sysAppLaunchCmdNormalLaunch,
                   NULL);
}



/*
 * Set the field text.  It allocates a copy on the heap, so you can give
 * a local variable as text.
 */
void
UpdateTextField(UInt16 objid, Char *text)
{
    FormPtr  formPtr = FrmGetActiveForm();
    FieldPtr fld = (FieldPtr) FrmGetObjectPtrByID(formPtr, objid);
    MemHandle handle;

    FldFreeMemory(fld);
    handle = MemHandleNew(StrLen(text) + 1);
    if (handle==NULL)
        return;
    StrCopy (MemHandleLock(handle), text);
    MemHandleUnlock(handle);
    FldSetTextHandle(fld, handle);
}



/*
 * Update a text label.  This can't be used for control labels.
 */
void
SetTextLabel(UInt16 objid, Char *t)
{
    FormPtr       formPtr = FrmGetActiveForm();
    UInt16        objIndex = FrmGetObjectIndex(formPtr, objid);
    
    if (objIndex == frmInvalidObjectId) {
        Char str[100];
        StrPrintF(str, "Missing form element! %d, '%s'", objIndex, t);
        ErrNonFatalDisplayIf(objIndex == frmInvalidObjectId, str);
        return;
    }
    
    /* Hide to force redraw in case label text is smaller now. */
    FrmHideObject(formPtr, objIndex);
    FrmCopyLabel(formPtr, objid, t);
    FrmShowObject(formPtr, objIndex);
}



/*
 * Update a text label.  This can't be used for control labels.
 */
void
SetTextLabelAligned(UInt16 objid, Char *t, Int16 horzpos, JustificationType align)
{
    FormPtr       formPtr = FrmGetActiveForm();
    UInt16        objIndex = FrmGetObjectIndex(formPtr, objid);
    RectangleType   rect;
    
    if (objIndex == frmInvalidObjectId) {
        Char str[100];
        StrPrintF(str, "Missing form element! %d, '%s'", objIndex, t);
        ErrNonFatalDisplayIf(objIndex == frmInvalidObjectId, str);
        return;
    }
    
    /* Hide to force redraw in case label text is smaller now. */
    FrmHideObject(formPtr, objIndex);
    
    FrmCopyLabel(formPtr, objid, t);

    FrmGetObjectBounds(formPtr, objIndex, &rect);

    HostTraceOutputTL(appErrorClass, "SetTextLabelAligned %hd %hd >%s<", horzpos, rect.extent.x, t);

    FrmSetObjectPosition(formPtr, objIndex,
                         horzpos + (align == leftAlign ? 0 : -(rect.extent.x / ((align == centerAlign) ? 2 : 1))),
                         rect.topLeft.y);

    FrmShowObject(formPtr, objIndex);
}



/*
 * Update a control label.  This can't be used for text labels.
 * This doesn't copy it, you must make sure that it doesn't get invalidated
 * while the control is visible.
 */
void
SetControlLabel(UInt16 objid, Char *t)
{
    FormPtr     formPtr = FrmGetActiveForm();
    ControlPtr  fld;
    UInt16      objIndex = FrmGetObjectIndex(formPtr, objid);

    if (objIndex == frmInvalidObjectId) {
        Char str[100];
        StrPrintF(str, "SCL: Missing form element! id=%d, value='%s'", objid, t);
        ErrNonFatalDisplayIf(objIndex == frmInvalidObjectId, str);
        return;
    }

    fld = (ControlPtr)FrmGetObjectPtrByID(formPtr, objid);
  
    /* Hide to force redraw in case label text is smaller now, as PalmOS < 3.5
     * does not redraw correctly. 
     */
    FrmHideObject(formPtr, objIndex);
    CtlSetLabel(fld, t);
    FrmShowObject(formPtr, objIndex);
}



void SetControlValue(UInt16 objid, Int16 num)
{
    FormPtr     formPtr = FrmGetActiveForm();
    ControlPtr  fld = (ControlPtr)FrmGetObjectPtrByID(formPtr, objid);

    if (objid == 0)
        return;

    if (fld == NULL) {
        Char str[100];
        StrPrintF(str, "SCV: Missing form element! id=%d, value='%d'", objid, num);
        ErrDisplay(str);
        return;
    }
    HostTraceOutputTL(appErrorClass, "SCV %hu, %h", objid, num);

    CtlSetValue(fld, num);
}



Int16 GetControlValue(UInt16 objid)
{
    FormPtr     formPtr = FrmGetActiveForm();
    ControlPtr  fld = (ControlPtr)FrmGetObjectPtrByID(formPtr, objid);
    if (fld == NULL) {
        Char str[100];
        StrPrintF(str, "SCV: Missing form element! id=%d", objid);
        ErrDisplay(str);
        return -1;
    }

    return CtlGetValue(fld);
}



/*
 * Update an field with a text value.
 */
void
SetFieldValue(UInt16 objid, Int16 num)
{
    Char text[FIELDLENGTH];
    StrIToA(text, num);
    
    UpdateTextField(objid, text);
}



Char * 
GetStringResource(UInt16 stringid, Char *buffer)
{
    Char* text;
    MemHandle handle = DmGetResource('tSTR', stringid);

    if (handle) {
        text = MemHandleLock(handle);
        StrNCopy(buffer, text, FIELDLENGTH);
        MemHandleUnlock(handle);
        DmReleaseResource(handle);
    } else {
        *buffer = 0;
    }
    return buffer;
}



/*
 * This color table defines a set things to be drawn in red.
 */
static const UIColorTableEntries redList[] = {
    UIObjectFrame, UIObjectForeground, UIObjectSelectedFill,
    UIMenuFrame, UIMenuForeground, UIMenuSelectedFill,
    UIFieldText, UIFieldTextLines, UIFieldCaret, UIFieldTextHighlightBackground,
    UIFieldFepRawText, UIFieldFepConvertedText,
    UIFieldFepUnderline, UIFormFrame,
    UIDialogFrame, UIAlertFrame, UIOK, UICaution, UIWarning,
};



/*
 * My version of a color table.
 */
typedef struct MyColorTableType
{
    UInt16 numEntries;
    RGBColorType rgb[UILastColorTableEntry];
    UInt32 appid;
} MyColorTableType;


/*
 * Sets or clears the red-on-black color scheme for color palm pilots
 * to match the userPrefs.optFlags.nightMode flag.
 *
 * Color tables are saved by renaming them from type sysResTSysPref
 * to type MWH2.
 */
 
static void RestoreUIColors(void)
{
    UInt16 i;

    for (i=0; i < sizeof(savedUIColors) / sizeof(savedUIColors[0]); i++) {
         UIColorSetTableEntry(i, &(savedUIColors[i]));
    }
}

/* 
 * Check if night mode is set with NightMode DAs from 1src.com
 */

static Boolean ExternalNightMode(void)
{
    UInt32 ignore;

    return errNone == FtrGet( 'NghA', 2000, &ignore ) ||
           errNone == FtrGet( 'Nigh', 2000, &ignore );
}


static void SaveUIColors(void)
{
    UInt16 i;

    for (i=0; i < sizeof(savedUIColors) / sizeof(savedUIColors[0]); i++) {
         UIColorGetTableEntryRGB(i, &(savedUIColors[i]));
    }
}


void
NightMode()
{
    Int16 i,j;

    RGBColorType redColor = { 0, 0xA0, 0, 0 };
    RGBColorType blackColor = { 0, 0, 0, 0 };

    if (!CheckMinRomVersion(sysVersion35))
        return;

    if (!userPrefs.optFlags.nightMode) {
        if (inNightMode) {
            RestoreUIColors();
            if (!ExternalNightMode())
               SetDisplayBorder(true);
            inNightMode=false;
        }
        return;
    }

    if (!inNightMode)
        SaveUIColors();

    j = 0;
    for (i=0; i < sizeof(redList) / sizeof(redList[0]); i++) {
        while (j < redList[i])
            UIColorSetTableEntry(j++, &blackColor);
        UIColorSetTableEntry(j++, &redColor);
    }
    while (j < UILastColorTableEntry)
        UIColorSetTableEntry(j++, &blackColor);

    SetDisplayBorder(false);
    inNightMode = true;
}



void drawChars(Char *str,Int32 x,Int32 y, JustificationType align, Boolean bkgblack)
{
    Int16 len = StrLen(str);

    ErrFatalDisplayIf(!str,"str is NULL in drawChars");

    if (align != leftAlign)
    {
        Int16 width = FntCharsWidth(str,len);
        x -= (align == centerAlign) ? width / 2 : width;
    }

    if (!bkgblack)
        WinDrawChars(str,len,x,y);
    else
        WinEraseChars(str,len,x,y);
}


/* This works on the Palm TX and LifeDrive, and maybe on some other devices.
   It removes the 2-pixel border around the screen which is a bad thing in
   night mode. */
static void SetDisplayBorder(Boolean value)
{
    Err    err;
    UInt16 ref;
    UInt32 deviceID;

    FtrGet( sysFtrCreator, sysFtrNumOEMDeviceID, &deviceID );

    if ( deviceID == 'Zi72' ) {
        /* apparently on the Z72, you can get a hang if an alarm occurs with
        the device asleep without the border.  We could monitor for the device
        going to sleep and fix this, but that would be a lot of work. */
        return;
    }

    err = SysLibFind("PmSystemLib-PmSy", &ref);
    if (err == sysErrLibNotFound) {
        err = SysLibLoad('libr', 'PmSy', &ref);
    }
    if ( err == errNone ) {
        _UI_BLACK( ref, ! value );
        SysLibRemove( ref );
    }
}


void NightModeCleanup()
{
    if ( inNightMode ) {
        if (! ExternalNightMode())
            SetDisplayBorder( true );
        inNightMode = false;
    }
}


Char* TrimName(Char* source, const Char* prefix)
{
    UInt16 len;

    len = StrLen(prefix);

    if (!StrNCompare(source, prefix, len))
        return source + len;
    else
        return source;
}



/* Check if we're in 5-way interaction mode */
Boolean InteractionMode(void)
{
    FrmNavStateFlagsType flags;

    return FrmGlueNavIsSupported() &&
        errNone == FrmGetNavState(FrmGetActiveForm(), &flags) &&
        (flags & kFrmNavStateFlagsInteractionMode);
}



void Show( UInt16 id, Boolean state )
{
    UInt16 index = GetObjectIndex( id );
    FormType* currentForm = FrmGetActiveForm();

    if ( state )
        FrmShowObject( currentForm, index );
    else
        FrmHideObject( currentForm, index );
    if ( FrmGetObjectType( currentForm, index ) == frmControlObj )
        CtlSetUsable( GetObjectPtr( id ), state);
}



UInt16 WinGlueSetCoordinateSystem(UInt16 newSystem)
{
    if (winVersion >= 4)
        return WinSetCoordinateSystem(newSystem);
    else
        return kCoordinatesStandard;
}
